﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;


namespace Jypeli
{
    /// <summary>
    /// Sarja kuvia, jotka vaihtuvat halutulla nopeudella. 
    /// Yksi animaatio koostuu yhdestä tai useammasta kuvasta.
    /// </summary>
    /// <remarks>
    /// Eri peliolioille kannattaa yleensä tehdä oma animaatio-olio,
    /// jotta niiden animaatiot voivat edetä eri tahtiin. Jos animaatioilla
    /// on kuitenkin samat kuvat, kannattaa samaa kuvataulukkoa käyttää
    /// molemmille.
    /// <example>
    /// Image[] kuvat = LoadImages("kuva1", "kuva2", "kuva3");
    /// 
    /// void LuoKentta()
    /// {
    ///     // ...
    ///     o1.Animation = new animation(kuvat);
    ///     o2.Animation = new Animation(kuvat);
    /// }
    /// </example>
    /// </remarks>
    public class Animation
    {
        private double secondsPerFrame;
        private TimeSpan startTime;
        private bool isAnimated = false;
        private int currentIndex;
        private int repeatCount = -1; // When -1, animation repeats forever.

        // NOTES:
        //
        // Animation frames should be immutable, because that is what
        // the Layer class assumes. So, let's not provide
        // any methods to modify the images of an existing animation.
        //
        // This is not a big deal, as there rarely is any need
        // to modify animations at runtime and constructing new
        // animation objects is quite easy anyway.

#if !XBOX && !WINDOWS_PHONE
        [NonSerialized]
#endif
        internal Image[] frames = null;

        /// <summary>
        /// Ruutujen määrä.
        /// </summary>
        public int FrameCount
        {
            get { return frames.Length; }
        }

        int lastRepeat = 0;

        /// <summary>
        /// Aktiivisen animaatioruudun indeksi (alkaa nollasta).
        /// </summary>
        public int CurrentFrameIndex
        {
            get
            {
                if ( !isAnimated ) return currentIndex;

                double secondsNow = Game.Time.SinceStartOfGame.TotalSeconds;
                double secondsAdvanced = secondsNow - startTime.TotalSeconds;
                int currentRepeat = (int)( secondsAdvanced / ( FrameCount * secondsPerFrame ) );

                if ( currentRepeat > lastRepeat )
                {
                    OnPlayed();
                    lastRepeat = currentRepeat;
                }
                    
                if ( repeatCount >= 0 && currentRepeat >= repeatCount )
                    Stop();

                return ( (int)( secondsAdvanced * FPS ) ) % FrameCount;
            }
        }

        /// <summary>
        /// Animaation päivitysnopeus ruutuina sekunnissa (frames per second).
        /// </summary>
        public double FPS
        {
            get { return 1 / secondsPerFrame; }
            set
            {
                if ( value == 0 )
                    throw new ArgumentException( "FPS can not be zero" );
                secondsPerFrame = 1 / value;
            }
        }

        /// <summary>
        /// Tällä hetkellä näkyvä ruutu animaatiosta.
        /// </summary>
        public Image CurrentFrame
        {
            get { return frames[CurrentFrameIndex]; }
        }

        /// <summary>
        /// Jos <c>true</c>, animaatio ei pysähtyessä palaa ensimmäiseen
        /// ruutuun.
        /// </summary>
        public bool StopOnLastFrame { get; set; }

        /// <summary>
        /// Tapahtuma, joka tapahtuu kun animaatio on suoritettu.
        /// </summary>
        public event Action Played;

        private void OnPlayed()
        {
            if ( Played != null ) Played();
        }

        /// <summary>
        /// Luo uuden animaation.
        /// </summary>
        /// <param name="frames">Animaation ruudut.</param>
        public Animation( params Image[] frames )
        {
            if ( frames.Length == 0 )
                throw new ArgumentException( "Animation must have at least one frame." );
            FPS = 25;
            currentIndex = 0;
            startTime = Game.Time.SinceStartOfGame;
            this.frames = frames;
        }

        /// <summary>
        /// Luo uuden animaation kuvasta.
        /// </summary>
        /// <param name="image">Kuva.</param>
        public static implicit operator Animation(Image image)
        {
            if ( image == null ) return null;
            return new Animation( image );
        }

        /// <summary>
        /// Luo kopion jo tunnetusta animaatiosta.
        /// </summary>
        /// <param name="src">Kopioitava animaatio.</param>
        internal Animation( Animation src )
        {
            FPS = src.FPS;
            isAnimated = src.isAnimated;
            startTime = src.startTime;
            currentIndex = src.currentIndex;
            repeatCount = src.repeatCount;
            frames = new Image[src.FrameCount];

            // Copy only the references to images.
            for ( int i = 0; i < src.FrameCount; i++ )
                frames[i] = src.frames[i];
        }

        /// <summary>
        /// Käynnistää animaation alusta.
        /// </summary>
        public void Start()
        {
            Start( -1 );
        }

        /// <summary>
        /// Käynnistää animaation alusta.
        /// </summary>
        /// <param name="repeatCount">Kuinka monta kertaa animaatio suoritetaan.</param>
        public void Start( int repeatCount )
        {
            this.repeatCount = repeatCount;
            startTime = Game.Time.SinceStartOfGame;
            isAnimated = true;
            lastRepeat = 0;
        }

        /// <summary>
        /// Jatkaa animaatiota siitä, mihin viimeksi jäätiin.
        /// </summary>
        public void Resume()
        {
            isAnimated = true;
        }

        /// <summary>
        /// Pysäyttää animaation asettaen sen ensimmäiseen ruutuun.
        /// </summary>
        public void Stop()
        {
            isAnimated = false;
            currentIndex = 0;
            if ( StopOnLastFrame )
                currentIndex = FrameCount - 1;
            repeatCount = -1;
        }

        /// <summary>
        /// Etenee animaatiossa halutun määrän ruutuja. Käytä negatiivisia arvoja, jos haluat
        /// kulkea taaksepäin.
        /// </summary>
        /// <param name="numberOfFrames">Edettävä määrä ruutuja.</param>
        public void Step( int numberOfFrames )
        {
            if ( isAnimated )
                Stop();
            
            //currentIndex = ( currentIndex + numberOfFrames ) % FrameCount;
            currentIndex = currentIndex + numberOfFrames;

            if ( currentIndex >= FrameCount )
            {
                // Animation has reached its final frame
                if ( repeatCount > 0 ) repeatCount--;

                if ( repeatCount == 0 )
                    Stop();
                else
                {
                    // Play it again, Sam
                    currentIndex %= FrameCount;
                }

                OnPlayed();
            }
        }

        /// <summary>
        /// Etenee animaatiossa yhden ruudun eteenpäin.
        /// </summary>
        public void Step()
        {
            Step( 1 );
        }
    }
}
