﻿using System;
using System.ComponentModel;

namespace Jypeli.Assets
{
    [EditorBrowsable( EditorBrowsableState.Never )]
    public abstract class AbstractMoverBrain : Brain
    {
        private double _speed = 10;

        /// <summary>
        /// Nopeus, jolla liikutaan.
        /// </summary>
        /// <value>Nopeus.</value>
        public double Speed
        {
            get { return _speed; }
            set { _speed = value; }
        }

        public AbstractMoverBrain()
        {
        }

        public AbstractMoverBrain( double speed )
        {
            this.Speed = speed;
        }
    }

    /// <summary>
    /// Aivot, joiden tehtävänä on laittaa omistajansa seuraamaan jotakin kohdetta.
    /// </summary>
    public class FollowerBrain : AbstractMoverBrain
    {
        #region Properties
        /// <summary>
        /// Kohde, jota seurataan.
        /// </summary>
        /// <value>Kohde.</value>
        public GameObject Target { get; set; }

        /// <summary>
        /// Tapahtuma, joka suoritetaan, kun ollaan tarpeeksi lähellä seurattavaa.
        /// </summary>
        public event EventHandler TargetClose;

        /// <summary>
        /// Etäisyys, jolloin ollaan lähellä kohdetta.
        /// </summary>
        /// <value>Etäisyys.</value>
        public double TargetCloseDistance { get; set; }

        /// <summary>
        /// Pysähdytäänkö, kun ollaan lähellä kohdetta.
        /// </summary>
        /// <value>
        /// 	<c>true</c> jos pysähdytään; jos ei, niin <c>false</c>.
        /// </value>
        public bool StopWhenTargetClose { get; set; }

        /// <summary>
        /// Seuraako kohdetta aina, riippumatta etäisyydestä.
        /// </summary>
        /// <value>
        /// 	<c>true</c> jos seuraa; jos ei, niin <c>false</c>.
        /// </value>
        public bool FollowAlways { get; set; }

        /// <summary>
        /// Etäisyys kohteeseen. Jos tämä tai alle, niin jolloin aletaan seuraamaan kohdetta.
        /// </summary>
        public double TargetFollowDistance { get; set; }
        #endregion

        /// <summary>
        /// Luo aivot.
        /// </summary>
        public FollowerBrain()
            : this( null )
        {
        }

        /// <summary>
        /// Luo aivot ja asettaa ne seuraamaan kohdetta.
        /// </summary>
        /// <param name="target">Seurattava olio.</param>
        public FollowerBrain( GameObject target )
            : base()
        {
            this.Target = target;
            StopWhenTargetClose = false;
            FollowAlways = false;
            TargetFollowDistance = double.PositiveInfinity;
            TargetCloseDistance = 100.0;
        }

        /// <summary>
        /// Luo aivot ja asettaa ne seuraamaan kohdetta tietyllä nopeudella.
        /// </summary>
        /// <param name="target">Seurattava olio.</param>
        /// <param name="speed">Nopeus.</param>
        public FollowerBrain( GameObject target, double speed )
            : base( speed )
        {
            this.Target = target;
            StopWhenTargetClose = false;
            FollowAlways = false;
            TargetFollowDistance = 500.0;
            TargetCloseDistance = 100.0;
        }

        /// <summary>
        /// Kutsutaan, kun tilaa päivitetään.
        /// Suurin osa päätöksenteosta tapahtuu täällä.
        /// </summary>
        protected override void Update( Time time )
        {
            if ( Target == null ) return;

            double distance = Vector.Distance( this.Owner.Position, Target.Position );

            bool targetClose = Math.Abs( distance ) < TargetCloseDistance;
            bool targetFar = Math.Abs( distance ) > TargetFollowDistance;
            if ( targetClose && TargetClose != null ) TargetClose( this, EventArgs.Empty );

            if ( this.Owner is PhysicsObject )
            {
                PhysicsObject po = this.Owner as PhysicsObject;
                po.Stop();
            }

            if ( !FollowAlways )
            {
                if ( StopWhenTargetClose && targetClose || targetFar )
                    return;
            }

            Vector movement = Target.Position - this.Owner.Position;
            Owner.Move( movement * Speed / movement.Magnitude );

            base.Update( time );
        }
    }

    /// <summary>
    /// Aivot, jotka laittavat omistajansa liikkumaan satunnaisesti eri suuntiin törmäyksissä ja tietyn ajan jälkeen.
    /// </summary>
    public class RandomMoverBrain : AbstractMoverBrain
    {
        private Timer changeDirectionTimer;
        private Vector currentDirection = Vector.Zero;

        /// <summary>
        /// Aika sekunteina, jonka kuluttua muutetaan liikesuuntaa
        /// </summary>
        public double ChangeMovementSeconds
        {
            get { return changeDirectionTimer.Interval; }
            set { changeDirectionTimer.Interval = value; }
        }

        /// <summary>
        /// Luo uudet satunnaisliikkujan aivot.
        /// </summary>
        public RandomMoverBrain()
        {
            changeDirectionTimer = new Timer();
            changeDirectionTimer.Timeout += ChangeDirection;
            changeDirectionTimer.Interval = 3;

            currentDirection = Vector.FromLengthAndAngle( 1, RandomGen.NextAngle() );
        }

        /// <summary>
        /// Luo uudet satunnaisliikkujan aivot ja asettaa niille nopeuden.
        /// </summary>
        public RandomMoverBrain( double speed )
            : base( speed )
        {
            changeDirectionTimer = new Timer();
            changeDirectionTimer.Timeout += ChangeDirection;
            changeDirectionTimer.Interval = 3;

            currentDirection = Vector.FromLengthAndAngle( 1, RandomGen.NextAngle() );
        }

        protected override void OnAdd( GameObject newOwner )
        {
            changeDirectionTimer.Start();
            base.OnAdd( newOwner );
        }

        protected override void OnRemove( GameObject prevOwner )
        {
            changeDirectionTimer.Stop();
            base.OnRemove( prevOwner );
        }

        void ChangeDirection()
        {
            if ( Owner == null ) return;
            //currentDirection = Vector.FromLengthAndAngle( double.MaxValue, RandomGen.NextAngle() );
            //Owner.MoveTo( Owner.Position + currentDirection, Speed );
            currentDirection = Vector.FromLengthAndAngle( 1, RandomGen.NextAngle() );
        }

        /// <summary>
        /// Kutsutaan, kun tapahtuu törmäys.
        /// </summary>
        /// <param name="target">Olio, johon törmätään.</param>
        public override void OnCollision( GameObject target )
        {
            Vector d = target.Position - Owner.Position;
            Angle n1 = d.Angle - Angle.RightAngle;
            Angle n2 = d.Angle + Angle.RightAngle;

            //currentDirection = Vector.FromLengthAndAngle( double.MaxValue, RandomGen.NextAngle( n1, n2 ) );
            //Owner.MoveTo( Owner.Position + currentDirection, Speed );
            currentDirection = Vector.FromLengthAndAngle( 1, RandomGen.NextAngle( n1, n2 ) );
            
            base.OnCollision( target );
        }

        protected override void Update( Time time )
        {
            if ( Owner == null ) 
                return;

            Vector moveVector = currentDirection * Speed * time.SinceLastUpdate.TotalSeconds;
            Owner.Move( moveVector );
            base.Update( time );
        }
    }
}
