﻿
#if DEBUG
// #define VISUALIZE
#endif

using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Linq;
using Jypeli;
using Jypeli.Assets;


/// <summary>
/// Tasohyppelypelin hahmo. Voi liikkua ja hyppiä. Lisäksi sillä voi olla ase.
/// </summary>
public class PlatformCharacter : PhysicsObject
{
    private class CollisionHelper
    {
#if VISUALIZE
        private struct Appearance
        {
            public Color Color;
            public Animation Animation;
        }

        private static Dictionary<PhysicsObject, Appearance> originalAppearances = new Dictionary<PhysicsObject, Appearance>();

        private static void ResetAppearance(PhysicsObject o)
        {
            if (!originalAppearances.ContainsKey(o))
                return;
            Appearance original = originalAppearances[o];
            o.Color = original.Color;
            o.Animation = original.Animation;
            originalAppearances.Remove(o);
        }
#endif

        private PhysicsObject parent;
        public PhysicsObject Object;
        public PhysicsObject LastHitObject;

        public CollisionHelper( PhysicsObject parent )
        {
            this.parent = parent;
        }

        public void SetObjectBeingHit(PhysicsObject collisionHelper, PhysicsObject target)
        {
            if (target != parent && IsPlatform(target))
            {
#if VISUALIZE
                if (! originalAppearances.ContainsKey(target))
                {
                    originalAppearances.Add(target, new Appearance { Animation = target.Animation, Color = target.Color });
                }

                target.Color = Color.Red;
                target.Animation = null;
                Timer.SingleShot(1.0, delegate() { ResetAppearance(target); });
#endif

                LastHitObject = target;
            }
        }
    }


    private Weapon weapon = null;
    private CollisionHelper[] collisionHelpers = new CollisionHelper[3];
    private Animation weaponFacingLeft;
    private Animation weaponFacingRight;
    private Timer jumpTimer = new Timer { Interval = 0.5 };
    private bool isJumping = false;
    private bool isWalking = false;
    private Direction _facingDirection = Direction.Right;

    private double lowTolerance { get { return Height * 0.1; } }
    private double highTolerance { get { return Height * 0.2; } }

    /// <summary>
    /// Hahmon rintamasuunta (vasen tai oikea).
    /// </summary>
    public Direction FacingDirection
    {
        get { return _facingDirection; }
        set { Turn( value ); }
    }

    public override Vector Size
    {
        get
        {
            return base.Size;
        }
        set
        {
            base.Size = value;
            for (int i = 0; i < collisionHelpers.Length; i++)
                collisionHelpers[i].Object.Size = new Vector(value.X / 3, value.Y);
        }
    }

    /// <summary>
    /// Hahmolla oleva ase.
    /// </summary>
    public Weapon Weapon
    {
        get { return weapon; }
        set
        {
            // First check: same weapon
            if ( weapon == value ) return;

            // Remove the previous weapon if any
            if ( weapon != null )
            {
                // Reset the weapon when removing

                if ( !IsWeaponFacingRight() )
                {
                    weapon.X *= -1;
                    weapon.Animation = weaponFacingRight;
                }

                weapon.Angle = Angle.Zero;
                this.Remove( weapon );
            }

            this.weapon = value;
            if ( value == null )
                return;

            this.Add( value );

            if ( value.Animation != null )
            {
                // Set the weapon animations
                weaponFacingRight = value.Animation;
                weaponFacingLeft = new Animation( Image.Flip( value.Animation.frames ) );

                if ( FacingDirection == Direction.Left )
                {
                    // If facing left, set the weapon to match the direction
                    weapon.X *= -1;
                    weapon.Angle = Angle.StraightAngle;
                    weapon.Animation = weaponFacingLeft;
                }
            }
        }
    }

    /// <summary>
    /// Jos <c>false</c>, hahmoa ei voi liikuttaa kun se on ilmassa.
    /// </summary>
    public bool CanMoveOnAir { get; set; }

    /// <summary>
    /// Jos <c>true</c>, hahmon liike jatkuu hidastuen vaikka kävelemisen lopettaa.
    /// </summary>
    public bool MaintainMomentum { get; set; }

    private Animation _leftWalkingAnim = null;
    private Animation _rightWalkingAnim = null;

    /// <summary>
    /// Kävelyanimaatio vasemmalle.
    /// </summary>
    public Animation LeftWalkingAnimation
    {
        get { return _leftWalkingAnim; }
        set
        {
            _leftWalkingAnim = value;
            if ( value != null && value.FrameCount > 0 && LeftIdleAnimation == null )
                LeftIdleAnimation = value.frames[0];
        }
    }

    /// <summary>
    /// Kävelyanimaatio oikealle.
    /// </summary>
    public Animation RightWalkingAnimation
    {
        get { return _rightWalkingAnim; }
        set
        {
            _rightWalkingAnim = value;
            if ( value != null && value.FrameCount > 0 && RightIdleAnimation == null )
                RightIdleAnimation = value.frames[0];
        }
    }

    /// <summary>
    /// Animaatio, jota käytetään kun hahmo on paikallaan,
    /// kääntyneenä vasemmalle.
    /// </summary>
    public Animation LeftIdleAnimation { get; set; }

    /// <summary>
    /// Animaatio, jota käytetään kun hahmo on paikallaan,
    /// kääntyneenä oikealle.
    /// </summary>
    public Animation RightIdleAnimation { get; set; }

    /// <summary>
    /// Hahmon suunnan muutos.
    /// </summary>
    public event Action<Direction> DirectionChanged;

    /// <summary>
    /// Luo uuden tasohyppelyhahmon.
    /// </summary>
    /// <param name="width">Hahmon leveys</param>
    /// <param name="height">Hahmon korkeus</param>
    public PlatformCharacter(double width, double height)
        : this(width, height, Shape.Rectangle)
    {
    }

    /// <summary>
    /// Luo uuden tasohyppelyhahmon.
    /// <param name="width">Hahmon leveys</param>
    /// <param name="height">Hahmon korkeus</param>
    /// </summary>
    public PlatformCharacter(double width, double height, Shape shape)
        : base(width, height, shape, CollisionShapeQuality.FromValue(0.7))
    {
        KineticFriction = 0.0;
        Restitution = 0.0;
        CanRotate = false;
        CanMoveOnAir = true;

        // This avoids high speeds, particularly when falling. This then avoids
        // going through objects.
        LinearDamping = 0.96;

        jumpTimer.Timeout += delegate { isJumping = false; };

        for (int i = 0; i < collisionHelpers.Length; i++)
        {
            collisionHelpers[i] = new CollisionHelper( this );
            collisionHelpers[i].Object = new PhysicsObject(width / 3, height)
            {
                IgnoresGravity = true,
                IgnoresCollisionResponse = true,
#if VISUALIZE
                IsVisible = true,
#else
                IsVisible = false,
#endif
            };
        }

#if VISUALIZE
        collisionHelpers[0].Object.Color = new Color(150, 150, 0, 20);
        collisionHelpers[1].Object.Color = new Color(150, 180, 0, 20);
        collisionHelpers[2].Object.Color = new Color(150, 210, 0, 20);
#endif

        AddedToGame += AddCollisionHelpers;

        IsUpdated = true;
    }

    private void AddCollisionHelpers()
    {
        if (Game is PhysicsGame)
        {
            PhysicsGame physicsGame = (PhysicsGame)Game;

            for (int i = 0; i < collisionHelpers.Length; i++)
            {
                physicsGame.Add(collisionHelpers[i].Object);
            }

            physicsGame.AddCollisionHandler(this, OnCollision);

            for (int i = 0; i < collisionHelpers.Length; i++)
            {
                physicsGame.AddCollisionHandler(collisionHelpers[i].Object, collisionHelpers[i].SetObjectBeingHit);
            }
        }
    }

    private static bool IsPlatform(PhysicsObject o)
    {
        return !o.IgnoresCollisionResponse;
    }

    /// <summary>
    /// Liikuttaa hahmoa.
    /// </summary>
    /// <param name="horizontalVelocity">Nopeus vaakasuunnassa.</param>
    public void Walk(double horizontalVelocity)
    {
        // TODO: Don't walk when against a wall. The character doesn't
        // fall down when walking against the wall (because of the friction,
        // I guess).

        if ( horizontalVelocity > 0 )
            Turn( Direction.Right );
        else if ( horizontalVelocity < 0 )
            Turn( Direction.Left );

        if (CanMoveOnAir || collisionHelpers.Any(c => IsStandingOn(c.LastHitObject, lowTolerance)))
        {
            this.Velocity = new Vector(horizontalVelocity, this.Velocity.Y);
        }

        isWalking = true;
    }

    /// <summary>
    /// Kääntyy.
    /// </summary>
    /// <param name="direction">Suunta</param>
    public void Turn( Direction direction )
    {
        if ( direction == FacingDirection || ( direction != Direction.Left && direction != Direction.Right ) )
            return;

        if ( direction == Direction.Right )
        {
            if ( RightWalkingAnimation != null )
            {
                Animation = RightWalkingAnimation;
                Animation.Resume();
            }

            if ( Weapon != null && !IsWeaponFacingRight() )
            {
                Weapon.Angle = Angle.Supplement( Weapon.Angle );
                Weapon.X = -Weapon.X;
                if ( weaponFacingRight != null )
                    Weapon.Animation = weaponFacingRight;
            }
        }
        else
        {
            if ( LeftWalkingAnimation != null )
            {
                Animation = LeftWalkingAnimation;
                Animation.Resume();
            }

            if ( Weapon != null && IsWeaponFacingRight() )
            {
                Weapon.Angle = Angle.Supplement( Weapon.Angle );
                Weapon.X = -Weapon.X;
                if ( weaponFacingLeft != null )
                    Weapon.Animation = weaponFacingLeft;
            }
        }

        _facingDirection = direction;

        if ( DirectionChanged != null )
        {
            DirectionChanged( direction );
        }
    }

    private bool IsWeaponFacingRight()
    {
        return (-Math.PI / 2) < Weapon.Angle.Radians
            && Weapon.Angle.Radians < (Math.PI / 2);
    }
    
    /// <summary>
    /// Hyppää, jos hahmo on staattisen olion päällä.
    /// </summary>
    /// <param name="speed">Lähtönopeus maasta.</param>
    /// <returns><code>true</code> jos hyppäys onnistui.</returns>
    public bool Jump(double speed)
    {
        for (int i = 0; i < collisionHelpers.Length; i++)
        {
            if (IsStandingOn(collisionHelpers[i].LastHitObject, highTolerance))
            {
                isJumping = true;
                this.IgnoresGravity = false;
                jumpTimer.Start(1);
                this.Hit(new Vector(0, speed * Mass));
                return true;
            }
        }
        return false;
    }

    /// <summary>
    /// Checks if the character is located on top of the target.
    /// </summary>
    /// <param name="target">Another object</param>
    /// <param name="yTolerance">How close must the distance on the y-axis must be</param>
    private bool IsStandingOn(PhysicsObject target, double yTolerance)
    {
        if (target == null || target.IsDestroyed) return false;

        // This prevents jumping when on the side of the wall, not on top of it.
        double epsilon = Width / 6;

        return (target.Position.Y <= this.Position.Y
            && (this.Bottom - target.Top) < yTolerance
            && (target.Left + epsilon) < this.Right
            && this.Left < (target.Right - epsilon));
    }

    private void OnCollision(PhysicsObject collisionHelperObject, PhysicsObject target)
    {
        if (!isJumping && IsPlatform(target) && IsStandingOn(target, highTolerance))
        {
            IgnoresGravity = true;
            StopVertical();
        }
    }

    public override void Destroy()
    {
        for (int i = 0; i < collisionHelpers.Length; i++)
        {
            collisionHelpers[i].Object.Destroy();
        }
        base.Destroy();
    }

    /// <summary>
    /// Ajetaan kun pelitilannetta päivitetään. Päivityksen voi toteuttaa omassa luokassa toteuttamalla tämän
    /// metodin. Perityn luokan metodissa tulee kutsua kantaluokan metodia.
    /// </summary>
    /// <param name="time">Peliaika.</param>
    [EditorBrowsable(EditorBrowsableState.Never)]
    public override void Update(Time time)
    {
        AdjustPosition();

        if (!isWalking)
            StopWalking();
        isWalking = false;

        base.Update(time);
    }

    private void AdjustPosition()
    {
        collisionHelpers[0].Object.Position = this.Position - new Vector(-Width / 4, Height * 1 / 8);
        collisionHelpers[1].Object.Position = this.Position - new Vector(0, Height * 1 / 8);
        collisionHelpers[2].Object.Position = this.Position - new Vector(Width / 4, Height * 1 / 8);

        if (isJumping)
        {
            return;
        }

        PhysicsObject platform = null;
        for (int i = 0; i < collisionHelpers.Length; i++)
        {
            if (IsStandingOn(collisionHelpers[i].LastHitObject, lowTolerance))
                platform = collisionHelpers[i].LastHitObject;
        }

        if (platform != null && IsStandingOn(platform, lowTolerance))
        {
            this.Y = platform.Top + this.Height / 2;
        }
        else
        {
            IgnoresGravity = false;
        }
    }

    private void StopWalking()
    {
        if ( !MaintainMomentum )
        {
            StopHorizontal();
        }

        if (FacingDirection == Direction.Left && LeftIdleAnimation != null)
        {
            Animation = LeftIdleAnimation;
            Animation.Resume();
        }
        else if (RightIdleAnimation != null)
        {
            Animation = RightIdleAnimation;
            Animation.Resume();
        }
    }

    /// <summary>
    /// Siirtää oliota.
    /// </summary>
    /// <param name="movement">Vektori, joka määrittää kuinka paljon siirretään.</param>
    public override void Move( Vector movement )
    {
        Vector dv = movement - this.Velocity;
        Walk( dv.X );
    }

    protected override void MoveToTarget()
    {
        if ( !moveTarget.HasValue )
        {
            Stop();
            moveTimer.Stop();
            return;
        }

        Vector d = moveTarget.Value - Position;
        double vt = moveSpeed * moveTimer.Interval;

        if ( d.Magnitude < vt )
        {
            Vector targetLoc = moveTarget.Value;
            Stop();
            moveTimer.Stop();
            moveTarget = null;
            OnArrivedAt( targetLoc );
        }
        else
        {
            Vector dv = Vector.FromLengthAndAngle( moveSpeed, d.Angle ) - this.Velocity;
            Walk( dv.X );
        }
    }
}
