﻿
#if DEBUG
// #define VISUALIZE
#endif

using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Linq;
using Jypeli;
using Jypeli.Assets;
using Physics2DDotNet;


/// <summary>
/// Tasohyppelypelin hahmo. Voi liikkua ja hyppiä. Lisäksi sillä voi olla ase.
/// </summary>
public class PlatformCharacter2 : PhysicsObject
{
    private int _tolerance = 5;
    private double _accel = 2;
    private double _maxVel = 2;

    private Animation _leftWalkingAnim = null;
    private Animation _rightWalkingAnim = null;

    private Weapon weapon = null;
    private Animation weaponFacingLeft;
    private Animation weaponFacingRight;

    private Vector _cachedGravity = Vector.Zero;
    private double _cachedGravityMagnitude = 0;
    private Vector _cachedGravityNormal = Vector.Zero;

    private bool collisionDetected = false;
    private int noCollisionCount = 0;
    private int walkSteps = 0;
    private Direction _curDirection = Direction.Right;
    private Timer detachTimer = new Timer();

    private double gravityMagnitude
    {
        get
        {
            updateGravity();
            return _cachedGravityMagnitude;
        }
    }

    private Vector gravityNormal
    {
        get
        {
            updateGravity();
            return _cachedGravityNormal;
        }
    }

    /// <summary>
    /// Toleranssiarvo, joka määrittää kuinka monta pelin päivitysjaksoa hahmo voi
    /// olla irti tasosta ennen kuin se lasketaan ilmassa olevaksi.
    /// </summary>
    public int PlatformTolerance
    {
        get { return _tolerance; }
        set { _tolerance = value; }
    }

    /// <summary>
    /// Hahmon rintamasuunta (vasen tai oikea).
    /// </summary>
    public Direction FacingDirection
    {
        get { return _curDirection; }
        set { Turn( value ); }
    }

    /// <summary>
    /// Hahmon kiihtyvyys.
    /// </summary>
    public double Acceleration
    {
        get { return _accel; }
        set { _accel = value; }
    }

    /// <summary>
    /// Hahmon maksiminopeus.
    /// </summary>
    public double MaxVelocity
    {
        get { return _maxVel; }
        set { _maxVel = value; }
    }

#if VISUALIZE
    private GameObject ssurface = new GameObject( 1, 2, Shape.Rectangle ) { Color = Color.Blue };
    private PhysicsObject _platform = null;

    /// <summary>
    /// Fysiikkaolio jonka päällä seisotaan.
    /// </summary>
    public PhysicsObject Platform
    {
        get { return _platform; }
        set
        {
            if ( value != null )
            {
                Vector distNormal = this.Position - value.Position;
                double sideWidth = Math.Sqrt( Width * Width + Height * Height );
                ssurface.Position = this.Position - Vector.FromLengthAndAngle( sideWidth, distNormal.Angle );
                ssurface.Color = Color.Red;
            }
            else
            {
                ssurface.Color = Color.Blue;
                PlatformNormal = Vector.Zero;
            }

            _platform = value;
        }
    }
#else
    /// <summary>
    /// Fysiikkaolio jonka päällä seisotaan.
    /// </summary>
    public PhysicsObject Platform { get; set; }
#endif

    /// <summary>
    /// Suuntavektori, joka osoittaa tason suuntaan.
    /// Nollavektori, jos ilmassa.
    /// </summary>
    public Vector PlatformNormal { get; private set; }

    /// <summary>
    /// Hahmolla oleva ase.
    /// </summary>
    public Weapon Weapon
    {
        get { return weapon; }
        set
        {
            // First check: same weapon
            if ( weapon == value ) return;

            // Remove the previous weapon if any
            if ( weapon != null )
            {
                // Reset the weapon when removing

                if ( !IsWeaponFacingRight() )
                {
                    weapon.X *= -1;
                    weapon.Animation = weaponFacingRight;
                }

                weapon.Angle = Angle.Zero;
                this.Remove( weapon );
            }

            this.weapon = value;
            if ( value == null )
                return;

            this.Add( value );

            if ( value.Animation != null )
            {
                // Set the weapon animations
                weaponFacingRight = value.Animation;
                weaponFacingLeft = new Animation( Image.Flip( value.Animation.frames ) );

                if ( FacingDirection == Direction.Left )
                {
                    // If facing left, set the weapon to match the direction
                    weapon.X *= -1;
                    weapon.Angle = Angle.StraightAngle;
                    weapon.Animation = weaponFacingLeft;
                }
            }
        }
    }

    /// <summary>
    /// Jos <c>false</c>, hahmoa ei voi liikuttaa kun se on ilmassa.
    /// </summary>
    public bool CanMoveOnAir { get; set; }

    /// <summary>
    /// Kävelyanimaatio vasemmalle.
    /// </summary>
    public Animation LeftWalkingAnimation
    {
        get { return _leftWalkingAnim; }
        set
        {
            _leftWalkingAnim = value;
            if ( value != null && value.FrameCount > 0 && LeftIdleAnimation == null )
                LeftIdleAnimation = value.frames[0];
        }
    }

    /// <summary>
    /// Kävelyanimaatio oikealle.
    /// </summary>
    public Animation RightWalkingAnimation
    {
        get { return _rightWalkingAnim; }
        set
        {
            _rightWalkingAnim = value;
            if ( value != null && value.FrameCount > 0 && RightIdleAnimation == null )
                RightIdleAnimation = value.frames[0];
        }
    }

    /// <summary>
    /// Animaatio, jota käytetään kun hahmo on paikallaan,
    /// kääntyneenä vasemmalle.
    /// </summary>
    public Animation LeftIdleAnimation { get; set; }

    /// <summary>
    /// Animaatio, jota käytetään kun hahmo on paikallaan,
    /// kääntyneenä oikealle.
    /// </summary>
    public Animation RightIdleAnimation { get; set; }

    /// <summary>
    /// Animaatio, jota käytetään kun hahmo on paikallaan.
    /// </summary>
    private Animation IdleAnimation
    {
        get
        {
            return _curDirection == Direction.Left ? LeftIdleAnimation : RightIdleAnimation;
        }
    }

    /// <summary>
    /// Animaatio, jota käytetään kun hahmo kävelee.
    /// </summary>
    private Animation WalkingAnimation
    {
        get
        {
            return _curDirection == Direction.Left ? LeftWalkingAnimation : RightWalkingAnimation;
        }
    }

    /// <summary>
    /// Tapahtuu kun suunta vaihtuu.
    /// </summary>
    public event Action<Direction> DirectionChanged;

    /// <summary>
    /// Luo uuden tasohyppelyhahmon.
    /// </summary>
    public PlatformCharacter2(double width, double height)
        : this(width, height, Shape.Circle)
    {
    }

    /// <summary>
    /// Luo uuden tasohyppelyhahmon.
    /// </summary>
    public PlatformCharacter2(double width, double height, Shape shape)
        : base(width, height, shape /*, CollisionShapeQuality.FromValue(0.7)*/ )
    {
        KineticFriction = 0.0;
        Restitution = 0.2;
        CanRotate = false;
        CanMoveOnAir = true;

        // This avoids high speeds, particularly when falling. This then avoids
        // going through objects.
        LinearDamping = 0.96;

        AddedToGame += AddCollisionHandler;
        IsUpdated = true;
    }

    private void updateGravity()
    {
        PhysicsGame physGame = Game.Instance as PhysicsGame;
        if ( physGame == null ) return;

        if ( physGame.Gravity != _cachedGravity )
        {
            _cachedGravity = physGame.Gravity;
            _cachedGravityMagnitude = _cachedGravity.Magnitude;
            _cachedGravityNormal = _cachedGravity / _cachedGravityMagnitude;

#if VISUALIZE
            ssurface.Angle = gravityNormal.LeftNormal.Angle;
#endif
        }
    }

    private void AddCollisionHandler()
    {
        if ( !( Game is PhysicsGame ) )
            return;

        this.Body.Colliding += delegate( object sender, CollisionEventArgs args )
        {
            PhysicsObject other = (PhysicsObject)args.Other.Tag;
            Vector normal = new Vector( args.Contact.Points[0].Normal.X, args.Contact.Points[0].Normal.Y );
            OnColliding( this, other, normal );
        };

        #if VISUALIZE
        ssurface.Width = 2 * Math.Sqrt( Game.Level.Width * Game.Level.Width + Game.Level.Height * Game.Level.Height );
        Game.Instance.Add( ssurface );
        #endif
    }

    /// <summary>
    /// Kävelee.
    /// </summary>
    public void Walk()
    {
        walkSteps++;
    }

    /// <summary>
    /// Kävelee tiettyyn suuntaan.
    /// </summary>
    /// <param name="direction">Rintamasuunta. Direction.Left tai Direction.Right</param>
    public void Walk( Direction direction )
    {
        Turn( direction );
        walkSteps++;
    }

    /// <summary>
    /// Pysähtyy ja kääntyy vasemmalle tai oikealle.
    /// </summary>
    /// <param name="direction">Rintamasuunta. Direction.Left tai Direction.Right</param>
    public void Turn( Direction direction )
    {
        if ( direction == _curDirection || direction == Direction.Up || direction == Direction.Down )
            return;

        walkSteps = 0;
        _curDirection = direction;
        Animation = IdleAnimation;
        if (Animation != null) Animation.Resume();
        if ( Weapon != null )
        {
            Weapon.Angle = Angle.Supplement( weapon.Angle );
            weapon.X *= -1;
            weapon.Animation = IsWeaponFacingRight() ? weaponFacingRight : weaponFacingLeft;
        }

        if ( DirectionChanged != null ) DirectionChanged( direction );
    }

    /// <summary>
    /// Pysähtyy.
    /// </summary>
    public void StopWalking()
    {
        walkSteps = 0;
        Animation = IdleAnimation;
        Animation.Resume();
    }

    private bool IsWeaponFacingRight()
    {
        return (-Math.PI / 2) < Weapon.Angle.Radians
            && Weapon.Angle.Radians < (Math.PI / 2);
    }

    /// <summary>
    /// Hyppää tietyllä nopeudella, jos hahmo seisoo tukevalla pohjalla.
    /// </summary>
    /// <param name="speed">Lähtönopeus.</param>
    public void Jump( double speed )
    {
        if ( Platform == null ) return;
        Platform = null;
        this.Hit( Mass * speed * -gravityNormal );
    }

    private void OnColliding( PhysicsObject collisionHelperObject, PhysicsObject target, Vector normal )
    {
        if ( target.IgnoresCollisionResponse )
            return;

        double dot = Vector.DotProduct( normal, gravityNormal );
        if ( Math.Abs( dot ) < 0.5 )
            return;

        if ( Platform != null && Platform != target )
            return;

        if ( Platform == null )
            Platform = target;
            
        collisionDetected = true;
        noCollisionCount = 0;
        PlatformNormal = normal;
    }

    /// <summary>
    /// Ajetaan kun pelitilannetta päivitetään. Päivityksen voi toteuttaa omassa luokassa toteuttamalla tämän
    /// metodin. Perityn luokan metodissa tulee kutsua kantaluokan metodia.
    /// </summary>
    /// <param name="time">Peliaika.</param>
    [EditorBrowsable(EditorBrowsableState.Never)]
    public override void Update(Time time)
    {
        if ( Platform != null && !collisionDetected )
        {
            if ( ++noCollisionCount > PlatformTolerance )
            {
                Platform = null;
            }
        }

        collisionDetected = false;

        if ( walkSteps > 0 )
        {
            // Walking
            Animation = WalkingAnimation;
            if (Animation != null) Animation.Resume();

            double impulse = Mass * Acceleration * time.SinceLastUpdate.TotalSeconds;
            Vector unitX = gravityMagnitude > 0 ? gravityNormal.LeftNormal : Vector.UnitX;

            if ( _curDirection == Direction.Left && -Velocity.X < MaxVelocity )
                //this.Push( Mass * Acceleration * -Vector.UnitX );
                this.Push( Mass * Acceleration * -unitX );
            else if ( _curDirection == Direction.Right && Velocity.X < MaxVelocity )
                //this.Push( Mass * Acceleration * Vector.UnitX );
                this.Push( Mass * Acceleration * unitX );

            walkSteps--;
        }
        else
        {
            // Not walking
            Animation = IdleAnimation;
            if ( Animation != null ) Animation.Resume();
        }

        base.Update(time);
    }

    /// <summary>
    /// Siirtää oliota.
    /// </summary>
    /// <param name="movement">Vektori, joka määrittää kuinka paljon siirretään.</param>
    public override void Move( Vector movement )
    {
        if ( movement.X > 0 ) Walk( Direction.Right );
        else if ( movement.X < 0 ) Walk( Direction.Left );
        if ( movement.Y > 0 ) Jump( movement.Y );
    }
}