﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Jypeli.Assets
{
    /// <summary>
    /// Laittaa omistajansa seuraamaan oliota, jolla on tietty Tag-arvo.
    /// Jos olioita on useita, lähdetään seuraamaan lähintä.
    /// </summary>
    public class TagFollowerBrain : AbstractMoverBrain
    {
        private double _followDist = double.PositiveInfinity;
        private double directionChangeInterval = 1.0;
        private double timeToDirectionChange = 0;
        private Vector currentStep = Vector.Zero;

        /// <summary>
        /// Tag-arvo, jota seurata.
        /// </summary>
        public string TagToFollow { get; set; }

        /// <summary>
        /// Tämänhetkinen seurattava.
        /// </summary>
        public GameObject CurrentTarget { get; set; }

        /// <summary>
        /// Etäisyys kohteeseen. Jos tämä tai alle, niin jolloin aletaan seuraamaan kohdetta.
        /// </summary>
        public double TargetFollowDistance
        {
            get { return _followDist; }
            set { _followDist = value; }
        }

        /// <summary>
        /// Etäisyys, jolloin ollaan lähellä kohdetta.
        /// </summary>
        /// <value>Etäisyys.</value>
        public double TargetCloseDistance { get; set; }

        /// <summary>
        /// Haahuillaanko ympäriinsä jos ei ole tarpeeksi lähellä olevaa kohdetta.
        /// </summary>
        public bool WanderIfNotClose { get; set; }

        /// <summary>
        /// Aika sekunteina, jonka kuluttua muutetaan liikesuuntaa harhaillessa
        /// </summary>
        public double WanderChangeMovementSeconds
        {
            get { return directionChangeInterval; }
            set { directionChangeInterval = value; timeToDirectionChange = value; }
        }

#if VISUALIZE
        GameObject distanceMarker;
        GameObject targetMarker;
#endif

        /// <summary>
        /// Alustaa uudet aivot.
        /// </summary>
        /// <param name="tag">Tag-arvo</param>
        /// <param name="speed">Nopeus</param>
        public TagFollowerBrain(string tag, double speed)
            : base(speed)
        {
            TagToFollow = tag;
            TargetFollowDistance = Double.PositiveInfinity;
            TargetCloseDistance = 0;
#if VISUALIZE
            distanceMarker = new GameObject( 1.0, 1.0, Shape.Circle );
            distanceMarker.Color = new Color( 80, 80, 80, 100 );
            Game.Instance.Add( distanceMarker );

            targetMarker = new GameObject( 1.0, 2.0, Shape.Rectangle );
            targetMarker.Color = Color.Transparent;
            Game.Instance.Add( targetMarker );
#endif
        }

        protected override void Update( Time time )
        {
            if ( Owner == null || TagToFollow == null ) return;

#if VISUALIZE
            if ( CurrentTarget != null && CurrentTarget.IsDestroyed )
            {
                Game.Instance.Remove( distanceMarker );
                Game.Instance.Remove( targetMarker );
            }
#endif

            if ( CurrentTarget != null && CurrentTarget.IsDestroyed )
                CurrentTarget = null;

#if VISUALIZE
            distanceMarker.Width = distanceMarker.Height = 2 * TargetFollowDistance;
            distanceMarker.Position = Owner.Position;

            if ( CurrentTarget != null )
            {
                targetMarker.Color = new Color( 255, 0, 0, 200 );
                targetMarker.Width = CurrentTarget.Width;
                targetMarker.X = CurrentTarget.X;
                targetMarker.Top = CurrentTarget.Bottom;
            }
            else
            {
                targetMarker.Color = Color.Transparent;
            }
#endif

            double currentDistance = CurrentTarget != null ? Vector.Distance( Owner.Position, CurrentTarget.Position ) : double.PositiveInfinity;

            if ( currentDistance < TargetCloseDistance )
            {
                // Disengage, we are too close
                CurrentTarget = null;
                currentDistance = double.PositiveInfinity;
            }

            for ( int i = 0; i < Game.allGameObjects.Count; i++ )
            {
                object tag = Game.allGameObjects[i].Tag;

                if ( tag != null && tag.ToString() == TagToFollow )
                {
                    if ( Game.allGameObjects[i] == null || Game.allGameObjects[i].IsDestroyed) 
                        continue;

                    double newDistance = Vector.Distance( Owner.Position, Game.allGameObjects[i].Position );

                    if ( newDistance < currentDistance && newDistance > TargetCloseDistance )
                    {
                        CurrentTarget = Game.allGameObjects[i];
                        currentDistance = newDistance;
                    }
                }
            }

            double t = time.SinceLastUpdate.TotalSeconds;

            if ( CurrentTarget == null || currentDistance > TargetFollowDistance )
            {
                if ( !WanderIfNotClose ) return;

                timeToDirectionChange -= t;
                if ( timeToDirectionChange <= 0 )
                {
                    currentStep = Vector.FromLengthAndAngle( Speed, RandomGen.NextAngle() );
                    timeToDirectionChange += directionChangeInterval;
                }

                Owner.Move( currentStep );
                return;
            }

            Vector d = CurrentTarget.Position - Owner.Position;
            Owner.Move( Vector.FromLengthAndAngle( Speed * time.SinceLastUpdate.TotalSeconds, d.Angle ) );
            
            base.Update( time );
        }
    }
}
