﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using System.ComponentModel;
using Physics2DDotNet;
using Physics2DDotNet.Ignorers;


namespace Jypeli.Assets
{
    /// <summary>
    /// Yksinkertainen tykki, joka ampuu kuulia tai muuta ammuksia.
    /// </summary>
    public class Cannon : Weapon
    {       
        /// <summary>
        /// Alustaa uuden tykin.
        /// </summary>
        public Cannon( double width, double height )
            : base( width, height )
        {
            Power.MaxValue = 15000;
            Ammo.DefaultValue = 100;
            TimeBetweenUse = TimeSpan.FromSeconds( 1 );

            Image = Game.LoadImageFromResources( "Cannon" );
            AttackSound = Game.LoadSoundEffectFromResources( "CannonFire" );
        }

        protected override PhysicsObject CreateProjectile()
        {
            return new CannonBall( 5 );
        }
    }

    /// <summary>
    /// Rynnäkkökivääri.
    /// </summary>
    public class AssaultRifle : Weapon
    {
        [Obsolete( "Use ProjectileCollision" )]
        public CollisionHandler BulletCollision
        {
            get { return ProjectileCollision; }
            set { ProjectileCollision = value; }
        }

        /// <summary>
        /// Alustaa uuden rynnäkkökiväärin.
        /// </summary>
        public AssaultRifle( double width, double height )
            : base( width, height )
        {
            AmmoIgnoresGravity = true;
            Ammo.DefaultValue = 90;
            Power.DefaultValue = 200;
            TimeBetweenUse = TimeSpan.FromSeconds( 0.2 );

            Image = Game.LoadImageFromResources( "AsRifle" );
            AttackSound = Game.LoadSoundEffectFromResources( "MGAttack" );
        }

        protected override PhysicsObject CreateProjectile()
        {
            return new Bullet( 5 );
        }
    }

    /// <summary>
    /// Plasmatykki.
    /// </summary>
    public class PlasmaCannon : Weapon
    {
        [Obsolete( "Use ProjectileCollision" )]
        public CollisionHandler PlasmaParticleCollision
        {
            get { return ProjectileCollision; }
            set { ProjectileCollision = value; }
        }

        /// <summary>
        /// Alustaa uuden plasmakiväärin.
        /// </summary>
        public PlasmaCannon( double width, double height )
            : base( width, height )
        {
            Power.DefaultValue = 6000;
            AmmoIgnoresGravity = true;
            AmmoIgnoresExplosions = true;
            MaxAmmoLifetime = TimeSpan.FromSeconds( 0.7 );

            Image = Game.LoadImageFromResources( "PlasmaCannon" );
            AttackSound = Game.LoadSoundEffectFromResources( "PlasmaAttack" );
            TimeBetweenUse = TimeSpan.FromSeconds( 0.1 );
        }

        protected override PhysicsObject  CreateProjectile()
        {
            return new Projectile( 3, 5, Color.Green );
        }
    }

    /// <summary>
    /// Lasertykki
    /// </summary>
    public class LaserGun : Weapon
    {
        [Obsolete( "Use ProjectileCollision" )]
        public CollisionHandler LaserCollision
        {
            get { return ProjectileCollision; }
            set { ProjectileCollision = value; }
        }

        /// <summary>
        /// Alustaa uuden laser-tykin.
        /// </summary>
        public LaserGun( double width, double height )
            : base( width, height )
        {
            Power.DefaultValue = 500;
            AmmoIgnoresGravity = true;
            AmmoIgnoresExplosions = true;
            //MaxAmmoLifetime = TimeSpan.FromSeconds( 0.7 );

            Image = Game.LoadImageFromResources( "PlasmaCannon" );
            AttackSound = Game.LoadSoundEffectFromResources( "laser" );
            TimeBetweenUse = TimeSpan.FromSeconds( 0 );
        }

        protected override PhysicsObject CreateProjectile()
        {
            //return new Projectile( 10, 1, 5, Color.Red );
            PhysicsObject beam = new PhysicsObject( new RaySegment( Vector.Zero, Vector.UnitX, 10 ) );
            beam.Color = Color.Red;
            return beam;
        }
    }    

    /// <summary>
    /// Tykinkuula.
    /// </summary>
    public class CannonBall : Projectile
    {
        /// <summary>
        /// Alustaa uuden tykinkuulan.
        /// </summary>
        public CannonBall( double radius )
            : base( radius, 20, "CannonBall" )
        {
        }
    }

    /// <summary>
    /// Luoti.
    /// </summary>
    public class Bullet : Projectile
    {
        /// <summary>
        /// Alustaa uuden luodin.
        /// </summary>
        public Bullet( double radius )
            : base( radius, 0.2, "Bullet" )
        {
        }
    }

    /// <summary>
    /// Kranaatti.
    /// </summary>
    public class Grenade : Projectile
    {
        /// <summary>
        /// Räjähdyksen säde.
        /// </summary>
        public double ExplosionRadius { get; set; }

        /// <summary>
        /// Räjähdyksen nopeus.
        /// </summary>
        public double ExplosionSpeed { get; set; }

        /// <summary>
        /// Räjähdyksen voima.
        /// </summary>
        public double ExplosionForce { get; set; }

        /// <summary>
        /// Aika, jonka päästä ammus räjähtää itsestään.
        /// </summary>
        public TimeSpan FuseTime { get; set; }

        /// <summary>
        /// Luo uuden kranaatin, joka räjähtää kolmen sekunnin päästä.
        /// </summary>
        /// <param name="radius"></param>
        public Grenade( double radius )
            : this( radius, TimeSpan.FromSeconds( 3 ) )
        {
        }

        /// <summary>
        /// Luo uuden kranaatin.
        /// </summary>
        /// <param name="radius">Säde.</param>
        /// <param name="fuseTime">Kuinka kauan kestää ennen räjähdystä.</param>
        public Grenade( double radius, TimeSpan fuseTime )
            : base( radius, 20, "Grenade" )
        {
            FuseTime = fuseTime;
            IsUpdated = true;
            ExplosionRadius = 150;
            ExplosionSpeed = 150;
            ExplosionForce = 1000;
        }

        /// <summary>
        /// Räjäytä kranaatti.
        /// </summary>
        public virtual void Explode()
        {
            if ( IsAddedToGame )
            {
                this.Destroy();
                var explosion = new Explosion( ExplosionRadius );
                explosion.Position = this.Position;
                explosion.Speed = ExplosionSpeed;
                explosion.Force = ExplosionForce;
                Game.Add( explosion );
            }
        }

        [EditorBrowsable( EditorBrowsableState.Never )]
        public override void Update( Time time )
        {
            if ( ( time.SinceStartOfGame - CreationTime ) > FuseTime && IsAddedToGame )
            {
                // Boom!
                Explode();
            }

            base.Update( time );
        }
    }

    /// <summary>
    /// Rypälepommi. Hajoaa räjähtäessään pienempiin osiin, jotka
    /// voivat edelleen räjähtää pienempiin osiin.
    /// </summary>
    public class ClusterGrenade : Grenade
    {
        private int clusterlevel = 0;

        /// <summary>
        /// Sirpaleiden määrä, joka räjähdyksestä syntyy.
        /// </summary>
        public int NumberOfClusters { get; set; }

        /// <summary>
        /// Sirpaleiden hajontasuunta.
        /// </summary>
        public Angle ClusterDirection { get; set; }

        /// <summary>
        /// Sirpaleiden hajontakaari.
        /// </summary>
        public double ClusterArc { get; set; }

        /// <summary>
        /// Alustaa uuden rypälepommin.
        /// </summary>
        /// <param name="radius">Säde.</param>
        /// <param name="cl">Kuinka monta kertaa rypäleet hajoavat edelleen. Kuitenkin vähintään yhden kerran.</param>
        public ClusterGrenade( double radius, int cl )
            : base( radius )
        {
            Mass = 20;
            Body.CollisionIgnorer = new ObjectIgnorer();
            NumberOfClusters = 3;
            ClusterDirection = Angle.FromRadians( Math.PI / 2 );
            ClusterArc = Math.PI;
            clusterlevel = cl - 1;
        }

        /// <summary>
        /// Räjäyttää kranaatin sirpaleiksi.
        /// </summary>
        public override void Explode()
        {
            if ( !IsAddedToGame ) return;

            Vector posOffset;
            double direction;

            base.Explode();

            for ( int i = 0; i < NumberOfClusters; i++ )
            {
                double currentRadius = Width / 2;
                double r = currentRadius * 0.6;
                // TODO: The mass could be evenly distributed to the clusters and the main projectile.
                Grenade g = ( clusterlevel > 0 ) ?
                    new ClusterGrenade( r, clusterlevel - 1 ) : new Grenade( r );

                direction = ClusterDirection.Radians - ( ( i - NumberOfClusters / 2 ) * ClusterArc / NumberOfClusters );
                posOffset = Vector.FromLengthAndAngle( 2 + this.Size.Magnitude, Angle.FromRadians( direction ) );

                g.Position = this.Position + posOffset;
                g.Mass = this.Mass / NumberOfClusters;
                g.Image = this.Image;
                g.Color = this.Color;
                g.FuseTime = this.FuseTime;
                g.ExplosionForce = this.ExplosionForce / 2;
                g.ExplosionRadius = this.ExplosionRadius / 2;
                g.ExplosionSpeed = this.ExplosionSpeed / 2;
                g.Body.CollisionIgnorer = this.Body.CollisionIgnorer;
                Game.Add( g );
            }
        }
    }
}
