﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using Microsoft.Xna.Framework;
using AdvanceMath;
using Microsoft.Xna.Framework.Graphics;

namespace Jypeli
{
    /// <summary>
    /// Kamera. Määrittää mikä osa pelitasosta on kerralla näkyvissä.
    /// </summary>
    [Save]
    public class Camera
    {
        [Save] internal bool _stayInLevel = false;
        [Save] internal double _zoomFactor = 1.0;

        /// <summary>
        /// Kameran sijainti.
        /// </summary>
        [Save] public Vector Position = Vector.Zero;

        /// <summary>
        /// Kameran paikan X-koordinaatti.
        /// </summary>
        public double X
        {
            get
            {
                return Position.X;
            }
            set
            {
                Position = new Vector(value, Position.Y);
            }
        }

        /// <summary>
        /// Kameran paikan Y-koordinaatti.
        /// </summary>
        public double Y
        {
            get
            {
                return Position.Y;
            }
            set
            {
                Position = new Vector(Position.X, value);
            }
        }

        /// <summary>
        /// Kameran zoomauskerroin.
        /// Oletuksena 1.0. Mitä suurempi zoomauskerroin, sitä lähempänä kamera on (esim 2.0 on 2 x lähempänä) ja toisinpäin.
        /// </summary>
        public double ZoomFactor
        {
            get { return _zoomFactor; }
            set { _zoomFactor = value; }
        }

        /// <summary>
        /// Jos tosi, kamera ei mene koskaan kentän ulkopuolelle.
        /// </summary>
        public bool StayInLevel
        {
            get { return _stayInLevel; }
            set { _stayInLevel = value; }
        }

        /// <summary>
        /// Olio, jota kamera seuraa. Jos <c>null</c>, mitään oliota ei seurata.
        /// </summary>
        public GameObject FollowedObject { get; set; }

        /// <summary>
        /// Luo uuden kameran.
        /// </summary>
        internal Camera()
        {
        }

        /// <summary>
        /// Muuntaa annetun pisteen ruutukoordinaateista maailmankoordinaatteihin.
        /// </summary>
        public Vector ScreenToWorld(Vector point)
        {
            return Position + (1 / ZoomFactor) * point;
        }

        /// <summary>
        /// Muuntaa annetun pisteen maailmankoordinaateista ruutukoordinaatteihin.
        /// </summary>
        /// <param name="point"></param>
        /// <returns></returns>
        public Vector WorldToScreen(Vector point)
        {
            return (point - Position) * ZoomFactor;
        }

        /// <summary>
        /// Kohdistaa kameran.
        /// </summary>
        /// <param name="v">Koordinaatti vektorina johon kamera kohdistetaan.</param>
        public void Move( Vector v )
        {
            Position += new Vector( v.X / ZoomFactor, v.Y / ZoomFactor );
        }

        /// <summary>
        /// Zoomaa.
        /// </summary>
        /// <param name="zoom">
        /// Zoomauskerroin. Ykköstä suurempi (esim. 1.5) lähentää ja
        /// ykköstä pienempi (esim. 0.5) zoomaa kauemmas.
        /// </param>
        public void Zoom( double zoom )
        {
            ZoomFactor *= zoom;
        }

        /// <summary>
        /// Resetoi kameran (keskittää ja laittaa zoomin oletusarvoon).
        /// </summary>
        public void Reset()
        {
            Position = Vector.Zero;
            ZoomFactor = 1.0f;
        }

        /// <summary>
        /// Seuraa jotakin peliobjektia.
        /// </summary>
        /// <param name="gameobject">Seurattava peliobjekti.</param>
        public void Follow( GameObject gameobject )
        {
            FollowedObject = gameobject;
        }

        /// <summary>
        /// Zoomaa ja keskittää kameran siten, että koko kenttä on näkyvissä kerralla.
        /// </summary>
        public void ZoomToLevel()
        {
            ZoomToLevel( 0 );
        }

        /// <summary>
        /// Zoomaa ja keskittää kameran siten, että koko kenttä on näkyvissä kerralla. Tällöin kamera ei seuraa mitään oliota.
        /// </summary>
        /// <param name="bordersize">Reunalle jätettävä tila (jos negatiivinen, niin osa kentästä jää piiloon).</param>
        public void ZoomToLevel( double borderSize )
        {
            FollowedObject = null;
            Position = Vector.Zero;
            Level level = Game.Instance.Level;

            double screenWidth = (double)Game.Screen.Width;
            double screenHeight = (double)Game.Screen.Height;

            if ( ( level.Width / level.Height ) >= ( screenWidth / screenHeight ) )
                this.ZoomFactor = screenWidth / ( level.Width + borderSize );
            else
                this.ZoomFactor = screenHeight / ( level.Height + borderSize );
        }

        /// <summary>
        /// Ajetaan kun pelitilannetta päivitetään.
        /// </summary>
        /// <param name="gameTime">Peliaika</param>
        internal void Update( Time time )
        {
            if ( FollowedObject != null )
            {
                Position = FollowedObject.Position;
            }

            if ( StayInLevel )
            {
                double screenWidth = (double)Game.Screen.Width;
                double screenHeight = (double)Game.Screen.Height;
                Level level = Game.Instance.Level;

                double screenAspectRatio = screenWidth / screenHeight;
                double levelAspectRatio = level.Width / level.Height;

                double zoomedWidth = level.Width * ZoomFactor;
                double zoomedHeight = level.Height * ZoomFactor;

                double viewAreaWidth = screenWidth / ZoomFactor;
                double viewAreaHeight = screenHeight / ZoomFactor;

                if ( zoomedWidth < screenWidth || zoomedHeight < screenHeight )
                {
                    ZoomFactor = Math.Max( screenWidth / level.Width, screenHeight / level.Height );
                }

                if ( ( Position.X - ( viewAreaWidth / 2 ) ) < level.Left )
                {
                    Position.X = level.Left + ( viewAreaWidth / 2 );
                }
                else if ( Position.X + ( viewAreaWidth / 2 ) > level.Right )
                {
                    Position.X = level.Right - ( viewAreaWidth / 2 );
                }

                if ( Position.Y - ( viewAreaHeight / 2 ) < level.Bottom )
                {
                    Position.Y = level.Bottom + ( viewAreaHeight / 2 );
                }
                else if ( Position.Y + ( viewAreaHeight / 2 ) > level.Top )
                {
                    Position.Y = level.Top - ( viewAreaHeight / 2 );
                }
            }
        }
    }
}
