﻿using System;

using XnaColor = Microsoft.Xna.Framework.Color;

namespace Jypeli
{
    /// <summary>
    /// Väri.
    /// </summary>
    [Save]
    public struct Color
    {
        [Save] public byte RedComponent;
        [Save] public byte GreenComponent;
        [Save] public byte BlueComponent;
        [Save] public byte AlphaComponent;

        internal Color( XnaColor c )
        {
            RedComponent = c.R;
            GreenComponent = c.G;
            BlueComponent = c.B;
            AlphaComponent = c.A;
        }

        internal XnaColor AsXnaColor()
        {
            return new XnaColor(
                RedComponent, GreenComponent, BlueComponent, AlphaComponent );
        }

        public Color( byte red, byte green, byte blue )
            : this( red, green, blue, byte.MaxValue )
        {
        }

        public Color( byte red, byte green, byte blue, byte alpha )
        {
            RedComponent = red;
            GreenComponent = green;
            BlueComponent = blue;
            AlphaComponent = alpha;
        }

        public Color( Color rgb, byte alpha )
            : this( rgb.RedComponent, rgb.GreenComponent, rgb.BlueComponent, alpha )
        {
        }

        public Color( double red, double green, double blue )
            : this( red, green, blue, 1.0 )
        {
        }

        public Color( double red, double green, double blue, double alpha )
        {
            var xnaColor = new XnaColor(
                (float)red,
                (float)green,
                (float)blue,
                (float)alpha );
            RedComponent = xnaColor.R;
            GreenComponent = xnaColor.G;
            BlueComponent = xnaColor.B;
            AlphaComponent = xnaColor.A;
        }

        public static bool operator ==( Color c1, Color c2 )
        {
            return c1.RedComponent == c2.RedComponent
                && c1.GreenComponent == c2.GreenComponent
                && c1.BlueComponent == c2.BlueComponent
                && c1.AlphaComponent == c2.AlphaComponent;
        }

        public static bool operator !=( Color c1, Color c2 )
        {
            return !( c1 == c2 );
        }

        public static Color Lerp( Color value1, Color value2, double amount )
        {
            return new Color( XnaColor.Lerp(
                value1.AsXnaColor(),
                value2.AsXnaColor(),
                (float)amount ) );
        }

        /// <summary>
        /// Antaa tummemman värin. Vähentaa jokaista kolmea osaväriä arvon <c>howMuch</c>
        /// verran.
        /// </summary>
        /// <param name="c">Alkuperäinen väri.</param>
        /// <param name="howMuch">Kuinka paljon tummempi (positiivinen luku).</param>
        /// <returns>Tummempi väri.</returns>
        public static Color Darker( Color c, int howMuch )
        {
            int r = (int)c.RedComponent - howMuch;
            int g = (int)c.GreenComponent - howMuch;
            int b = (int)c.BlueComponent - howMuch;
            if ( r < 0 ) r = 0;
            if ( g < 0 ) g = 0;
            if ( b < 0 ) b = 0;
            Microsoft.Xna.Framework.Color x = c.AsXnaColor();
            return new Color( (byte)r, (byte)g, (byte)b, c.AlphaComponent );
        }

        /// <summary>
        /// Antaa kirkkaamman värin. Kasvattaa jokaista kolmea osaväriä arvon <c>howMuch</c>
        /// verran.
        /// </summary>
        /// <param name="c">Alkuperäinen väri.</param>
        /// <param name="howMuch">Kuinka paljon vaaleampi.</param>
        /// <returns>Vaaleampi väri.</returns>
        public static Color Lighter( Color c, int howMuch )
        {
            int r = (int)c.RedComponent + howMuch;
            int g = (int)c.GreenComponent + howMuch;
            int b = (int)c.BlueComponent + howMuch;
            if ( r > byte.MaxValue ) r = byte.MaxValue;
            if ( g > byte.MaxValue ) g = byte.MaxValue;
            if ( b > byte.MaxValue ) b = byte.MaxValue;
            Microsoft.Xna.Framework.Color x = c.AsXnaColor();
            return new Color( (byte)r, (byte)g, (byte)b, c.AlphaComponent );
        }

        public static readonly Color Aqua = new Color( 0, 255, 255, 255 );

        public static readonly Color Beige = new Color( 245, 245, 220, 255 );

        public static readonly Color Black = new Color( 0, 0, 0, 255 );

        public static readonly Color Blue = new Color( 0, 0, 255, 255 );

        public static readonly Color Brown = new Color( 165, 42, 42, 255 );

        public static readonly Color Cyan = new Color( 0, 255, 255, 255 );

        public static readonly Color DarkBlue = new Color( 0, 0, 139, 255 );

        public static readonly Color DarkCyan = new Color( 0, 139, 139, 255 );

        public static readonly Color DarkGray = new Color( 169, 169, 169, 255 );

        public static readonly Color DarkGreen = new Color( 0, 100, 0, 255 );

        public static readonly Color DarkMagenta = new Color( 139, 0, 139, 255 );

        public static readonly Color DarkOrange = new Color( 255, 140, 0, 255 );

        public static readonly Color DarkRed = new Color( 139, 0, 0, 255 );

        public static readonly Color DarkTurquoise = new Color( 0, 206, 209, 255 );

        public static readonly Color DarkViolet = new Color( 148, 0, 211, 255 );

        public static readonly Color ForestGreen = new Color( 34, 139, 34, 255 );

        public static readonly Color Fuchsia = new Color( 255, 0, 255, 255 );

        public static readonly Color Gray = new Color( 128, 128, 128, 255 );

        public static readonly Color Green = new Color( 0, 128, 0, 255 );

        public static readonly Color GreenYellow = new Color( 173, 255, 47, 255 );

        public static readonly Color LightBlue = new Color( 173, 216, 230, 255 );

        public static readonly Color LightCyan = new Color( 224, 255, 255, 255 );

        public static readonly Color LightGray = new Color( 211, 211, 211, 255 );

        public static readonly Color LightGreen = new Color( 144, 238, 144, 255 );

        public static readonly Color LightPink = new Color( 255, 182, 193, 255 );

        public static readonly Color LightYellow = new Color( 255, 255, 224, 255 );

        public static readonly Color Lime = new Color( 0, 255, 0, 255 );

        public static readonly Color LimeGreen = new Color( 50, 205, 50, 255 );

        public static readonly Color Magenta = new Color( 255, 0, 255, 255 );

        public static readonly Color Maroon = new Color( 128, 0, 0, 255 );

        public static readonly Color MediumBlue = new Color( 0, 0, 205, 255 );

        public static readonly Color MediumPurple = new Color( 147, 112, 219, 255 );

        public static readonly Color MediumTurquoise = new Color( 72, 209, 204, 255 );

        public static readonly Color MediumVioletRed = new Color( 199, 21, 133, 255 );

        public static readonly Color Orange = new Color( 255, 165, 0, 255 );

        public static readonly Color OrangeRed = new Color( 255, 69, 0, 255 );

        public static readonly Color Pink = new Color( 255, 192, 203, 255 );

        public static readonly Color Purple = new Color( 128, 0, 128, 255 );

        public static readonly Color Red = new Color( 255, 0, 0, 255 );

        public static readonly Color SkyBlue = new Color( 135, 206, 235, 255 );

        /// <summary>
        /// Läpinäkyvä väri.
        /// </summary>
        public static readonly Color Transparent = new Color(XnaColor.Transparent);

        public static readonly Color Turquoise = new Color( 64, 224, 208, 255 );

        public static readonly Color White = new Color( 255, 255, 255, 255 );

        public static readonly Color Violet = new Color( 238, 130, 238, 255 );

        public static readonly Color Wheat = new Color( 245, 222, 179, 255 );

        public static readonly Color Yellow = new Color( 255, 255, 0, 255 );

        public static readonly Color YellowGreen = new Color( 154, 205, 50, 255 );
    }
}
