﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using Microsoft.Xna.Framework.Input;

using XnaKeyboard = Microsoft.Xna.Framework.Input.Keyboard;
using System.Collections.Generic;

namespace Jypeli.Controls
{
    /// <summary>
    /// Näppäimistö peliohjaimena.
    /// </summary>
    public class Keyboard : Controller<KeyboardState>
    {
#if WINDOWS
        internal static KeyboardTextBuffer Buffer = new KeyboardTextBuffer();
#endif

        internal Keyboard()
        {
            oldState = XnaKeyboard.GetState();
        }

        private static Keys ToXnaKeys( Key k )
        {
            return (Keys)k;
        }

        protected override bool IsTriggered( Listener listener )
        {
            if ( listener.Type == ListeningType.KeyboardKey )
            {
                return KeyTriggered( listener );
            }

            if ( listener.Type == ListeningType.KeyboardAll )
            {
                List<Key> keys = KeysTriggered( listener.State );
                return keys.Count > 0;
            }

            return base.IsTriggered( listener );
        }

        private bool KeyTriggered( Listener listener )
        {
            Keys k = ToXnaKeys( listener.Key );

            switch ( listener.State )
            {
                case ButtonState.Released:
                    return ( oldState.IsKeyDown( k ) && newState.IsKeyUp( k ) );

                case ButtonState.Pressed:
                    return ( oldState.IsKeyUp( k ) && newState.IsKeyDown( k ) );

                case ButtonState.Up:
                    return ( newState.IsKeyUp( k ) );

                case ButtonState.Down:
                    return ( newState.IsKeyDown( k ) );
            }

            return false;
        }

        private List<Key> KeysTriggered( ButtonState state )
        {
            List<Keys> pressedBefore = new List<Keys>( oldState.GetPressedKeys() );
            List<Keys> pressedNow = new List<Keys>( newState.GetPressedKeys() );
            List<Key> triggered = new List<Key>();

            switch ( state )
            {
                case ButtonState.Released:
                    for ( int i = 0; i < pressedBefore.Count; i++ )
                    {
                        if ( !pressedNow.Contains( pressedBefore[i] ) )
                            triggered.Add( (Key)pressedBefore[i] );
                    }
                    break;

                case ButtonState.Pressed:
                    for ( int i = 0; i < pressedNow.Count; i++ )
                    {
                        if ( !pressedBefore.Contains( pressedNow[i] ) )
                            triggered.Add( (Key)pressedNow[i] );
                    }
                    break;

                case ButtonState.Up:
                    throw new NotImplementedException( "Not implemented" );

                case ButtonState.Down:
                    for ( int i = 0; i < pressedNow.Count; i++ )
                    {
                        triggered.Add( (Key)pressedNow[i] );
                    }
                    break;
            }

            return triggered;
        }

        internal override bool IsBufferEmpty()
        {
            return ( newState.GetPressedKeys().Length == 0 );
        }

        /// <summary>
        /// Ottaa käytöstä poistetun napin <c>k</c> takaisin käyttöön.
        /// </summary>
        public void Enable( Key k )
        {
            base.Enable( listener => ( listener.Key == k ) );
        }

        /// <summary>
        /// Poistaa napin <c>k</c> käytöstä.
        /// </summary>
        public void Disable( Key k )
        {
            base.Disable( listener => ( listener.Key == k ) );
        }

        internal override KeyboardState GetCurrentState()
        {
            return XnaKeyboard.GetState();
        }

        /// <summary>
        /// Palauttaa annetun näppäimen tilan (ks. <c>ButtonState</c>).
        /// </summary>
        /// <param name="k">Näppäin.</param>
        /// <returns>Näppäimen tila</returns>
        public ButtonState GetKeyState( Key k )
        {
            Keys key = ToXnaKeys( k );
            bool down = newState.IsKeyDown( key );
            bool lastdown = oldState.IsKeyDown( key );

            if ( lastdown && down )
                return ButtonState.Down;
            if ( !lastdown && down )
                return ButtonState.Pressed;
            if ( lastdown && !down )
                return ButtonState.Released;

            return ButtonState.Up;
        }

        /// <summary>
        /// Tarkistaa, onko kumpikaan shift-näppäimistä painettuna.
        /// </summary>
        /// <returns>
        /// 	<c>true</c> jos alhaalla, muuten <c>painettuna</c>.
        /// </returns>
        public bool IsShiftDown()
        {
            return newState.IsKeyDown( Keys.LeftShift ) || newState.IsKeyDown( Keys.RightShift );
        }

        /// <summary>
        /// Tarkistaa, onko kumpikaan ctrl-näppäimistä painettuna.
        /// </summary>
        /// <returns>
        /// 	<c>true</c> jos alhaalla, muuten <c>painettuna</c>.
        /// </returns>
        public bool IsCtrlDown()
        {
            return newState.IsKeyDown( Keys.LeftControl ) || newState.IsKeyDown( Keys.RightControl );
        }

        /// <summary>
        /// Tarkistaa, onko kumpikaan alt-näppäimistä painettuna.
        /// </summary>
        /// <returns>
        /// 	<c>true</c> jos alhaalla, muuten <c>painettuna</c>.
        /// </returns>
        public bool IsAltDown()
        {
            return newState.IsKeyDown( Keys.LeftAlt ) || newState.IsKeyDown( Keys.RightAlt );
        }

        internal override string GetControlText( Listener listener )
        {
            if ( listener.Key == Key.Add ) return "NumPad+";
            if ( listener.Key == Key.Subtract ) return "NumPad-";
            if ( listener.Key == Key.Multiply ) return "NumPad*";
            if ( listener.Key == Key.Divide ) return "NumPad/";
            if ( listener.Key == Key.OemQuotes ) return "Ä";
            if ( listener.Key == Key.OemTilde ) return "Ö";

            return listener.Key.ToString();
        }

        public Listener Listen( Key k, ButtonState state, Handler handler, String helpText )
        {
            Listener l = new SimpleListener( this, ListeningType.KeyboardKey, helpText, handler );
            l.Key = k;
            l.State = state;
            Add( l );
            return l;
        }

        public Listener Listen<T1>( Key k, ButtonState state, Handler<T1> handler, String helpText, T1 p1 )
        {
            Listener l = new SimpleListener<T1>( this, ListeningType.KeyboardKey, helpText, handler, p1 );
            l.Key = k;
            l.State = state;
            Add( l );
            return l;
        }

        public Listener Listen<T1, T2>( Key k, ButtonState state, Handler<T1, T2> handler, String helpText, T1 p1, T2 p2 )
        {
            Listener l = new SimpleListener<T1, T2>( this, ListeningType.KeyboardKey, helpText, handler, p1, p2 );
            l.Key = k;
            l.State = state;
            Add( l );
            return l;
        }

        public Listener Listen<T1, T2, T3>( Key k, ButtonState state, Handler<T1, T2, T3> handler, String helpText, T1 p1, T2 p2, T3 p3 )
        {
            Listener l = new SimpleListener<T1, T2, T3>( this, ListeningType.KeyboardKey, helpText, handler, p1, p2, p3 );
            l.Key = k;
            l.State = state;
            Add( l );
            return l;
        }

        public Listener Listen<T1, T2, T3, T4>( Key k, ButtonState state, Handler<T1, T2, T3, T4> handler, String helpText, T1 p1, T2 p2, T3 p3, T4 p4 )
        {
            Listener l = new SimpleListener<T1, T2, T3, T4>( this, ListeningType.KeyboardKey, helpText, handler, p1, p2, p3, p4 );
            l.Key = k;
            l.State = state;
            Add( l );
            return l;
        }

        public Listener ListenAll( ButtonState state, MultiKeyHandler handler, String helpText )
        {
            Listener l = new MultiKeyListener( this, helpText, handler );
            l.State = state;
            Add( l );
            return l;
        }

        public Listener ListenAll<T1>( ButtonState state, MultiKeyHandler<T1> handler, String helpText, T1 p1 )
        {
            Listener l = new MultiKeyListener<T1>( this, helpText, handler, p1 );
            l.State = state;
            Add( l );
            return l;
        }

        public Listener ListenAll<T1, T2>( ButtonState state, MultiKeyHandler<T1, T2> handler, String helpText, T1 p1, T2 p2 )
        {
            Listener l = new MultiKeyListener<T1, T2>( this, helpText, handler, p1, p2 );
            l.State = state;
            Add( l );
            return l;
        }

        public Listener ListenAll<T1, T2, T3>( ButtonState state, MultiKeyHandler<T1, T2, T3> handler, String helpText, T1 p1, T2 p2, T3 p3 )
        {
            Listener l = new MultiKeyListener<T1, T2, T3>( this, helpText, handler, p1, p2, p3 );
            l.State = state;
            Add( l );
            return l;
        }

        public Listener ListenAll<T1, T2, T3, T4>( ButtonState state, MultiKeyHandler<T1, T2, T3, T4> handler, String helpText, T1 p1, T2 p2, T3 p3, T4 p4 )
        {
            Listener l = new MultiKeyListener<T1, T2, T3, T4>( this, helpText, handler, p1, p2, p3, p4 );
            l.State = state;
            Add( l );
            return l;
        }
    }
}
