﻿using System;
using System.Diagnostics;
using Microsoft.Xna.Framework.Graphics;
using System.Text;

using XnaRectangle = Microsoft.Xna.Framework.Rectangle;
using XnaV2 = Microsoft.Xna.Framework.Vector2;
using XnaColor = Microsoft.Xna.Framework.Color;

namespace Jypeli
{
    internal enum ContentSource
    {
        GameContent,
        ResourceContent,
    }


    /// <summary>
    /// Fontti.
    /// </summary>
    public class Font
    {
        /// <summary>
        /// OletusFontti.
        /// </summary>
        public static readonly Font Default = new Font("MediumFont", ContentSource.ResourceContent);

        /// <summary>
        /// Pieni oletusfontti.
        /// </summary>
        public static readonly Font DefaultSmall = new Font("SmallFont", ContentSource.ResourceContent);

        /// <summary>
        /// Suuri oletusfontti.
        /// </summary>
        public static readonly Font DefaultLarge = new Font("LargeFont", ContentSource.ResourceContent);

        /// <summary>
        /// Paksunnettu oletusfontti.
        /// </summary>
        public static readonly Font DefaultBold = new Font("MediumFontBold", ContentSource.ResourceContent);

        /// <summary>
        /// Paksunnettu pieni oletusfontti.
        /// </summary>
        public static readonly Font DefaultSmallBold = new Font("SmallFontBold", ContentSource.ResourceContent);

        /// <summary>
        /// Paksunnettu suuri oletusfontti.
        /// </summary>
        public static readonly Font DefaultLargeBold = new Font("LargeFontBold", ContentSource.ResourceContent);

        private SpriteFont xnaFont;
        private string name;
        private ContentSource source;
        private Vector[] charsizes;

        internal SpriteFont XnaFont
        {
            get { DoLoad(); return xnaFont; }
        }

        /// <summary>
        /// Merkin leveys.
        /// </summary>
        public double CharacterWidth
        {
            get { return XnaFont.MeasureString( "X" ).X; }
        }

        /// <summary>
        /// Merkin korkeus.
        /// </summary>
        public double CharacterHeight
        {
            get { return XnaFont.MeasureString( "X" ).Y; }
        }

        internal Font(string name, ContentSource source)
        {
            this.xnaFont = null;
            this.charsizes = null;
            this.name = name;
            this.source = source;
        }

        internal Font(SpriteFont xnaFont)
        {
            this.xnaFont = xnaFont;
            this.charsizes = new Vector[xnaFont.Characters.Count];
            this.name = null;
            this.source = ContentSource.ResourceContent;
        }

        private void DoLoad()
        {
            if (xnaFont == null)
            {
                if (source == ContentSource.ResourceContent)
                    xnaFont = Game.ResourceContent.Load<SpriteFont>(name);
                else
                    xnaFont = Game.Instance.Content.Load<SpriteFont>(name);

                charsizes = new Vector[xnaFont.Characters.Count];
            }
        }

        /// <summary>
        /// Palauttaa annetun merkin koon tässä fontissa.
        /// </summary>
        /// <param name="c">Merkki</param>
        /// <returns>Kokovektori, nollavektori jos merkkiä ei ole määritelty</returns>
        public Vector GetCharacterSize( char c )
        {
            int index = XnaFont.Characters.IndexOf( c );
            if ( index < 0 ) return Vector.Zero;

            if ( charsizes[index] == Vector.Zero )
            {
                XnaV2 xnaSize = XnaFont.MeasureString( c.ToString() );
                charsizes[index] = new Vector( xnaSize.X, xnaSize.Y );
            }

            return charsizes[index];
        }

        /// <summary>
        /// Katkaisee merkkijonon loppupäästä niin että se sopii annettuun
        /// pikselileveyteen fontilla kirjoitettuna.
        /// </summary>
        /// <param name="str">Merkkijono</param>
        /// <param name="maxLineWidth">Maksimipikselimäärä merkkijonolle</param>
        /// <returns></returns>
        public string TruncateText( string str, double maxLineWidth )
        {
            StringBuilder builder = new StringBuilder( str );
            double realWidth = XnaFont.MeasureString( str ).X;

            while ( realWidth > maxLineWidth )
            {
                builder.Remove( builder.Length - 1, 1 );
                realWidth = XnaFont.MeasureString( builder ).X;
            }

            return builder.ToString();
        }

        /// <summary>
        /// Rivittää tekstin.
        /// </summary>
        /// <param name="text">Rivitettävä teksti.</param>
        /// <param name="softLineWidth">Leveys jonka jälkeen seuraava sana rivitetään seuraavalle riville.</param>
        /// <param name="hardLineWidth">Leveys jonka jälkeen sana katkaistaan keskeltä.</param>
        public string WrapText( string text, double softLineWidth, double hardLineWidth )
        {
            if (softLineWidth <= 0 || hardLineWidth <= 0)
            {
                throw new ArgumentException("Width must be positive.");
            }

            StringBuilder sb = new StringBuilder();
            int space = text.IndexOf( ' ' );
            int startIndex = 0;

            while ( startIndex < text.Length )
            {
                string line = space < 0 ? text.Substring( startIndex ) : text.Substring( startIndex, space - startIndex );
                double lineWidth = XnaFont.MeasureString( line ).X;

                if ( lineWidth > hardLineWidth )
                {
                    // Hard break
                    int breakIndex = FindWrapIndex( line, hardLineWidth, false );
                    int prevLength = sb.Length;
                    sb.Append( line );
                    sb.Length = prevLength + breakIndex;
                    sb.Append( "\n" );
                    startIndex += breakIndex;
                }
                else if ( space >= 0 && lineWidth > softLineWidth )
                {
                    // Soft break
                    sb.Append( line );
                    sb.Append( "\n" );
                    startIndex = space + 1;
                }
                else if ( space < 0 )
                {
                    // End of text
                    sb.Append( line );
                    break;
                }

                // If space >= 0 and the line has not been wrapped, we just jump to
                // the next space, incorporating this one into the line

                if ( space >= 0 && space + 1 < text.Length )
                    space = text.IndexOf( ' ', space + 1 );
                else
                    space = -1;
            }

            return sb.ToString();
        }

        /// <summary>
        /// Etsii katkaisuindeksin merkkijonolle merkki kerrallaan.
        /// Välilyönneillä ei ole erikoisasemaaa.
        /// </summary>
        /// <param name="text">Merkkijono</param>
        /// <param name="maxWidth">Maksimileveys pikseleinä</param>
        /// <param name="fromRight">Oikealta vasemmalle (oletus vasemmalta oikealle)</param>
        /// <returns>Katkaisukohdan indeksi</returns>
        private int FindWrapIndex( string text, double maxWidth, bool fromRight )
        {
            double currentWidth = -XnaFont.Spacing;
            int i = fromRight ? text.Length - 1 : 0;
            int step = fromRight ? -1 : 1;

            //for ( int i = 0; i < text.Length; i++ )
            while ( ( fromRight && i >= 0 ) || ( !fromRight && i < text.Length ) )
            {
                currentWidth += XnaFont.Spacing + GetCharacterSize( text[i] ).X;
                if ( currentWidth >= maxWidth ) return i;
                i += step;
            }

            return fromRight ? -1 : text.Length;
        }
    }
}
