﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using System.ComponentModel;
using System.Xml;
using System.Xml.Serialization;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework.Content;

#if XBOX
using Microsoft.Xna.Framework.GamerServices;
#endif

using Jypeli.Widgets;
using Jypeli.Effects;
using Jypeli.Controls;
using Jypeli.WP7;

using XnaColor = Microsoft.Xna.Framework.Color;
using XnaSoundEffect = Microsoft.Xna.Framework.Audio.SoundEffect;
using XnaRectangle = Microsoft.Xna.Framework.Rectangle;
using System.Reflection;


namespace Jypeli
{
    /// <summary>
    /// Peliluokka reaaliaikaisille peleille.
    /// </summary>
    [Save]
    public class Game : Microsoft.Xna.Framework.Game, ControlContexted, IDisposable
    {
        /// <summary>
        /// Pienin mahdollinen kerros.
        /// </summary>
        public const int MinLayer = -3;

        /// <summary>
        /// Suurin mahdollinen kerros.
        /// </summary>
        public const int MaxLayer = 3;

        /// <summary>
        /// Kerrosten määrä.
        /// </summary>
        public const int LayerCount = MaxLayer - MinLayer + 1;

        /// <summary>
        /// Pelin nimi.
        /// </summary>
        public static string Name { get; private set; }

        // TJ: Since there is only one isntance of the
        // game class in practice, its useful to have it available as
        // a static member. This way, we avoid passing a reference to the
        // game in lots of places.
        //
        // In addition, some often-used things, such as Time or GraphicsDevice,
        // are available as static properties (public or internal) as well.
        public static Game Instance { get; private set; }

        /// <summary>
        /// Tapahtuu kun Game.Instance on alustettu.
        /// </summary>
        public static event Action InstanceInitialized;

        /// <summary>
        /// Kamera, joka näyttää ruudulla näkyvän osan kentästä.
        /// Kameraa voidaan siirtää, zoomata tai asettaa seuraamaan tiettyä oliota.
        /// </summary>
        [Save]
        public Camera Camera { get; set; }

        /// <summary>
        /// Kentän reunat näkyvissä tai pois näkyvistä.
        /// Huomaa, että tämä ominaisuus ei vaikuta reunojen törmäyskäsittelyyn.
        /// </summary>
        public bool DrawPerimeter { get; set; }

        /// <summary>
        /// Väri, jolla kentän reunat piirretään.
        /// </summary>
        public Color PerimeterColor { get; set; }

        /// <summary>
        /// Kirjaston mukana tuleva sisältö.
        /// Voidaan käyttää esimerkiksi tekstuurien lataamiseen.
        /// </summary>
        internal static ResourceContentManager ResourceContent { get; private set; }

        /// <summary>
        /// Näytön dimensiot, eli koko ja reunat.
        /// </summary>
        public static ScreenView Screen
        {
            get { return Instance.screen; }
        }

        /// <summary>
        /// Pelin kontrollit.
        /// </summary>
        public static Jypeli.Controls.Controls Controls
        {
            get { return Instance.controls; }
        }

        private ListenContext _context = new ListenContext() { Active = true };

        /// <summary>
        /// Pelin pääohjainkonteksti.
        /// </summary>
        public ListenContext ControlContext
        {
            get { return Instance._context; }
        }

        /// <summary>
        /// Viestinäyttö, johon voi laittaa viestejä.
        /// </summary>
        /// <value>Viestinäyttö.</value>
        public MessageDisplay MessageDisplay { get; set; }

        /// <summary>
        /// Tietovarasto, johon voi tallentaa tiedostoja pidempiaikaisesti.
        /// Sopii esimerkiksi pelitilanteen lataamiseen ja tallentamiseen.
        /// </summary>
        public static FileManager DataStorage { get { return Instance.dataStorage; } }

        /// <summary>
        /// Onko olio valittavissa.
        /// Vain valittu (fokusoitu) olio voii kuunnella näppäimistöä ja muita ohjainlaitteita.
        /// Peliolio on aina valittavissa.
        /// </summary>
        public bool AcceptsFocus { get { return true; } }

        /// <summary>
        /// Näppäimistö.
        /// </summary>
        public Keyboard Keyboard { get { return controls.Keyboard; } }

        /// <summary>
        /// Hiiri.
        /// </summary>
        public Mouse Mouse { get { return controls.Mouse; } }

        /// <summary>
        /// Kosketusnäyttö. Vain kännykässä.
        /// </summary>
        public TouchPanel TouchPanel { get { return controls.TouchPanel; } }

        public PhoneBackButton PhoneBackButton { get { return controls.PhoneBackButton; } }

        /// <summary>
        /// Peliohjain yksi.
        /// </summary>
        public GamePad ControllerOne { get { return controls.GameControllers[0]; } }

        /// <summary>
        /// Peliohjain kaksi.
        /// </summary>
        public GamePad ControllerTwo { get { return controls.GameControllers[1]; } }

        /// <summary>
        /// Peliohjain kolme.
        /// </summary>
        public GamePad ControllerThree { get { return controls.GameControllers[2]; } }

        /// <summary>
        /// Peliohjain neljä.
        /// </summary>
        public GamePad ControllerFour { get { return controls.GameControllers[3]; } }

        /// <summary>
        /// Kiihtyvyysanturi. Vain kännykässä.
        /// </summary>
        public Accelerometer Accelerometer { get { return controls.Accelerometer; } }

        private Phone phone;

        /// <summary>
        /// Phone-olio esim. puhelimen tärisyttämiseen.
        /// </summary>
        public Phone Phone { get { return phone; } }


        /// <summary>
        /// Aktiivinen kenttä.
        /// </summary>
        public Level Level
        {
            get { return theLevel; }
        }

        /// <summary>
        /// Peliaika. Tätä päivitetään noin 30 kertaa sekunnissa.
        /// </summary>
        public static Time Time
        {
            get { return currentTime; }
        }

        /// <summary>
        /// Tuuli. Vaikuttaa vain efekteihin
        /// </summary>
        public static Vector Wind { get; set; }

        /// <summary>
        /// Teksti, joka näkyy pelin ikkunassa (jos peli ei ole koko ruudun tilassa).
        /// </summary>
        public string Title
        {
            get { return Window.Title; }
            set { Window.Title = value; }
        }

        /// <summary>
        /// Onko peli kokoruututilassa.
        /// </summary>
        public bool IsFullScreen
        {
            get { return isFullScreenRequested; }
            set
            {
                if ( GraphicsDevice == null )
                {
                    // GraphicsDevice is not initialized yet.
                    isFullScreenRequested = value;
                }
                else if ( ( Graphics.IsFullScreen != value ) )
                {
#if WINDOWS_PHONE
                    Phone.ResetScreen();
#else
                    SetWindowSize( GraphicsDevice.DisplayMode.Width, GraphicsDevice.DisplayMode.Height, value );
#endif
                }
            }
        }

        /// <summary>
        /// Mediasoitin.
        /// </summary>
        public MediaPlayer MediaPlayer { get; private set; }

        private Jypeli.Controls.Controls controls;
        private FileManager dataStorage;
        private ScreenView screen;

        private static Renderer renderer = new Renderer();

        internal static Renderer Renderer { get { return renderer; } }

        [EditorBrowsable( EditorBrowsableState.Never )]
        public static new GraphicsDevice GraphicsDevice
        {
            get { return ( (Microsoft.Xna.Framework.Game)Instance ).GraphicsDevice; }
        }

        internal static GraphicsDeviceManager Graphics { get; private set; }

#if !WINDOWS_PHONE
        internal static List<Light> Lights { get { return lights; } }
#endif

#if DEBUG
        private Widgets.BarGauge[] objectCountDisplays = new Widgets.BarGauge[LayerCount];
        private Label fpsDisplay;
        private string fpsText = "00";
        private int fpsSkipCounter;
        private bool isDebugScreenShown = false;
#endif

        private List<Widget> widgets = new List<Widget>();
        private Queue<Widget> widgetsToAdd = new Queue<Widget>();
        private List<Widget> widgetsToRemove = new List<Widget>();

        struct PendingObject
        {
            public GameObject ObjectBeingAdded;
            public int Layer;
        }

        private Layer[] layers;

        internal static List<GameObject> allGameObjects = new List<GameObject>();

        /// <summary>
        /// This list contains objects that have been removed (not destroyed).
        /// Layers take care of removing destroyed objects.
        /// </summary>
        internal static List<GameObject> gameObjectsToRemove = new List<GameObject>();

        private Queue<PendingObject> gameObjectsToAdd = new Queue<PendingObject>();
        private List<GameObject> gameObjectsThatNeedUpdating = new List<GameObject>();

#if !WINDOWS_PHONE
        private static List<Light> lights = new List<Light>();
#endif

        private Level theLevel;

        private bool loadContentHasBeenCalled = false;

        // fullscreen isn't used as default, because debug mode doesn't work well with it
        private bool isFullScreenRequested = false;

        private static Time currentTime = new Time();


        /// <summary>
        /// Alustaa uuden peliluokan.
        /// </summary>
        public Game()
        {
            if ( Instance != null )
                throw new Exception( "Only one instance of the Game class can be created." );

            Instance = this;
            Name = Assembly.GetCallingAssembly().FullName.Split( ',' )[0];

            layers = new Layer[Game.LayerCount];
            for ( int i = 0; i < Game.LayerCount; i++ )
            {
                layers[i] = new Layer();
            }

#if WINDOWS_PHONE
            
            ResourceContent = new ResourceContentManager( this.Services, WindowsPhoneResources.ResourceManager );
#elif XBOX
            ResourceContent = new ResourceContentManager( this.Services, XBox360Resources.ResourceManager );
#else
            ResourceContent = new ResourceContentManager( this.Services, Resources.ResourceManager );
#endif

            Content.RootDirectory = "Content";

            Camera = new Camera();
            controls = new Jypeli.Controls.Controls();

            Graphics = new GraphicsDeviceManager( this );
            Graphics.PreferredDepthStencilFormat = Jypeli.Graphics.SelectStencilMode();

#if WINDOWS
            dataStorage = new WindowsFileManager();
#elif WINDOWS_PHONE
            dataStorage = new IsolatedStorageManager();
#elif XBOX
            Components.Add( new GamerServicesComponent( this ) );
            dataStorage = new XboxFileManager();
#endif

            DrawPerimeter = false;
            PerimeterColor = Color.DarkGray;

            phone = new Phone();
            Phone.Tombstoning = true;
        }

        /// <summary>
        /// Suorittaa aliohjelman kun peli on varmasti alustettu.
        /// </summary>
        /// <param name="actionMethod">Suoritettava aliohjelma.</param>
        public static void AssertInitialized( Action actionMethod )
        {
            if ( Instance != null )
                actionMethod();
            else
                InstanceInitialized += actionMethod;
        }

        public Layer GetLayer( int layer )
        {
            CheckForInvalidLayer( layer );
            return layers[LayerToIndex( layer )];
        }

        /// <summary>
        /// Lisää olion peliin, kerrokseen 0.
        /// </summary>
        public void Add( GameObject o )
        {
            Add( o, 0 );
        }

#if !WINDOWS_PHONE
        /// <summary>
        /// Lisää valon peliin. Nykyisellään valoja voi olla ainoastaan
        /// yksi kappale.
        /// </summary>
        public void Add( Light light )
        {
            if ( light == null ) throw new NullReferenceException( "Tried to add a null light to game" );

            if ( lights.Count >= 1 )
                throw new NotSupportedException( "Only one light is supported" );

            lights.Add( light );
        }
#endif

        private void CheckForInvalidLayer( int layer )
        {
            if ( layer < MinLayer || MaxLayer < layer )
            {
                string message = string.Format( "The layer must be between {0} and {1}", MinLayer, MaxLayer );
                throw new ArgumentException( message );
            }
        }

        /// <summary>
        /// Lisää peliolion peliin, tiettyyn kerrokseen.
        /// </summary>
        /// <param name="o">Lisättävä olio.</param>
        /// <param name="layer">Kerros, luku väliltä [-3, 3].</param>
        public virtual void Add( GameObject o, int layer )
        {
            if ( o == null ) throw new NullReferenceException( "Tried to add a null object to game" );
            CheckForInvalidLayer( layer );

            if (o is ParticleSystem)
            {
                layers[LayerToIndex(layer)].AddEffect((ParticleSystem)o);
            }
            else
            {
                bool wasAboutToBeRemoved = gameObjectsToRemove.Contains( o );

                if (allGameObjects.Contains(o) && !wasAboutToBeRemoved)
                     throw new ArgumentException("The object has already been added");

                if ( wasAboutToBeRemoved )
                {
                    gameObjectsToRemove.Remove( o );
                }

                PendingObject po;
                po.ObjectBeingAdded = o;
                po.Layer = layer;
                gameObjectsToAdd.Enqueue(po);
            }
        }

        internal static int LayerToIndex( int layer )
        {
            return layer + ( 0 - MinLayer );
        }

        internal static int IndexToLayer( int index )
        {
            return index - ( 0 - MinLayer );
        }

        /// <summary> 
        /// Poistaa olion pelistä. Jos haluat tuhota olion, 
        /// kutsu mielummin olion <c>Destroy</c>-metodia. 
        /// </summary> 
        /// <remarks> 
        /// Oliota ei poisteta välittömästi, vaan viimeistään seuraavan 
        /// päivityksen jälkeen. 
        /// </remarks> 
        public void Remove( GameObject o )
        {
            /*
             * The object is not removed immediately, because the removal
             * can happen in an update loop, while the objects are being
             * iterated over.
             */
            gameObjectsToRemove.Add( o );
        }

        /// <summary>
        /// Palauttaa peliolion, joka on annetussa paikassa.
        /// Jos paikassa ei ole mitään pelioliota, palautetaan null.
        /// Jos olioita on useampia, palautetaan ensin lisätty.
        /// </summary>
        /// <param name="position">Paikkakoordinaatit</param>
        /// <returns>Mahdollinen olio</returns>
        public GameObject GetObjectAt( Vector position )
        {
            for ( int i = 0; i < allGameObjects.Count; i++ )
            {
                if ( allGameObjects[i].IsInside( position ) )
                    return allGameObjects[i];
            }

            return null;
        }

        /// <summary>
        /// Palauttaa peliolion, joka on annetussa paikassa tietyllä säteellä.
        /// Jos paikassa ei ole mitään pelioliota, palautetaan null.
        /// Jos olioita on useampia, palautetaan ensin lisätty.
        /// </summary>
        /// <param name="position">Paikkakoordinaatit</param>
        /// <returns>Mahdollinen olio</returns>
        public GameObject GetObjectAt( Vector position, double radius )
        {
            for ( int i = 0; i < allGameObjects.Count; i++ )
            {
                Vector positionUp = new Vector( position.X, position.Y + radius );
                Vector positionDown = new Vector( position.X, position.Y - radius );
                Vector positionLeft = new Vector( position.X - radius, position.Y );
                Vector positionRight = new Vector( position.X + radius, position.Y );

                if ( allGameObjects[i].IsInside( position ) ) return allGameObjects[i];
                if ( allGameObjects[i].IsInside( positionUp ) ) return allGameObjects[i];
                if ( allGameObjects[i].IsInside( positionDown ) ) return allGameObjects[i];
                if ( allGameObjects[i].IsInside( positionLeft ) ) return allGameObjects[i];
                if ( allGameObjects[i].IsInside( positionRight ) ) return allGameObjects[i];
            }

            return null;
        }

        /// <summary>
        /// Palauttaa peliolion, joka on annetussa paikassa.
        /// Vain annetulla tagilla varustetut oliot huomioidaan.
        /// Jos paikassa ei ole mitään pelioliota, palautetaan null.
        /// Jos olioita on useampia, palautetaan ensin lisätty.
        /// </summary>
        /// <param name="position">Paikkakoordinaatit</param>
        /// <param name="tag">Etsittävän olion tagi.</param>
        /// <returns>Mahdollinen olio</returns>
        public GameObject GetObjectAt( Vector position, object tag )
        {
            for ( int i = 0; i < allGameObjects.Count; i++ )
            {
                if ( allGameObjects[i].IsInside( position ) && allGameObjects[i].Tag.Equals( tag ) )
                    return allGameObjects[i];
            }

            return null;
        }

        /// <summary>
        /// Palauttaa peliolion, joka on annetussa paikassa tietyllä säteellä.
        /// Vain annetulla tagilla varustetut oliot huomioidaan.
        /// Jos paikassa ei ole mitään pelioliota, palautetaan null.
        /// Jos olioita on useampia, palautetaan ensin lisätty.
        /// </summary>
        /// <param name="position">Paikkakoordinaatit</param>
        /// <returns>Mahdollinen olio</returns>
        public GameObject GetObjectAt( Vector position, object tag, double radius )
        {
            for ( int i = 0; i < allGameObjects.Count; i++ )
            {
                if ( !allGameObjects[i].Tag.Equals( tag ) )
                    continue;

                Vector positionUp = new Vector( position.X, position.Y + radius );
                Vector positionDown = new Vector( position.X, position.Y - radius );
                Vector positionLeft = new Vector( position.X - radius, position.Y );
                Vector positionRight = new Vector( position.X + radius, position.Y );

                if ( allGameObjects[i].IsInside( position ) ) return allGameObjects[i];
                if ( allGameObjects[i].IsInside( positionUp ) ) return allGameObjects[i];
                if ( allGameObjects[i].IsInside( positionDown ) ) return allGameObjects[i];
                if ( allGameObjects[i].IsInside( positionLeft ) ) return allGameObjects[i];
                if ( allGameObjects[i].IsInside( positionRight ) ) return allGameObjects[i];
            }

            return null;
        }

        /// <summary> 
        /// Poistaa widgetin pelistä. Jos haluat tuhota widgetin, 
        /// kutsu mielummin sen <c>Destroy</c>-metodia.
        /// </summary> 
        /// <remarks> 
        /// Widgettiä ei poisteta välittömästi, vaan viimeistään seuraavan 
        /// päivityksen jälkeen. 
        /// </remarks> 
        public void Remove( Widget w )
        {
            widgetsToRemove.Add( w );
        }

        internal void AddUpdatedObject( GameObject o )
        {
            if ( !gameObjectsThatNeedUpdating.Contains( o ) )
            {
                gameObjectsThatNeedUpdating.Add( o );
            }
        }

        internal void RemoveUpdatedObject( GameObject o )
        {
            gameObjectsThatNeedUpdating.Remove( o );
        }

        /// <summary>
        /// Toteuttaa olion poistamisen pelistä. Älä kutsu tätä suoraan pelissä, käytä mielummin
        /// pelioliolle <c>o</c> metodia <c>o.Destroy()</c>.
        /// </summary>
        internal virtual void DoRemove( GameObject o )
        {
            allGameObjects.Remove( o );
            o.IsAddedToGame = false;
        }

        /// <summary>
        /// Lisää widgetin peliin.
        /// </summary>
        public void Add( Widget w )
        {
            if ( w == null ) throw new NullReferenceException( "Tried to add a null widget to game" );

            bool wasAboutToBeRemoved = widgetsToRemove.Contains( w );
            if ( !wasAboutToBeRemoved && widgets.Contains(w) )
                throw new ArgumentException( "The widget has already been added" );

            if ( wasAboutToBeRemoved )
            {
                widgetsToRemove.Remove( w );
                w.OnRemovedFromGame();
            }

            widgetsToAdd.Enqueue( w );
        }

        /// <summary>
        /// Lataa kuvan contentista.
        /// </summary>
        /// <param name="name">Kuvan nimi (ei tarkennetta)</param>
        /// <returns>Image-olio</returns>
        public static Image LoadImage( string name )
        {
            return new Image( name );
        }

        /// <summary>
        /// Lataa taulukon kuvia contentista.
        /// </summary>
        /// <param name="name">Kuvien nimet ilman tarkennetta pilkuin eroiteltuna</param>
        /// <returns>Taulukko Image-olioita</returns>
        public static Image[] LoadImages( params string[] names )
        {
            Image[] result = new Image[names.Length];
            for ( int i = 0; i < names.Length; i++ )
                result[i] = LoadImage( names[i] );
            return result;
        }

        /// <summary>
        /// Lataa ääniefektin contentista.
        /// </summary>
        /// <param name="name">Äänen nimi (ei tarkennetta)</param>
        /// <returns>SoundEffect-olio</returns>
        public static SoundEffect LoadSoundEffect( string name )
        {
            return new SoundEffect( name );
        }

        /// <summary>
        /// Lataa taulukon ääniefektejä contentista.
        /// </summary>
        /// <param name="names">Äänien nimet ilman tarkennetta pilkuin eroiteltuna</param>
        /// <returns>Taulukko SoundEffect-olioita</returns>
        public static SoundEffect[] LoadSoundEffects( params string[] names )
        {
            SoundEffect[] result = new SoundEffect[names.Length];
            for ( int i = 0; i < names.Length; i++ )
                result[i] = LoadSoundEffect( names[i] );
            return result;
        }

        /// <summary>
        /// Nollaa kaiken.
        /// </summary>
        public virtual void ClearAll()
        {
            Level.Clear();
            ClearGameObjects();
            ClearWidgets();
            Timer.ClearAll();
#if !WINDOWS_PHONE
            ClearLights();
#endif
            ClearControls();
            GC.Collect();
            addMessageDisplay();
        }

        /// <summary>
        /// Palauttaa kontrollit alkutilaansa.
        /// </summary>
        public void ClearControls()
        {
            controls.Clear();
            TouchPanel.Clear();
#if DEBUG && !WINDOWS_PHONE
            Keyboard.Listen( Key.F12, ButtonState.Pressed, ToggleDebugScreen, null );
#endif
        }

#if DEBUG
        private void ToggleDebugScreen()
        {
            if ( isDebugScreenShown )
                isDebugScreenShown = false;
            else
                isDebugScreenShown = true;
        }
#endif

        /// <summary>
        /// Tuohoaa ja poistaa pelistä kaikki pelioliot (ml. fysiikkaoliot).
        /// </summary>
        public void ClearGameObjects()
        {
            foreach ( var layer in layers )
                layer.Clear();

            foreach ( var o in allGameObjects )
            {
                if ( !gameObjectsToRemove.Contains(o) )
                {
                    o.Destroy();
                    gameObjectsToRemove.Add( o );
                }
            }
        }

#if !WINDOWS_PHONE
        public void ClearLights()
        {
            lights.Clear();
        }
#endif

        public void ClearWidgets()
        {
            foreach ( var o in widgets )
            {
                if ( !widgetsToRemove.Contains( o ) )
                {
                    o.Destroy();
                    widgetsToRemove.Add( o );
                }
            }
        }

        /// <summary>
        /// Ajetaan kun pelin tilannetta päivitetään. Päivittämisen voi toteuttaa perityssä luokassa
        /// toteuttamalla tämän metodin. Perityn luokan metodissa tulee kutsua kantaluokan metodia.
        /// </summary>
        protected virtual void Update( Time time )
        {
        }
        /// <summary>
        /// This gets called after the GraphicsDevice has been created. So, this is
        /// the place to initialize the resources needed in the game. Except the graphics content,
        /// which should be called int LoadContent(), according to the XNA docs.
        /// </summary>
        [EditorBrowsable( EditorBrowsableState.Never )]
        protected override void Initialize()
        {
            Jypeli.Graphics.Initialize();

#if WINDOWS_PHONE
            isFullScreenRequested = true;
            Phone.ResetScreen();
#else
            SetWindowSize( GraphicsDevice.DisplayMode.Width, GraphicsDevice.DisplayMode.Height, isFullScreenRequested );
#endif

            screen = new ScreenView( GraphicsDevice.Viewport );

            theLevel = new Level( this );

            MediaPlayer = new MediaPlayer( Content );

            addMessageDisplay();

#if DEBUG
            double barWidth = 20;
            double barHeight = Screen.Height;
            fpsDisplay = new Label( "00" );
            fpsDisplay.Color = Color.Gray;
            fpsDisplay.X = Level.Right - barWidth / 2 - fpsDisplay.Width;
            fpsDisplay.Y = Screen.Top - fpsDisplay.Height / 2;

            double left = Screen.Right - LayerCount * barWidth;

            for ( int i = 0; i < LayerCount; i++ )
            {
                var gauge = new Widgets.BarGauge( barWidth, Screen.Height );
                gauge.X = left + i * barWidth + barWidth / 2;
                gauge.Y = Screen.Center.Y;
                gauge.BarColor = Color.DarkGreen;
                gauge.BindTo( layers[i].ObjectCount );
                objectCountDisplays[i] = gauge;
            }

            Keyboard.Listen( Key.F12, ButtonState.Pressed, ToggleDebugScreen, null );
#endif

            base.Initialize();
        }

        /// <summary>
        /// XNA calls this when graphics resources need to be loaded.
        /// Note that this can be called multiple times (whenever the graphics device is reset).
        /// </summary>
        [EditorBrowsable( EditorBrowsableState.Never )]
        protected override void LoadContent()
        {
            if ( !loadContentHasBeenCalled )
            {
                if ( InstanceInitialized != null )
                    InstanceInitialized();

#if !WINDOWS_PHONE
                Begin();
#endif
                loadContentHasBeenCalled = true;
            }

#if DEBUG && !WINDOWS_PHONE
            MessageDisplay.Add( "F12 - Debug view" );
#endif
            base.LoadContent();

            GC.Collect();
        }

        /// <summary>
        /// Tässä alustetaan peli.
        /// </summary>
        public virtual void Begin()
        {
        }

        /// <summary>
        /// Tässä alustetaan peli tombstoning-tilasta.
        /// </summary>
        public virtual void Continue()
        {
        }


        private void addMessageDisplay()
        {
            MessageDisplay = new MessageDisplay();
            Add( MessageDisplay );
        }

        private void RemoveExpiredObjects()
        {
            foreach ( var layer in layers )
            {
                layer.RemoveDestroyedObjects();
            }

            while ( gameObjectsToRemove.Count > 0 )
            {
                GameObject o = gameObjectsToRemove[0];
                gameObjectsToRemove.RemoveAt( 0 );

                for ( int i = 0; i < layers.Length; i++ )
                {
                    if ( layers[i].DoRemove( o ) )
                        break;
                }
                DoRemove( o );
            }

            while ( widgetsToRemove.Count > 0 )
            {
                Widget w = widgetsToRemove[0];
                w.Game = null;
                widgetsToRemove.RemoveAt( 0 );
                widgets.Remove( w );

                if ( w.IsModal )
                {
                    Game.Instance.ControlContext.RestoreFocus();

                    for ( int i = 0; i < widgets.Count; i++ )
                    {
                        widgets[i].ControlContext.RestoreFocus();
                    }
                }
            }
        }

        private void AddPendingObjects()
        {
            while ( gameObjectsToAdd.Count > 0 )
            {
                PendingObject po = gameObjectsToAdd.Dequeue();

                allGameObjects.Add( po.ObjectBeingAdded );
                layers[LayerToIndex( po.Layer )].Add( po.ObjectBeingAdded );

                po.ObjectBeingAdded.IsAddedToGame = true;
                po.ObjectBeingAdded.OnAddedToGame();
            }

            Widget modalWidget = null;

            while ( widgetsToAdd.Count > 0 )
            {
                Widget w = widgetsToAdd.Dequeue();
                widgets.Add( w );
                w.Game = this;
                w.ControlContext.Active = true;

                if ( w.IsModal )
                {
                    Game.Instance.ControlContext.SaveFocus();
                    Game.Instance.ControlContext.Active = false;
                    modalWidget = w;
                }

                w.OnAddedToGame();
            }

            if ( modalWidget != null )
            {
                for ( int i = 0; i < widgets.Count; i++ )
                {
                    if ( widgets[i] == modalWidget )
                        continue;

                    widgets[i].ControlContext.SaveFocus();
                    widgets[i].ControlContext.Active = false;
                }
            }
        }

        [EditorBrowsable( EditorBrowsableState.Never )]
        protected override void Update( GameTime gameTime )
        {
            Time time = currentTime = new Time( gameTime );

            RemoveExpiredObjects();
            AddPendingObjects();

            if ( this.IsActive ) controls.Update();

            // The update in derived classes.
            this.Update( time );

            this.Camera.Update( time );

            foreach ( var obj in gameObjectsThatNeedUpdating )
            {
                obj.Update( time );
            }

            foreach ( Widget w in widgets )
            {
                w.Update( time );
                if ( w.IsDestroyed )
                {
                    widgetsToRemove.Add( w );
                }
            }

            foreach ( var layer in layers )
            {
                layer.Update( time );
            }

            Timer.UpdateAll();

            base.Update( gameTime );
        }

        [EditorBrowsable( EditorBrowsableState.Never )]
        protected override void Draw( GameTime gameTime )
        {
            Time time = new Time( gameTime );

            GraphicsDevice.Clear( ClearOptions.Target | ClearOptions.DepthBuffer, Level.BackgroundColor.AsXnaColor(), 1.0f, 0 );

            if ( Level.Background.Image != null && !Level.Background.MovesWithCamera )
            {
                // TJ: When using drawing.DrawTexture(), polygons that are drawn
                // after the background don't get drawn at all.

                SpriteBatch spriteBatch = Jypeli.Graphics.SpriteBatch;
                //spriteBatch.Begin( SpriteBlendMode.AlphaBlend, SpriteSortMode.Deferred, SaveStateMode.SaveState );
                spriteBatch.Begin( SpriteSortMode.Deferred, BlendState.AlphaBlend );
                spriteBatch.Draw( Level.Background.Image.XNATexture, new XnaRectangle( 0, 0, (int)screen.Width, (int)screen.Height ), XnaColor.White );
                spriteBatch.End();
            }

            // The world matrix adjusts the position and size of objects according to the camera angle.
            var worldMatrix =
                Matrix.CreateTranslation( (float)-Camera.Position.X, (float)-Camera.Position.Y, 0 )
                * Matrix.CreateScale( (float)Camera.ZoomFactor, (float)Camera.ZoomFactor, 1f );

            // If the background should move with camera, draw it here.
            Level.Background.Draw( worldMatrix, Matrix.Identity );

            if ( DrawPerimeter )
            {
                Matrix m = Matrix.CreateScale( (float)Level.Width, (float)Level.Height, 1 ) * worldMatrix;
                Renderer.DrawRectangle( ref m, PerimeterColor );
            }

            for ( int i = 0; i < LayerCount; i++ )
            {
                layers[i].Draw( Camera );
            }

            foreach ( Widget w in widgets )
            {
                w.Draw( Matrix.Identity );
            }

#if DEBUG
            if ( isDebugScreenShown )
            {
                DrawDebugView();
            }
#endif

            base.Draw( gameTime );
        }

#if DEBUG
        private void DrawDebugView()
        {
            Texture2D cachedTexture = layers[LayerToIndex( 0 )].cachedTexture;

            if ( cachedTexture != null )
            {
                const double scale = 0.1;

                Image cache = new Image( cachedTexture );
                Vector cachePosition = new Vector( Screen.Center.X, Screen.Top - cache.Height * scale / 2 );
                Vector cacheSize = new Vector(cache.Width * scale, cache.Height * scale);

                Matrix m1 = Matrix.CreateScale( (float)cacheSize.X, (float)cacheSize.Y, 1 )
                    * Matrix.CreateTranslation( (float)cachePosition.X, (float)cachePosition.Y, 0f );

                Renderer.DrawImage( cache, ref m1, new Vector( 1, 1 ) );

                Renderer.DrawRectangle( ref m1, Color.Green );

                //Vector maximumImageSize = new Vector( cap.MaxTextureWidth, cap.MaxTextureHeight ) * scale;
                Vector maximumImageSize = new Vector( 2048, 2048 ) * scale;
                Vector maxImagePosition;
                maxImagePosition.X = cachePosition.X - cacheSize.X / 2 + maximumImageSize.X / 2;
                maxImagePosition.Y = Screen.Top - maximumImageSize.Y / 2;

                Matrix m2 = Matrix.CreateScale( (float)maximumImageSize.X, (float)maximumImageSize.Y, 1 )
                    * Matrix.CreateTranslation( (float)maxImagePosition.X, (float)maxImagePosition.Y, 0 );
                renderer.DrawRectangle( ref m2, Color.Red );
            }

            for ( int i = 0; i < LayerCount; i++ )
            {
                objectCountDisplays[i].Draw( Matrix.Identity );
            }

            if ( fpsSkipCounter++ > 10 )
            {
                fpsSkipCounter = 0;
                fpsText = ( 1.0 / Time.SinceLastUpdate.TotalSeconds ).ToString( "F2" );
            }

            fpsDisplay.Text = fpsText;
            fpsDisplay.Draw( Matrix.Identity );
        }
#endif

        /// <summary>
        /// Asettaa ikkunan koon.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        public bool SetWindowSize(int width, int height)
        {
            return SetWindowSize( width, height, IsFullScreen );
        }

        /// <summary>
        /// Asettaa ikkunan koon ja alustaa pelin käyttämään joko ikkunaa tai koko ruutua.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        /// <param name="fullscreen">Koko ruutu jos <c>true</c>, muuten ikkuna.</param>
        /// <returns></returns>
        public bool SetWindowSize( int width, int height, bool fullscreen )
        {
            // copy-paste code from: http://forums.xna.com/forums/p/1031/5461.aspx#5461
            // If we aren't using a full screen mode, the height and width of the window can
            // be set to anything equal to or smaller than the actual screen size.
            if ( fullscreen == false )
            {
                if ( ( width <= GraphicsAdapter.DefaultAdapter.CurrentDisplayMode.Width )
                    && ( height <= GraphicsAdapter.DefaultAdapter.CurrentDisplayMode.Height ) )
                {
                    // The mode is supported, so set the buffer formats, apply changes and return
                    doSetWindowSize( Graphics, width, height, false );
                    return true;
                }
            }
            else
            {
                // If we are using full screen mode, we should check to make sure that the display
                // adapter can handle the video mode we are trying to set.  To do this, we will
                // iterate thorugh the display modes supported by the adapter and check them against
                // the mode we want to set.
                foreach ( DisplayMode dm in GraphicsAdapter.DefaultAdapter.SupportedDisplayModes )
                {
                    // Check the width and height of each mode against the passed values
                    if ( ( dm.Width == width ) && ( dm.Height == height ) )
                    {
                        // The mode is supported, so set the buffer formats, apply changes and return
                        doSetWindowSize( Graphics, width, height, true );

                        return true;
                    }
                }
            }

            return false;
        }

        private void doSetWindowSize( GraphicsDeviceManager graphics, int width, int height, bool fullscreen )
        {
            graphics.PreferredBackBufferWidth = width;
            graphics.PreferredBackBufferHeight = height;
            graphics.IsFullScreen = fullscreen;
            graphics.ApplyChanges();

            if ( GraphicsDevice != null )
            {
                Viewport viewPort = GraphicsDevice.Viewport;
                Mouse.Viewport = viewPort;
                if ( screen != null )
                {
                    screen.viewPort = viewPort;
                }
            }
        }

        /// <summary>
        /// Lopettaa pelin.
        /// </summary>
        public void Exit()
        {
            Phone.StopVibrating();
            base.Exit();
        }

        /// <summary>
        /// Tallentaa pelin.
        /// </summary>
        /// <param name="tagName">Pelitilanteen nimi.</param>
        public void SaveGame( string tagName )
        {
            Type gameType = this.GetType();
            SaveState state = DataStorage.BeginSave( tagName );

            foreach ( PropertyInfo property in gameType.GetProperties( BindingFlags.GetProperty | StorageFile.AllOfInstance ) )
            {
                if ( property.GetCustomAttributes( typeof( SaveAttribute ), true ).Length == 0 )
                    continue;

                object propValue = property.GetValue( this, null );
                Type propType = property.PropertyType;
                //DataStorage.Save( propValue, propType, tagName + FileManager.SanitizeFileName( property.Name ) );
                state.Save( propValue, propType, FileManager.SanitizeFileName( property.Name ) + "Property" );
            }

            foreach ( FieldInfo field in gameType.GetFields( BindingFlags.GetField | StorageFile.AllOfInstance ) )
            {
                if ( field.GetCustomAttributes( typeof( SaveAttribute ), true ).Length == 0 )
                    continue;

                object fieldValue = field.GetValue( this );
                Type fieldType = field.FieldType;
                //DataStorage.Save( fieldValue, fieldType, tagName + FileManager.SanitizeFileName( field.Name ) );
                state.Save( fieldValue, fieldType, FileManager.SanitizeFileName( field.Name ) + "Field" );
            }

            state.EndSave();
        }

        /// <summary>
        /// Lataa pelin.
        /// </summary>
        /// <param name="tagName">Pelitilanteen nimi.</param>
        public void LoadGame( string tagName )
        {
            Type gameType = this.GetType();

            LoadState state = DataStorage.BeginLoad( tagName );

            foreach ( PropertyInfo property in gameType.GetProperties( BindingFlags.GetProperty | StorageFile.AllOfInstance ) )
            {
                if ( property.GetCustomAttributes( typeof( SaveAttribute ), true ).Length == 0 )
                    continue;

                object oldValue = property.GetValue( this, null );
                Type propType = property.PropertyType;
                //object newValue = DataStorage.Load( oldValue, propType, tagName + FileManager.SanitizeFileName( property.Name ) );
                object newValue = state.Load( oldValue, propType, FileManager.SanitizeFileName( property.Name ) + "Property" );
                property.SetValue( this, newValue, null );
            }

            foreach ( FieldInfo field in gameType.GetFields( BindingFlags.GetField | StorageFile.AllOfInstance ) )
            {
                if ( field.GetCustomAttributes( typeof( SaveAttribute ), true ).Length == 0 )
                    continue;

                object oldValue = field.GetValue( this );
                Type fieldType = field.FieldType;
                //object newValue = DataStorage.Load( oldValue, fieldType, tagName + FileManager.SanitizeFileName( field.Name ) );
                object newValue = state.Load( oldValue, fieldType, FileManager.SanitizeFileName( field.Name ) + "Field" );
                field.SetValue( this, newValue );
            }

            state.EndLoad();
        }

        public void AddFactory<T>( string tag, Factory.FactoryMethod method )
        {
            Factory.AddFactory<T>( tag, method );
        }

        public void RemoveFactory<T>( string tag, Factory.FactoryMethod method )
        {
            Factory.RemoveFactory<T>( tag, method );
        }

        public T FactoryCreate<T>( string tag )
        {
            return Factory.FactoryCreate<T>( tag );
        }

        /// <summary>
        /// Näyttää kontrollien ohjetekstit.
        /// </summary>
        public void ShowControlHelp()
        {
            MessageDisplay.Add( "Ohjeet:" );

            foreach ( String message in controls.GetHelpTexts() )
            {
                MessageDisplay.Add( message );
            }
        }

        [EditorBrowsable( EditorBrowsableState.Never )]
        public void Dispose()
        {
        }

        /// <summary>
        /// Sitoo kontrollien ohjeet viestinäyttöön ja haluttuihin nappeihin.
        /// Tämän jälkeen nappeja painamalla pelaaja saa automaattisesti ohjeen esille.
        /// </summary>
        /// <param name="keysOrButtons">Napit, joita painamalla ohjeen saa näkyviin.</param>
        private void BindControlHelp( params object[] keysOrButtons )
        {
            String nappaimet = "";

            foreach ( object o in keysOrButtons )
            {
                if ( o is Key )
                {
                    Key k = (Key)o;
                    controls.Keyboard.Listen( k, ButtonState.Pressed, ShowControlHelp, null );

                    nappaimet += k.ToString();
                }

                if ( o is Button )
                {
                    Button b = (Button)o;
                    for ( int i = 0; i < controls.GameControllers.Length; i++ )
                    {
                        controls.GameControllers[i].Listen( b, ButtonState.Pressed, ShowControlHelp, null );
                    }

                    nappaimet += b.ToString();
                }

                nappaimet += " / ";
            }

            MessageDisplay.Add( "Katso näppäinohje painamalla " + nappaimet.Substring( 0, nappaimet.Length - 3 ) );
        }

        internal static Image LoadImageFromResources( string name )
        {
            return new Image( ResourceContent.Load<Texture2D>( name ) );
        }

        internal static SoundEffect LoadSoundEffectFromResources( string name )
        {
            return new SoundEffect( ResourceContent.Load<XnaSoundEffect>( name ) );
        }

        /// <summary>
        /// Lataa fontin. Fontin tulee olla lisätty content-hakemistoon.
        /// </summary>
        /// <param name="name">Fontin tiedoston nimi, ilman päätettä.</param>
        public static Font LoadFont( string name )
        {
            return new Font( name, ContentSource.GameContent );
        }
    }
}

