﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;

#if !XBOX && !WINDOWS_PHONE
using System.Runtime.Serialization.Formatters.Binary;
#endif

using Microsoft.Xna.Framework;
using Jypeli.Controls;
using System.Reflection;


namespace Jypeli
{
    /// <summary>
    /// Pelialueella liikkuva olio.
    /// Käytä fysiikkapeleissä <c>PhysicsObject</c>-olioita.
    /// </summary>
#if !XBOX && !WINDOWS_PHONE
    [Serializable]
#endif
    [Save]
    public class GameObject : Destroyable, Tagged
    {
        #region Fields

#if !XBOX && !WINDOWS_PHONE
        [NonSerialized]
#endif
        private GameObject parent;

        private Brain brain = Brain.None;

        private Shape shape;
        private Animation animation = null;
        private Color color = Color.White;
        private Vector size;
        private Vector textureWrapSize = new Vector( 1, 1 );

        internal List<GameObject> childObjects = null;
        private Queue<GameObject> childObjectsToRemove = null;

        internal List<Listener> associatedListeners = new List<Listener>();

        private bool isThisObjectUpdated = false;
        private bool textureFillsShape = false;
        private bool isAddedToGame = false;

        protected Timer moveTimer = null;
        protected Vector? moveTarget = null;
        protected double moveSpeed;

        #endregion

        /// <summary>
        /// Tarvitseeko olio päivittämistä. Kun perit oman luokkasi
        /// tästä luokasta, aseta tämä arvoon <c>true</c>, kun luokan olio
        /// tarvitsee päivittämistä.
        /// </summary>
        /// <remarks>
        /// Jos tämä asetetaan arvoon <code>true</code>, olion <c>Update</c>-metodia
        /// kutsutaan säännöllisesti (noin 25 kertaa sekunnissa) sen jälkeen, kun olio
        /// on lisätty peliin.
        /// 
        /// Tämä ei ole oletuksena <c>true</c>, jotta vältetään pelin hidastuminen
        /// turhien päivityskutsujen takia. Jos esimerkiksi kenttä koostuu tuhansista
        /// seinäpalikoista, näiden kaikkien päivittäminen koko ajan olisi turhaa.
        /// </remarks>
        protected bool IsUpdated
        {
            get { return isThisObjectUpdated; }
            set
            {
                if ( value )
                {
                    if ( IsAddedToGame )
                        MakeThisUpdated();
                    else
                        AddedToGame += MakeThisUpdated;
                }
                else
                {
                    if ( IsAddedToGame )
                        Game.RemoveUpdatedObject( this );
                    AddedToGame -= MakeThisUpdated;
                }

                isThisObjectUpdated = value;
            }
        }

        private void MakeThisUpdated()
        {
            Game.AddUpdatedObject( this );
        }

        /// <summary>
        /// Olio, jonka lapsiolio tämä olio on. Jos <c>null</c>, olio ei
        /// ole minkään olion lapsiolio.
        /// </summary>
        public GameObject Parent
        {
            get { return parent; }
            private set { parent = value; }
        }

        /// <summary>
        /// Vapaasti asetettava muuttuja.
        /// </summary>
        /// <example>
        /// Tagia voi käyttää vaikkapa olion keräämien pisteiden tallettamiseen:
        /// <code>
        /// hahmo.Tag = 5;
        /// </code>
        /// Kun luet pisteitä, täytyy arvo muuntaa oikean tyyppiseksi kertomalla tyypin
        /// nimi suluissa:
        /// <code>
        /// int pisteitä = (int)hahmo.Tag;
        /// </code>
        /// </example>
        public object Tag { get; set; }

        /// <summary>
        /// Piirretäänkö oliota ruudulle.
        /// </summary>
        [Save]
        public bool IsVisible { get; set; }

        /// <summary>
        /// Olion aivot.
        /// Ks. <c>Brain</c>
        /// </summary>
        public virtual Brain Brain
        {
            get
            {
                return brain;
            }
            set
            {
                brain.Owner = null;
                brain = ( ( value != null ) ? value : Brain.None );
                brain.Owner = this;
                brain.AddToGameEvent();
                if (value != null)
                    IsUpdated = true;
            }
        }

        /// <summary>
        /// Olion luomisaika.
        /// </summary>
        public TimeSpan CreationTime { get; private set; }

        /// <summary>
        /// Olion elinaika. Lasketaan siitä lähtien, kun olio luodaan.
        /// </summary>
        public TimeSpan Lifetime
        {
            get { return Game.Time.SinceStartOfGame - CreationTime; }
        }

        /// <summary>
        /// Olion suurin mahdollinen elinaika.
        /// Kun <c>Lifetime</c> on suurempi kuin tämä, olio kuolee.
        /// </summary>
        public TimeSpan MaximumLifetime { get; set; }

        /// <summary>
        /// Olion koko pelimaailmassa.
        /// Kertoo olion äärirajat, ei muotoa.
        /// </summary>
        public virtual Vector Size
        {
            get { return size; }
            set
            {
                if ( value.X < 0.0 || value.Y < 0.0 )
                    throw new ArgumentException( "The size must be positive." );

                // TODO: this doesn't work properly.

                Vector oldSize = size;
                Vector newSize = value;
                size = value;

                double xFactor = newSize.X / oldSize.X;
                double yFactor = newSize.Y / oldSize.Y;
                if (childObjects != null)
                {
                    foreach (var o in childObjects)
                    {
                        Vector oldChildSize = o.Size;
                        o.Size = new Vector(oldChildSize.X * xFactor, oldChildSize.Y * yFactor);

                        //                    Vector direction = o.Position.Normalize();
                        //                    double distance = o.Position.Magnitude;
                        Vector oldChildPosition = o.Position;
                        o.Position = new Vector(oldChildPosition.X * xFactor, oldChildPosition.Y * yFactor);
                    }
                }
            }
        }

        /// <summary>
        /// Olion leveys (X-suunnassa, leveimmässä kohdassa).
        /// </summary>
        public double Width
        {
            get { return Size.X; }
            set { Size = new Vector( value, Size.Y ); }
        }

        /// <summary>
        /// Olion korkeus (Y-suunnassa, korkeimmassa kohdassa).
        /// </summary>
        public double Height
        {
            get { return Size.Y; }
            set { Size = new Vector( Size.X, value ); }
        }

        /// <summary>
        /// Olion paikka. Jos olio on jonkun toisen peliolion lapsi, paikka on suhteessa
        /// tämän vanhempaan (<c>Parent</c>). Muuten paikka on paikka pelimaailmassa.
        /// </summary>
        [Save] public virtual Vector Position { get; set; }

        /// <summary>
        /// Olion absoluuttinen paikka pelimaailmassa.
        /// Jos olio ei ole minkään toisen peliolion lapsiolio,
        /// tämä on sama kuin <c>Position</c>.
        /// </summary>
        public Vector AbsolutePosition
        {
            get
            {
                if ( Parent != null )
                    return Parent.AbsolutePosition + this.Position;
                return Position;
            }
        }

        /// <summary>
        /// Olion absoluuttinen kulma pelimaailmassa.
        /// Jos olio ei ole minkään toisen peliolion lapsiolio,
        /// tämä on sama kuin <c>Angle</c>.
        /// </summary>
        public Angle AbsoluteAngle
        {
            get
            {
                if ( Parent != null )
                    return Parent.AbsoluteAngle + this.Angle;
                return Angle;
            }
        }

        /// <summary>
        /// Olion vasemman reunan x-koordinaatti.
        /// </summary>
        public double Left
        {
            get { return Position.X - Size.X / 2; }
            set { Position = new Vector( value + Size.X / 2, Position.Y ); }
        }

        /// <summary>
        /// Olion oikean reunan x-koordinaatti.
        /// </summary>
        public double Right
        {
            get { return Position.X + Size.X / 2; }
            set { Position = new Vector( value - Size.X / 2, Position.Y ); }
        }

        /// <summary>
        /// Olion yläreunan y-koordinaatti.
        /// </summary>
        public double Top
        {
            get { return Position.Y + Size.Y / 2; }
            set { Position = new Vector( Position.X, value - Size.Y / 2 ); }
        }

        /// <summary>
        /// Olion alareunan y-koordinaatti.
        /// </summary>
        public double Bottom
        {
            get { return Position.Y - Size.Y / 2; }
            set { Position = new Vector( Position.X, value + Size.Y / 2 ); }
        }

        /// <summary>
        /// Olion paikan X-koordinaatti.
        /// </summary>
        public double X
        {
            get
            {
                return Position.X;
            }
            set
            {
                Position = new Vector( value, Position.Y );
            }
        }

        /// <summary>
        /// Olion paikan Y-koordinaatti.
        /// </summary>
        public double Y
        {
            get
            {
                return Position.Y;
            }
            set
            {
                Position = new Vector( Position.X, value );
            }
        }

        /// <summary>
        /// Olion kulma tai rintamasuunta.
        /// Nolla = osoittaa oikealle.
        /// </summary>      
        public virtual Angle Angle { get; set; }

        /// <summary>
        /// Animaatio. Voi olla <c>null</c>, jolloin piirretään vain väri.
        /// </summary>
        public Animation Animation
        {
            get { return animation; }
            set { animation = value; }
            }

        /// <summary>
        /// Olion kuva. Voi olla <c>null</c>, jolloin piirretään vain väri.
        /// </summary>
        public Image Image
        {
            get
            {
                if ( Animation != null )
                    return Animation.CurrentFrame;
                return null;
            }
            set
            {
                Animation = value;
            }
        }

        /// <summary>
        /// Määrittää kuinka moneen kertaan kuva piirretään. Esimerkiksi (3.0, 2.0) piirtää
        /// kuvan 3 kertaa vaakasuunnassa ja 2 kertaa pystysuunnassa.
        /// </summary>
        public Vector TextureWrapSize
        {
            get { return textureWrapSize; }
            set { textureWrapSize = value; }
            }

        /// <summary>
        /// Väri, jonka värisenä olio piirretään, jos tekstuuria ei ole määritelty.
        /// </summary>
        public Color Color
        {
            get { return color; }
            set { color = value; }
        }

        /// <summary>
        /// Jos <c>true</c>, kuva piirretään niin, ettei se mene olion muodon
        /// ääriviivojen yli. Toisin sanoen kuva piirretään tarkasti vain
        /// muodon määrittämälle alueelle.
        /// </summary>
        /// <remarks>
        /// Tämän asettaminen tekee olion piirtämisestä hitaamman. Jos
        /// muoto on yksinkertainen, harkitse voisiko kuvan piirtää niin, että
        /// läpinäkyvyyttä käyttämällä saadaan kuvasta halutun muotoinen.
        /// </remarks>
        public bool TextureFillsShape
        {
            get { return textureFillsShape; }
            set { textureFillsShape = value; }
            }

        /// <summary>
        /// Olion muoto.
        /// </summary>
        public virtual Shape Shape
        {
            get { return shape; }
            set  { shape = value; }
        }

        /// <summary>
        /// Olion muoto merkkijonona (kenttäeditorin käyttöön)
        /// </summary>
        internal string ShapeString
        {
            get { return Shape.GetType().Name; }
            set { Shape = Shape.FromString( value ); }
        }

        /// <summary>
        /// Peli, johon olio on lisätty. <c>null</c>, jos
        /// oliota ei ole lisätty peliin.
        /// </summary>
        public Game Game
        {
            get { return isAddedToGame ? Game.Instance : null; }
        }

        /// <summary>
        /// Onko olio lisätty peliin.
        /// </summary>
        public bool IsAddedToGame
        {
            get { return isAddedToGame; }
            internal set
            {
                isAddedToGame = value;
                if ( childObjects != null )
                {
                    foreach ( var o in childObjects )
                        o.isAddedToGame = value;
                }
            }
        }

        /// <summary>
        /// Tapahtuu, kun olio lisätään peliin.
        /// </summary>
        public event Action AddedToGame;


        internal void OnAddedToGame()
        {
            if ( AddedToGame != null )
                AddedToGame();
            brain.AddToGameEvent();
        }

        /// <summary>
        /// Tapahtuu, kun on saavuttu haluttuun paikkaan (MoveTo-metodi)
        /// </summary>
        public event Action<GameObject, Vector> ArrivedAt;

        protected void OnArrivedAt( Vector location )
        {
            if ( ArrivedAt != null )
                ArrivedAt( this, location );
        }

        #region Destroyable

        /// <summary>
        /// Onko olio tuhottu.
        /// </summary>
        public bool IsDestroyed
        {
            get { return Lifetime > MaximumLifetime; }
        }

        /// <summary> 
        /// Tapahtuu, kun olio tuhotaan. 
        /// </summary> 
        public event Action Destroyed;

        private void OnDestroyed()
        {
            if ( Destroyed != null )
                Destroyed();
        }

        /// <summary>
        /// Tuhoaa olion. Tuhottu olio poistuu pelistä.
        /// </summary>
        public virtual void Destroy()
        {
            this.MaximumLifetime = new TimeSpan( 0 );
            OnDestroyed();

            if ( childObjects != null )
            {
                foreach ( GameObject child in childObjects )
                {
                    child.Destroy();
                }
            }

            if ( associatedListeners != null )
            {
                foreach ( Listener listener in associatedListeners )
                {
                    listener.Destroy();
                }
            }
        }

        #endregion

        /// <summary>
        /// Alustaa uuden peliolion.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        public GameObject( double width, double height )
            : this( width, height, Shape.Rectangle )
        {
        }

        /// <summary>
        /// Alustaa uuden peliolion.
        /// Kappaleen koko ja ulkonäkö ladataan parametrina annetusta kuvasta.
        /// </summary>
        /// <param name="image">Kuva</param>
        public GameObject(Image image)
            : this( image.Width, image.Height, Shape.Rectangle )
        {
            this.Image = image;
        }

        /// <summary>
        /// Alustaa uuden peliolion.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        /// <param name="shape">Muoto.</param>
        public GameObject( double width, double height, Shape shape )
        {
            this.CreationTime = Game.Time.SinceStartOfGame;
            this.MaximumLifetime = TimeSpan.MaxValue;
            this.size = new Vector( width, height );
            this.shape = shape;
            this.IsVisible = true;
            this.Tag = "";
        }

#if !XBOX && !WINDOWS_PHONE
        /// <summary>
        /// Tekee oliosta kopion.
        /// </summary>
        /// <returns>
        /// Uusi <see cref="GameObject"/>-olio samoilla arvoilla.
        /// </returns>
        public object Clone()
        {
            MemoryStream ms = new MemoryStream();
            BinaryFormatter bf = new BinaryFormatter();

            bf.Serialize( ms, this );
            ms.Position = 0;

            object obj = bf.Deserialize( ms );
            ms.Close();

            // Copy nonserializable data
            ( (GameObject)obj ).Animation = new Animation( this.Animation );

            return obj;
        }
#endif

        /// <summary>
        /// Onko piste <c>p</c> tämän olion sisäpuolella.
        /// </summary>
        public bool IsInside(Vector point)
        {
            // TODO: onko lapsiolioiden sisällä?

            Vector p = this.AbsolutePosition;

            return point.X >= (p.X - Width / 2)
                && point.X <= (p.X + Width / 2)
                && point.Y >= (p.Y - Height / 2)
                && point.Y <= (p.Y + Height / 2);
        }

        /// <summary>
        /// Lisää annetun peliolion tämän olion lapseksi. Lapsiolio liikkuu tämän olion mukana,
        /// ja sen paikka ja koko ilmaistaan suhteessa tähän olioon.
        /// </summary>
        /// <remarks>
        /// <c>PhysicsObject</c>-tyyppisiä olioita ei voi lisätä lapsiolioksi.
        /// </remarks>
        public void Add( GameObject childObject )
        {
            if (childObjects == null)
            {
                childObjects = new List<GameObject>();
                childObjectsToRemove = new Queue<GameObject>();
            }

            if ( childObjects.Contains( childObject ) )
                throw new ArgumentException( "The child object has already been added" );
            if ( childObject is PhysicsObject )
                throw new NotImplementedException( "Having a PhysicsObject as a child object is not supported." );

            childObjects.Add( childObject );
            childObject.Parent = this;

            if ( IsAddedToGame )
            {
                childObject.IsAddedToGame = true;
                childObject.OnAddedToGame();
            }
            else
                AddedToGame += childObject.AddedToGame;


            // Let's keep it simple and assume that child objects need updating,
            // even if that is not the case.
            if ( !IsUpdated )
                IsUpdated = true;
        }

        /// <summary> 
        /// Poistaa lapsiolion. Jos haluat tuhota olion, 
        /// kutsu mielummin olion <c>Destroy</c>-metodia. 
        /// </summary> 
        /// <remarks> 
        /// Oliota ei poisteta välittömästi, vaan viimeistään seuraavan 
        /// päivityksen jälkeen. 
        /// </remarks> 
        public void Remove( GameObject childObject )
        {
            childObjectsToRemove.Enqueue( childObject );
        }

        /// <summary>

        /// Peliolion päivitys. Tätä kutsutaan, kun <c>IsUpdated</c>-ominaisuuden
        /// arvoksi on asetettu <c>true</c> ja olio on lisätty peliin.
        /// <see cref="IsUpdated"/>
        /// </summary>
        /// <param name="time">Peliaika.</param>
        [EditorBrowsable( EditorBrowsableState.Never )]
        public virtual void Update( Time time )
        {
            Brain.DoUpdate( time );

            if (childObjects != null)
            {
                foreach (GameObject child in childObjects)
                {
                    child.Update(time);
                    if (child.IsDestroyed)
                    {
                        childObjectsToRemove.Enqueue(child);
                    }
                }

                while (childObjectsToRemove.Count > 0)
                {
                    GameObject o = childObjectsToRemove.Dequeue();
                    childObjects.Remove(o);
                    this.AddedToGame -= o.AddedToGame;
                    o.IsAddedToGame = false;
                }
            }
        }

        /// <summary>
        /// Siirtää oliota.
        /// </summary>
        /// <param name="movement">Vektori, joka määrittää kuinka paljon siirretään.</param>
        public virtual void Move( Vector movement )
        {
            Position += movement;
        }

        /// <summary>
        /// Yrittää siirtyä annettuun paikkaan annetulla nopeudella.
        /// Laukaisee tapahtuman ArrivedAt, kun paikkaan on päästy.
        /// </summary>
        /// <param name="location">Paikka johon siirrytään</param>
        /// <param name="speed">
        /// Nopeus (paikkayksikköä sekunnissa) jolla liikutaan.
        /// Nopeus on maksiminopeus. Jos välissä on hitaampaa maastoa tai
        /// esteitä, liikkumisnopeus voi olla alle sen.
        /// </param>
        public virtual void MoveTo( Vector location, double speed )
        {
            if ( moveTimer == null )
            {
                moveTimer = new Timer();
                moveTimer.Timeout += MoveToTarget;
                moveTimer.Interval = 0.01;
            }
            else if ( moveTimer.Enabled )
                moveTimer.Stop();

            moveSpeed = speed;
            moveTarget = location;
            moveTimer.Start();
        }

        protected virtual void MoveToTarget()
        {
            if ( !moveTarget.HasValue )
            {
                moveTimer.Stop();
                return;
            }

            Vector d = moveTarget.Value - Position;
            double vt = moveSpeed * moveTimer.Interval;

            if ( d.Magnitude < vt )
            {
                Vector targetLoc = moveTarget.Value;
                moveTimer.Stop();
                Position = moveTarget.Value;
                moveTarget = null;
                OnArrivedAt( targetLoc );
            }
            else
                //Position += d.Normalize() * moveSpeed * sender.Interval;
                Position += Vector.FromLengthAndAngle( vt, d.Angle );
        }
    }
}
