﻿using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;
using System;
using System.IO;
using System.Collections.Generic;
using System.Diagnostics;

using XnaRectangle = Microsoft.Xna.Framework.Rectangle;
using XnaV2 = Microsoft.Xna.Framework.Vector2;
using XnaColor = Microsoft.Xna.Framework.Color;
using System.ComponentModel;

namespace Jypeli
{
    /// <summary>
    /// Kuva.
    /// </summary>
    public class Image
    {
        private string assetName;
        private Texture2D xnaTexture;

        internal Texture2D XNATexture
        {
            get { DoLoad(); return xnaTexture; }
        }

        /// <summary>
        /// Leveys pikseleinä.
        /// </summary>
        public int Width
        {
            get { DoLoad(); return XNATexture.Width; }
        }

        /// <summary>
        /// Korkeus pikseleinä.
        /// </summary>
        public int Height
        {
            get { DoLoad(); return XNATexture.Height; }
        }

        /// <summary>
        /// Nimi.
        /// </summary>
        public string Name
        {
            get { DoLoad(); return xnaTexture.Name; }
        }

        internal Image( int width, int height )
        {
            assetName = null;
            xnaTexture = new Texture2D( Game.GraphicsDevice, width, height );
        }

        internal Image( string assetName )
        {
            this.assetName = assetName;
            this.xnaTexture = null;
        }

        [EditorBrowsable(EditorBrowsableState.Never)]
        public Image( Microsoft.Xna.Framework.Graphics.Texture2D texture )
        {
            assetName = null;
            this.xnaTexture = texture;
        }

        private void DoLoad()
        {
            // Some duct tape around the fact that in XNA,
            // content can not be loaded before LoadContent().
            if ( xnaTexture == null )
            {
                Debug.Assert( assetName != null );
                xnaTexture = Game.Instance.Content.Load<Texture2D>( assetName );
            }
        }

        #region static methods

        /// <summary>
        /// Lataa kuvan tiedostosta. Kuvan ei tarvitse olla lisättynä
        /// Content-projektiin.
        /// </summary>
        /// <param name="path">Tiedoston polku.</param>
#if !XBOX
        public static Image FromFile( string path )
        {
            StreamReader sr = new StreamReader(path);
            Image img = new Image( Texture2D.FromStream( Game.GraphicsDevice, sr.BaseStream) );
            return img;
        }
#endif
        
        /// <summary>
        /// Luo tähtitaivaskuvan.
        /// </summary>
        /// <param name="width">Tekstuurin leveys.</param>
        /// <param name="height">Tekstuurin korkeus.</param>
        /// <param name="stars">Tähtien määrä.</param>
        /// <returns>Tekstuuri.</returns>
        public static Image CreateStarSky( int width, int height, int stars )
        {
            XnaColor[] textureColors = new XnaColor[width * height];

            // Black background
            int i = 0;
            for ( int ver = 0; ver < height; ver++ )
            {
                for ( int hor = 0; hor < width; hor++ )
                {
                    textureColors[i++] = XnaColor.Black;
                }
            }

            // Random stars
            for ( int j = 0; j < stars; j++ )
            {
                int star = RandomGen.NextInt( 0, width * height );
                int size = RandomGen.NextInt( 1, 5 );

                for ( int k = 0; k < size / 2; k++ )
                {
                    XnaColor starcolor = RandomGen.NextColor( Jypeli.Color.White, new Color( 192, 192, 192, 255 ) ).AsXnaColor();

                    textureColors[star + k] = starcolor;

                    if ( size % 2 != 0 || size == 2 )
                        continue;

                    int nextStar = star + k + width;

                    if ( nextStar < ( width * height ) )
                    {
                        textureColors[nextStar] = starcolor;
                    }
                }
            }

            //Texture2D newTexture = new Texture2D( Game.GraphicsDevice, width, height, 1, TextureUsage.None, SurfaceFormat.Color );
            Texture2D newTexture = new Texture2D(Game.GraphicsDevice, width, height, false, SurfaceFormat.Color);
            newTexture.SetData<XnaColor>( textureColors );

            return new Image( newTexture );
        }

        /// <summary>
        /// Luo kuvan tekstistä.
        /// </summary>
        /// <param name="text">Teksti josta kuva luodaan</param>
        /// <param name="font">Fontti</param>
        /// <param name="textColor">Tekstin väri</param>
        /// <param name="backgroundColor">Tekstin taustaväri</param>
        /// <returns>Teksti kuvana</returns>
        public static Image FromText( string text, Font font, Color textColor, Color backgroundColor )
        {
            if ( text == null )
                text = "";

            var spriteBatch = new SpriteBatch( Game.GraphicsDevice );
            var device = spriteBatch.GraphicsDevice;

            XnaV2 textDims = font.XnaFont.MeasureString( text );
            int textw = ( textDims.X > 1 ) ? Convert.ToInt32( textDims.X ) : 1;
            int texth = ( textDims.Y > 1 ) ? Convert.ToInt32( textDims.Y ) : 1;

            //RenderTarget2D rt = new RenderTarget2D( device, textw, texth, 1, device.DisplayMode.Format );
            RenderTarget2D rt = new RenderTarget2D( device, textw, texth, false, SurfaceFormat.Color, DepthFormat.Depth24Stencil8 );

            //device.SetRenderTarget( 0, rt );
            device.SetRenderTarget( rt );
            device.Clear( ClearOptions.Target | ClearOptions.DepthBuffer, backgroundColor.AsXnaColor(), 1.0f, 0 );

            spriteBatch.Begin();
            spriteBatch.DrawString( font.XnaFont, text, XnaV2.Zero, textColor.AsXnaColor() );
            spriteBatch.End();

            //device.SetRenderTarget( 0, null );
            device.SetRenderTarget( null );

            //return new Image( rt.GetTexture() );
            return new Image( rt );
        }

        /// <summary>
        /// Piirtää tekstiä kuvan päälle.
        /// </summary>
        /// <param name="img">Kuva jonka päälle piirretään</param>
        /// <param name="text">Piirrettävä teksti</param>
        /// <param name="position">Piirtokohta (origo keskellä kuvaa)</param>
        /// <param name="font">Fontti</param>
        /// <param name="textColor">Tekstin väri</param>
        /// <param name="backgroundColor">Tekstin taustaväri</param>
        /// <returns>Kuva tekstin kanssa</returns>
        public static Image DrawTextOnImage( Image img, string text, Vector position, Font font, Color textColor, Color backgroundColor )
        {
            if ( text == null )
                text = "";

            var spriteBatch = new SpriteBatch( Game.GraphicsDevice );
            var device = spriteBatch.GraphicsDevice;

            XnaV2 textDims = font.XnaFont.MeasureString( text );
            int textw = ( textDims.X > 1 ) ? Convert.ToInt32( textDims.X ) : 1;
            int texth = ( textDims.Y > 1 ) ? Convert.ToInt32( textDims.Y ) : 1;

            //RenderTarget2D rt = new RenderTarget2D( device, textw, texth, 1, device.DisplayMode.Format );
            RenderTarget2D rt = new RenderTarget2D( device, img.Width, img.Height, false, SurfaceFormat.Color, DepthFormat.Depth24Stencil8 );

            //device.SetRenderTarget( 0, rt );
            device.SetRenderTarget(rt);
            device.Clear( ClearOptions.Target | ClearOptions.DepthBuffer, backgroundColor.AsXnaColor(), 1.0f, 0 );

            float xpos = 0.5f * ( img.Width - textw ) + (float)position.X;
            float ypos = 0.5f * ( img.Height - texth ) - (float)position.Y;

            spriteBatch.Begin();
            spriteBatch.Draw( img.XNATexture, rt.Bounds, XnaColor.White );
            spriteBatch.DrawString( font.XnaFont, text, new XnaV2( xpos, ypos ), textColor.AsXnaColor() );
            spriteBatch.End();

            //device.SetRenderTarget( 0, null );
            device.SetRenderTarget(null);
            

            //return new Image( rt.GetTexture() );
            return new Image(rt);
        }

        /// <summary>
        /// Piirtää tekstiä kuvan päälle keskelle kuvaa.
        /// </summary>
        /// <param name="img">Kuva jonka päälle piirretään</param>
        /// <param name="text">Piirrettävä teksti</param>
        /// <param name="font">Fontti</param>
        /// <param name="textColor">Tekstin väri</param>
        /// <returns>Kuva tekstin kanssa</returns>
        public static Image DrawTextOnImage( Image img, string text, Font font, Color textColor )
        {
            return DrawTextOnImage( img, text, Vector.Zero, font, textColor, Jypeli.Color.Transparent );
        }

        /// <summary>
        /// Luo pystysuuntaisen liukuväritetyn kuvan.
        /// </summary>
        /// <param name="imageWidth">kuvan leveys.</param>
        /// <param name="imageHeight">kuvan korkeus.</param>
        /// <param name="lowerColor">Alareunassa käytettävä väri.</param>
        /// <param name="upperColor">Yläreunassa käytettävä väri.</param>
        /// <returns>Väritetty kuva.</returns>
        public static Image FromGradient( int imageWidth, int imageHeight, Color lowerColor, Color upperColor )
        {
            XnaColor lower = lowerColor.AsXnaColor();
            XnaColor upper = upperColor.AsXnaColor();
            XnaColor[] textureColors = new XnaColor[imageWidth * imageHeight];
            int i = 0;

            for ( int ver = 0; ver < imageHeight; ver++ )
            {
                for ( int hor = 0; hor < imageWidth; hor++ )
                {
                    textureColors[i++] = XnaColor.Lerp( upper, lower, ( (float)ver / (float)imageHeight ) );
                }
            }

            //Texture2D newTexture = new Texture2D( Game.GraphicsDevice, imageWidth, imageHeight, 1, TextureUsage.None, SurfaceFormat.Color );
            Texture2D newTexture = new Texture2D(Game.GraphicsDevice, imageWidth, imageHeight, false, SurfaceFormat.Color);

            newTexture.SetData<XnaColor>( textureColors );

            return new Image( newTexture );
        }

        /// <summary>
        /// Luo yksivärisen kuvan.
        /// </summary>
        /// <param name="imageWidth">Kuvan leveys.</param>
        /// <param name="imageHeight">Kuvan korkeus.</param>
        /// <param name="color">Kuvan väri.</param>
        /// <returns>Väritetty kuva.</returns>
        public static Image FromColor(int imageWidth, int imageHeight, Color color)
        {
            return Image.FromGradient(imageWidth, imageHeight, color, color);
        }

        private static XnaColor[] MirrorLine( XnaColor[] scanline, int width )
        {
            XnaColor[] res = new XnaColor[width];
            int l = 0;
            int r = width - 1;

            while ( l < r )
            {
                res[l] = scanline[r];
                res[r] = scanline[l];
                l++; r--;
            }

            if ( l == r )
            {
                // Center pixel
                res[l] = scanline[l];
            }

            return res;
        }

        /// <summary>
        /// Peilaa kuvan X-suunnassa.
        /// </summary>
        /// <param name="image">Peilattava kuva.</param>
        /// <returns>Peilattu kuva.</returns>
        public static Image Mirror( Image image )
        {
            Texture2D newTex = new Texture2D( image.XNATexture.GraphicsDevice, image.Width, image.Height, false, image.XNATexture.Format );
            XnaColor[] scanline = new XnaColor[image.Width];
            var scanRect = new XnaRectangle( 0, 0, image.Width, 1 );

            for ( scanRect.Y = 0; scanRect.Y < image.Height; scanRect.Y++ )
            {
                image.XNATexture.GetData<XnaColor>( 0, scanRect, scanline, 0, image.Width );
                scanline = MirrorLine( scanline, image.Width );
                newTex.SetData<XnaColor>( 0, scanRect, scanline, 0, image.Width );
            }

            return new Image( newTex );
        }

        /// <summary>
        /// Peilaa kuvat X-suunnassa.
        /// </summary>
        /// <param name="images">Peilattavat kuvat.</param>
        /// <returns>Peilatut kuvat.</returns>
        public static Image[] Mirror( Image[] images )
        {
            Image[] result = new Image[images.Length];
            for ( int i = 0; i < images.Length; i++ )
                result[i] = Mirror( images[i] );
            return result;
        }

        /// <summary>
        /// Peilaa kuvan Y-suunnassa.
        /// </summary>
        /// <param name="image">Peilattava kuva.</param>
        /// <returns>Peilattu kuva.</returns>
        public static Image Flip( Image image )
        {
            Texture2D newTex = new Texture2D( image.XNATexture.GraphicsDevice, image.Width, image.Height, false, image.XNATexture.Format );
            XnaColor[] scanlineUpper = new XnaColor[image.Width];
            XnaColor[] scanlineLower = new XnaColor[image.Width];
            var scanRectUpper = new XnaRectangle( 0, 0, image.Width, 1 );
            var scanRectLower = new XnaRectangle( 0, 0, image.Width, 1 );

            for ( int i = 0; i < image.Height / 2; i++ )
            {
                scanRectUpper.Y = i;
                scanRectLower.Y = image.Height - 1 - i;

                image.XNATexture.GetData<XnaColor>( 0, scanRectUpper, scanlineUpper, 0, image.Width );
                image.XNATexture.GetData<XnaColor>( 0, scanRectLower, scanlineLower, 0, image.Width );

                newTex.SetData<XnaColor>( 0, scanRectUpper, scanlineLower, 0, image.Width );
                newTex.SetData<XnaColor>( 0, scanRectLower, scanlineUpper, 0, image.Width );
            }

            if ( image.Height % 2 == 1 )
            {
                // Center line
                scanRectUpper.Y = image.Height / 2;
                image.XNATexture.GetData<XnaColor>( 0, scanRectUpper, scanlineUpper, 0, image.Width );
                newTex.SetData<XnaColor>( 0, scanRectUpper, scanlineUpper, 0, image.Width );
            }

            return new Image( newTex );
        }

        /// <summary>
        /// Peilaa kuvat Y-suunnassa.
        /// </summary>
        /// <param name="images">Peilattavat kuvat.</param>
        /// <returns>Peilatut kuvat.</returns>
        public static Image[] Flip( Image[] images )
        {
            Image[] result = new Image[images.Length];
            for ( int i = 0; i < images.Length; i++ )
                result[i] = Flip( images[i] );
            return result;
        }

        /// <summary>
        /// Värittää kuvan.
        /// </summary>
        /// <param name="image">Väritettävä kuva.</param>
        /// <param name="color">Väri, jolla väritetään.</param>
        /// <returns>Väritetty kuva.</returns>
        public static Image Color( Image image, Color color )
        {
            Texture2D newTex = new Texture2D( image.XNATexture.GraphicsDevice, image.Width, image.Height, false, image.XNATexture.Format );
            XnaColor[] scanline = new XnaColor[image.Width];
            var scanRect = new XnaRectangle( 0, 0, image.Width, 1 );
            XnaColor xnaColor = color.AsXnaColor();

            for ( scanRect.Y = 0; scanRect.Y < image.Height; scanRect.Y++ )
            {
                image.XNATexture.GetData<XnaColor>( 0, scanRect, scanline, 0, image.Width );

                for ( int i = 0; i < image.Width; i++ )
                {
                    if ( scanline[i].A < 255 )
                    {
                        scanline[i].R = (byte)( ( 255 - scanline[i].A ) * xnaColor.R + scanline[i].A * scanline[i].R );
                        scanline[i].G = (byte)( ( 255 - scanline[i].A ) * xnaColor.G + scanline[i].A * scanline[i].G );
                        scanline[i].B = (byte)( ( 255 - scanline[i].A ) * xnaColor.B + scanline[i].A * scanline[i].B );

                        if ( scanline[i].A > 10 )
                        {
                            scanline[i].A = color.AlphaComponent;
                        }
                    }
                }

                newTex.SetData<XnaColor>( 0, scanRect, scanline, 0, image.Width);
            }

            return new Image( newTex );
        }

        /// <summary>
        /// Värittää kuvat.
        /// </summary>
        /// <param name="images">Väritettävät kuvat.</param>
        /// <param name="color">Väri, jolla väritetään.</param>
        /// <returns>Väritetyt kuvat.</returns>
        public static Image[] Color( Image[] images, Color color )
        {
            Image[] result = new Image[images.Length];
            for ( int i = 0; i < images.Length; i++ )
                result[i] = Color( images[i], color );
            return result;
        }

        public static Image Color(Image image, byte alpha)
        {
            Texture2D newTex = new Texture2D(image.XNATexture.GraphicsDevice, image.Width, image.Height, false, image.XNATexture.Format);
            XnaColor[] scanline = new XnaColor[image.Width];
            var scanRect = new XnaRectangle(0, 0, image.Width, 1);

            for (scanRect.Y = 0; scanRect.Y < image.Height; scanRect.Y++)
            {
                image.XNATexture.GetData<XnaColor>(0, scanRect, scanline, 0, image.Width);
                for (int i = 0; i < image.Width; i++)
                {
                    scanline[i].A = alpha;
                }
                newTex.SetData<XnaColor>(0, scanRect, scanline, 0, image.Width);
            }
            return new Image(newTex);
        }

        #endregion
    }
}
