﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Jypeli
{
    /// <summary>
    /// Rakenne, joka pitää fysiikkaoliot kiinteän matkan päässä toisistaan.
    /// </summary>
    public class PhysicsStructure : Destroyable
    {
        private double _softness = 0;

        /// <summary>
        /// Onko rakenne lisätty peliin.
        /// </summary>
        public bool IsAddedToGame { get; internal set; }

        /// <summary>
        /// Rakenteeseen kuuluvat oliot.
        /// </summary>
        public List<PhysicsObject> Objects { get; private set; }

        /// <summary>
        /// Rakenteeseen kuuluvat liitokset.
        /// </summary>
        public List<AxleJoint> Joints { get; private set; }

        /// <summary>
        /// Olioiden välisten liitosten pehmeys.
        /// </summary>
        public double Softness
        {
            get { return _softness; }
            set
            {
                _softness = value;

                for ( int i = 0; i < Joints.Count; i++ )
                {
                    Joints[i].Softness = value;
                }
            }
        }

        /// <summary>
        /// Luo uuden tyhjän rakenteen.
        /// </summary>
        public PhysicsStructure()
        {
            Objects = new List<PhysicsObject>();
            Joints = new List<AxleJoint>();
        }

        /// <summary>
        /// Luo uuden rakenteen ja varustaa sen fysiikkaolioilla.
        /// </summary>
        /// <param name="objs">Fysiikkaoliot</param>
        public PhysicsStructure( params PhysicsObject[] objs )
            : this()
        {
            for ( int i = 0; i < objs.Length; i++ )
            {
                Add( objs[i] );
            }
        }

        /// <summary>
        /// Lisää fysiikkaolion rakenteeseen.
        /// </summary>
        /// <param name="obj">Lisättävä olio</param>
        public void Add( PhysicsObject obj )
        {
            if ( IsAddedToGame )
                Game.Instance.Add( obj );

            foreach ( var existing in Objects )
            {
                AxleJoint joint = new AxleJoint( obj, existing );
                joint.Softness = _softness;
                Joints.Add( joint );

                if ( IsAddedToGame )
                    ( (PhysicsGame)Game.Instance ).Add( joint );
            }

            Objects.Add( obj );
        }

        #region Destroyable

        /// <summary>
        /// Onko olio tuhottu
        /// </summary>
        public bool IsDestroyed { get; private set; }

        /// <summary>
        /// Tapahtuu kun olio tuhotaan
        /// </summary>
        public event Action Destroyed;

        /// <summary>
        /// Tuhoaa olion
        /// </summary>
        public void Destroy()
        {
            foreach ( var joint in Joints ) joint.Destroy();
            foreach ( var obj in Objects ) obj.Destroy();
            IsDestroyed = true;
        }

        #endregion
    }
}
