﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Jypeli.Effects;

using XnaRectangle = Microsoft.Xna.Framework.Rectangle;
using XnaColor = Microsoft.Xna.Framework.Color;

namespace Jypeli
{
    /// <summary>
    /// Piirtojärjestys.
    /// </summary>
    public enum DrawOrder
    {
        /// <summary>
        /// Piirtojärjestyksellä ei ole väliä. Oliot piirretään mahdollisimman tehokkaassa järjestyksessä.
        /// Tämä on oletus.
        /// </summary>
        Irrelevant,

        /// <summary>
        /// Oliot piirretään siinä järjestyksessä missä ne on lisätty peliin.
        /// </summary>
        FirstToLast,
    }

    
    /// <summary>
    /// Kerros. Vastaa olioiden piirtämisestä.
    /// </summary>
    public class Layer
    {
        /// <summary>
        /// Vertices for drawing a filled square.
        /// </summary>
        static readonly Vector[] squareVertices =
        {
            new Vector( -0.5, -0.5 ),
            new Vector( -0.5, 0.5 ),
            new Vector( 0.5, -0.5 ),
            new Vector( 0.5, 0.5 )
        };

        /// <summary>
        /// Indices for the vertex array of the square.
        /// </summary>
        static readonly Int16[] squareIndices =
        {
            0, 1, 2,
            2, 1, 3
        };

        /// <summary>
        /// Vertices for drawing a filled triangle.
        /// </summary>
        static readonly Vector[] triangleVertices =
        {
            new Vector( 0f, 0.5f ),
            new Vector( 0.5f, -0.5f ),
            new Vector( -0.5f, -0.5f ),
        };

        /// <summary>
        /// Indices for the vertex array of the triangle.
        /// </summary>
        static readonly Int16[] triangleIndices = { 0, 1, 2 };

        /// <summary>
        /// The maximum width of a texture. 2048 is the maximum that is supported
        /// in the reach profile. However, we don't use the full value
        /// so that one big texture does not fill the entire cache.
        /// </summary>
        private const int MaxTextureWidth = 2048 / 2;

        /// <summary>
        /// <see cref="MaxTextureWidth"/>
        /// </summary>
        private const int MaxTextureHeight = 2048 / 2;

        static readonly TextureCoordinates defaultCoords = new TextureCoordinates()
        {
            TopLeft = new Vector2(0.0f, 0.0f),
            TopRight = new Vector2(1.0f, 0.0f),
            BottomLeft = new Vector2(0.0f, 1.0f),
            BottomRight = new Vector2(1.0f, 1.0f),
        };

        internal Texture2D cachedTexture;
        private Dictionary<Image, TextureCoordinates> textureCoordinates = new Dictionary<Image, TextureCoordinates>();

        private List<GameObject> objects = new List<GameObject>();
        private Queue<GameObject> destroyedObjects = new Queue<GameObject>();
        private List<ParticleSystem> effects = new List<ParticleSystem>();
        private Queue<ParticleSystem> destroyedEffects = new Queue<ParticleSystem>();

        private Vector _relativeTransition = new Vector(1, 1);

        /// <summary>
        /// Olioiden piirtojärjestys.
        /// </summary>
        public DrawOrder DrawOrder { get; set; }

#if DEBUG
        [EditorBrowsable(EditorBrowsableState.Never)]
        internal DoubleMeter ObjectCount = new DoubleMeter(0.0) { MaxValue = 10000 };
#endif

        /// <summary>
        /// Kuinka paljon tämän kerroksen olioiden paikka muuttuu kameran siirtyessä suhteessa muihin kerroksiin.
        /// Esimerkiksi arvo 0.5 aiheuttaa siirtymän joka on puolet normaalista.
        /// </summary>
        public Vector RelativeTransition
        {
            get { return _relativeTransition; }
            set { _relativeTransition = value; }
        }

        /// <summary>
        /// Ruudukko. Ruudukko piirretään asettamalla tähän <c>Grid</c>-olio.
        /// Jos <c>null</c>, ruudukkoa ei piirretä.
        /// </summary>
        public Grid Grid { get; set; }


        internal Layer()
        {
            DrawOrder = DrawOrder.Irrelevant;
        }

        internal void AddEffect(ParticleSystem p)
        {
            effects.Add(p);
        }

        internal void Add(GameObject o)
        {
            if (o.Image != null)
            {
                AddNewImages(o.Animation.frames);
            }

            objects.Add(o);
        }

        /// <summary>
        /// Adds new images to the cache.
        /// </summary>
        /// <returns><c>true</c> if images were successfully added or they were already in the cache.</returns>
        private bool AddNewImages(Image[] images)
        {
            List<Image> newImages = new List<Image>();

            foreach (Image img in images)
            {
                if (img.Width > MaxTextureWidth || img.Height > MaxTextureHeight)
                    return false;

                if (!textureCoordinates.ContainsKey(img))
                    newImages.Add(img);
            }

            if (newImages.Count > 0)
            {
                if (!ComposeCachedTexture(newImages))
                    return false;
            }

            return true;
        }

        /// <summary>
        /// Composes individual textures into one big texture, that can
        /// then be sent to the graphics card in one batch.
        /// </summary>
        /// <remarks>
        /// This method also sets texture coordinates for each image. The texture
        /// coordinates refer to a small part of the image.
        /// </remarks>
        private bool ComposeCachedTexture(List<Image> newImages)
        {
            Image[] images = new Image[textureCoordinates.Keys.Count + newImages.Count];
            int maxWidth = 0;
            int maxHeight = 0;
            int i = 0;

            foreach (Image img in textureCoordinates.Keys)
            {
                images[i++] = img;
                if (img.Width > maxWidth)
                    maxWidth = img.Width;
                if (img.Height > maxHeight)
                    maxHeight = img.Height;
            }
            foreach (Image img in newImages)
            {
                images[i++] = img;
                if (img.Width > maxWidth)
                    maxWidth = img.Width;
                if (img.Height > maxHeight)
                    maxHeight = img.Height;
            }

            Comparison<Image> compareHeight = delegate(Image a, Image b) { return b.Height - a.Height; };
            Array.Sort(images, compareHeight);

            Texture2D[] textures = new Texture2D[images.Length];
            for (i = 0; i < images.Length; i++)
                textures[i] = images[i].XNATexture;

            int totalWidth;
            XnaRectangle[] locations = CalculateLocations(textures, out totalWidth);
            if (locations == null)
                return false;

            cachedTexture = TileTextures(textures, locations, totalWidth, textures[0].Height, maxWidth, maxHeight);

            SetTextureCoordinates(images, locations);
            return true;
        }

        /// <summary>
        /// Calculates locations for individual images in the composed image, so that
        /// as little space is wasted as possible.
        /// </summary>
        /// <param name="textures">Array of textures sorted by height.</param>
        /// <param name="totalWidth">The width for the composed texture.</param>
        /// <returns>Locations for individual textures, or null if there is not enough space.</returns>
        private static XnaRectangle[] CalculateLocations(Texture2D[] textures, out int totalWidth)
        {
            XnaRectangle[] locations = new XnaRectangle[textures.Length];
            List<XnaRectangle> openLocations = new List<XnaRectangle>();
            int rowHeight = textures[0].Height;
            int top = 0;
            int left = 0;

            for (int i = 0; i < textures.Length; i++)
            {
                XnaRectangle rect;
                int foundIndex;

                if (FindFittingLocation(textures[i], openLocations, out foundIndex))
                {
                    XnaRectangle openRect = openLocations[foundIndex];
                    rect = new XnaRectangle(
                        openRect.X,
                        openRect.Y,
                        textures[i].Width,
                        textures[i].Height);

                    openLocations.RemoveAt(foundIndex);

                    if (rect.Height < openRect.Height)
                    {
                        XnaRectangle leftOverSpace = new XnaRectangle(
                            openRect.X,
                            openRect.Y + rect.Height,
                            openRect.Width,
                            openRect.Height - rect.Height);

                        openLocations.Add(leftOverSpace);
                    }
                    if (rect.Width < openRect.Width)
                    {
                        XnaRectangle leftOverSpace = new XnaRectangle(
                            openRect.X + rect.Width,
                            openRect.Y,
                            openRect.Width - rect.Width,
                            rect.Height);

                        openLocations.Add(leftOverSpace);
                    }
                }
                else
                {
                    rect = new XnaRectangle(left, top, textures[i].Width, textures[i].Height);

                    if (rect.Height < rowHeight)
                    {
                        XnaRectangle leftOverSpace = new XnaRectangle(
                            left,
                            top + rect.Height,
                            rect.Width,
                            rowHeight - rect.Height);

                        openLocations.Add(leftOverSpace);
                    }

                    if ((left + textures[i].Width) > MaxTextureWidth)
                    {
                        totalWidth = 0;
                        return null;
                    }
                    else
                    {
                        left += textures[i].Width;
                    }
                }


                locations[i] = rect;
            }

            totalWidth = left;
            return locations;
        }

        private static bool FindFittingLocation(Texture2D tex, List<XnaRectangle> openLocations, out int foundIndex)
        {
            for (int i = 0; i < openLocations.Count; i++)
            {
                XnaRectangle r = openLocations[i];

                if (tex.Width <= r.Width && tex.Height <= r.Height)
                {
                    foundIndex = i;
                    return true;
                }
            }

            foundIndex = -1;
            return false;
        }

        private static Texture2D TileTextures(Texture2D[] textures, XnaRectangle[] locations, int totalWidth, int totalHeight, int maxWidth, int maxHeight)
        {
            Debug.Assert(textures.Length == locations.Length);

            Texture2D composite = new Texture2D(textures[0].GraphicsDevice, totalWidth, totalHeight, false, textures[0].Format);
            XnaColor[] buffer = new XnaColor[maxWidth * maxHeight];

            for (int i = 0; i < textures.Length; i++)
            {
                int dataLength = textures[i].Width * textures[i].Height;
                XnaRectangle srcRect = new XnaRectangle(0, 0, textures[i].Width, textures[i].Height);
                XnaRectangle destRect = locations[i];

                textures[i].GetData<XnaColor>(0, srcRect, buffer, 0, dataLength);
                composite.SetData<XnaColor>(0, destRect, buffer, 0, dataLength);
            }

            return composite;
        }

        private void SetTextureCoordinates(Image[] images, XnaRectangle[] locations)
        {
            Debug.Assert(images.Length == locations.Length);

            for (int i = 0; i < images.Length; i++)
            {
                Image img = images[i];
                XnaRectangle rect = locations[i];

                // This adjustment avoids visible edges
                // around individual images.
                int leftPixel = rect.X + 1;
                int topPixel = rect.Y + 1;
                int rightPixel = rect.X + rect.Width - 1;
                int bottomPixel = rect.Y + rect.Height - 1;

                float left = ((float)leftPixel) / cachedTexture.Width;
                float right = ((float)(rightPixel)) / cachedTexture.Width;
                float top = ((float)topPixel) / cachedTexture.Height;
                float bottom = ((float)(bottomPixel)) / cachedTexture.Height;

                TextureCoordinates c = new TextureCoordinates();
                c.TopLeft = new Vector2(left, top);
                c.TopRight = new Vector2(right, top);
                c.BottomLeft = new Vector2(left, bottom);
                c.BottomRight = new Vector2(right, bottom);

                if (!textureCoordinates.ContainsKey(img))
                {
                    textureCoordinates.Add(img, c);
                }
                else
                {
                    textureCoordinates[img] = c;
                }
            }
        }

        internal void RemoveDestroyedObjects()
        {
            while (destroyedObjects.Count > 0)
            {
                GameObject o = destroyedObjects.Dequeue();
                objects.Remove(o);
                Game.Instance.DoRemove(o);
            }

            while (destroyedEffects.Count > 0)
            {
                ParticleSystem p = destroyedEffects.Dequeue();
                effects.Remove(p);
            }
        }

        /// <summary>
        /// Removes the object immediately.
        /// </summary>
        /// <param name="o">The object to remove</param>
        /// <returns><c>true</c> if the object was found and removed.</returns>
        internal bool DoRemove( GameObject o )
        {
            return objects.Remove( o );
        }

        internal void Clear()
        {
            foreach ( var o in objects )
            {
                if ( !Game.gameObjectsToRemove.Contains( o ) )
                {
                    o.Destroy();
                    destroyedObjects.Enqueue( o );
                }
            }

            foreach ( var e in effects )
            {
                e.Destroy();
                destroyedEffects.Enqueue( e );
            }

            cachedTexture = null;
            textureCoordinates.Clear();
        }

        internal void Update( Time time )
        {
#if DEBUG
            ObjectCount.Value = objects.Count;
#endif
            foreach ( var effect in effects )
            {
                effect.Update( time );
            }
        }

        internal void Draw(Camera camera)
        {
            var worldMatrix =
                Matrix.CreateTranslation((float)(-camera.Position.X * RelativeTransition.X), (float)(-camera.Position.Y * RelativeTransition.Y), 0)
                * Matrix.CreateScale((float)(camera.ZoomFactor), (float)(camera.ZoomFactor), 1f);

            switch (DrawOrder)
            {
                case DrawOrder.Irrelevant:
                    DrawEfficientlyInNoParticularOrder(ref worldMatrix);
                    break;
                case DrawOrder.FirstToLast:
                    DrawInOrderFromFirstToLast(ref worldMatrix);
                    break;
                default:
                    break;
            }

            foreach (var effect in effects)
            {
                if (effect.IsDestroyed)
                {
                    destroyedEffects.Enqueue(effect);
                }
                
                effect.Draw(worldMatrix);
            }

            if ( Grid != null )
            {
                DrawGrid( ref worldMatrix );
            }
        }

        private void DrawGrid( ref Matrix matrix )
        {
            Graphics.LineBatch.Begin( ref matrix );

            var camera = Game.Instance.Camera;
            var screen = Game.Screen;
            Vector topLeft = camera.ScreenToWorld( new Vector( screen.Left, screen.Top ) );
            Vector topRight = camera.ScreenToWorld( new Vector( screen.Right, screen.Top ) );
            Vector bottomLeft = camera.ScreenToWorld( new Vector( screen.Left, screen.Bottom ) );
            Vector bottomRight = camera.ScreenToWorld( new Vector( screen.Right, screen.Bottom ) );

            int horizontalCount = (int)Math.Ceiling( ( topRight.X - topLeft.X ) / Grid.CellSize.X );
            int leftmostLine = (int)Math.Ceiling( topLeft.X / Grid.CellSize.X );
            double leftmostX = leftmostLine * Grid.CellSize.X;

            for ( int i = 0; i < horizontalCount; i++ )
            {
                double x = leftmostX + i * Grid.CellSize.X;
                Vector startPoint = new Vector( x, topLeft.Y );
                Vector endPoint = new Vector( x, bottomLeft.Y );
                Graphics.LineBatch.Draw( startPoint, endPoint, Grid.Color );
            }

            int verticalCount = (int)Math.Ceiling( ( topRight.Y - bottomRight.Y ) / Grid.CellSize.Y );
            int bottommostLine = (int)Math.Ceiling( bottomRight.Y / Grid.CellSize.Y );
            double bottommostY = bottommostLine * Grid.CellSize.Y;

            for ( int i = 0; i < verticalCount; i++ )
            {
                double y = bottommostY + i * Grid.CellSize.Y;

                // Doesn't draw the line when y is 0! Wonder why...
                if ( y == 0.0 ) y += 0.1;

                Vector startPoint = new Vector( bottomLeft.X, y );
                Vector endPoint = new Vector( bottomRight.X, y );
                Graphics.LineBatch.Draw( startPoint, endPoint, Grid.Color );
            }

            Graphics.LineBatch.End();
        }

        private void DrawInOrderFromFirstToLast(ref Matrix worldMatrix)
        {
            Game.Renderer.LightingEnabled = true;

            foreach (var o in objects)
            {
                if ( o.IsDestroyed )
                {
                    destroyedObjects.Enqueue( o );
                }

                Draw(o, ref worldMatrix);
            }

            Game.Renderer.LightingEnabled = false;
        }

        private void DrawEfficientlyInNoParticularOrder(ref Matrix worldMatrix)
        {
            Game.Renderer.LightingEnabled = true;
            Graphics.ShapeBatch.Begin(ref worldMatrix);
            Graphics.ImageBatch.Begin(ref worldMatrix, cachedTexture);

            foreach (var o in objects)
            {
                if ( o.IsDestroyed )
                    destroyedObjects.Enqueue(o);

                if (!o.IsVisible)
                    continue;

                bool isSimple = (o.childObjects == null) && !o.TextureFillsShape;

                if (isSimple && (o.Image != null) && textureCoordinates.ContainsKey(o.Image))
                {
                     DrawTexture(o, ref worldMatrix);
                }
                else if (isSimple && (o.Image == null) && (o.Shape == Shape.Rectangle || o.Shape == Shape.Triangle))
                {
                    DrawShape(o, ref worldMatrix);
                }
                else
                {
                    Draw(o, ref worldMatrix);
                }
            }

            Graphics.ImageBatch.End();
            Graphics.ShapeBatch.End();
            Game.Renderer.LightingEnabled = false;
        }

        private void DrawTexture(GameObject o, ref Matrix parentTransformation)
        {
            Vector2 position = new Vector2((float)o.Position.X, (float)o.Position.Y);
            Vector2 scale = new Vector2((float)o.Size.X, (float)o.Size.Y);
            float rotation = (float)o.Angle.Radians;

            if (o.IsVisible)
            {
                TextureCoordinates c = defaultCoords;
                if (textureCoordinates.ContainsKey(o.Image))
                    c = textureCoordinates[o.Image];

                Graphics.ImageBatch.Draw(c, position, scale, rotation);
            }
        }

        private void DrawShape(GameObject o, ref Matrix parentTransformation)
        {
            Vector2 position = new Vector2((float)o.Position.X, (float)o.Position.Y);
            Vector2 scale = new Vector2((float)o.Size.X, (float)o.Size.Y);
            float rotation = (float)o.Angle.Radians;

            if (o.IsVisible)
            {
                Vector[] vertices;
                Int16[] indices;

                if (o.Shape == Shape.Rectangle)
                {
                    vertices = squareVertices;
                    indices = squareIndices;
                }
                else if (o.Shape == Shape.Triangle)
                {
                    vertices = triangleVertices;
                    indices = triangleIndices;
                }
                else
                {
                    vertices = new Vector[] { };
                    indices = new Int16[] { };
                }

                Graphics.ShapeBatch.Draw(vertices, indices, o.Color, position, scale, rotation);
            }
        }

        private void Draw(GameObject o, ref Matrix parentTransformation)
        {
            Vector drawScale = new Vector(1, 1);
            if (o.Shape.IsUnitSize)
                drawScale = o.Size;

            Vector2 position = new Vector2((float)o.Position.X, (float)o.Position.Y);
            Vector2 scale = new Vector2((float)drawScale.X, (float)drawScale.Y);
            float rotation = (float)o.Angle.Radians;

            if (o.IsVisible)
            {
                Matrix transformation =
                    Matrix.CreateScale(scale.X, scale.Y, 1f)
                    * Matrix.CreateRotationZ(rotation)
                    * Matrix.CreateTranslation(position.X, position.Y, 0f)
                    * parentTransformation;

                if (o.Image != null && (!o.TextureFillsShape))
                {
                    Game.Renderer.DrawImage(o.Image, ref transformation, o.TextureWrapSize);
                }
                else if (o.Image != null)
                {
                    Game.Renderer.DrawShape(o.Shape, ref transformation, ref transformation, o.Image, o.TextureWrapSize, o.Color);
                }
                else
                {
                    Game.Renderer.DrawShape(o.Shape, ref transformation, o.Color);
                }
            }

            if (o.childObjects != null)
            {
                Matrix childTransformation =
                    Matrix.CreateRotationZ(rotation)
                    * Matrix.CreateTranslation(position.X, position.Y, 0f)
                    * parentTransformation;

                // light positioning for child objects has not been implemented, so
                // let's not use it.
                Game.Renderer.LightingEnabled = false;

                foreach (var child in o.childObjects)
                {
                    Draw(child, ref childTransformation);
                }

                Game.Renderer.LightingEnabled = true;
            }
        }
    }
}
