﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using System.Linq;
using Physics2DDotNet.Ignorers;
using Physics2DDotNet.Shapes;
using Jypeli.LevelEditor;
using Jypeli.Widgets;

namespace Jypeli
{
    /// <summary>
    /// Pelikenttä, johon voi lisätä olioita. Kentällä voi myös olla reunat ja taustaväri tai taustakuva.
    /// </summary>
    [Save]
    public class Level
    {
        [Save] double _width = 1000;
        [Save] double _height = 800;

        private Game game;
        private ObjectIgnorer ignorerForBorders = new ObjectIgnorer();

        public double AmbientLight { get; set; }

        PhysicsObject ground = null;
        double[] heights = null;
        double scale = 1.0;

        /// <summary>
        /// Kentän keskipiste.
        /// </summary>
        public readonly Vector Center = Vector.Zero;

        /// <summary>
        /// Kentän taustaväri.
        /// </summary>
        public Color BackgroundColor { get; set; }

        /// <summary>
        /// Kentän taustakuva.
        /// </summary>
        public Background Background { get; set; }

        /// <summary>
        /// Kentän leveys.
        /// </summary>
        public double Width
        {
            get { return _width; }
            set { _width = value; }
        }

        /// <summary>
        /// Kentän korkeus.
        /// </summary>
        public double Height
        {
            get { return _height; }
            set { _height = value; }
        }

        /// <summary>
        /// Kentän vasemman reunan x-koordinaatti.
        /// </summary>
        public double Left
        {
            get { return -Width / 2; }
        }

        /// <summary>
        /// Kentän oikean reunan x-koordinaatti.
        /// </summary>
        public double Right
        {
            get { return Width / 2; }
        }

        /// <summary>
        /// Kentän yläreunan y-koordinaatti.
        /// </summary>
        public double Top
        {
            get { return Height / 2; }
        }

        /// <summary>
        /// Kentän alareunan y-koordinaatti.
        /// </summary>
        public double Bottom
        {
            get { return -Height / 2; }
        }

        internal Level( Game game )
        {
            this.game = game;
            AmbientLight = 1.0;
            BackgroundColor = Color.LightBlue; // default color

            // creates a null background
            this.Background = new Background( Vector.Zero );
        }

        internal void Clear()
        {
            Background.Image = null;
            ground = null;
        }

        /// <summary>
        /// Lisää kaikille kentän sivuille reunat, joihin oliot voivat törmätä.
        /// </summary>
        public void CreateBorders()
        {
            CreateLeftBorder();
            CreateRightBorder();
            CreateTopBorder();
            CreateBottomBorder();
        }

        /// <summary>
        /// Lisää kaikille kentän sivuille reunat, joihin oliot voivat törmätä.
        /// </summary>
        /// <param name="isVisible">Reunan näkyvyys <c>true</c>, jos näkyvät reunat, muuten <c>false</c>.</param>
        public void CreateBorders( bool isVisible )
        {
            CreateBorders( PhysicsObject.DefaultCoefficients.Restitution, isVisible );
        }

        /// <summary>
        /// Lisää kaikille kentän sivuille reunat, joihin oliot voivat törmätä.
        /// </summary>
        /// <param name="restitution">Reunojen kimmoisuus (0.0 = pysäyttää, 1.0 = kimpoaa täydellä voimalla)</param>
        /// <param name="isVisible">Reunan näkyvyys <c>true</c>, jos näkyvät reunat, muuten <c>false</c>.</param>
        public void CreateBorders( double restitution, bool isVisible )
        {
            CreateTopBorder( restitution, isVisible );
            CreateBottomBorder( restitution, isVisible );
            CreateLeftBorder( restitution, isVisible );
            CreateRightBorder( restitution, isVisible );
        }

        private PhysicsObject CreateBorder( double width, double height )
        {
            PhysicsObject b = PhysicsObject.CreateStaticObject( width, height );
            b.Color = Color.Gray;
            b.Body.CollisionIgnorer = ignorerForBorders;
            game.Add( b );
            return b;
        }

        /// <summary>
        /// Lisää kenttään vasemman reunan.
        /// </summary>
        /// <returns>Uusi reuna.</returns>
        /// <param name="restitution">Reunojen kimmoisuus (0.0 = pysäyttää, 1.0 = kimpoaa täydellä voimalla)</param>
        /// <param name="isVisible">Reunan näkyvyys <c>true</c>, jos näkyvät reunat, muuten <c>false</c>.</param>
        public PhysicsObject CreateLeftBorder( double restitution, bool isVisible )
        {
            double thickness = this.GetBorderThickness();
            PhysicsObject b = CreateBorder( this.Height, thickness );
            b.Angle = Angle.FromRadians( -Math.PI / 2 );
            b.Position = new Vector( Left - ( thickness / 2 ), Center.Y );
            b.Restitution = restitution;
            b.IsVisible = isVisible;
            return b;
        }

        /// <summary>
        /// Lisää kenttään oikean reunan.
        /// </summary>
        /// <returns>Uusi reuna.</returns>
        /// <param name="restitution">Reunojen kimmoisuus (0.0 = pysäyttää, 1.0 = kimpoaa täydellä voimalla)</param>
        /// <param name="isVisible">Reunan näkyvyys <c>true</c>, jos näkyvät reunat, muuten <c>false</c>.</param>
        public PhysicsObject CreateRightBorder( double restitution, bool isVisible )
        {
            double thickness = this.GetBorderThickness();
            PhysicsObject b = CreateBorder( this.Height, thickness );
            b.Angle = Angle.FromRadians( Math.PI / 2 );
            b.Position = new Vector( Right + ( thickness / 2 ), Center.Y );
            b.Restitution = restitution;
            b.IsVisible = isVisible;
            return b;
        }

        /// <summary>
        /// Lisää kenttään yläreunan.
        /// </summary>
        /// <returns>Uusi reuna.</returns>
        /// <param name="restitution">Reunojen kimmoisuus (0.0 = pysäyttää, 1.0 = kimpoaa täydellä voimalla)</param>
        /// <param name="isVisible">Reunan näkyvyys <c>true</c>, jos näkyvät reunat, muuten <c>false</c>.</param>
        public PhysicsObject CreateTopBorder( double restitution, bool isVisible )
        {
            double thickness = this.GetBorderThickness();
            PhysicsObject b = CreateBorder( this.Width + ( 2 * thickness ), thickness );
            b.Angle = Angle.FromRadians( Math.PI );
            b.Position = new Vector( Center.X, Top + ( thickness / 2 ) );
            b.Restitution = restitution;
            b.IsVisible = isVisible;
            return b;
        }

        /// <summary>
        /// Lisää kenttään alareunan.
        /// </summary>
        /// <returns>Uusi reuna.</returns>
        /// <param name="restitution">Reunojen kimmoisuus (0.0 = pysäyttää, 1.0 = kimpoaa täydellä voimalla)</param>
        /// <param name="isVisible">Reunan näkyvyys <c>true</c>, jos näkyvät reunat, muuten <c>false</c>.</param>
        public PhysicsObject CreateBottomBorder( double restitution, bool isVisible )
        {
            double thickness = GetBorderThickness();
            PhysicsObject b = CreateBorder( Width + ( 2 * thickness ), thickness );
            b.Angle = Angle.Zero;
            b.Position = new Vector( Center.X, Bottom - ( thickness / 2 ) );
            b.Restitution = restitution;
            b.IsVisible = isVisible;
            return b;
        }

        /// <summary>
        /// Lisää kenttään vasemman reunan.
        /// </summary>
        /// <returns>Uusi reuna.</returns>
        public PhysicsObject CreateLeftBorder()
        {
            return CreateLeftBorder( PhysicsObject.DefaultCoefficients.Restitution, true );
        }

        /// <summary>
        /// Lisää kenttään oikean reunan.
        /// </summary>
        /// <returns>Uusi reuna.</returns>
        public PhysicsObject CreateRightBorder()
        {
            return CreateRightBorder( PhysicsObject.DefaultCoefficients.Restitution, true );
        }

        /// <summary>
        /// Lisää kenttään yläreunan.
        /// </summary>
        /// <returns>Uusi reuna.</returns>
        public PhysicsObject CreateTopBorder()
        {
            return CreateTopBorder( PhysicsObject.DefaultCoefficients.Restitution, true );
        }

        /// <summary>
        /// Lisää kenttään alareunan.
        /// </summary>
        /// <returns>Uusi reuna.</returns>
        public PhysicsObject CreateBottomBorder()
        {
            return CreateBottomBorder( PhysicsObject.DefaultCoefficients.Restitution, true );
        }

        private double GetBorderThickness()
        {
            return this.Width / 10;
        }

        public double GetGroundHeight(double x)
        {
            if (heights == null || x < Left || x > Right) return Bottom;

            int n = heights.Length;
            double step = Width / (n - 1);
            double maxHeight = heights.Max() * scale;

            double indexX = (Width / 2 + x) / step;
            int lowerIndex = (int)Math.Floor(indexX);
            int upperIndex = (int)Math.Ceiling(indexX);

            if (upperIndex >= n) return Bottom; // DEBUG
            if (lowerIndex == upperIndex) return Bottom + scale * heights[lowerIndex];

            double k = (heights[upperIndex] - heights[lowerIndex]) / step;
            double relX = (Width / 2 + x) % step;

            return Bottom + heights[lowerIndex] + relX * k;
        }

        public Vector GetGroundNormal(double x)
        {
            if (heights == null || x < Left || x > Right) return Vector.UnitY;

            int n = heights.Length;
            double step = Width / (n - 1);
            double maxHeight = heights.Max() * scale;

            double indexX = (Width / 2 + x) / step;
            int lowerIndex = (int)Math.Floor(indexX);
            int upperIndex = (int)Math.Ceiling(indexX);

            if (upperIndex >= n) return Vector.UnitY; // DEBUG
            if (lowerIndex == upperIndex)
            {
                return (GetGroundNormal(x - step / 2) + GetGroundNormal(x + step / 2)) / 2;
            }

            double k = (heights[upperIndex] - heights[lowerIndex]) / step;
            return Vector.FromLengthAndAngle(1, Angle.ArcTan(k) + Angle.RightAngle);
        }

        /// <summary>
        /// Helppo tapa lisätä kenttään epätasainen maasto.
        /// Maasto kuvataan luettelemalla Y-koordinaatteja vasemmalta oikealle lukien. Kahden Y-koordinaatin
        /// väli on aina sama.
        /// </summary>
        /// <param name="heights">Y-koordinaatit lueteltuna vasemmalta oikealle.</param>
        /// <param name="scale">Vakio, jolla jokainen Y-koordinaatti kerrotaan. Hyödyllinen,
        /// jos halutaan muuttaa koko maaston korkeutta muuttamatta jokaista pistettä yksitellen.
        /// Tavallisesti arvoksi kelpaa 1.0.</param>
        /// <remarks>
        /// Huomaa, että maastossa ei voi olla kahta pistettä päällekkäin.
        /// </remarks>
        public PhysicsObject CreateGround( double[] heights, double scale )
        {
            if ( heights.Length < 2 )
                throw new Exception( "At least two Y-points needed in order to create ground" );
            if ( heights.Min() < 1 )
                throw new Exception( "The heights must be positive and at least 1" ); // JN: doesn't work well with 0 values (slow and memory usage problems)

            int n = heights.Length;
            double step = Width / ( n - 1 );
            double maxHeight = heights.Max() * scale;

            // Each point should be adjusted by this amount in order to get the center of the shape at (0, 0)
            // This way, drawing a single texture on top of it covers the whole shape.
            double halfHeight = maxHeight / 2;

            Vector[] vertexes = new Vector[n * 2];

            // Let's start from the bottom right corner and head clockwise from there.
            Vector bottomRight = new Vector( Width / 2, -halfHeight );

            // Bottom vertexes, right to left
            for ( int i = 0; i < n; i++ )
            {
                vertexes[i] = new Vector( bottomRight.X - ( i * step ), bottomRight.Y );
            }

            double left = -Width / 2;

            // Top vertexes, left to right
            for ( int i = 0; i < n; i++ )
            {
                vertexes[n + i] = new Vector( left + ( i * step ), heights[i] * scale - halfHeight );
            }

            IndexTriangle[] triangles = new IndexTriangle[( n - 1 ) * 2];
            Int16[] outlineIndices = new Int16[n * 2];

            for ( int i = 0; i < n * 2; i++ )
                outlineIndices[i] = (Int16)i;
            for ( int i = 0; i < n - 1; i++ )
            {
                triangles[2 * i] = new IndexTriangle( (Int16)i, (Int16)( 2 * n - i - 2 ), (Int16)( 2 * n - i - 1 ) );
                triangles[2 * i + 1] = new IndexTriangle( (Int16)i, (Int16)( i + 1 ), (Int16)( 2 * n - i - 2 ) );
            }

            // A crude estimate for the correct value. Too small a value can slow down PolygonShape constructor
            // significantly.
            double gridSpacing = Math.Min( heights.Min(), 5 );

            Vector[] outlineVertices = new Vector[outlineIndices.Length];
            for ( int i = 0; i < outlineIndices.Length; i++ )
                outlineVertices[i] = vertexes[outlineIndices[i]];

            Vector size = new Vector( this.Width, maxHeight );
            Polygon shape = new Polygon( new ShapeCache( vertexes, triangles, outlineIndices ), false );
            PolygonShape physicsShape = (PolygonShape)PhysicsObject.CreatePhysicsShape( shape, size, 20, gridSpacing );

            if (ground == null)
            {
                ground = PhysicsObject.CreateStaticObject(size.X, size.Y, shape, physicsShape);
                ground.Color = Color.ForestGreen;
                game.Add(ground);
            }

            ground.Position = new Vector( Center.X, Bottom + halfHeight );

            this.heights = heights;
            this.scale = scale;

            return ground;
        }

        /// <summary>
        /// Helppo tapa lisätä kenttään epätasainen maasto.
        /// Maasto kuvataan luettelemalla Y-koordinaatteja vasemmalta oikealle lukien. Kahden Y-koordinaatin
        /// väli on aina sama.
        /// </summary>
        /// <param name="heights">Y-koordinaatit lueteltuna vasemmalta oikealle.</param>
        /// <param name="scale">Vakio, jolla jokainen Y-koordinaatti kerrotaan. Hyödyllinen,
        /// jos halutaan muuttaa koko maaston korkeutta muuttamatta jokaista pistettä yksitellen.
        /// Tavallisesti arvoksi kelpaa 1.0.</param>
        /// <param name="image">Maastossa käytettävä kuva.</param>
        /// <returns></returns>
        public PhysicsObject CreateGround( double[] heights, double scale, Image image )
        {
            PhysicsObject ground = this.CreateGround( heights, scale );
            ground.Image = image;
            ground.TextureFillsShape = true;
            return ground;
        }

        /// <summary>
        /// Luo satunnaisen maaston.
        /// </summary>
        /// <param name="min">Matalin kohta.</param>
        /// <param name="max">Korkein kohta.</param>
        /// <param name="points">Pisteiden määrä.</param>
        public PhysicsObject CreateGround( double min, double max, int points )
        {
            return this.CreateGround( min, max, points, Color.ForestGreen, null );
        }

        /// <summary>
        /// Luo satunnaisen maaston.
        /// </summary>
        /// <param name="min">Matalin kohta.</param>
        /// <param name="max">Korkein kohta.</param>
        /// <param name="points">Pisteiden määrä.</param>
        /// <param name="color">Maaston väri.</param>
        /// <returns></returns>
        public PhysicsObject CreateGround( double min, double max, int points, Color color )
        {
            return this.CreateGround( min, max, points, color, null );
        }

        /// <summary>
        /// Luo satunnaisen maaston.
        /// </summary>
        /// <param name="min">Matalin kohta.</param>
        /// <param name="max">Korkein kohta.</param>
        /// <param name="points">Pisteiden määrä.</param>
        /// <param name="image">Maaston kuva.</param>
        /// <returns></returns>
        public PhysicsObject CreateGround( double min, double max, int points, Image image )
        {
            return this.CreateGround( min, max, points, Color.ForestGreen, image );
        }

        /// <summary>
        /// Luo satunnaisen maaston.
        /// </summary>
        /// <param name="min">Matalin kohta.</param>
        /// <param name="max">Korkein kohta.</param>
        /// <param name="points">Pisteiden määrä.</param>
        /// <param name="color">Maaston väri.</param>
        /// <param name="image">Maaston kuva.</param>
        /// <returns></returns>
        public PhysicsObject CreateGround( double min, double max, int points, Color color, Image image )
        {
            if ( min < 1.0f ) min = 1.0f; // has to be at least this to work well
            double[] heights = RandomGen.NextDoubleArray( min, max, points );
            PhysicsObject ground = this.CreateGround( heights, 1.0 );

            ground.Image = image;
            ground.TextureFillsShape = true;
            ground.Color = color;
            return ground;
        }

        /// <summary>
        /// Luo satunnaisen maaston.
        /// </summary>
        /// <param name="min">Matalin kohta.</param>
        /// <param name="max">Korkein kohta.</param>
        /// <param name="points">Pisteiden määrä.</param>
        /// <param name="maxchange">Suurin sallittu muutos korkeudessa pisteiden välillä.</param>
        public PhysicsObject CreateGround( double min, double max, int points, int maxchange )
        {
            return this.CreateGround( min, max, points, Color.ForestGreen, null, maxchange );
        }

        /// <summary>
        /// Luo satunnaisen maaston.
        /// </summary>
        /// <param name="min">Matalin kohta.</param>
        /// <param name="max">Korkein kohta.</param>
        /// <param name="points">Pisteiden määrä.</param>
        /// <param name="color">Maaston väri.</param>
        /// <param name="maxchange">Suurin sallittu erotus kahden pisteen välillä.</param>
        /// <returns></returns>
        public PhysicsObject CreateGround( double min, double max, int points, Color color, int maxchange )
        {
            return this.CreateGround( min, max, points, color, null, maxchange );
        }

        /// <summary>
        /// Luo satunnaisen maaston.
        /// </summary>
        /// <param name="min">Matalin kohta.</param>
        /// <param name="max">Korkein kohta.</param>
        /// <param name="points">Pisteiden määrä.</param>
        /// <param name="image">Maaston kuva.</param>
        /// <param name="maxchange">Suurin sallittu erotus kahden pisteen välillä.</param>
        /// <returns></returns>
        public PhysicsObject CreateGround( double min, double max, int points, Image image, int maxchange )
        {
            return this.CreateGround( min, max, points, Color.ForestGreen, image, maxchange );
        }

        /// <summary>
        /// Luo satunnaisen maaston.
        /// </summary>
        /// <param name="min">Matalin kohta.</param>
        /// <param name="max">Korkein kohta.</param>
        /// <param name="points">Pisteiden määrä.</param>
        /// <param name="color">Maaston väri.</param>
        /// <param name="image">Maaston kuva.</param>
        /// <param name="maxchange">Suurin sallittu erotus kahden pisteen välillä.</param>
        /// <returns></returns>
        public PhysicsObject CreateGround( double min, double max, int points, Color color, Image image, int maxchange )
        {
            if ( min < 1.0f ) min = 1.0f; // has to be at least this to work well
            double[] heights = RandomGen.NextDoubleArray( min, max, points, maxchange );
            PhysicsObject ground = this.CreateGround( heights, 1.0 );

            ground.Image = image;
            ground.TextureFillsShape = true;
            ground.Color = color;
            return ground;
        }

        /// <summary>
        /// Palauttaa satunnaisen kohdan kentän reunojen sisältä.
        /// </summary>
        /// <returns>Vektori.</returns>
        public Vector GetRandomPosition()
        {
            return new Vector( RandomGen.NextDouble( Left, Right ), RandomGen.NextDouble( Bottom, Top ) );
        }

        /// <summary>
        /// Lataa kentän tiedostosta.
        /// Kenttätiedostoja voi tehdä Jypelin mukana tulevalle editorilla.
        /// </summary>
        /// <param name="fileName">Kenttätiedoston nimi</param>
        public void LoadFromFile( string fileName )
        {
#if WINDOWS
            LevelData levelData = new LevelData();
            Factory.FactoryMethod createObject = delegate { return new LevelObject( levelData, 0 ); };
            Game.Instance.AddFactory<LevelObject>( null, createObject );
            levelData = Game.DataStorage.Load<LevelData>( new LevelData(), fileName );
            Game.Instance.RemoveFactory<LevelObject>( null, createObject );
            DoLoadLevel( levelData );
#else
            throw new NotImplementedException( "LoadFromFile not implemented for WP7 / Xbox360" );
#endif
        }

        /// <summary>
        /// Lataa kentän contentista.
        /// Kenttätiedostoja voi tehdä Jypelin mukana tulevalle editorilla.
        /// </summary>
        /// <param name="assetName">Kentän nimi contentissa</param>
        public void LoadFromContent( string assetName )
        {
            LevelData levelData = new LevelData();
            Factory.FactoryMethod createObject = delegate { return new LevelObject( levelData, 0 ); };
            Game.Instance.AddFactory<LevelObject>( null, createObject );
            levelData = Game.DataStorage.LoadContent<LevelData>( levelData, assetName );
            Game.Instance.RemoveFactory<LevelObject>( null, createObject );
            DoLoadLevel( levelData );
        }

        private void DoLoadLevel( LevelData levelData )
        {
            foreach ( var levelObj in levelData.Objects )
            {
                Game.Instance.Add( levelObj.ConstructObject() );
            }
        }
    }
}
