﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;

namespace Jypeli
{
    /// <summary>
    /// Mittari, joka mittaa erityyppisiä arvoja.
    /// Sidottavissa näyttöihin, kuten <c>ValueDisplay</c> ja <c>BarGauge</c>.
    /// </summary>
    [Save]
    public abstract class Meter<ValueType> where ValueType : struct, IComparable<ValueType>, IEquatable<ValueType>
    {
        bool valueSet;

        ValueType val;
        ValueType minval;
        ValueType maxval;
        ValueType defval;

        /// <summary>
        /// Mittarin arvo.
        /// </summary>
        [Save]
        public ValueType Value
        {
            get { return val; }
            set
            {
                if ( value.Equals( val ) )
                    return;

                ValueType oldval = val;
                valueSet = true;
                
                if ( value.CompareTo( minval ) <= 0 )
                {
                    val = minval;
                    OnChange( oldval, val );
                    OnLowerLimit();
                    return;
                }

                if ( value.CompareTo( maxval ) >= 0 )
                {
                    val = maxval;
                    OnChange( oldval, val );
                    OnUpperLimit();
                    return;
                }

                val = value;
                OnChange( oldval, value );

                if ( val.CompareTo( LowerAlertValue ) < 0 && oldval.CompareTo( LowerAlertValue ) >= 0 )
                    OnLowerAlert();

                if ( val.CompareTo( UpperAlertValue ) > 0 && oldval.CompareTo( UpperAlertValue ) <= 0 )
                    OnUpperAlert();
            }
        }

        /// <summary>
        /// Mittarin oletusarvo.
        /// </summary>
        public ValueType DefaultValue
        {
            get { return defval; }
            set
            {
                defval = clampValue( value, minval, maxval );
                if ( !valueSet ) Value = value;
            }
        }

        /// <summary>
        /// Mittarin pienin sallittu arvo.
        /// Kun mittari saavuttaa tämän arvon, laukeaa tapahtuma <c>LowerLimit</c>.
        /// </summary>
        public ValueType MinValue
        {
            get { return minval; }
            set { minval = value; updateBounds(); }
        }

        /// <summary>
        /// Mittarin suurin sallittu arvo.
        /// Kun mittari saavuttaa tämän arvon, laukeaa tapahtuma <c>UpperLimit</c>.
        /// </summary>
        public ValueType MaxValue
        {
            get { return maxval; }
            set { maxval = value; updateBounds(); }
        }

        /// <summary>
        /// Mittarin hälytysalaraja.
        /// Kun raja alittuu, laukeaa tapahtuma <c>LowerAlert</c>.
        /// </summary>
        public ValueType UpperAlertValue { get; set; }

        /// <summary>
        /// Mittarin hälytysalaraja.
        /// Kun raja ylittyy, laukeaa tapahtuma <c>UpperAlert</c>.
        /// </summary>
        public ValueType LowerAlertValue { get; set; }

        #region Events

        /// <summary>
        /// Mittarin muutostapahtumankäsittelijä.
        /// </summary>
        public delegate void OnChangeHandler( ValueType oldValue, ValueType newValue );

        /// <summary>
        /// Mittarin hälytystapahtumankäsittelijä.
        /// </summary>
        public delegate void OnAlertHandler( ValueType value );

        /// <summary>
        /// Mittarin ääriarvotapahtumankäsittelijä.
        /// </summary>
        public delegate void OnLimitHandler( ValueType value );

        /// <summary>
        /// Tapahtuu, kun mittarin arvo muuttuu.
        /// </summary>
        public event OnChangeHandler Changed;

        /// <summary>
        /// Tapahtuu, kun arvo alittaa hälytysalarajan.
        /// </summary>
        public event OnAlertHandler LowerAlert;

        /// <summary>
        /// Tapahtuu, kun arvo alittaa hälytysylärajan.
        /// </summary>
        public event OnAlertHandler UpperAlert;

        /// <summary>
        /// Tapahtuu, kun arvo saavuttaa pienimmän sallitun arvonsa.
        /// </summary>
        public event OnLimitHandler LowerLimit;

        /// <summary>
        /// Tapahtuu, kun arvo saavuttaa suurimman sallitun arvonsa.
        /// </summary>
        public event OnLimitHandler UpperLimit;

        private void OnChange( ValueType oldValue, ValueType newValue )
        {
            if ( Changed != null )
                Changed( oldValue, newValue );
        }

        private void OnLowerAlert()
        {
            if ( LowerAlert != null )
                LowerAlert( val );
        }

        private void OnUpperAlert()
        {
            if ( UpperAlert != null )
                UpperAlert( val );
        }

        private void OnLowerLimit()
        {
            if ( LowerLimit != null )
                LowerLimit( val );
        }

        private void OnUpperLimit()
        {
            if ( UpperLimit != null )
                UpperLimit( val );
        }

        #endregion

        /// <summary>
        /// Luo uuden mittarin.
        /// </summary>
        /// <param name="defaultVal">Oletusarvo.</param>
        /// <param name="minVal">Pienin sallittu arvo.</param>
        /// <param name="maxVal">Suurin sallittu arvo.</param>
        public Meter( ValueType defaultVal, ValueType minVal, ValueType maxVal )
        {
            minval = minVal;
            maxval = maxVal;
            defval = defaultVal;
            val = defaultVal;

            LowerAlertValue = minVal;
            UpperAlertValue = maxVal;

            updateBounds();
        }

        /// <summary>
        /// Luo uuden mittarin kopiona parametrina annetusta.
        /// </summary>
        /// <param name="src">Kopioitava mittari.</param>
        public Meter( Meter<ValueType> src )
        {
            minval = src.minval;
            maxval = src.maxval;
            defval = src.defval;
            val = src.val;

            LowerAlertValue = src.LowerAlertValue;
            UpperAlertValue = src.UpperAlertValue;

            updateBounds();
        }

        /// <summary>
        /// Palauttaa mittarin arvon oletusarvoonsa.
        /// </summary>
        public void Reset()
        {
            Value = DefaultValue;
        }

        private static ValueType clampValue( ValueType v, ValueType min, ValueType max )
        {
            if ( v.CompareTo( min ) < 0 )
                return min;

            if ( v.CompareTo( max ) > 0 )
                return max;

            return v;
        }

        private static void clampValue( ref ValueType v, ValueType min, ValueType max )
        {
            if ( v.CompareTo( min ) < 0 )
                v = min;

            else if ( v.CompareTo( max ) > 0 )
                v = max;
        }

        private void updateBounds()
        {
            clampValue( ref val, minval, maxval );
            clampValue( ref defval, minval, maxval );
        }

        /// <summary>
        /// Palauttaa mittarin arvon merkkijonona.
        /// </summary>
        public override String ToString()
        {
            return Value.ToString();
        }
    }


    /// <summary>
    /// Mittari, joka mittaa int-tyyppisiä arvoja.
    /// Sidottavissa näyttöihin, kuten <c>ValueDisplay</c> ja <c>BarGauge</c>.
    /// </summary>
    public class IntMeter : Meter<int>
    {
        public IntMeter( int defaultValue )
            : base( defaultValue, 0, int.MaxValue )
        {
        }

        /// <summary>
        /// Antaa mittariolion <c>m</c> arvon, kun mittaria käytetään
        /// sellaisessa yhteydessä, jossa vaaditaan tavallista <c>int</c>-
        /// tyyppistä kokonaislukua.
        /// </summary>
        public static implicit operator int( IntMeter m )
        {
            return m.Value;
        }

        /// <summary>
        /// Antaa mittariolion <c>m</c> arvon, kun mittaria käytetään
        /// sellaisessa yhteydessä, jossa vaaditaan tavallista <c>double</c>-
        /// tyyppistä liukulukua.
        /// </summary>
        public static implicit operator double( IntMeter m )
        {
            return (double)m.Value;
        }
    }

    /// <summary>
    /// Mittari, joka mittaa double-tyyppisiä arvoja.
    /// Sidottavissa näyttöihin, kuten <c>ValueDisplay</c> ja <c>BarGauge</c>.
    /// </summary>
    public class DoubleMeter : Meter<double>
    {
        public DoubleMeter( double defaultValue )
            : base( defaultValue, 0.0, double.MaxValue )
        {
        }

        /// <summary>
        /// Antaa mittariolion <c>m</c> arvon, kun mittaria käytetään
        /// sellaisessa yhteydessä, jossa vaaditaan tavallista <c>double</c>-
        /// tyyppistä liukulukua.
        /// </summary>
        public static implicit operator double( DoubleMeter m )
        {
            return m.Value;
        }
    }
}
