﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using Physics2DDotNet;
using Physics2DDotNet.PhysicsLogics;
using Physics2DDotNet.Joints;
using Physics2DDotNet.Solvers;
using AdvanceMath;
using System.Collections.Generic;
using Physics2DDotNet.Ignorers;

namespace Jypeli
{
    /// <summary>
    /// Törmäystapahtumiin liitettävän metodin tyyppi. Törmäyksen käsittelevällä
    /// metodilla ei ole paluuarvoa ja se ottaa yhden <code>Collision</code>-tyyppisen
    /// parametrin.
    /// </summary>
    public delegate void CollisionHandler( PhysicsObject collidingObject, PhysicsObject otherObject );


    /// <summary>
    /// Peli, jossa on fysiikan laskenta mukana. Peliin lisätyt <code>PhysicsObject</code>-oliot
    /// käyttäytyvät fysiikan lakien mukaan.
    /// </summary>
    public class PhysicsGame : Game
    {
        private PhysicsEngine phsEngine;
        private GravityField gravityfield;
        private Vector gravity = Vector.Zero;
        
        static internal Dictionary<int, ObjectIgnorer> IgnoreGroups = null;

        /// <summary>
        /// Onko fysiikan laskenta käytössä vai ei.
        /// </summary>
        public bool PhysicsEnabled { get; set; }

        /// <summary>
        /// Painovoima. Voimavektori, joka vaikuttaa kaikkiin ei-staattisiin kappaleisiin.
        /// </summary>
        public Vector Gravity
        {
            get
            {
                return gravity;
            }
            set
            {
                gravity = value;
                updatePhysicsConstants();
            }
        }

        /// <summary>
        /// Alustaa uuden fysiikkapelin.
        /// </summary>
        public PhysicsGame()
            : base()
        {
            PhysicsEnabled = true;
            phsEngine = new PhysicsEngine();
            phsEngine.BroadPhase = new Physics2DDotNet.Detectors.SelectiveSweepDetector();
            //phsEngine.BroadPhase = new Physics2DDotNet.Detectors.SpatialHashDetector();

            SequentialImpulsesSolver phsSolver = new SequentialImpulsesSolver();
            phsSolver.Iterations = 12;
            phsSolver.SplitImpulse = true;
            //phsSolver.BiasFactor = 0.7;
            phsSolver.BiasFactor = 0.0;
            //phsSolver.AllowedPenetration = 0.1;
            phsSolver.AllowedPenetration = 0.01;
            phsEngine.Solver = (CollisionSolver)phsSolver;
        }

        /// <summary>
        /// Luo uuden liitoksen, jolla voi liittää kaksi oliota joustavasti toisiinsa.
        /// </summary>
        /// <param name="o1">Ensimmäinen liitettävä olio.</param>
        /// <param name="o2">Toinen liitettävä olio.</param>
        /// <param name="anchor">Liitoksen paikka.</param>
        /// <returns>Liitos.</returns>
        [Obsolete("Käytä mieluummin AxleJoint-luokkaa")]
        public static Physics2DDotNet.Joints.HingeJoint CreateHingeJoint( PhysicsObject o1, PhysicsObject o2, Vector anchor )
        {
            return new Physics2DDotNet.Joints.HingeJoint( o1.Body, o2.Body, new Vector2D( anchor.X, anchor.Y ), new Lifespan() );
        }

        public override void Add( GameObject o, int layer )
        {
            if ( o is PhysicsObject )
            {
                AddToEngine( (PhysicsObject)o );
            }
            base.Add( o, layer );
        }

        internal override void DoRemove( GameObject o )
        {
            if ( o is PhysicsObject )
            {
                RemoveFromEngine( (PhysicsObject)o );
            }
            base.DoRemove( o );
        }

        /// <summary>
        /// Nollaa kaiken (kontrollit, näyttöobjektit, ajastimet ja fysiikkamoottorin).
        /// </summary>
        public override void ClearAll()
        {
            ClearPhysics();
            base.ClearAll();
        }

        /// <summary>
        /// Nollaa fysiikkamoottorin.
        /// </summary>
        private void ClearPhysics()
        {
            phsEngine.Clear();
        }

        private void updatePhysicsConstants()
        {
            if ( gravityfield != null ) gravityfield.Lifetime.IsExpired = true;

            if ( gravity != Vector.Zero )
            {
                gravityfield = new GravityField( new Vector2D( gravity.X, gravity.Y ), new Lifespan() );
                phsEngine.AddLogic( gravityfield );
            }
        }

        private void AddToEngine( PhysicsObject po )
        {
            if ( po.Body.Engine != null )
                return;

            po.Body.Lifetime.IsExpired = false;
            phsEngine.AddBody( po.Body );
        }

        private void RemoveFromEngine( PhysicsObject po )
        {
            if ( po.Body.Engine == null )
                return;

            po.Body.Lifetime.IsExpired = true;
        }

        /// <summary>
        /// Lisää liitoksen peliin.
        /// </summary>
        public void Add( Physics2DDotNet.Joints.Joint j )
        {
            j.Lifetime.IsExpired = false;
            phsEngine.AddJoint( j );
        }

        /// <summary>
        /// Poistaa liitoksen pelistä.
        /// </summary>
        /// <param name="j"></param>
        internal void Remove( Physics2DDotNet.Joints.Joint j )
        {
            j.Lifetime.IsExpired = true;
        }

        /// <summary>
        /// Lisää liitoksen peliin.
        /// </summary>
        public void Add( AxleJoint j )
        {
            if ( !j.Object1.IsAddedToGame ) Add( j.Object1 );
            if ( j.Object2 != null && !j.Object2.IsAddedToGame ) Add( j.Object2 );

            Add( j.innerJoint );
        }

        /// <summary>
        /// Lisää rakenteisen olion peliin.
        /// </summary>
        /// <param name="structure">Rakenne</param>
        public void Add( PhysicsStructure structure )
        {
            foreach ( var joint in structure.Joints )
            {
                Add( joint );
            }

            structure.IsAddedToGame = true;
        }

        /// <summary>
        /// Ajetaan kun pelin tilannetta päivitetään. Päivittämisen voi toteuttaa perityssä luokassa
        /// toteuttamalla tämän metodin. Perityn luokan metodissa tulee kutsua kantaluokan metodia.
        /// </summary>
        /// <param name="time"></param>
        protected override void Update( Time time )
        {
            double dt = time.SinceLastUpdate.TotalSeconds;

            if ( PhysicsEnabled )
            {
                phsEngine.Update(dt);
            }

            base.Update( time );
        }

        /// <summary>
        /// Määrittää metodin, jota kutsutaan kun
        /// olio <code>o</code> törmää johonkin.
        /// </summary>
        /// <param name="o">Törmäävä olio.</param>
        /// <param name="handler">Metodi, joka käsittelee törmäyksen.</param>
        public void AddCollisionHandler( PhysicsObject o, CollisionHandler handler )
        {
            if ( handler == null )
                throw new NullReferenceException( "handler must not be null" );

            o.Body.Collided += delegate( object sender, CollisionEventArgs args )
            {
                if ( !o.IsDestroyed )
                {
                    PhysicsObject other = (PhysicsObject)args.Other.Tag;
                    handler( o, other );
                }
            };
        }
    }
}
