﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using System.Diagnostics;
using AdvanceMath;
using Physics2DDotNet;
using Physics2DDotNet.Shapes;
using Physics2DDotNet.Ignorers;
using System.Collections.Generic;


#if !XBOX

#endif

namespace Jypeli
{
    /// <summary>
    /// Kappaleen kuvion laatu törmäyksentunnistuksessa.
    /// </summary>
    public struct CollisionShapeQuality
    {
        /// <summary>
        /// Laatuarvo.
        /// </summary>
        public double Value;

        internal CollisionShapeQuality( double value )
        {
            this.Value = value;
        }

        /// <summary>
        /// Alustaa uuden laatuattribuutin.
        /// </summary>
        /// <param name="value">Lukuarvo välillä <c>0.0</c> (huonoin) ja <c>1.0</c> (paras).</param>
        public static CollisionShapeQuality FromValue( double value )
        {
            if ( value < 0.0 || 1.0 < value )
                throw new ArgumentException( "The value must be between 0.0 and 1.0." );
            return new CollisionShapeQuality( value );
        }

        /// <summary>
        /// Huonoin mahdollinen laatu, nopea mutta epätarkka.
        /// </summary>
        public static readonly CollisionShapeQuality Worst = FromValue( 0.0 );

        /// <summary>
        /// Välttävä laatu.
        /// </summary>
        public static readonly CollisionShapeQuality Tolerable = FromValue( 0.3 );

        /// <summary>
        /// Hyvä laatu.
        /// </summary>
        public static readonly CollisionShapeQuality Good = FromValue( 0.5 );

        /// <summary>
        /// Paras mahdollinen laatu, tarkka mutta hidas.
        /// </summary>
        public static readonly CollisionShapeQuality Best = FromValue( 1.0 );

        internal static readonly CollisionShapeQuality Unspecified = new CollisionShapeQuality( -1.0 );

        internal bool IsUnspecified
        {
            get { return ( Value + double.Epsilon ) < 0; }
        }
    }

    internal class Force
    {
        /// <summary>
        /// Voiman suunta ja suuruus.
        /// </summary>
        public Vector Value { get; set; }

        /// <summary>
        /// Voiman luomisaika.
        /// </summary>
        public TimeSpan CreationTime { get; private set; }

        /// <summary>
        /// Elinaika. Lasketaan siitä lähtien, kun voima luodaan.
        /// </summary>
        public TimeSpan Lifetime
        {
            get { return Game.Time.SinceStartOfGame - CreationTime; }
        }

        /// <summary>
        /// Suurin mahdollinen elinaika.
        /// Kun <c>Lifetime</c> on suurempi kuin tämä, voima lopettaa vaikuttamasta.
        /// </summary>
        public TimeSpan MaximumLifetime { get; set; }

        public Force( Vector f )
        {
            Value = f;
            CreationTime = Game.Time.SinceStartOfGame;
            MaximumLifetime = TimeSpan.MaxValue;
        }

        public Force( Vector f, TimeSpan t )
            : this( f )
        {
            MaximumLifetime = CreationTime + t;
        }

        /// <summary>
        /// Lopettaa voiman vaikutuksen.
        /// </summary>
        public virtual void Destroy()
        {
            this.MaximumLifetime = new TimeSpan( 0 );
        }

        public bool IsDestroyed()
        {
            return Lifetime > MaximumLifetime;
        }
    }


    /// <summary>
    /// Peliolio, joka noudattaa fysiikkamoottorin määräämiä fysiikan lakeja.
    /// Voidaan kuitenkin myös laittaa noudattamaan lakeja valikoidusti.
    /// </summary>
    public class PhysicsObject : GameObject
    {
        internal static readonly Coefficients DefaultCoefficients = new Coefficients( 0.5, 0.4, 0.4 );
        private const double DefaultMass = 1.0;

        private CollisionShapeQuality requestedQuality = CollisionShapeQuality.Unspecified;

        /// <summary>
        /// Olioon vaikuttavat voimat
        /// </summary>
        internal List<Force> ActiveForces = new List<Force>();

        /// <summary>
        /// Fysiikkamoottorin käyttämä tietorakenne.
        /// </summary>
        public Body Body { get; set; }

        /// <summary>
        /// Olio, jolla voi välttää oliota osumasta tiettyihin muihin olioihin.
        /// </summary>
        public Ignorer CollisionIgnorer
        {
            get { return Body.CollisionIgnorer; }
            set { Body.CollisionIgnorer = value; }
        }

        int _ignoreGroup;

        /// <summary>
        /// Törmäysryhmä.
        /// Oliot jotka ovat samassa törmäysryhmässä menevät toistensa läpi.
        /// Jos ryhmä on nolla tai negatiivinen, sillä ei ole vaikutusta.
        /// </summary>
        public int CollisionIgnoreGroup
        {
            get { return _ignoreGroup; }
            set
            {
                _ignoreGroup = value;
                if ( Game == null ) AddedToGame += AddCollisionIgnoreGroup;
                else AddCollisionIgnoreGroup();
            }
        }

        /// <summary>
        /// Olion paikka koordinaatistossa. Käsittää sekä X- että Y-koordinaatin.
        /// </summary>
        [Save]
        public override Vector Position
        {
            get
            {
                Vector2D v = Body.State.Position.Linear;
                return new Vector( v.X, v.Y );
            }
            set { Body.State.Position.Linear = new Vector2D( value.X, value.Y ); }
        }

        /// <summary>
        /// Olion koko (x on leveys, y on korkeus).
        /// </summary>
        [Save]
        public override Vector Size
        {
            get
            {
                return base.Size;
            }
            set
            {
                IShape shape;
                if ( !requestedQuality.IsUnspecified )
                    shape = CreatePhysicsShape( base.Shape, value, requestedQuality );
                else
                    shape = CreatePhysicsShape( base.Shape, value );

                Body.Shape = shape;
                base.Size = value;
            }
        }

        /// <summary>
        /// Kulma, jossa olio on. Oliota voi pyörittää kulmaa vaihtamalla.
        /// </summary>
        [Save]
        public override Angle Angle
        {
            get { return Angle.FromRadians( Body.State.Position.Angular ); }
            set { Body.State.Position.Angular = value.Radians; }
        }

        /// <summary>
        /// Jos <c>false</c>, olio ei voi pyöriä.
        /// </summary>
        public bool CanRotate
        {
            get { return !double.IsPositiveInfinity( MomentOfInertia ); }
            set
            {
                if ( !value )
                {
                    MomentOfInertia = double.PositiveInfinity;
                }
                else
                {
                    UpdateBody( this.Mass );
                }
            }
        }

        /// <summary>
        /// Olion massa. Mitä suurempi massa, sitä suurempi voima tarvitaan olion liikuttamiseksi.
        /// </summary>
        /// <remarks>
        /// Massan asettaminen muuttaa myös hitausmomenttia (<c>MomentOfInertia</c>).
        /// </remarks>
        public double Mass
        {
            get { return Body.Mass.Mass; }
            set
            {
                // We should change the moment of intertia as well. If the mass is changed like, from
                // 1.0 to 100000.0, it would look funny if a heavy object would spin wildly from a small
                // touch.

                if ( !CanRotate )
                {
                    // The moment of inertia has been set to positive infinity,
                    // let's keep it that way and just set the mass.
                    Body.Mass.Mass = value;
                }
                else
                {
                    UpdateBody( value );
                }
            }
        }

        /// <summary>
        /// Olion hitausmomentti. Mitä suurempi hitausmomentti, sitä enemmän vääntöä tarvitaan
        /// olion pyörittämiseksi.
        /// </summary>
        public double MomentOfInertia
        {
            get { return Body.Mass.MomentOfInertia; }
            set { Body.Mass.MomentOfInertia = value; }
        }

        /// <summary>
        /// Lepokitka. Liikkeen alkamista vastustava voima, joka ilmenee kun olio yrittää lähteä liikkeelle
        /// toisen olion pinnalta (esim. laatikkoa yritetään työntää eteenpäin).
        /// </summary>
        public double StaticFriction
        {
            get { return Body.Coefficients.StaticFriction; }
            set { Body.Coefficients.StaticFriction = value; }
        }

        /// <summary>
        /// Liikekitka. Liikettä vastustava voima joka ilmenee kun kaksi oliota liikkuu toisiaan vasten
        /// (esim. laatikko liukuu maata pitkin). Arvot välillä 0.0 (ei kitkaa) ja 1.0 (täysi kitka).
        /// </summary>
        public double KineticFriction
        {
            get { return Body.Coefficients.DynamicFriction; }
            set { Body.Coefficients.DynamicFriction = value; }
        }

        /// <summary>
        /// Olion kimmoisuus. Arvo välillä 0.0-1.0.
        /// Arvolla 1.0 olio säilyttää kaiken vauhtinsa törmäyksessä. Mitä pienempi arvo,
        /// sitä enemmän olion vauhti hidastuu törmäyksessä.
        /// </summary>
        public double Restitution
        {
            get { return Body.Coefficients.Restitution; }
            set { Body.Coefficients.Restitution = value; }
        }

        /// <summary>
        /// Olion nopeus.
        /// </summary>
        [Save]
        public Vector Velocity
        {
            get
            {
                Vector2D v = Body.State.Velocity.Linear;
                return new Vector( v.X, v.Y );
            }
            set { Body.State.Velocity.Linear = new Vector2D(value.X, value.Y); }
        }

        /// <summary>
        /// Olion kulmanopeus.
        /// </summary>
        [Save]
        public double AngularVelocity
        {
            get { return Body.State.Velocity.Angular; }
            set { Body.State.Velocity.Angular = value; }
        }

        /// <summary>
        /// Olion kiihtyvyys.
        /// </summary>
        [Save]
        public Vector Acceleration
        {
            get
            {
                Vector2D v = Body.State.Acceleration.Linear;
                return new Vector( v.X, v.Y );
            }
            set { Body.State.Acceleration.Linear = new Vector2D( value.X, value.Y ); }
        }

        /// <summary>
        /// Olion kulmakiihtyvyys.
        /// </summary>
        [Save]
        public double AngularAcceleration
        {
            get { return Body.State.Acceleration.Angular; }
            set { Body.State.Acceleration.Angular = value; }
        }

        [Save]
        internal Vector ForceAccumulator
        {
            get { return new Vector( Body.State.ForceAccumulator.Linear.X, Body.State.ForceAccumulator.Linear.Y ); }
            set { Body.State.ForceAccumulator.Linear = new Vector2D( value.X, value.Y ); }
        }

        [Save]
        internal double AngularForceAccumulator
        {
            get { return Body.State.ForceAccumulator.Angular; }
            set { Body.State.ForceAccumulator.Angular = value; }
        }

        /// <summary>
        /// Olion muoto.
        /// </summary>
        public override Shape Shape
        {
            get { return base.Shape; }
            set
            {
                IShape newShape;
                if ( !requestedQuality.IsUnspecified )
                    newShape = CreatePhysicsShape( value, Size, requestedQuality );
                else
                    newShape = CreatePhysicsShape( value, Size );

                Body.Shape = newShape;
                base.Shape = value;
            }
        }

        /// <summary>
        /// Olion hidastuminen. Hidastaa olion vauhtia, vaikka se ei
        /// osuisi mihinkään. Vähän kuin väliaineen (esim. ilman tai veden)
        /// vastus. Oletusarvo on 1.0, jolloin hidastumista ei ole. Mitä
        /// pienempi arvo, sitä enemmän kappale hidastuu.
        /// 
        /// Yleensä kannattaa käyttää arvoja, jotka ovat lähellä ykköstä,
        /// esim. 0.95.
        /// </summary>
        public double LinearDamping
        {
            get { return Body.LinearDamping; }
            set { Body.LinearDamping = value; }
        }

        /// <summary>
        /// Olion pyörimisen hidastuminen.
        /// 
        /// <see cref="LinearDamping"/>
        /// </summary>
        public double AngularDamping
        {
            get { return Body.AngularDamping; }
            set { Body.AngularDamping = value; }
        }

        public override Brain Brain
        {
            get
            {
                return base.Brain;
            }
            set
            {
                if ( value != null )
                    Body.Collided += OnCollision;
                else
                    Body.Collided -= OnCollision;

                base.Brain = value;
            }
        }

        /// <summary>
        /// Jättääkö olio räjähdyksen paineaallon huomiotta.
        /// </summary>
        public bool IgnoresExplosions { get; set; }

        /// <summary>
        /// Jättääkö olio törmäyksen huomioimatta. Jos tosi, törmäyksestä
        /// tulee tapahtuma, mutta itse törmäystä ei tapahdu.
        /// </summary>
        public bool IgnoresCollisionResponse
        {
            get { return Body.IgnoresCollisionResponse; }
            set { Body.IgnoresCollisionResponse = value; }
        }

        /// <summary>
        /// Jättääkö olio painovoiman huomioimatta.
        /// </summary>
        public bool IgnoresGravity
        {
            get { return Body.IgnoresGravity; }
            set { Body.IgnoresGravity = value; }
        }

        /// <summary>
        /// Jättääkö olio kaikki fysiikkalogiikat (ks. <c>AddPhysicsLogic</c>)
        /// huomiotta. Vaikuttaa esim. painovoimaan, mutta ei törmäyksiin.
        /// </summary>
        public bool IgnoresPhysicsLogics
        {
            get { return Body.IgnoresPhysicsLogics; }
            set { Body.IgnoresPhysicsLogics = value; }
        }

        /// <summary>
        /// Tekee suorakulmaisen kappaleen, joka on staattinen (eli pysyy paikallaan).
        /// </summary>
        /// <param name="width">Kappaleen leveys</param>
        /// <param name="height">Kappaleen korkeus</param>
        /// <returns>Fysiikkaolio</returns>
        public static PhysicsObject CreateStaticObject( double width, double height )
        {
            PhysicsObject o = new PhysicsObject( width, height );
            o.MakeStatic();
            return o;
        }

        /// <summary>
        /// Tekee suorakulmaisen kappaleen, joka on staattinen (eli pysyy paikallaan).
        /// Kappaleen koko ja ulkonäkö ladataan parametrina annetusta kuvasta.
        /// </summary>
        /// <param name="width">Kappaleen leveys</param>
        /// <param name="height">Kappaleen korkeus</param>
        /// <returns>Fysiikkaolio</returns>
        public static PhysicsObject CreateStaticObject( Image image )
        {
            PhysicsObject o = new PhysicsObject( image );
            o.MakeStatic();
            return o;
        }

        /// <summary>
        /// Tekee vapaamuotoisen kappaleen, joka on staattinen (eli pysyy paikallaan).
        /// </summary>
        /// <param name="width">Kappaleen leveys</param>
        /// <param name="height">Kappaleen korkeus</param>
        /// <param name="shape">Kappaleen muoto</param>
        /// <returns>Fysiikkaolio</returns>
        public static PhysicsObject CreateStaticObject( double width, double height, Shape shape )
        {
            PhysicsObject o = new PhysicsObject( width, height, shape );
            o.MakeStatic();
            return o;
        }

        /// <summary>
        /// Tekee vapaamuotoisen kappaleen, joka on staattinen (eli pysyy paikallaan).
        /// Tällä rakentajalla voidaan määrittää myös törmäystarkistuksen laatu.
        /// </summary>
        /// <param name="width">Kappaleen leveys</param>
        /// <param name="height">Kappaleen korkeus</param>
        /// <param name="shape">Kappaleen muoto</param>
        /// <param name="quality">Törmäystarkistuksen laatu. CollisionShapeQuality.Best = paras (hitain), CollisionShapeQuality.Worst = huonoin (nopein)</param>
        /// <returns>Fysiikkaolio</returns>
        public static PhysicsObject CreateStaticObject( double width, double height, Shape shape, CollisionShapeQuality quality )
        {
            PhysicsObject o = new PhysicsObject( width, height, shape, quality );
            o.MakeStatic();
            return o;
        }

        public static PhysicsObject CreateStaticObject( double width, double height, Shape shape, double maxDistanceBetweenVertices, double gridSpacing )
        {
            PhysicsObject o = new PhysicsObject( width, height, shape, maxDistanceBetweenVertices, gridSpacing );
            o.MakeStatic();
            return o;
        }

        internal static PhysicsObject CreateStaticObject( double width, double height, Shape shape, IShape physicsShape )
        {
            PhysicsObject o = new PhysicsObject( width, height, shape, physicsShape );
            o.MakeStatic();
            return o;
        }

        /// <summary>
        /// Luo uuden fysiikkaolion.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        public PhysicsObject( double width, double height )
            : this( width, height, Shape.Rectangle )
        {
        }

        /// <summary>
        /// Luo uuden fysiikkaolion.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        /// <param name="shape">Muoto.</param>
        public PhysicsObject( double width, double height, Shape shape )
            : this( width, height, shape, CreatePhysicsShape( shape, new Vector( width, height ) ) )
        {
        }

        /// <summary>
        /// Luo uuden fysiikkaolion.
        /// Kappaleen koko ja ulkonäkö ladataan parametrina annetusta kuvasta.
        /// </summary>
        /// <param name="image">Kuva</param>
        public PhysicsObject( Image image )
            : this( image.Width, image.Height, Shape.Rectangle )
        {
            this.Image = image;
        }

        /// <summary>
        /// Luo uuden fysiikkaolion asettaen laadun törmäyskappaleelle. Käytä tätä rakentajaa vain,
        /// jos törmäystunnistuksen laatu ei ole tyydyttävää.
        /// </summary>
        /// <remarks>
        /// Mitä parempi laatu, sen tarkempi törmäysten tunnistus mutta toisaalta sen hitaampi
        /// laskenta. Jos laatu on liian huono, kappaleet saattavat mennä toistensa läpi.
        /// 
        /// Sopiva arvo laadulle riippuu tilanteesta. Jos tarvitsee vain tietää, milloin kappale törmää
        /// johonkin toiseen, riitää huonompikin laatu. Jos taas esimerkiksi halutaan, että kappaleita
        /// voi pinota päällekkäin, tarvitaan parempaa laatua. Toisaalta, jos peli pyörii liian hitaasti,
        /// voi olla parempi heikentää joidenkin kappaleiden laatua. Sopiva arvo löytyy kokeilemalla.
        /// </remarks>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        /// <param name="shape">Muoto.</param>
        /// <param name="quality">Törmäystunnistuksen laatu.</param>
        /// <example>
        /// PhysicsObject laatikko = new PhysicsObject( 20.0, 30.0, Shape.Rectangle, CollisionShapeQuality.FromValue(0.6) );
        /// </example>
        public PhysicsObject( double width, double height, Shape shape, CollisionShapeQuality quality )
            : this( width, height, shape, CreatePhysicsShape( shape, new Vector( width, height ), quality ) )
        {
            requestedQuality = quality;
        }

        /// <summary>
        /// Luo uuden fysiikkaolion antaen parametreja fysiikan laskentaan. Käytä tätä rakentajaa vain,
        /// jos on tarvetta kokeilla eri parametrien vaikutusta törmäyksen laatuun.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        /// <param name="shape">Muoto.</param>
        /// <param name="maxDistanceBetweenVertices">Suurin sallittu etäisyys fysiikkakuvion verteksien välillä.</param>
        /// <param name="gridSpacing">Kuinka pientä ruudukkoa fysiikkakuvion laskennassa käytetään. Ei ole käytössä ympyrälle.</param>
        /// <remarks>
        /// Sekä <c>maxDistanceBetweenVertices</c> että <c>gridSpacing</c> -parametrit vaikuttavat fysiikan
        /// laskentaan siten, että pienemmällä arvolla saadaan tarkempi laskenta, suuremmalla nopeampi. Oikea
        /// arvo on siis kompromissi fysiikan tarkkuuden ja nopean laskennan välillä.
        /// 
        /// Suositeltavat (ei välttämättä parhaat mahdolliset) arvot olisivat seuraavat: <c>GridSpacing</c>
        /// korkeintaan puolet kuvion kahden kulmapisteen etäisyydestä (esim. neliölle puolet etäisyydestä
        /// ja <c>maxDistanceBetweenVertices</c> vastaavasti korkeintaan kolmannes.
        /// 
        /// Sopivat arvot löytynevät parhaiten kokeilemalla.
        /// </remarks>
        public PhysicsObject( double width, double height, Shape shape, double maxDistanceBetweenVertices, double gridSpacing )
            : this( width, height, shape, CreatePhysicsShape( shape, new Vector( width, height ), maxDistanceBetweenVertices, gridSpacing ) )
        {
        }

        /// <summary>
        /// Luo fysiikkaolion, jonka muotona on säde.
        /// </summary>
        /// <param name="raySegment">Säde.</param>
        public PhysicsObject( RaySegment raySegment )
            : this( 1, 1, raySegment )
        {
        }

        /// <summary>
        /// Initializes the object with the given physics shape. The size of
        /// the physicsShape must be the one given.
        /// </summary>
        internal PhysicsObject( double width, double height, Shape shape, IShape physicsShape )
            : base( width, height, shape )
        {
            Coefficients c = new Coefficients( DefaultCoefficients.Restitution, DefaultCoefficients.StaticFriction, DefaultCoefficients.DynamicFriction );
            Body = new Body( new PhysicsState( ALVector2D.Zero ), physicsShape, DefaultMass, c, new Lifespan() );
            Body.Tag = this;
            if ( Game is PhysicsGame )
            {
                PhysicsGame pg = (PhysicsGame)Game;
            }
        }

        private void AddCollisionIgnoreGroup()
        {
            if ( _ignoreGroup == 0 )
            {
                Body.CollisionIgnorer = null;
            }
            else
            {
                if ( PhysicsGame.IgnoreGroups == null ) PhysicsGame.IgnoreGroups = new Dictionary<int, ObjectIgnorer>();
                if ( !PhysicsGame.IgnoreGroups.ContainsKey( _ignoreGroup ) ) PhysicsGame.IgnoreGroups.Add( _ignoreGroup, new ObjectIgnorer() );
                Body.CollisionIgnorer = PhysicsGame.IgnoreGroups[_ignoreGroup];
            }
        }

        private static double GetDefaultMaxDistanceBetweenVertices( double width, double height )
        {
            return Math.Min( width, height ) / 3;
        }

        private static double GetDefaultGridSpacing( double width, double height )
        {
            return Math.Min( width, height ) / 2;
        }

        /// <summary>
        /// Tekee oliosta staattisen. Staattinen olio ei liiku muiden olioiden törmäyksistä,
        /// vaan ainoastaan muuttamalla suoraan sen paikkaa tai nopeutta.
        /// </summary>
        protected void MakeStatic()
        {
            Mass = double.PositiveInfinity;
            CanRotate = false;
            IgnoresGravity = true;
        }

        /// <summary>
        /// This updates the mass and momentOfInertia based on the new mass.
        /// </summary>
        private void UpdateBody( double mass )
        {
            Body.Mass = Body.GetMassInfo( mass, Body.Shape );
        }

        internal static IShape CreatePhysicsShape( Shape shape, Vector size )
        {
            double vertexDistance = GetDefaultMaxDistanceBetweenVertices( size.X, size.Y );
            double gridSpacing = GetDefaultGridSpacing( size.X, size.Y );
            return CreatePhysicsShape( shape, size, vertexDistance, gridSpacing );
        }

        internal static IShape CreatePhysicsShape( Shape shape, Vector size, CollisionShapeQuality quality )
        {
            double smaller = Math.Min( size.X, size.Y );

            // The CollisionTest project (a stack of boxes) shows that gridspacing value can be
            // a little bigger than vertexDistance. Thus the differences here.
            double minimumGridSpacing = 0.05 * smaller;
            double maximumGridSpacing = 0.8 * smaller;
            double minimumVertexDistance = smaller * 0.02;
            double maximumVertexDistance = smaller * 0.5;

            // The logic is inverted here: the smaller the values, the better the quality.
            double maxLengthBetweenVertexes = maximumVertexDistance - quality.Value * ( maximumVertexDistance - minimumVertexDistance );
            double gridSpacing = maximumGridSpacing - quality.Value * ( maximumGridSpacing - minimumGridSpacing );

            return CreatePhysicsShape( shape, size, maxLengthBetweenVertexes, gridSpacing );
        }

        /// <summary>
        /// Creates a shape to be used in the Physics Body. A physics shape is scaled to the
        /// size of the object. In addition, it has more vertices and some additional info
        /// that is used in collision detection. The <c>maxLengthBetweenVertexes</c> and
        /// <c>gridSpacing</c> parameters affect (for most shapes) how accurately the collision
        /// detection is calculated for the shape.
        /// </summary>
        internal static IShape CreatePhysicsShape( Shape shape, Vector size, double maxLengthBetweenVertexes, double gridSpacing )
        {
            if ( shape is RaySegment )
            {
                RaySegment raySegment = (RaySegment)shape;
                Physics2DDotNet.Shapes.RaySegment singleSegment = new Physics2DDotNet.Shapes.RaySegment(
                    new Vector2D(raySegment.Origin.X, raySegment.Origin.Y),
                    new Vector2D(raySegment.Direction.X, raySegment.Direction.Y),
                    raySegment.Length );
                return new RaySegmentsShape( singleSegment );
            }
            else if ( shape is Circle )
            {
                Debug.Assert( shape.IsUnitSize );

                double smaller = Math.Min( size.X, size.Y );
                double bigger = Math.Max( size.X, size.Y );
                // Average between width and height.
                double r = smaller + ( bigger - smaller ) / 2;
                int vertexCount = (int)Math.Ceiling( ( 2 * Math.PI * r ) / maxLengthBetweenVertexes );

                //return new CircleShape(hw, Math.Max(16, vertexCount));
                Vector2D[] vertexes = new Vector2D[vertexCount];

                double a = 0.5 * size.X;
                double b = 0.5 * size.Y;

                for ( int i = 0; i < vertexCount; i++ )
                {
                    double t = ( i * 2 * Math.PI ) / vertexCount;
                    double x = a * Math.Cos( t );
                    double y = b * Math.Sin( t );
                    vertexes[i] = new Vector2D( x, y );
                }

                return new PolygonShape( vertexes, gridSpacing );
            }
            else
            {
                Vector2D[] originalVertexes = new Vector2D[shape.Cache.OutlineVertices.Length];
                for ( int i = 0; i < shape.Cache.OutlineVertices.Length; i++ )
                {
                    Vector v = shape.Cache.OutlineVertices[i];
                    if ( shape.IsUnitSize )
                    {
                        v.X *= size.X;
                        v.Y *= size.Y;
                    }
                    originalVertexes[i] = new Vector2D(v.X, v.Y);
                }

                Vector2D[] polyVertexes = VertexHelper.Subdivide(originalVertexes, maxLengthBetweenVertexes);

                return new PolygonShape(polyVertexes, gridSpacing);
            }
        }

        /// <summary>
        /// Työntää oliota.
        /// </summary>
        /// <param name="force">Voima, jolla oliota työnnetään.</param>
        public virtual void Push( Vector force )
        {
            Body.ApplyForce(new Vector2D(force.X, force.Y));
        }

        /// <summary>
        /// Työntää oliota tietyn ajan tietyllä voimalla.
        /// </summary>
        /// <param name="force">Voima, jolla oliota työnnetään sekunnissa.</param>
        /// <param name="time">Aika, kuinka kauan voimaa pidetään yllä.</param>
        public virtual void Push( Vector force, TimeSpan time )
        {
            Body.ApplyForce( new Vector2D( force.X, force.Y ) );
            ActiveForces.Add( new Force( force, time ) );
        }

        /// <summary>
        /// Kohdistaa kappaleeseen impulssin. Tällä kappaleen saa nopeasti liikkeeseen.
        /// </summary>
        public virtual void Hit( Vector impulse )
        {
            Body.ApplyImpulse(new Vector2D(impulse.X, impulse.Y));
        }

        /// <summary>
        /// Kohdistaa kappaleeseen vääntövoiman. Voiman suunta riippuu merkistä.
        /// </summary>
        /// <param name="torque">Vääntövoima.</param>
        public virtual void ApplyTorque( double torque )
        {
            Body.ApplyTorque(torque);
        }

        /// <summary>
        /// Pysäyttää olion.
        /// </summary>
        public virtual void Stop()
        {
            Body.ClearForces();
            Body.State.Acceleration = ALVector2D.Zero;
            Body.State.Velocity = ALVector2D.Zero;
            Body.State.ForceAccumulator = ALVector2D.Zero;
        }

        /// <summary>
        /// Pysäyttää olion liikkeen vaakasuunnassa.
        /// </summary>
        public void StopHorizontal()
        {
            ALVector2D oldAcc = Body.State.Acceleration;
            Body.State.Acceleration = new ALVector2D( oldAcc.Angular, 0, oldAcc.Y );

            ALVector2D oldVel = Body.State.Velocity;
            Body.State.Velocity = new ALVector2D( oldVel.Angular, 0, oldVel.Y );

            ALVector2D oldForce = Body.State.ForceAccumulator;
            Body.State.ForceAccumulator = new ALVector2D( oldForce.Angular, 0, oldForce.Y );
        }

        /// <summary>
        /// Pysäyttää olion liikkeen pystysuunnassa.
        /// </summary>
        public void StopVertical()
        {
            ALVector2D oldAcc = Body.State.Acceleration;
            Body.State.Acceleration = new ALVector2D( oldAcc.Angular, oldAcc.X, 0 );

            ALVector2D oldVel = Body.State.Velocity;
            Body.State.Velocity = new ALVector2D( oldVel.Angular, oldVel.X, 0 );

            ALVector2D oldForce = Body.State.ForceAccumulator;
            Body.State.ForceAccumulator = new ALVector2D( oldForce.Angular, oldForce.X, 0 );
        }

        public override void Destroy()
        {
            Body.Lifetime.IsExpired = true;
            base.Destroy();
        }

        private void OnCollision( object sender, CollisionEventArgs e )
        {
            PhysicsObject target = (PhysicsObject)e.Other.Tag;
            Brain.OnCollision( target );
        }

        public override void Update( Time time )
        {
            for ( int i = ActiveForces.Count - 1; i >= 0; i-- )
            {
                if ( ActiveForces[i].IsDestroyed() )
                {
                    ActiveForces.RemoveAt( i );
                    continue;
                }

                // Apply the force
                Push( ActiveForces[i].Value * time.SinceLastUpdate.TotalSeconds );
            }

            base.Update( time );
        }

        /// <summary>
        /// Siirtää oliota.
        /// </summary>
        /// <param name="movement">Vektori, joka määrittää kuinka paljon siirretään.</param>
        public override void Move( Vector movement )
        {
            Vector dv = movement - this.Velocity;
            Hit( Mass * dv );
        }

        protected override void MoveToTarget()
        {
            if ( !moveTarget.HasValue )
            {
                Stop();
                moveTimer.Stop();
                return;
            }

            Vector d = moveTarget.Value - Position;
            double vt = moveSpeed * moveTimer.Interval;

            if ( d.Magnitude < vt )
            {
                Vector targetLoc = moveTarget.Value;
                Stop();
                moveTimer.Stop();
                moveTarget = null;
                OnArrivedAt( targetLoc );
            }
            else
            {
                Vector dv = Vector.FromLengthAndAngle( moveSpeed, d.Angle ) - this.Velocity;
                Hit( Mass * dv );
            }
        }
    }
}
