﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using AdvanceMath;

namespace Jypeli
{
    /// <summary>
    /// Satunnaisgeneraattori. Luo satunnaisia arvoja, mm. lukuja, vektoreita sekä kulmia.
    /// </summary>
    public static class RandomGen
    {
        private static Random rand = new Random();

        /// <summary>
        /// Palauttaa satunnaisen totuusarvon.
        /// </summary>
        public static bool NextBool()
        {
            return ( rand.NextDouble() >= 0.5 );
        }

        /// <summary>
        /// Palauttaa satunnaisen kokonaisluvun, joka on vähintään 0 ja pienempi kuin <code>max</code>.
        /// </summary>
        /// <returns></returns>
        public static int NextInt( int maxValue )
        {
            return rand.Next( maxValue );
        }

        /// <summary>
        /// Palauttaa satunnaisen kokonaisluvun, joka on vähintään <code>min</code> ja pienempi kuin <code>max</code>.
        /// </summary>
        public static int NextInt( int min, int max )
        {
            return rand.Next( min, max );
        }        

        /// <summary>
        /// Palauttaa satunnaisen liukuluvun parametrien <code>min</code> ja <code>max</code> väliltä.
        /// </summary>
        public static double NextDouble( double min, double max )
        {
            return min + rand.NextDouble() * ( max - min );
        }

        /// <summary>
        /// Palauttaa satunnaisen suunnan.
        /// </summary>
        public static Direction NextDirection()
        {
            double randdir = rand.NextDouble();

            if ( randdir <= 0.25 )
                return Direction.Up;

            if ( randdir <= 0.5 )
                return Direction.Down;

            if ( randdir <= 0.75 )
                return Direction.Left;

            return Direction.Right;
        }

        /// <summary>
        /// Palauttaa double-taulukon.
        /// </summary>
        /// <param name="min">Pienin arvo.</param>
        /// <param name="max">Suurin arvo.</param>
        /// <param name="size">Taulukon koko.</param>
        /// <returns>Taulukko.</returns>
        public static double[] NextDoubleArray( double min, double max, int size )
        {
            double[] array = new double[size];

            for ( int i = 0; i < size; i++ )
            {
                array[i] = NextDouble( min, max );
            }

            return array;
        }

        /// <summary>
        /// Palauttaa double-taulukon.
        /// </summary>
        /// <param name="min">Pienin arvo.</param>
        /// <param name="max">Suurin arvo.</param>
        /// <param name="size">Taulukon koko.</param>
        /// <param name="maxchange">Suurin sallittu muutos kahden luvun välillä.</param>
        /// <returns>Taulukko.</returns>
        public static double[] NextDoubleArray( double min, double max, int size, int maxchange )
        {
            double[] array = new double[size];
            double curmin = min;
            double curmax = max;

            for ( int i = 0; i < size; i++ )
            {
                array[i] = NextDouble( curmin, curmax );

                curmin = AdvanceMath.MathHelper.Max((float)min, (float)(array[i] - maxchange));
                curmax = AdvanceMath.MathHelper.Min((float)max, (float)(array[i] + maxchange));
            }

            return array;
        }

        /// <summary>
        /// Palauttaa satunnaisen värin.
        /// </summary>
        /// <returns>Väri.</returns>
        public static Color NextColor()
        {
            return new Color( (float)rand.NextDouble(), (float)rand.NextDouble(), (float)rand.NextDouble(), 1 );
        }

        /// <summary>
        /// Palauttaa satunnaisen värin.
        /// </summary>
        /// <returns>Väri.</returns>
        public static Color NextColor( Color first, Color second )
        {
            return Color.Lerp( first, second, (float)rand.NextDouble());
        }

        /// <summary>
        /// Palauttaa satunnaisen kulman.
        /// </summary>
        /// <returns>Kulma.</returns>
        public static Angle NextAngle()
        {
            return Angle.FromRadians( NextDouble( 0, 2 * Math.PI ) );
        }

        /// <summary>
        /// Palauttaa satunnaisen kulman nollasta annettuun maksimiin.
        /// </summary>
        /// <param name="max">Maksimikulma.</param>
        /// <returns>Kulma.</returns>
        public static Angle NextAngle( Angle max )
        {
            return Angle.FromRadians( NextDouble( 0, max.Radians ) );
        }

        /// <summary>
        /// Palauttaa satunnaisen kulman tietyltä väliltä.
        /// </summary>
        /// <param name="min">Minimikulma.</param>
        /// <param name="max">Maksimikulma.</param>
        /// <returns>Kulma.</returns>
        public static Angle NextAngle( Angle min, Angle max )
        {
            double a1 = min.Radians;
            double a2 = max.Radians;
            while ( a2 < a1 ) a2 += 2 * Math.PI;
            return Angle.FromRadians( NextDouble( a1, a2 ) );
        }

        /// <summary>
        /// Palauttaa satunnaisen kokonaisluvun annettujen todennäköisyyksien
        /// perusteella.
        /// </summary>
        /// <param name="p">Todennäköisyydet. 0 = ei koskaan, 1 = varmasti, 0.5 = 50% jne.</param>
        /// <example>
        /// int luku = RandomGen.NextIntWithProbabilities( 0.4 );  // palauttaa 40% tod.näk. nollan, muuten ykkösen (60%)
        /// int luku2 = RandomGen.NextIntWithProbabilities( 0.6, 0.2 );  // palauttaa 60% tod.näk. nollan, 20% tn. ykkösen ja muuten kakkosen (40%)
        /// int luku3 = RandomGen.NextIntWithProbabilities( 0.6, 0.4 );  // palauttaa 60% tod.näk. nollan ja 40% tn. ykkösen
        /// </example>
        /// <returns>Kokonaisluku väliltä 0 - (p+1)</returns>
        public static int NextIntWithProbabilities( params double[] p )
        {
            double randomNum = rand.NextDouble();
            double accumulator = 0;

            for ( int i = 0; i < p.Length; i++ )
            {
                accumulator += p[i];
                if ( randomNum < accumulator ) return i;
            }

            return p.Length;
        }
    }
}
