﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using System.Diagnostics;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Jypeli.Effects;

using XnaV2 = Microsoft.Xna.Framework.Vector2;
using XnaV3 = Microsoft.Xna.Framework.Vector3;


namespace Jypeli
{
    /// <summary>
    /// Luokka, joka sisältää metodeita kuvioiden ja tekstuurien piirtämiseen 2D-tasossa.
    /// </summary>
    /// <remarks>
    /// Toteutus on yksinkertainen ja siten hidas. Jos on paljon samankaltaisia piirrettäviä
    /// kohteita, niin käytä mielummin Batch-luokkia.
    /// </remarks>
    internal class Renderer
    {
        /// <summary>
        /// Vertices that form a rectangle on which to draw textures.
        /// </summary>
        static readonly VertexPositionTexture[] textureVertices = new VertexPositionTexture[]
        {
            new VertexPositionTexture( new XnaV3(-0.5f, 0.5f, 0), new XnaV2(0.0f, 0.0f) ),
            new VertexPositionTexture( new XnaV3(-0.5f, -0.5f, 0), new XnaV2(0.0f, 1.0f) ),
            new VertexPositionTexture( new XnaV3(0.5f, 0.5f, 0), new XnaV2(1.0f, 0.0f) ),
            new VertexPositionTexture( new XnaV3(0.5f, -0.5f, 0), new XnaV2(1.0f, 1.0f) )
        };

        /// <summary>
        /// Indices that form two triangles from the vertex array.
        /// </summary>
        static readonly Int16[] textureTriangleIndices = new short[]
        {
            0, 1, 2,
            1, 3, 2
        };

        public bool LightingEnabled { get; set; }

        static readonly RasterizerState rsSolidCounterClockwise = new RasterizerState
        {
            CullMode = CullMode.CullCounterClockwiseFace,
            FillMode = FillMode.Solid,
        };

        public Renderer()
        {
        }

        public void DrawImage( Image texture, ref Matrix matrix, Vector wrapSize )
        {
            var device = Game.GraphicsDevice;

            VertexPositionTexture[] tempVertices = new VertexPositionTexture[textureVertices.Length];
            for (int i = 0; i < textureVertices.Length; i++)
            {
                tempVertices[i].Position = textureVertices[i].Position;
            }

            // Since the origo in texture coordinates is at upper left corner,
            // but at center in an object's coordinates, we need to make some
            // adjustments here:
            float left = (float)( -wrapSize.X / 2 + 0.5 );
            float right = (float)( wrapSize.X / 2 + 0.5 );
            float top = (float)( -wrapSize.Y / 2 + 0.5 );
            float bottom = (float)( wrapSize.Y / 2 + 0.5 );

            tempVertices[0].TextureCoordinate = new XnaV2( left, top );
            tempVertices[1].TextureCoordinate = new XnaV2( left, bottom );
            tempVertices[2].TextureCoordinate = new XnaV2( right, top );
            tempVertices[3].TextureCoordinate = new XnaV2( right, bottom );

            device.RasterizerState = RasterizerState.CullClockwise;
#if WINDOWS_PHONE
            // The WP7 emulator interprets cullmodes incorectly sometimes.
            device.RasterizerState = RasterizerState.CullNone;
#endif

            device.BlendState = BlendState.AlphaBlend;
            
            Effect effect = Graphics.GetTextureEffect( ref matrix, texture.XNATexture, LightingEnabled );
            foreach ( EffectPass pass in effect.CurrentTechnique.Passes )
            {
                pass.Apply();
                device.SamplerStates[0] = SamplerState.LinearClamp;
                Game.GraphicsDevice.DrawUserIndexedPrimitives<VertexPositionTexture>(
                    PrimitiveType.TriangleList,
                    tempVertices, 0, tempVertices.Length,
                    textureTriangleIndices, 0, textureTriangleIndices.Length / 3
                    );
            }
        }

        /// <summary>
        /// Piirtää kuvion niin, että tekstuuri täyttää sen.
        /// </summary>
        public void DrawShape( Shape shape, ref Matrix transformation, ref Matrix textureTransformation, Image texture, Vector textureWrapSize, Color color )
        {
            var device = Game.GraphicsDevice;

            // Set up for drawing the shape to the stencil buffer

            device.Clear( ClearOptions.Stencil, Color.Black.AsXnaColor(), 0, 0 );
            /*device.RenderState.StencilEnable = true;
            //device.RenderState.ColorWriteChannels = ColorWriteChannels.None;
            device.RenderState.ColorWriteChannels = ColorWriteChannels.All;
            device.RenderState.ReferenceStencil = 0;
            device.RenderState.StencilFunction = CompareFunction.Equal;
            device.RenderState.StencilPass = StencilOperation.IncrementSaturation;
            device.RenderState.DepthBufferWriteEnable = false;*/

            device.DepthStencilState.StencilEnable = true;
            device.DepthStencilState.ReferenceStencil = 0;
            device.DepthStencilState.StencilFunction = CompareFunction.Equal;
            device.DepthStencilState.StencilPass = StencilOperation.IncrementSaturation;
            device.DepthStencilState.DepthBufferWriteEnable = false;
            device.BlendState.ColorWriteChannels = ColorWriteChannels.All;

            // Draw the shape to the stencil buffer
            DrawShape( shape, ref transformation, color );

            // Set up for drawing a texture according to the stencil buffer
            
            /*device.RenderState.DepthBufferWriteEnable = true;
            device.RenderState.ColorWriteChannels = ColorWriteChannels.All;
            device.RenderState.ReferenceStencil = 1;
            device.RenderState.StencilPass = StencilOperation.Keep;
            device.RenderState.StencilFunction = CompareFunction.LessEqual;*/

            device.DepthStencilState.DepthBufferWriteEnable = true;
            device.BlendState.ColorWriteChannels = ColorWriteChannels.All;
            device.DepthStencilState.ReferenceStencil = 1;
            device.DepthStencilState.StencilPass = StencilOperation.Keep;
            device.DepthStencilState.StencilFunction = CompareFunction.LessEqual;
            
            
            

            DrawImage( texture, ref textureTransformation, textureWrapSize );

            // Done with the stencil buffer
            //device.RenderState.StencilEnable = false;
            device.DepthStencilState.StencilEnable = false;
        }

        public void DrawShape( Shape shape, ref Matrix matrix, Color color )
        {
            if ( shape is RaySegment )
            {
                DrawRaySegment( (RaySegment)shape, ref matrix, color );
            }
            else if ( shape.Cache.Triangles != null )
            {
                DrawFilledShape( shape.Cache, ref matrix, color );
            }
            else
            {
                DrawPolygon( shape.Cache.OutlineVertices, ref matrix, color );
            }
        }

        public void DrawRaySegment( RaySegment segment, ref Matrix matrix, Color color )
        {
            var device = Game.GraphicsDevice;

            Vector endPoint = segment.Origin + segment.Direction * segment.Length;

            VertexPositionColor[] colorVertices = new VertexPositionColor[2];
            colorVertices[0].Position = new Vector3( (float)segment.Origin.X, (float)segment.Origin.Y, 0 );
            colorVertices[0].Color = color.AsXnaColor();
            colorVertices[1].Position = new Vector3( (float)endPoint.X, (float)endPoint.Y, 0 );
            colorVertices[1].Color = color.AsXnaColor();

            BasicEffect effect = Graphics.BasicColorEffect;
            effect.World = matrix;
            foreach ( EffectPass pass in effect.CurrentTechnique.Passes )
            {
                pass.Apply();
                device.SamplerStates[0] = SamplerState.LinearClamp;
                device.DrawUserPrimitives<VertexPositionColor>( PrimitiveType.LineStrip, colorVertices, 0, 1 );
            }
        }

        /// <summary>
        /// Piirtää suorakulmion.
        /// </summary>
        internal void DrawRectangle( ref Matrix matrix, Color color )
        {
            Vector[] vertices = new Vector[]
            {
                new Vector(-0.5, 0.5),
                new Vector(-0.5, -0.5),
                new Vector(0.5, -0.5),
                new Vector(0.5, 0.5),
            };
            Game.Renderer.DrawPolygon( vertices, ref matrix, color );
        }

        internal void DrawFilledShape( ShapeCache cache, ref Matrix matrix, Color color )
        {
            var device = Game.GraphicsDevice;

            VertexPositionColor[] vertices = new VertexPositionColor[cache.Vertices.Length];
            for ( int i = 0; i < vertices.Length; i++ )
            {
                Vector v = cache.Vertices[i];
                vertices[i] = new VertexPositionColor( new XnaV3( (float)v.X, (float)v.Y, 0 ), color.AsXnaColor() );
            }

            Int16[] indices = new Int16[cache.Triangles.Length * 3];
            for ( int i = 0; i < cache.Triangles.Length; i++ )
            {
                indices[3 * i] = cache.Triangles[i].i1;
                indices[3 * i + 1] = cache.Triangles[i].i2;
                indices[3 * i + 2] = cache.Triangles[i].i3;
            }

            device.RasterizerState = RasterizerState.CullCounterClockwise;
            device.BlendState = BlendState.NonPremultiplied;
#if WINDOWS_PHONE
            // The WP7 emulator interprets cullmodes incorectly sometimes.
            device.RasterizerState = RasterizerState.CullNone;
#endif

            Effect effect = Graphics.GetColorEffect( ref matrix, LightingEnabled );
            foreach ( EffectPass pass in effect.CurrentTechnique.Passes )
            {
                pass.Apply();
                device.SamplerStates[0] = SamplerState.LinearClamp;
                device.DrawUserIndexedPrimitives<VertexPositionColor>(
                    PrimitiveType.TriangleList,
                    vertices, 0, vertices.Length,
                    indices, 0, indices.Length / 3
                    );
            }
        }

        public void DrawPolygon( Vector[] vertices, ref Matrix matrix, Color color )
        {
            if ( vertices.Length < 3 )
                throw new ArgumentException( "Polygon must have at least three vertices" );

            var device = Game.GraphicsDevice;

            VertexPositionColor[] colorVertices = new VertexPositionColor[vertices.Length];
            for ( int i = 0; i < colorVertices.Length; i++ )
            {
                Vector p = vertices[i];
                colorVertices[i] = new VertexPositionColor(
                    new XnaV3( (float)p.X, (float)p.Y, 0 ),
                    color.AsXnaColor()
                    );
            }

            int n = colorVertices.Length;
            Int16[] indices = new Int16[2 * n];
            for ( int i = 0; i < ( n - 1 ); i++ )
            {
                indices[2 * i] = (Int16)i;
                indices[2 * i + 1] = (Int16)( i + 1 );
            }
            indices[2 * ( n - 1 )] = (Int16)( n - 1 );
            indices[2 * ( n - 1 ) + 1] = (Int16)0;

            Effect effect = Graphics.GetColorEffect( ref matrix, LightingEnabled );
            foreach ( EffectPass pass in effect.CurrentTechnique.Passes )
            {
                pass.Apply();
                device.SamplerStates[0] = SamplerState.LinearClamp;
                device.DrawUserIndexedPrimitives<VertexPositionColor>(
                    PrimitiveType.LineStrip,
                    colorVertices, 0, colorVertices.Length,
                    indices, 0, indices.Length - 1
                    );

            }
        }


        internal void DrawVertices( Vector[] vertices, Matrix matrix, Color color )
        {
            VertexPositionColor[] pointVertices = new VertexPositionColor[vertices.Length];
            for ( int i = 0; i < pointVertices.Length; i++ )
            {
                Vector p = vertices[i];
                pointVertices[i] = new VertexPositionColor(
                    new XnaV3( (float)p.X, (float)p.Y, 0 ),
                    Color.Red.AsXnaColor()
                    );
            }

            var device = Game.GraphicsDevice;
            //device.RenderState.PointSize = 2;

            BasicEffect effect = Graphics.BasicColorEffect;
            effect.World = matrix;
            foreach ( var pass in effect.CurrentTechnique.Passes )
            {
                pass.Apply();
                device.SamplerStates[0] = SamplerState.LinearClamp;
                device.DrawUserPrimitives<VertexPositionColor>(
                    PrimitiveType.LineList,
                    pointVertices, 0, pointVertices.Length
                    );
            }
        }
    }
}
