﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using System.ComponentModel;
using AdvanceMath;
using Physics2DDotNet;
using Physics2DDotNet.Shapes;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using XnaRectangle = Microsoft.Xna.Framework.Rectangle;
using XnaColor = Microsoft.Xna.Framework.Color;
using System.Reflection;


namespace Jypeli
{
    /// <summary>
    /// Kuvio.
    /// </summary>
    public abstract class Shape
    {
        /// <summary>
        /// If true, the shape must be scaled by the size of the object
        /// that has the shape. Typically, an unit-sized object has width
        /// and height of 1.0.
        /// </summary>
        public abstract bool IsUnitSize { get; }

        internal abstract ShapeCache Cache { get; }

        /// <summary>
        /// Ympyrä.
        /// </summary>
        public static readonly Circle Circle = new Circle();

        /// <summary>
        /// Suorakulmio.
        /// </summary>
        public static readonly Rectangle Rectangle = new Rectangle();

        /// <summary>
        /// Tasasivuinen kolmio.
        /// </summary>
        public static readonly Triangle Triangle = new Triangle();

        /// <summary>
        /// Sydän.
        /// </summary>
        public static readonly Heart Heart = new Heart();

        /// <summary>
        /// Tähti.
        /// </summary>
        public static readonly Star Star = new Star();

        /// <summary>
        /// Timantti- / salmiakkikuvio
        /// </summary>
        public static readonly Shape Diamond = new RegularPolygon(4);

        /// <summary>
        /// Pentagoni eli viisikulmio.
        /// </summary>
        public static readonly Shape Pentagon = new RegularPolygon( 5 );

        /// <summary>
        /// Heksagoni eli kuusikulmio.
        /// </summary>
        public static readonly Shape Hexagon = new RegularPolygon( 6 );

        /// <summary>
        /// Oktagoni eli kahdeksankulmio.
        /// </summary>
        public static readonly Shape Octagon = new RegularPolygon( 8 );

        /// <summary>
        /// Luo kuvion annetusta kuvasta. Kuvassa tulee olla vain yksi
        /// yhtenäinen muoto (toisin sanoen kuvio ei voi koostua monesta osasta).
        /// </summary>
        /// <remarks>
        /// Kuvion luominen voi olla melko hidasta. Kannattaa luoda kuvio heti pelin alussa
        /// ja käyttää kerran luotua kuviota kaikille olioille.
        /// </remarks>
        /// <param name="image">Kuva, josta muoto luetaan.</param>
        public static Shape FromImage( Image image )
        {
            Texture2D texture = image.XNATexture;
            IBitmap bm = new TextureBitmap( texture );
            Vector2D[] vertices = VertexHelper.CreateFromBitmap( bm );

            for ( int i = 0; i < vertices.Length; i++ )
            {
                vertices[i].X /= texture.Width;
                vertices[i].Y /= texture.Height;

                // The center of the texture coordinates is at top left corner, but the center of
                // game objects is by default at the center of the object.
                vertices[i] -= new Vector2D( (float)0.5, (float)0.5 );
            }

            Vector[] polygonVertices = new Vector[vertices.Length];
            for ( int i = 0; i < vertices.Length; i++ )
                polygonVertices[i] = new Vector( vertices[i].X, vertices[i].Y );

            ShapeCache cache = new ShapeCache( polygonVertices );
            return new Polygon( cache );
        }

        /// <summary>
        /// Luo muodon merkkijonosta, esim. "Circle"
        /// </summary>
        /// <returns></returns>
        public static Shape FromString( string shapeStr )
        {
            Type shapeClass = typeof( Shape );
            BindingFlags flags = BindingFlags.GetField | BindingFlags.Public | BindingFlags.Static;
            FieldInfo selectedShape = shapeClass.GetField( shapeStr, flags );
            return (Shape)selectedShape.GetValue( null );
        }

        /// <summary>
        /// Luo säännöllisen monikulmion (polygonin)
        /// </summary>
        /// <param name="vertexCount">Kulmapisteiden määrä (3=kolmio, 4=neliö jne.)</param>
        /// <returns>Monikulmio</returns>
        public static Shape CreateRegularPolygon( int vertexCount )
        {
            if ( vertexCount < 3 ) throw new ArgumentException( "You need at least 3 vertices to create a polygon!" );
            return new RegularPolygon( vertexCount );
        }

        internal static ShapeCache CreateRegularPolygonCache( int vertexCount )
        {
            double angleStep = 2 * Math.PI / vertexCount;
            Int16 centerIndex = (Int16)vertexCount;

            Vector[] vertices = new Vector[vertexCount + 1];
            IndexTriangle[] triangles = new IndexTriangle[vertexCount];
            Int16[] outlineIndices = new Int16[vertexCount];

            for ( int i = 0; i < vertexCount; i++ )
            {
                double a = i * angleStep;
                vertices[i] = new Vector( 0.5 * Math.Cos( a ), 0.5 * Math.Sin( a ) );
                outlineIndices[i] = (Int16)i;
            }
            vertices[centerIndex] = Vector.Zero;

            for ( int i = 0; i < vertexCount - 1; i++ )
            {
                triangles[i] = new IndexTriangle( (Int16)i, centerIndex, (Int16)( i + 1 ) );
            }
            triangles[vertexCount - 1] = new IndexTriangle( (Int16)( vertexCount - 1 ), centerIndex, (Int16)0 );

            return new ShapeCache( vertices, triangles, outlineIndices );
        }
    }

    /// <summary>
    /// Ympyrä.
    /// </summary>
    public class Circle : Shape
    {
        private static ShapeCache _cache = CreateRegularPolygonCache( 64 );

        internal override ShapeCache Cache
        {
            get { return _cache; }
        }

        public override bool IsUnitSize { get { return true; } }

        internal Circle() { }
    }

    /// <summary>
    /// Suorakulmio.
    /// </summary>
    public class Rectangle : Shape
    {
        private static readonly Vector[] vertices = new Vector[]
        {
            new Vector( -0.5, -0.5 ),
            new Vector( -0.5, 0.5 ),
            new Vector( 0.5, -0.5 ),
            new Vector( 0.5, 0.5 ),
        };

        private static readonly IndexTriangle[] triangles = new IndexTriangle[]
        {
            new IndexTriangle( 0, 1, 2 ),
            new IndexTriangle( 2, 1, 3 ),
        };

        private static readonly Int16[] outlineIndices = new Int16[]
        {
            2, 3, 1, 0
        };

        private static readonly ShapeCache _cache = new ShapeCache( vertices, triangles, outlineIndices );

        internal override ShapeCache Cache { get { return _cache; } }

        public override bool IsUnitSize { get { return true; } }

        internal Rectangle() { }
    }

    /// <summary>
    /// Sydän.
    /// </summary>
    public class Heart : Shape
    {
        private static readonly Vector[] vertices = new Vector[]
        {
            new Vector( 0, -0.5 ),
            new Vector( 0.5, 0.2 ),
            new Vector( 0.4, 0.4 ),
            new Vector( 0.25, 0.5 ),
            new Vector( 0.1, 0.4 ),
            new Vector( 0, 0.2 ),
            new Vector( -0.1, 0.4 ),
            new Vector( -0.25, 0.5 ),
            new Vector( -0.4, 0.4 ),
            new Vector( -0.5, 0.2 )
       };

        private static readonly IndexTriangle[] triangles = new IndexTriangle[]
        {
            new IndexTriangle( 0, 9, 5 ),
            new IndexTriangle( 8, 7, 6 ),
            new IndexTriangle( 8, 6, 5 ),
            new IndexTriangle( 9, 8, 5 ),
            new IndexTriangle( 3, 2, 4 ),
            new IndexTriangle( 4, 2, 5 ),
            new IndexTriangle( 2, 1, 5 ),
            new IndexTriangle( 5, 1, 0 ),
        };

        private static readonly ShapeCache _cache = new ShapeCache( vertices, triangles );

        internal override ShapeCache Cache { get { return _cache; } }

        public override bool IsUnitSize { get { return true; } }

        internal Heart() { }
    }

    /// <summary>
    /// Tähti.
    /// </summary>
    public class Star : Shape
    {
        private static readonly Vector[] vertices = new Vector[]
        {
            new Vector( -0.5, -0.5 ),
            new Vector( -0.25, 0 ),
            new Vector( 0, -0.16 ),
            new Vector( -0.5, 0.2 ),
            new Vector( -0.15, 0.2 ),
            new Vector( 0, 0.5 ),
            new Vector( 0.15, 0.2 ),
            new Vector( 0.5, 0.2 ),
            new Vector( 0.25, 0 ),
            new Vector( 0.5, -0.5 ),
        };

        private static readonly IndexTriangle[] triangles = new IndexTriangle[]
        {
            new IndexTriangle( 0, 1, 2 ),
            new IndexTriangle( 1, 3, 4 ),
            new IndexTriangle( 1, 5, 8 ),
            new IndexTriangle( 8, 6, 7 ),
            new IndexTriangle( 9, 2, 8 ),
            new IndexTriangle( 1, 8, 2 ),
        };

        private static readonly Int16[] outlineIndices = new Int16[]
        {
            2, 9, 8, 7, 6, 5, 4, 3, 1, 0
        };

        private static readonly ShapeCache _cache = new ShapeCache( vertices, triangles, outlineIndices );

        internal override ShapeCache Cache { get { return _cache; } }

        public override bool IsUnitSize { get { return true; } }

        internal Star() { }
    }

    /// <summary>
    /// Tasasivuinen kolmio.
    /// </summary>
    public class Triangle : Shape
    {
        private static readonly Vector[] vertices = new Vector[]
        {
            new Vector( 0, 0.5 ),
            new Vector( 0.5, -0.5 ),
            new Vector( -0.5, -0.5 ),
        };

        private static readonly IndexTriangle[] triangles = new IndexTriangle[]
        {
            new IndexTriangle( 0, 1, 2 ),
        };

        private static readonly Int16[] outlineIndices = new Int16[]
        {
            2, 1, 0
        };

        private static readonly ShapeCache _cache = new ShapeCache( vertices, triangles, outlineIndices );

        internal override ShapeCache Cache { get { return _cache; } }

        public override bool IsUnitSize { get { return true; } }

        internal Triangle() { }
    }

    /// <summary>
    /// Jana.
    /// </summary>
    public class RaySegment : Shape
    {
        internal override ShapeCache Cache
        {
            get { throw new Exception( "Cache is not defined for RaySegment" ); }
        }

        public override bool IsUnitSize { get { return false; } }

        public Vector Origin;
        public Vector Direction;
        public double Length;

        public RaySegment( Vector origin, Vector direction, double length )
        {
            this.Origin = origin;
            this.Direction = direction;
            this.Length = length;
        }
    }

    /// <summary>
    /// Monikulmio.
    /// </summary>
    internal class Polygon : Shape
    {
        private bool isUnitSize = true;
        private ShapeCache _cache;

        /// <summary>
        /// If true, the shape must be scaled by the size of the object
        /// that has the shape. Typically, an unit-sized object has width
        /// and height of 1.0.
        /// </summary>
        public override bool IsUnitSize 
        {
            get { return isUnitSize; }
        }

        internal override ShapeCache Cache
        {
            get { return _cache; }
        }

        public Polygon( ShapeCache cache )
            : this( cache, true )
        {
        }

        public Polygon( ShapeCache cache, bool isUnitSize )
        {
            this._cache = cache;
            this.isUnitSize = isUnitSize;
        }
    }

    /// <summary>
    /// Säännöllinen monikulmio.
    /// </summary>
    internal class RegularPolygon : Polygon
    {
        public int CornerCount { get; internal set; }
        public override bool IsUnitSize { get { return true; } }

        public RegularPolygon( int vertexCount )
            : base( CreateRegularPolygonCache( vertexCount ) )
        {
            CornerCount = vertexCount;
        }
    }

    /// <summary>
    /// Muotojen määrityksessä käytettävä kolmio.
    /// </summary>
    [EditorBrowsable( EditorBrowsableState.Never )]
    public struct IndexTriangle
    {
        /// <summary>
        /// Kulmapisteet.
        /// </summary>
        public Int16 i1, i2, i3;

        /// <summary>
        /// Luo uuden kolmion. Parametreina kulmapisteiden indeksit lueteltuna myötäpäivään.
        /// </summary>
        public IndexTriangle( Int16 i1, Int16 i2, Int16 i3 )
        {
            this.i1 = i1;
            this.i2 = i2;
            this.i3 = i3;
        }
    }

    /// <summary>
    /// Sisältää valmiiksi lasketut kolmiot, joiden avulla piirtäminen on suoraviivaista.
    /// </summary>
    [EditorBrowsable( EditorBrowsableState.Never )]
    public class ShapeCache
    {
        // The vertices are in counter-clockwise order
        // because that is what the polygon shape of the physics
        // engine expects.

        /// <summary>
        /// Ulkoreunan verteksit, lueteltuna vastapäivään.
        /// </summary>
        public readonly Vector[] OutlineVertices;

        /// <summary>
        /// Kaikki verteksit, ml. kolmioiden kulmapisteet.
        /// </summary>
        public readonly Vector[] Vertices;

        /// <summary>
        /// Kolmiot, joiden avulla kuvio voidaan täyttää värillä.
        /// </summary>
        public readonly IndexTriangle[] Triangles;

        /// <summary>
        /// Luo kuvion kolmioilla, joiden avulla kuvio voidaan täyttää värillä. Kaikkien
        /// verteksien tulee olla kuvion ulkoreunalla.
        /// </summary>
        /// <param name="outlineVertices">Ulkoreunan verteksit, lueteltuna vastapäivään.</param>
        /// <param name="triangles">Kolmiot.</param>
        public ShapeCache( Vector[] outlineVertices, IndexTriangle[] triangles )
        {
            this.Vertices = this.OutlineVertices = outlineVertices;
            Triangles = triangles;
        }

        /// <summary>
        /// Luo kuvion pelkillä reuna-vertekseillä. Kuviolle ei tule tietoa kolmioista,
        /// näin ollen sitä ei voi täyttää värillä.
        /// </summary>
        /// <param name="outlineVertices">Ulkoreunan verteksit, lueteltuna vastapäivään.</param>
        public ShapeCache( Vector[] outlineVertices )
        {
            this.Vertices = this.OutlineVertices = outlineVertices;
            this.Triangles = null;
        }

        /// <summary>
        /// Luo kuvion, joka voidaan piirtää täytettynä värillä.
        /// </summary>
        /// <param name="vertices">Kaikki verteksit (ml. reunan verteksit sekä kolmioiden verteksit).</param>
        /// <param name="triangles">Kolmiot, joista kuvio koostuu.</param>
        /// <param name="outlineIndices">Ulkoreunan verteksit lueteltuna vastapäivään, indekseinä <c>vertices</c>-taulukkoon.</param>
        public ShapeCache( Vector[] vertices, IndexTriangle[] triangles, Int16[] outlineIndices )
        {
            Vertices = vertices;
            Triangles = triangles;

            OutlineVertices = new Vector[outlineIndices.Length];
            for ( int i = 0; i < outlineIndices.Length; i++ )
            {
                OutlineVertices[i] = vertices[outlineIndices[i]];
            }
        }
    }

    /// <summary>
    /// Tekstuuribittikartta muotojen luomiseen tekstuureista.
    /// Sisältää tekstuurin tiedot väritaulukkona.
    /// </summary>
    internal class TextureBitmap : IBitmap
    {
        /// <summary>
        /// Bittikarttadata.
        /// </summary>
        protected bool[,] bitmap;

        /// <summary>
        /// Luo uuden bittikartan tekstuurin pohjalta.
        /// </summary>
        /// <param name="texture">Tekstuuri.</param>
        /// <param name="isOpaque">Predikaatti, joka määrää, onko annettu väri läpinäkyvä.</param>
        public TextureBitmap( Texture2D texture, Predicate<XnaColor> isOpaque )
        {
            XnaColor[] scanline = new XnaColor[texture.Width];
            XnaRectangle srcRect = new XnaRectangle( 0, 0, texture.Width, 1 );

            bitmap = new bool[texture.Width, texture.Height];

            for ( int i = 0; i < texture.Height; i++ )
            {
                // Scan a line from the texture
                srcRect.Y = i;
                texture.GetData<XnaColor>( 0, srcRect, scanline, 0, texture.Width );

                for ( int j = 0; j < texture.Width; j++ )
                {
                    // Flip the y-coordinates because the y-coordinates of the texture
                    // increase downwards.
                    bitmap[j, texture.Height - i - 1] = isOpaque( scanline[j] );
                }
            }
        }

        /// <summary>
        /// Luo uuden bittikartan tekstuurin pohjalta oletusläpinäkyvyysehdoilla.
        /// Ks. <see cref="IsOpaqueColor"/>
        /// </summary>
        /// <param name="texture">Tekstuuri.</param>
        public TextureBitmap( Texture2D texture )
            : this( texture, IsOpaqueColor )
        {
        }

        /// <summary>
        /// Bittikartan leveys pikseleinä.
        /// </summary>
        public int Width
        {
            get { return bitmap.GetLength( 0 ); }
        }

        /// <summary>
        /// Bittikartan korkeus pikseleinä.
        /// </summary>
        public int Height
        {
            get { return bitmap.GetLength( 1 ); }
        }

        /// <summary>
        /// Palauttaa yksittäisen pikselin värin annetusta koordinaattipisteestä (ensin x, sitten y).
        /// </summary>
        public bool this[int x, int y]
        {
            get
            {
                if ( x < 0 || y < 0 || x >= Width || y >= Height ) { return false; }
                return bitmap[x, y];
            }
        }

        /// <summary>
        /// Päättelee pikselin läpinäkyvyyden sen värin perusteella.
        /// Tässä tapauksessa alfa-arvon tulee olla suurempi tai yhtäsuuri kuin 127.
        /// </summary>
        /// <param name="c">Pikselin väri.</param>        
        /// <returns>Läpinäkyvyys.</returns>
        public static bool IsOpaqueColor( XnaColor c )
        {
            return ( c.A >= 127 );
        }
    }
}
