﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Jypeli
{
    /// <summary>
    /// Ajastin, joka voidaan asettaa laukaisemaan tapahtumia tietyin väliajoin.
    /// </summary>
    public class Timer
    {
        #region Events

        /// <summary>
        /// Wanha ajastintapahtumankäsittelijä.
        /// </summary>
        [Obsolete( "Use Timeout event" )]
        public delegate void TriggerHandler( Timer sender );

        /// <summary>
        /// Wanha ajastintapahtuma.
        /// </summary>
        [Obsolete( "Use Timeout event" )]
        public event TriggerHandler Trigger;

        /// <summary>
        /// Tapahtuu väliajoin.
        /// </summary>
        public event Action Timeout;

        #endregion

        private static List<Timer> activeTimers = new List<Timer>();
        private static List<Timer> timersToAdd = new List<Timer>();
        private static List<Timer> timersToRemove = new List<Timer>();
        
        private bool _enabled = false;
        private TimeSpan savedValue = TimeSpan.Zero;
        private TimeSpan nextTriggerTime = TimeSpan.Zero;
        private TimeSpan interval = new TimeSpan( 0, 0, 1 );

        #region Properties

        /// <summary>
        /// Ajastin päällä/pois päältä.
        /// </summary>
        public bool Enabled
        {
            get { return _enabled; }
            set
            {
                if ( value == _enabled ) return;
                if ( value )
                    Start();
                else
                    Stop();
            }
        }

        /// <summary>
        /// Aika sekunneissa, jonka välein <c>TimeOut</c> tapahtuu.
        /// </summary>
        public double Interval
        {
            get { return SecondCounter.MaxValue; }
            set
            {
                interval = new TimeSpan( (int)( value * 1e7 ) );
                SecondCounter.MaxValue = value;
                if ( Enabled )
                    nextTriggerTime = Game.Time.SinceStartOfGame + interval;
            }
        }

        /// <summary>
        /// Menossa oleva hetki nollasta väliaikaan (<c>Interval</c>).
        /// </summary>
        public double CurrentTime
        {
            get { return SecondCounter.Value; }
            set { SecondCounter.Value = value; }
        }

        /// <summary>
        /// Ajastimen käynnistysaika.
        /// </summary>
        public TimeSpan StartTime { get; private set; }

        /// <summary>
        /// Sekuntilaskuri. Voidaan sitoa näyttöihin.
        /// </summary>
        public DoubleMeter SecondCounter { get; private set; }

        /// <summary>
        /// Määrää, kuinka monta kertaa tapahtuma suoritetaan.
        /// Kun tapahtumaa on suoritettu tarpeeksi, <c>Enabled</c> saa automaattisesti
        /// arvon <c>false</c>, jolloin ajastin pysähtyy.
        /// Kun laskuri nollataan, myös Times palautuu oletusarvoonsa.
        /// Huomaa, että <c>TimesLimited</c> tulee olla <c>true</c>, että arvo otetaan huomioon.
        /// </summary>
        /// <see cref="TimesLimited"/>
        public IntMeter Times { get; private set; }

        /// <summary>
        /// Ajastimen suorituskertojen rajoitus päälle/pois.
        /// </summary>
        /// <see cref="Times"/>
        public bool TimesLimited
        {
            get { return Times.MinValue == 0; }
            set
            {
                Times.MinValue = value ? 0 : 1;
            }
        }

        /// <summary>
        /// Vapaasti asetettava muuttuja.
        /// Arvo ei muutu, jos sitä ei muuteta.
        /// </summary>
        public object Tag { get; set; }

        #endregion

        #region Public methods

        /// <summary>
        /// Alustaa uuden ajastinluokan.
        /// </summary>
        public Timer()
        {
            SecondCounter = new DoubleMeter( 0 );
            Times = new IntMeter( 1 );
            Times.MinValue = 1;
            Times.MaxValue = 1;
            Enabled = false;
        }

        /// <summary>
        /// Kutsuu aliohjelmaa <c>onTimeout</c> annetun ajan kuluttua.
        /// Ajastin luodaan automaattisesti.
        /// </summary>
        /// <param name="seconds">Aika sekunteina.</param>
        /// <param name="onTimeout">Kutsuttava aliohjelma.</param>
        public static void SingleShot( double seconds, Action onTimeout )
        {
            Timer t = new Timer();
            t.Interval = seconds;
            t.Timeout += onTimeout;
            t.Start( 1 );
        }

        /// <summary>
        /// Käynnistää ajastimen.
        /// </summary>
        public void Start()
        {
            _enabled = true;
            StartTime = Game.Time.SinceStartOfGame;
            nextTriggerTime = StartTime + interval;
            timersToAdd.Add( this );
        }

        /// <summary> 
        /// Käynnistää ajastimen, rajoittaa suorituskerrat. 
        /// </summary> 
        /// <param name="times">Kuinka monta kertaa tulee ajastintapahtuma.</param> 
        public void Start( int times )
        {
            _enabled = true;
            LimitTimes( times );
            StartTime = Game.Time.SinceStartOfGame;
            nextTriggerTime = StartTime + interval;
            timersToAdd.Add( this );
        }

        /// <summary>
        /// Pysäyttää ajastimen tallentaen sen tilan.
        /// </summary>
        public void Pause()
        {
            _enabled = false;
            savedValue = Game.Time.SinceStartOfGame - StartTime;
            timersToRemove.Add( this );
        }

        /// <summary>
        /// Pysäyttää ajastimen ja nollaa sen tilan.
        /// </summary>
        public void Stop()
        {
            _enabled = false;
            savedValue = TimeSpan.Zero;
            timersToRemove.Add( this );
        }

        /// <summary>
        /// Nollaa ajastimen tilan.
        /// Myös suorituskerrat nollataan.
        /// </summary>
        public void Reset()
        {
            StartTime = Game.Time.SinceStartOfGame;
            SecondCounter.Reset();
            Times.Reset();
            nextTriggerTime = StartTime + interval;
        }

        private void LimitTimes( int numTimes )
        {
            TimesLimited = true;
            Times.DefaultValue = numTimes;
            Times.MaxValue = numTimes;
            Times.Value = numTimes;
        }

        /// <summary>
        /// Poistaa kaikki ajastimet.
        /// </summary>
        internal static void ClearAll()
        {
            timersToRemove.Clear();
            timersToRemove.AddRange( activeTimers );
        }

        internal static void UpdateAll()
        {
            TimeSpan currentTime = Game.Time.SinceStartOfGame;

            for ( int i = timersToAdd.Count - 1; i >= 0; i-- )
            {
                activeTimers.Add( timersToAdd[i] );
                timersToAdd.RemoveAt( i );
            }

            for ( int i = timersToRemove.Count - 1; i >= 0; i-- )
            {
                activeTimers.Remove( timersToRemove[i] );
                timersToRemove.RemoveAt( i );
            }

            for ( int i = 0; i < activeTimers.Count; i++ )
            {
                if ( !activeTimers[i].Enabled ) continue;

                activeTimers[i].SecondCounter.Value = ( currentTime - activeTimers[i].StartTime ).TotalSeconds;

                if ( currentTime >= activeTimers[i].nextTriggerTime )
                {
                    activeTimers[i].nextTriggerTime += activeTimers[i].interval;

                    if ( activeTimers[i].TimesLimited )
                    {
                        if ( --activeTimers[i].Times.Value <= 0 )
                        {
                            // The timer has executed its maximum times, stop it
                            activeTimers[i].Stop();
                        }
                    }

                    if ( activeTimers[i].Trigger != null )
                        activeTimers[i].Trigger( activeTimers[i] );
                    if ( activeTimers[i].Timeout != null )
                        activeTimers[i].Timeout();
                }
            }
        }

        #endregion
    }
}
