﻿#region MIT License
/*
 * Copyright (c) 2009 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */ 

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using AdvanceMath;

namespace Jypeli
{
    /// <summary>
    /// Sisältää näkymän leveyden ja korkeuden sekä reunojen koordinaatit.
    /// Y-koordinaatti kasvaa ylöspäin.
    /// </summary>
    public interface View
    {
        /// <summary>
        /// Näkymän keskipiste.
        /// </summary>
        Vector Center { get; }

        /// <summary>
        /// Kulma, johon näkymä on käännetty.
        /// Huom. ei välttämättä toimi kaikilla näkymillä.
        /// </summary>
        Angle Angle { get; }

        /// <summary>
        /// Näkymän leveys x-suunnassa.
        /// </summary>
        double Width { get; }

        /// <summary>
        /// Näkymän korkeus y-suunnassa.
        /// </summary>
        double Height { get; }

        /// <summary>
        /// Näkymän "turvallinen" ts. laiteriippumaton leveys x-suunnassa.
        /// </summary>
        double WidthSafe { get; }

        /// <summary>
        /// Näkymän "turvallinen" ts. laiteriippumaton korkeus y-suunnassa.
        /// </summary>
        double HeightSafe { get; }

        /// <summary>
        /// Näkymän yläreunan y-koordinaatti.
        /// </summary>
        double Top { get; }

        /// <summary>
        /// Näkymän alareunan y-koordinaatti.
        /// </summary>
        double Bottom { get; }

        /// <summary>
        /// Näkymän vasemman reunan x-koordinaatti.
        /// </summary>
        double Left { get; }

        /// <summary>
        /// Näkymän oikean reunan x-koordinaatti.
        /// </summary>
        double Right { get; }

        /// <summary>
        /// Näkymän yläreunan "turvallinen" ts. laiteriippumaton y-koordinaatti.
        /// </summary>
        double TopSafe { get; }

        /// <summary>
        /// Näkymän alareunan "turvallinen" ts. laiteriippumaton y-koordinaatti.
        /// </summary>
        double BottomSafe { get; }

        /// <summary>
        /// Näkymän vasemman reunan "turvallinen" ts. laiteriippumaton x-koordinaatti.
        /// </summary>
        double LeftSafe { get; }

        /// <summary>
        /// Näkymän oikean reunan "turvallinen" ts. laiteriippumaton x-koordinaatti.
        /// </summary>
        double RightSafe { get; }

        /// <summary>
        /// Muuntaa näkymään suhteessa olevat koordinaatit ylemmän tason koordinaateiksi.
        /// </summary>
        /// <param name="src">Suhteellinen vektori.</param>
        /// <returns>Ylemmän tason vektori.</returns>
        Vector TranslateTo( Vector src );

        /// <summary>
        /// Muuntaa ylemmän tason koordinaatit näkymään suhteellisiksi.
        /// </summary>
        /// <param name="src">Ylemmän tason vektori.</param>
        /// <returns>Suhteellinen vektori</returns>
        Vector TranslateFrom( Vector src );
    }

    /// <summary>
    /// Sisältää näytön leveyden ja korkeuden sekä reunojen koordinaatit.
    /// Y-koordinaatti kasvaa ylöspäin.
    /// Koordinaatteja ei voi muuttaa.
    /// </summary>
    public class ScreenView : View
    {
        internal Viewport viewPort;

        /// <summary>
        /// Alustaa uuden näyttönäkymän.
        /// </summary>
        /// <param name="viewPort">Näytön viewport.</param>
        public ScreenView( Viewport viewPort )
        {
            this.viewPort = viewPort;
        }

        /// <summary>
        /// Näytön keskipiste.
        /// </summary>
        public Vector Center
        {
            get { return Vector.Zero; }
        }

        /// <summary>
        /// Kulma, johon näkymä on käännetty.
        /// Näytöllä aina nolla.
        /// </summary>
        public Angle Angle
        {
            get { return Angle.Zero; }
        }

        /// <summary>
        /// Näytön leveys x-suunnassa.
        /// </summary>
        public double Width
        {
            get { return viewPort.Width; }
        }

        /// <summary>
        /// Näytön korkeus y-suunnassa.
        /// </summary>
        public double Height
        {
            get { return viewPort.Height; }
        }

        /// <summary>
        /// Näytön vasemman reunan x-koordinaatti.
        /// </summary>
        public double Left
        {
            get { return -Width / 2; }
        }

        /// <summary>
        /// Näytön oikean reunan x-koordinaatti.
        /// </summary>
        public double Right
        {
            get { return Width / 2; }
        }

        /// <summary>
        /// Näytön yläreunan y-koordinaatti.
        /// </summary>
        public double Top
        {
            get { return Height / 2; }
        }

        /// <summary>
        /// Näytön alareunan y-koordinaatti.
        /// </summary>
        public double Bottom
        {
            get { return -Height / 2; }
        }

        /// <summary>
        /// Näytön "turvallinen" ts. laiteriippumaton leveys x-suunnassa.
        /// </summary>
        public double WidthSafe
        {
            get { return viewPort.TitleSafeArea.Width; }
        }

        /// <summary>
        /// Näytön "turvallinen" ts. laiteriippumaton korkeus y-suunnassa.
        /// </summary>
        public double HeightSafe
        {
            get { return viewPort.TitleSafeArea.Height; }
        }

        /// <summary>
        /// Näytön vasemman reunan "turvallinen" ts. laiteriippumaton x-koordinaatti.
        /// </summary>
        public double LeftSafe
        {
            get { return -viewPort.TitleSafeArea.Width / 2; }
        }

        /// <summary>
        /// Näytön oikean reunan "turvallinen" ts. laiteriippumaton x-koordinaatti.
        /// </summary>
        public double RightSafe
        {
            get { return viewPort.TitleSafeArea.Width / 2; }
        }

        /// <summary>
        /// Näytön yläreunan "turvallinen" ts. laiteriippumaton y-koordinaatti.
        /// </summary>
        public double TopSafe
        {
            get { return viewPort.TitleSafeArea.Height / 2; }
        }

        /// <summary>
        /// Näytön alareunan "turvallinen" ts. laiteriippumaton y-koordinaatti.
        /// </summary>
        public double BottomSafe
        {
            get { return -viewPort.TitleSafeArea.Height / 2; }
        }

        internal Vector FromXnaScreenCoordinates(Vector2 position)
        {
            double x = position.X - (viewPort.Width / 2);
            double y = -position.Y + (viewPort.Height / 2);
            return new Vector(x, y);
        }

        /// <summary>
        /// Ei käytössä, sillä näyttökoordinaatit ovat kaikkein ylin koordinaatistotaso.
        /// </summary>
        /// <param name="src">Ylemmän tason vektori.</param>
        /// <returns>Suhteellinen vektori</returns>
        public Vector TranslateFrom( Vector src )
        {
            return src;
        }

        /// <summary>
        /// Ei käytössä, sillä näyttökoordinaatit ovat kaikkein ylin koordinaatistotaso.
        /// </summary>
        /// <param name="src">Suhteellinen vektori.</param>
        /// <returns>Ylemmän tason vektori.</returns>
        public Vector TranslateTo( Vector src )
        {
            return src;
        }
    }

    /// <summary>
    /// Sisältää ikkunan sijainnin, leveyden ja korkeuden sekä reunojen koordinaatit.
    /// Y-koordinaatti kasvaa ylöspäin.
    /// </summary>
    public class WindowView : View
    {
        /// <summary>
        /// Alustaa uuden ikkunanäkymän.
        /// </summary>
        /// <param name="x">Ikkunan x-koordinaatti.</param>
        /// <param name="y">Ikkunan y-koordinaatti.</param>
        /// <param name="width">Ikkunan leveys.</param>
        /// <param name="height">Ikkunan korkeus.</param>
        public WindowView( double x, double y, double width, double height )
        {
            XOffset = x;
            YOffset = y;
            Width = width;
            Height = height;
        }

        /// <summary>
        /// Ikkunan x-koordinaatti ylemmällä tasolla.
        /// </summary>
        public double XOffset { get; set; }

        /// <summary>
        /// Ikkunan y-koordinaatti ylemmällä tasolla.
        /// </summary>
        public double YOffset { get; set; }

        /// <summary>
        /// Ikkunan leveys x-suunnassa.
        /// </summary>
        public double Width { get; set; }

        /// <summary>
        /// Ikkunan korkeus y-suunnassa.
        /// </summary>
        public double Height { get; set; }

        /// <summary>
        /// Ikkunan keskipiste.
        /// </summary>
        public Vector Center
        {
            get { return Vector.Zero; }
        }

        /// <summary>
        /// Ikkunan origon X-paikka piirrettäessä (vasen/oikea reuna tai keskellä)
        /// </summary>
        public HorizontalAlignment HorizontalOrigin { get; set; }

        /// <summary>
        /// Ikkunan origon Y-paikka piirrettäessä (ylä-/alareuna tai keskellä)
        /// </summary>
        public VerticalAlignment VerticalOrigin { get; set; }

        /// <summary>
        /// Ikkunan keskipiste piirrettäessä ikkunan omissa koordinaateissa.
        /// </summary>
        public Vector DrawCenter
        {
            get
            {
                double cx = ( HorizontalOrigin == HorizontalAlignment.Center ) ? 0 : ( (HorizontalOrigin == HorizontalAlignment.Left) ? -Left : -Right );
                double cy = ( VerticalOrigin == VerticalAlignment.Center ) ? 0 : ( ( VerticalOrigin == VerticalAlignment.Top ) ? -Top : -Bottom );

                return new Vector( cx, cy );
            }
        }

        /// <summary>
        /// Kulma, johon ikkuna on käännetty.
        /// </summary>
        public Angle Angle { get; set; }

        /// <summary>
        /// Ikkunan vasemman reunan x-koordinaatti.
        /// </summary>
        public double Left
        {
            get { return -Width / 2; }
        }

        /// <summary>
        /// Ikkunan oikean reunan x-koordinaatti.
        /// </summary>
        public double Right
        {
            get { return Width / 2; }
        }

        /// <summary>
        /// Ikkunan yläreunan y-koordinaatti.
        /// </summary>
        public double Top
        {
            get { return Height / 2; }
        }

        /// <summary>
        /// Ikkunan alareunan y-koordinaatti.
        /// </summary>
        public double Bottom
        {
            get { return -Height / 2; }
        }

        /// <summary>
        /// Ikkunan piirtoalueen leveys x-suunnassa.
        /// </summary>
        public double WidthSafe
        {
            get { return Width - 20; }
        }

        /// <summary>
        /// Ikkunan piirtoalueen korkeus y-suunnassa.
        /// </summary>
        public double HeightSafe
        {
            get { return Height - 20; }
        }

        /// <summary>
        /// Ikkunan piirtoalueen vasemman reunan x-koordinaatti.
        /// </summary>
        public double LeftSafe
        {
            get { return Left + 20; }
        }

        /// <summary>
        /// Ikkunan piirtoalueen oikean reunan x-koordinaatti.
        /// </summary>
        public double RightSafe
        {
            get { return Right - 20; }
        }

        /// <summary>
        /// Ikkunan piirtoalueen yläreunan y-koordinaatti.
        /// </summary>
        public double TopSafe
        {
            get { return Top - 20; }
        }

        /// <summary>
        /// Ikkunan piirtoalueen alareunan y-koordinaatti.
        /// </summary>
        public double BottomSafe
        {
            get { return Bottom + 20; }
        }

        /// <summary>
        /// Muuntaa ylemmän tason koordinaatit ikkunanäkymään suhteellisiksi.
        /// </summary>
        /// <param name="src">Ylemmän tason vektori.</param>
        /// <returns>Suhteellinen vektori</returns>
        public Vector TranslateFrom( Vector src )
        {
            return src - new Vector( XOffset, YOffset );
        }

        /// <summary>
        /// Muuntaa ikkunanäkymään suhteessa olevat koordinaatit ylemmän tason koordinaateiksi.
        /// </summary>
        /// <param name="src">Suhteellinen vektori.</param>
        /// <returns>Ylemmän tason vektori.</returns>
        public Vector TranslateTo( Vector src )
        {
            return src + new Vector( XOffset, YOffset );
        }

        /// <summary>
        /// Muuntaa ikkunanäkymään suhteessa olevat koordinaatit ylemmän tason koordinaateiksi
        /// suhteellisen origon huomioon ottaen.
        /// Älä käytä muuten kuin piirtäessä!
        /// </summary>
        /// <param name="src">Suhteellinen vektori.</param>
        /// <returns>Ylemmän tason vektori.</returns>
        public Vector DrawTranslateTo( Vector src )
        {
            return DrawCenter + src + new Vector( XOffset, YOffset );
        }
    }

    /// <summary>
    /// Asemointi vaakasuunnassa.
    /// </summary>
    public enum HorizontalAlignment
    {
        /// <summary>
        /// Keskellä.
        /// </summary>
        Center,

        /// <summary>
        /// Vasemmassa reunassa.
        /// </summary>
        Left,

        /// <summary>
        /// Oikeassa reunassa.
        /// </summary>
        Right
    }

    /// <summary>
    /// Asemointi pystysuunnassa.
    /// </summary>
    public enum VerticalAlignment
    {
        /// <summary>
        /// Keskellä.
        /// </summary>
        Center,

        /// <summary>
        /// Yläreunassa.
        /// </summary>
        Top,

        /// <summary>
        /// Alareunassa.
        /// </summary>
        Bottom
    }
}

