using System;
using Microsoft.Xna.Framework;

#if WINDOWS_PHONE
using Microsoft.Phone.Shell;
using System.IO.IsolatedStorage;
using System.Xml.Serialization;
#endif


namespace Jypeli.WP7
{
    /// <summary>
    /// Puhelimen nytn asemointi.
    /// </summary>
    public enum DisplayOrientation
    {
        /// <summary>
        /// Vaakasuuntainen. nytt kntyy automaattisesti, jos puhelin knnetn toisinpin.
        /// </summary>
        Landscape,

        /// <summary>
        /// Vaakasuuntainen, vasemmalle knnetty.
        /// </summary>
        LandscapeLeft,

        /// <summary>
        /// Vaakasuuntainen, oikealle knnetty.
        /// </summary>
        LandscapeRight,

        /// <summary>
        /// Pystysuuntainen.
        /// </summary>
        Portrait,
    }

    /// <summary>
    /// Puhelimen nytn tarkkuus.
    /// </summary>
    public enum DisplayResolution
    {
        /// <summary>
        /// Pieni tarkkuus. Paremi suorituskyky ja pienempi akun kulutus.
        /// </summary>
        Small,

        /// <summary>
        /// Suuri tarkkuus.
        /// </summary>
        Large,
    }    

    /// <summary>
    /// Aliohjelmia ja ominaisuuksia, jotka toimivat vain puhelimessa. Voidaan kutsua mys PC:lle knnettess,
    /// mutta tllin mitn ei yksinkertaisesti tapahdu.
    /// </summary>
    public class Phone
    {
        private DisplayOrientation _displayOrientation = DisplayOrientation.Landscape;
        private DisplayResolution _displayResolution = DisplayResolution.Small;

        /// <summary>
        /// Vrisytt puhelinta.
        /// </summary>
        /// <param name="milliSeconds">Aika kuinka kauan Vrisytetn millisekunteina</param>
        public void Vibrate(int milliSeconds)
        {
#if WINDOWS_PHONE
            Microsoft.Devices.VibrateController.Default.Start(TimeSpan.FromMilliseconds(milliSeconds));
#endif
        }

        /// <summary>
        /// Lopettaa puhelimen vrinn.
        /// </summary>
        public void StopVibrating()
        {
#if WINDOWS_PHONE
            Microsoft.Devices.VibrateController.Default.Stop();
#endif
        }

        private static void GetScreenSize(DisplayResolution resolution, out int width, out int height)
        {
            switch (resolution)
            {
                case DisplayResolution.Small:
                    width = 240;
                    height = 400;
                    break;
                default:
                    width = 480;
                    height = 800;
                    break;
            }
        }

        /// <summary>
        /// Puhelimen nytn tarkkuus.
        /// </summary>
        public DisplayResolution DisplayResolution
        {
            get { return _displayResolution; }
            set
            {
                if (_displayResolution != value)
                {
                    _displayResolution = value;
                    ResetScreen(); 
                }
            }
        }

        /// <summary>
        /// Puhelimen Nytn asemointi.
        /// </summary>
        public DisplayOrientation DisplayOrientation
        {
            get { return _displayOrientation; }
            set
            {
                if (_displayOrientation != value)
                {
                    _displayOrientation = value;
                    Game.Instance.Accelerometer.DisplayOrientation = value;
                    ResetScreen();
                }
            }
        }

        private bool _tombstoning = false;

        /// <summary>
        /// Tallennetaanko pelin tilanne jos ruutu sammutetaan, vaihdetaan ohjelmaa tms.
        /// </summary>
        public bool Tombstoning
        {
            get { return _tombstoning; }
            set
            {
#if WINDOWS_PHONE
                if ( value && !_tombstoning )
                {
                    PhoneApplicationService.Current.Deactivated += SaveTombstone;
                    _tombstoning = true;
                }
                else if ( !value && _tombstoning )
                {
                    PhoneApplicationService.Current.Deactivated -= SaveTombstone;
                    _tombstoning = false;
                }
#endif
            }
        }

#if WINDOWS_PHONE
        internal Phone()
        {
            PhoneApplicationService.Current.Launching += NewGame;
            PhoneApplicationService.Current.Activated += LoadTombstone;
        }

        void NewGame( object sender, LaunchingEventArgs e )
        {
            Game.Instance.Begin();
        }

        internal void LoadTombstone( object sender, ActivatedEventArgs e )
        {
            Game.Instance.Continue();
            Game.Instance.LoadGame( "tombstone" );
        }

        internal void SaveTombstone( object sender, DeactivatedEventArgs e )
        {
            Game.Instance.SaveGame( "tombstone" );
        }
#endif


        internal void ResetScreen()
        {
#if WINDOWS_PHONE
            int screenWidth, screenHeight;
            GraphicsDeviceManager graphics = Game.Graphics;
            GetScreenSize(_displayResolution, out screenWidth, out screenHeight);

            switch (_displayOrientation)
            {
                case DisplayOrientation.Landscape:
                    graphics.SupportedOrientations = Microsoft.Xna.Framework.DisplayOrientation.LandscapeLeft | Microsoft.Xna.Framework.DisplayOrientation.LandscapeRight;
                    Game.Instance.SetWindowSize(screenHeight, screenWidth);
                    break;
                case DisplayOrientation.LandscapeLeft:
                    graphics.SupportedOrientations = Microsoft.Xna.Framework.DisplayOrientation.LandscapeLeft;
                    Game.Instance.SetWindowSize(screenHeight, screenWidth);
                    break;
                case DisplayOrientation.LandscapeRight:
                    graphics.SupportedOrientations = Microsoft.Xna.Framework.DisplayOrientation.LandscapeRight;
                    Game.Instance.SetWindowSize(screenHeight, screenWidth);
                    break;
                case DisplayOrientation.Portrait:
                    graphics.SupportedOrientations = Microsoft.Xna.Framework.DisplayOrientation.Portrait;
                    Game.Instance.SetWindowSize(screenWidth, screenHeight);
                    break;
                default:
                    break;
            }

            graphics.ApplyChanges();
#endif
        }
    }
}
