﻿#region MIT License
/*
 * Copyright (c) 2009-2011 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion


using System;
using System.Collections.Generic;
using System.Diagnostics;
using Microsoft.Xna.Framework;

#if WINDOWS_PHONE
using Microsoft.Xna.Framework.Input.Touch;

using XnaTouchPanel = Microsoft.Xna.Framework.Input.Touch.TouchPanel;
#endif

using Jypeli.Controls;
using Jypeli.Widgets;


namespace Jypeli.WP7
{
    /// <summary>
    /// Kosketuspaneeli.
    /// </summary>
    public class TouchPanel : Controller<TouchPanelState>, PointingDevice
    {
        /// <summary>
        /// Kuinka monta kosketusta näytöllä on aktiivisena.
        /// </summary>
        public int ActiveChannels
        {
            get { return newState.ActiveTouches.Count; }
        }

        /// <summary>
        /// Kuinka monta kosketusta näytöllä on aktiivisena.
        /// </summary>
        public int NumTouches
        {
            get { return newState.ActiveTouches.Count; }
        }

        internal override bool IsBufferEmpty()
        {
            return true;
        }

        /// <summary>
        /// Ottaa käytöstä poistetun kosketuskontrollin <c>k</c> takaisin käyttöön.
        /// </summary>
        /// <param name="state">Kosketuksen tila</param>
        public void Enable( ButtonState state )
        {
            base.Enable( listener => ( listener.State == state ) );
        }

        /// <summary>
        /// Poistaa kosketuskontrollin <c>k</c> käytöstä.
        /// </summary>
        /// <param name="state">Kosketuksen tila</param>
        public void Disable( ButtonState state )
        {
            base.Disable( listener => ( listener.State == state ) );
        }

        internal override string GetControlText( Listener listener )
        {
            if ( listener.Type != ListeningType.Touch )
                Debug.Assert( false, "Bad listener type for touch panel" );
            return "Touch panel";
        }

        internal override TouchPanelState GetCurrentState()
        {
#if WINDOWS_PHONE
            return new TouchPanelState( XnaTouchPanel.GetState(), newState );
#else
            return new TouchPanelState();
#endif
        }

        public bool IsTriggeredOnChannel( int channel, Listener listener )
        {
#if WINDOWS_PHONE
            if ( !listener.Context.Active || listener.Type != ListeningType.Touch || newState.ActiveTouches.Count <= channel )
                return false;

            if ( listener.Widget != null && !listener.Widget.IsInside( newState.ActiveTouches[channel].PositionOnScreen ) )
                return false;

            if ( listener.GameObject != null && !listener.GameObject.IsInside( newState.ActiveTouches[channel].PositionOnWorld ) )
                return false;

            switch ( newState.ActiveTouches[channel].State )
            {
                case TouchLocationState.Moved:
                    return listener.State == ButtonState.Down;

                case TouchLocationState.Pressed:
                    return listener.State == ButtonState.Pressed;

                case TouchLocationState.Released:
                    return listener.State == ButtonState.Released;
            }
#endif

            return false;
        }


        #region Listen with no parameters

        public Listener Listen( ButtonState state, TouchHandler handler, string helpText )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener( this, ListeningType.Touch, helpText, handler );
            l.State = state;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        public Listener ListenOn( Widget w, ButtonState state, TouchHandler handler, String helpText )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener( this, ListeningType.Touch, helpText, handler );
            l.State = state;
            l.Widget = w;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        public Listener ListenOn( GameObject o, ButtonState state, TouchHandler handler, String helpText )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener( this, ListeningType.Touch, helpText, handler );
            l.State = state;
            l.GameObject = o;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        #endregion

        #region Listen with 1 parameter

        public Listener Listen<T1>( ButtonState state, TouchHandler<T1> handler, string helpText, T1 p1 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1>( this, ListeningType.Touch, helpText, handler, p1 );
            l.State = state;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        public Listener ListenOn<T1>( Widget w, ButtonState state, TouchHandler<T1> handler, String helpText, T1 p1 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1>( this, ListeningType.Touch, helpText, handler, p1 );
            l.State = state;
            l.Widget = w;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        public Listener ListenOn<T1>( GameObject o, ButtonState state, TouchHandler<T1> handler, String helpText, T1 p1 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1>( this, ListeningType.Touch, helpText, handler, p1 );
            l.State = state;
            l.GameObject = o;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        #endregion

        #region Listen with 2 parameters

        public Listener Listen<T1, T2>( ButtonState state, TouchHandler<T1, T2> handler, string helpText, T1 p1, T2 p2 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1, T2>( this, ListeningType.Touch, helpText, handler, p1, p2 );
            l.State = state;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        public Listener ListenOn<T1, T2>( Widget w, ButtonState state, TouchHandler<T1, T2> handler, String helpText, T1 p1, T2 p2 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1, T2>( this, ListeningType.Touch, helpText, handler, p1, p2 );
            l.State = state;
            l.Widget = w;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        public Listener ListenOn<T1, T2>( GameObject o, ButtonState state, TouchHandler<T1, T2> handler, String helpText, T1 p1, T2 p2 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1, T2>( this, ListeningType.Touch, helpText, handler, p1, p2 );
            l.State = state;
            l.GameObject = o;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        #endregion

        #region Listen with 3 parameters

        public Listener Listen<T1, T2, T3>( ButtonState state, TouchHandler<T1, T2, T3> handler, string helpText, T1 p1, T2 p2, T3 p3 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1, T2, T3>( this, ListeningType.Touch, helpText, handler, p1, p2, p3 );
            l.State = state;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        public Listener ListenOn<T1, T2, T3>( Widget w, ButtonState state, TouchHandler<T1, T2, T3> handler, String helpText, T1 p1, T2 p2, T3 p3 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1, T2, T3>( this, ListeningType.Touch, helpText, handler, p1, p2, p3 );
            l.State = state;
            l.Widget = w;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        public Listener ListenOn<T1, T2, T3>( GameObject o, ButtonState state, TouchHandler<T1, T2, T3> handler, String helpText, T1 p1, T2 p2, T3 p3 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1, T2, T3>( this, ListeningType.Touch, helpText, handler, p1, p2, p3 );
            l.State = state;
            l.GameObject = o;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        #endregion

        #region Listen with 4 parameters

        public Listener Listen<T1, T2, T3, T4>( ButtonState state, TouchHandler<T1, T2, T3, T4> handler, string helpText, T1 p1, T2 p2, T3 p3, T4 p4 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1, T2, T3, T4>( this, ListeningType.Touch, helpText, handler, p1, p2, p3, p4 );
            l.State = state;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        public Listener ListenOn<T1, T2, T3, T4>( Widget w, ButtonState state, TouchHandler<T1, T2, T3, T4> handler, String helpText, T1 p1, T2 p2, T3 p3, T4 p4 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1, T2, T3, T4>( this, ListeningType.Touch, helpText, handler, p1, p2, p3, p4 );
            l.State = state;
            l.Widget = w;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        public Listener ListenOn<T1, T2, T3, T4>( GameObject o, ButtonState state, TouchHandler<T1, T2, T3, T4> handler, String helpText, T1 p1, T2 p2, T3 p3, T4 p4 )
        {
#if WINDOWS_PHONE
            Listener l = new TouchListener<T1, T2, T3, T4>( this, ListeningType.Touch, helpText, handler, p1, p2, p3, p4 );
            l.State = state;
            l.GameObject = o;
            Add( l );
            return l;
#else
            return Listener.Null;
#endif
        }

        #endregion
    }
}
