﻿#region MIT License
/*
 * Copyright (c) 2009-2011 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tomi Karppinen, Tero Jäntti
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;


namespace Jypeli.Widgets
{
    public class Background : Widget
    {
        public override Image Image
        {
            get { return base.Image; }
            set
            {
                base.Image = value;
                if ( value != null )
                    Size = new Vector(value.Width, value.Height);
            }
        }

        /// <summary>
        /// Liikkuuko taustakuva kameran mukana vai ei.
        /// </summary>
        /// <value><c>true</c> jos liikkuu; muuten <c>false</c>.</value>
        public bool MovesWithCamera { get; set; }

        /// <summary>
        /// Taustakuvan skaala.
        /// <example>
        /// Jos halutaan kaksinkertainen kuva, niin laitetaan skaalaksi 2, puolet pienemmän kuvan skaala on vastaavasti 0.5.
        /// </example>
        /// </summary>
        /// <value>The scale.</value>
        public double Scale { get; set; }

        /// <summary>
        /// Sovittaa taustakuvan kentän kokoiseksi.
        /// </summary>
        public void FitToLevel()
        {
            Level level = Game.Instance.Level;
            Size = new Vector( level.Width, level.Height );
        }

        /// <summary>
        /// Luo uuden taustakuvan.
        /// </summary>
        /// <param name="width">Taustakuvan leveys.</param>
        /// /// <param name="height">Taustakuvan korkeus.</param>
        public Background( double width, double height )
            : base( width, height )
        {
            this.Size = new Vector( width, height );
            this.Scale = 1.0f;
            this.MovesWithCamera = true;
        }

        /// <summary>
        /// Luo uuden taustakuvan.
        /// </summary>
        /// <param name="size">Taustakuvan koko.</param>
        public Background( Vector size )
            : this( size.X, size.Y )
        {
        }

        /// <summary>
        /// Luo avaruustaustakuvan.
        /// </summary>
        /// <param name="amount">Tähtien määrä.</param>
        /// <returns>Kuva.</returns>
        public Image CreateStars( int amount )
        {
            int imageWidth = (int)Game.Screen.Width;
            int imageHeight = (int)Game.Screen.Height;

            Image image = Image.CreateStarSky( imageWidth, imageHeight, amount );

            Image = image;
            MovesWithCamera = false;

            return image;
        }

        /// <summary>
        /// Luo avaruustaustakuvan.
        /// </summary>
        /// <returns>Kuva.</returns>
        public Image CreateStars()
        {
            int textureWidth = (int)Game.Screen.Width;
            int textureHeight = (int)Game.Screen.Height;
            int amount = ( textureWidth * textureHeight ) / 1000;

            Image image = this.CreateStars( amount );

            Image = image;
            MovesWithCamera = false;

            return image;
        }

        /// <summary>
        /// Luo liukuväritaustan taustakuvaksi.
        /// </summary>
        /// <param name="lowerColor">Alempi väri.</param>
        /// <param name="upperColor">Ylempi väri.</param>
        /// <returns>Kuva.</returns>
        public Image CreateGradient( Color lowerColor, Color upperColor )
        {
            int textureWidth = (int)Game.Screen.Width;
            int textureHeight = (int)Game.Screen.Height;
            int amount = ( textureWidth * textureHeight ) / 800;

            Image image = Image.FromGradient( textureWidth, textureHeight, lowerColor, upperColor );

            Image = image;
            MovesWithCamera = false;

            return image;
        }

        internal protected override void Draw( Matrix parentTransformation, Matrix transformation )
        {
            if ( Image == null ) return;

            if ( MovesWithCamera )
            {
                Matrix matrix =
                    Matrix.CreateScale( (float)Size.X, (float)Size.Y, 1f )
                    * transformation
                    * parentTransformation;
                Game.Renderer.DrawImage( Image, ref matrix, new Vector( 1, 1 ) );
            }
        }
    }
}
