﻿#region MIT License
/*
 * Copyright (c) 2009-2011 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using System.ComponentModel;
using Microsoft.Xna.Framework;

namespace Jypeli.Widgets
{
    /// <summary>
    /// Palkki, jonka korkeutta voi säätää.
    /// </summary>
    public class BarGauge : Widget
    {
        private static readonly Vector[] barVertices = new Vector[]
        {
            new Vector(-0.5, 0),
            new Vector(0.5, 0),
            new Vector(0.5, 1.0),
            new Vector(-0.5, 1.0)
        };

        private static readonly IndexTriangle[] barIndices = new IndexTriangle[]
        {
            new IndexTriangle(0, 3, 2),
            new IndexTriangle(0, 2, 1)
        };

        private static readonly ShapeCache shapeCache = new ShapeCache( barVertices, barIndices );

        private static readonly DoubleMeter defaultMeter = new DoubleMeter( 0 ) { MaxValue = 100 };

        private DoubleMeter boundMeter = defaultMeter;

        /// <summary>
        /// Palkin väri.
        /// </summary>
        public Color BarColor { get; set; }

        /// <summary>
        /// Palkin rakentaja.
        /// </summary>
        public BarGauge( double width, double height )
            : base( width, height )
        {
            Color = Color.Transparent;
            BarColor = Color.Red;
            boundMeter.MaxValue = 100.0;
        }

        /// <summary>
        /// Asettaa palkin näyttämään <c>meter</c>-olion arvoa.
        /// Palkin maksimiarvoksi tulee olion <c>meter</c> maksimiarvo.
        /// </summary>
        public void BindTo( DoubleMeter meter )
        {
            boundMeter = meter;
        }

        internal protected override void Draw( Matrix parentTransformation, Matrix transformation )
        {
            double barLength = Size.Y * ( boundMeter.Value / boundMeter.MaxValue );
            Renderer renderer = Game.Renderer;
            Matrix m =
                Matrix.CreateScale( (float)Size.X, (float)barLength, 1f )
                * Matrix.CreateTranslation( 0, (float)( -Height / 2 ), 0 )
                * Matrix.CreateRotationZ( (float)Angle.Radians )
                * Matrix.CreateTranslation( (float)Position.X, (float)Position.Y, 0.0f )
                * parentTransformation;

            renderer.DrawFilledShape( shapeCache, ref m, BarColor );

            Vector[] borderVertices = new Vector[]
            {
                new Vector(-0.5, 0.5),
                new Vector(-0.5, -0.5),
                new Vector(0.5, -0.5),
                new Vector(0.5, 0.5),
            };

            // The border that is drawn by base class gets obscured by the bar.
            // Let's draw it again.
            Game.Renderer.DrawPolygon( borderVertices, ref transformation, BorderColor );

            base.Draw( parentTransformation, transformation );
        }
    }
}
