﻿#region MIT License
/*
 * Copyright (c) 2009-2011 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tomi Karppinen, Tero Jäntti
 */

using System;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Jypeli.Widgets
{
    public enum TextSizeMode
    {
        /// <summary>
        /// Tekstikentän koko on käyttäjän asettama.
        /// Ylipitkä teksti katkaistaan.
        /// </summary>
        None,

        /// <summary>
        /// Tekstikentän koko asetetaan tekstin koon mukaan.
        /// </summary>
        AutoSize,

        /// <summary>
        /// Tekstin koko asetetaan tekstikentän koon mukaan.
        /// </summary>
        StretchText,

        /// <summary>
        /// Teksti rivitetään tekstikentän leveyden mukaan.
        /// </summary>
        Wrapped,
    }


    /// <summary>
    /// Tekstikenttä.
    /// </summary>
    public class Label : Widget
    {
        Font font;
        string title = "";
        string originalText = "";
        string visibleText = "";
        TextSizeMode sizeMode = TextSizeMode.None;
        object boundMeter = null;
        int decimalPlaces = 2;
        string doubleFormatString = "{0:N2}";
        double _xMargin;
        private double _yMargin;
        bool useDefaultHeight = false;

        const double DefaultWidth = 100;

        static double GetDefaultHeight()
        {
            Vector2 fontDims = Font.Default.XnaFont.MeasureString( "A" );
            return fontDims.Y;
        }

        /// <summary>
        /// Tekstikentän otsikko.
        /// </summary>
        public string Title
        {
            get { return title; }
            set
            {
                title = value;
                updateSize();
            }
        }

        /// <summary>
        /// Teksti.
        /// </summary>
        public virtual string Text
        {
            get { return originalText; }
            set
            {
                if ( value == null )
                    throw new ArgumentException( "Text must not be null" );
                originalText = value;
                updateSize();
            }
        }

        public bool IsTruncated
        {
            get { return visibleText.Length < ( originalText.Length + Title.Length ); }
        }
        
        Vector3 _textScale = new Vector3(1, 1, 1);

        /// <summary>
        /// Tekstin skaalaus. Oletus (1,1) ; isompi suurempi.
        /// </summary>
        public Vector TextScale
        {
            get { return new Vector(_textScale.X, _textScale.Y); }
            set
            {
                _textScale = new Vector3((float)value.X, (float)value.Y, 1.0f);
                updateSize();
            }
        }

        /// <summary>
        /// Kuinka monta desimaalia näytetään, kun tekstikenttä on
        /// sidottu näyttämään desimaalilukua.
        /// </summary>
        public int DecimalPlaces
        {
            get { return decimalPlaces; }
            set
            {
                decimalPlaces = value;
                doubleFormatString = "{0:N" + decimalPlaces + "}";
            }
        }

        /// <summary>
        /// Kuinka tekstikentän koko määräytyy.
        /// </summary>
        public TextSizeMode SizeMode
        {
            get { return sizeMode; }
            set { sizeMode = value; updateSize(); }
        }

        /// <summary>
        /// Näytettävän tekstin koko.
        /// Ei välttämättä sama kuin <c>Size</c>.
        /// </summary>
        public Vector TextSize { get; private set; }

        /// <summary>
        /// Tekstin väri.
        /// </summary>
        public Color TextColor { get; set; }

        /// <summary>
        /// Tekstin fontti.
        /// </summary>
        public Font Font
        {
            get { return font; }
            set { font = value; updateSize(); }
        }

        /// <summary>
        /// Tekstin sijoitus vaakasuunnassa.
        /// Vaikuttaa vain, jos tekstikentän koko on suurempi kuin tekstin koko
        /// ja <c>SizeMode</c> ei ole <c>SizeMode.AutoSize</c>.
        /// </summary>
        public HorizontalAlignment HorizontalAlignment { get; set; }

        /// <summary>
        /// Tekstin sijoitus pystysuunnassa.
        /// Vaikuttaa vain, jos tekstikentän koko on suurempi kuin tekstin koko
        /// ja <c>SizeMode</c> ei ole <c>SizeMode.AutoSize</c>.
        /// </summary>
        public VerticalAlignment VerticalAlignment { get; set; }

        /// <summary>
        /// Marginaali vasemmasta/oikeasta reunasta.
        /// </summary>
        public double XMargin
        {
            get { return _xMargin; }
            set { _xMargin = value; updateSize(); }
        }

        /// <summary>
        /// Marginaali ylä-/alareunasta.
        /// </summary>
        public double YMargin 
        {
            get { return _yMargin; }
            set { _yMargin = value; updateSize(); }
        }

        /// <summary>
        /// Tekstikentän koko.
        /// Jos SizeMode on SizeMode.StretchText, teksti
        /// venytetään kentän koon mukaiseksi.
        /// </summary>
        public override Vector Size
        {
            get
            {
                return base.Size;
            }
            set
            {
                if (sizeMode == TextSizeMode.StretchText)
                {
                    TextScale = new Vector(TextScale.X * value.X / base.Size.X, TextScale.Y * value.Y / base.Size.Y);
                }
                useDefaultHeight = false;
                base.Size = value;

                if ( sizeMode == TextSizeMode.None )
                {
                    TruncateText();
                }
            }
        }

        /// <summary>
        /// Luo uuden tekstikentän. Asettaa
        /// koon tekstin mukaan.
        /// </summary>
        public Label()
            : this( DefaultWidth, 10 )
        {
            sizeMode = TextSizeMode.AutoSize;

            if ( Game.Instance != null )
            {
                this.Height = GetDefaultHeight();
                updateSize();
            }
            else
                Game.InstanceInitialized += setDefaultHeight;
        }

        /// <summary>
        /// Luo uuden tekstikentän annetulla tekstillä. Asettaa
        /// koon tekstin mukaan.
        /// </summary>
        public Label( string text )
            : this( DefaultWidth, 10, text )
        {
            sizeMode = TextSizeMode.AutoSize;

            if ( Game.Instance != null )
            {
                this.Height = GetDefaultHeight();
                updateSize();
            }
            else
                Game.InstanceInitialized += setDefaultHeight;
        }

        /// <summary>
        /// Luo uuden tekstikentän animaatiolla.
        /// </summary>
        public Label( Animation animation )
            : base( animation )
        {
            Game.AssertInitialized( Initialize );
        }

        /// <summary>
        /// Luo uuden tekstikentän.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        public Label( double width, double height )
            : this( width, height, "" )
        {
        }

        /// <summary>
        /// Luo uuden tekstikentän.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        /// <param name="text">Teksti.</param>
        public Label( double width, double height, string text )
            : base( width, height, Shape.Rectangle )
        {
            this.sizeMode = TextSizeMode.None;
            this.originalText = text;
            Game.AssertInitialized( Initialize );
        }

        private void setDefaultHeight()
        {
            if ( !useDefaultHeight ) return;
            this.Height = GetDefaultHeight();
        }

        private void Initialize()
        {
            font = Font.Default;
            Color = Color.Transparent;
            TextColor = Color.Black;
            HorizontalAlignment = HorizontalAlignment.Center;
            VerticalAlignment = VerticalAlignment.Center;
            updateSize();
        }

        private void ReleaseOldMeter()
        {
            if ( boundMeter is IntMeter )
                UnBind( (IntMeter)boundMeter );
            else if ( boundMeter is DoubleMeter )
                UnBind( (DoubleMeter)boundMeter );

            boundMeter = null;
        }

        /// <summary>
        /// Asettaa tekstikentän näyttämään annetun mittarin arvoa.
        /// </summary>
        public void BindTo( IntMeter meter )
        {
            ReleaseOldMeter();
            boundMeter = meter;
            SetNumber( 0, meter.Value );
            meter.Changed += this.SetNumber;
        }

        /// <summary>
        /// Asettaa tekstikentän näyttämään annetun mittarin arvoa.
        /// </summary>
        public void BindTo( DoubleMeter meter )
        {
            ReleaseOldMeter();
            boundMeter = meter;
            SetNumber( 0, meter.Value );
            meter.Changed += this.SetNumber;
        }

        /// <summary>
        /// Irrottaa tekstikentän mittarista.
        /// </summary>
        public void Unbind()
        {
            ReleaseOldMeter();
        }

        public void UnBind( IntMeter meter )
        {
            meter.Changed -= this.SetNumber;
            if ( meter == boundMeter )
                Text = "";
        }

        public void UnBind( DoubleMeter meter )
        {
            meter.Changed -= this.SetNumber;
            if ( meter == boundMeter )
                Text = "";
        }

        private void SetNumber( int unused, int newNumber )
        {
            Text = newNumber.ToString();
        }

        private void SetNumber( double unused, double newNumber )
        {
            Text = string.Format( doubleFormatString, newNumber );
        }

        /// <summary>
        /// Päivittää tekstikentän/tekstin koon ja rivittää tekstin.
        /// </summary>
        private void updateSize()
        {
            SpriteFont xnaFont = this.Font.XnaFont;
            visibleText = (title.Length > 0) ? (title + ": " + originalText) : originalText;

            if ( visibleText.Length == 0 )
            {
                this.TextSize = Vector.Zero;
                NotifyParentAboutChangedSizingAttributes();
                return;
            }

            Vector2 rawTextDims = xnaFont.MeasureString( visibleText );
            Vector2 fullTextDims = new Vector2( _textScale.X * rawTextDims.X, _textScale.Y * rawTextDims.Y );

            if ( SizeMode == TextSizeMode.AutoSize )
            {
                Vector newSize = new Vector( fullTextDims.X + 2 * XMargin, fullTextDims.Y + 2 * YMargin );
                PreferredSize = newSize;
                Size = newSize;

                TextSize = new Vector( fullTextDims.X, fullTextDims.Y );
            }
            else if ( ( SizeMode == TextSizeMode.Wrapped ) && Width > 0 && fullTextDims.X > Width )
            {
                visibleText = Font.WrapText(visibleText, PreferredSize.X - 4 * Font.CharacterWidth - 2 * XMargin, PreferredSize.X - 2 * XMargin);
                Vector2 textDims = xnaFont.MeasureString( visibleText );

                Vector newSize = new Vector( base.Size.X, textDims.Y + 2 * YMargin );
                PreferredSize = newSize;
                Size = newSize;

                TextSize = new Vector( textDims.X, textDims.Y );
            }
            else if ( SizeMode == TextSizeMode.None )
            {
                TruncateText();
            }

            NotifyParentAboutChangedSizingAttributes();
        }

        private void TruncateText()
        {
            double textWidth = this.Width - 2 * XMargin;
            if ( textWidth <= 0 )
            {
                textWidth = this.Width;
            }

            visibleText = font.TruncateText( Text, textWidth );

            Vector2 textDims = Font.XnaFont.MeasureString( visibleText );
            TextSize = new Vector( textDims.X, textDims.Y );
        }

        private double GetHorizontalAlignment()
        {
            switch ( HorizontalAlignment )
            {
                case HorizontalAlignment.Center:
                    return -TextSize.X / 2;
                case HorizontalAlignment.Left:
                    return -Width / 2 + XMargin;
                case HorizontalAlignment.Right:
                    return Width / 2 - TextSize.X - XMargin;
                default:
                    return XMargin;
            }
        }

        private double GetVerticalAlignment()
        {
            switch ( VerticalAlignment )
            {
                case VerticalAlignment.Center:
                    return -TextSize.Y / 2;
                case VerticalAlignment.Top:
                    return -Size.Y / 2 - YMargin;
                case VerticalAlignment.Bottom:
                    return Size.Y / 2 - TextSize.Y + YMargin;
                default:
                    return -YMargin;
            }
        }

        internal protected override void Draw( Matrix parentTransformation, Matrix transformation )
        {
            Draw( parentTransformation, transformation, visibleText );
        }

        protected void Draw( Matrix parentTransformation, Matrix transformation, string text )
        {
            SpriteBatch spriteBatch = Graphics.SpriteBatch;
            Matrix m =
                Matrix.CreateScale(_textScale) *
                Matrix.CreateTranslation( (float)GetHorizontalAlignment(), (float)GetVerticalAlignment(), 0 )
                * Matrix.CreateRotationZ( (float)-Angle.Radians )
                * Matrix.CreateTranslation( (float)( Game.Screen.Width / 2 ), (float)( Game.Screen.Height / 2 - 2 * AbsolutePosition.Y ), 0 )
                * Matrix.CreateTranslation( (float)Position.X, (float)Position.Y, 0 )
                * parentTransformation;

            spriteBatch.Begin( SpriteSortMode.FrontToBack, BlendState.AlphaBlend, SamplerState.LinearClamp, DepthStencilState.Default, RasterizerState.CullCounterClockwise, null, m );
            spriteBatch.DrawString( this.Font.XnaFont, text, Vector2.Zero, TextColor.AsXnaColor() );
            spriteBatch.End();

            base.Draw( parentTransformation, transformation );
        }
    }
}
