﻿#region MIT License
/*
 * Copyright (c) 2009-2011 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tomi Karppinen, Tero Jäntti
 */

using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;

namespace Jypeli.Widgets
{
    [EditorBrowsable( EditorBrowsableState.Never )]
    public class ScrollableList<W> : Widget where W : Widget
    {
        public bool IsAtTop
        {
            get { return ( (VerticalScrollLayout)Layout ).StartIndex == 0; }
        }

        public bool IsAtBottom
        {
            get { return ( (VerticalScrollLayout)Layout ).EndIndex == childObjects.Count; }
        }

        public int ItemCount { get { return childObjects.Count; } }

        public ScrollableList()
            : base( new VerticalScrollLayout() )
        {
            Color = Color.Transparent;
            HorizontalSizing = Sizing.AlwaysExpanding;
            VerticalSizing = Sizing.AlwaysExpanding;
        }

        public void ScrollUp()
        {
            var layout = (VerticalScrollLayout)Layout;
            if ( layout.StartIndex > 0 )
            {
                layout.StartIndex--;
                NotifyParentAboutChangedSizingAttributes();
            }
        }

        public void ScrollDown()
        {
            var layout = (VerticalScrollLayout)Layout;
            if ( layout.StartIndex < childObjects.Count - 1 )
            {
                layout.StartIndex++;
                NotifyParentAboutChangedSizingAttributes();
            }
        }

        public override void Clear()
        {
            ( (VerticalScrollLayout)Layout ).StartIndex = 0;
            ( (VerticalScrollLayout)Layout ).EndIndex = 0;
            base.Clear();
        }

        public W this[int index]
        {
            get { return (W)childObjects[index]; }
            set { childObjects[index] = value; }
        }

        protected internal override void DrawChildObjects( ref Matrix parentTransformation, ref Matrix childTransformation )
        {
            var layout = (VerticalScrollLayout)Layout;
            for ( int i = layout.StartIndex; i < layout.EndIndex; i++ )
            {
                childObjects[i].Draw( childTransformation );
            }
        }
    }


    /// <summary>
    /// Listakomponentti. Voidaan liittää listaan, joka
    /// toteuttaa <c>INotifyList</c>-rajapinnan. Tällöin
    /// listaan tehdyt muutokset päivittyvät komponenttiin.
    /// Listaa voi skrollata, jos kaikki rivit eivät mahdu
    /// kerralla sen sisälle.
    /// </summary>
    /// <typeparam name="T">Listan alkion tyyppi.</typeparam>
    /// <typeparam name="L">Listan tyyppi.</typeparam>
    /// <typeparam name="W">Listan rivin tyyppi (esimerkiksi Label).</typeparam>
    [EditorBrowsable( EditorBrowsableState.Never )]
    public abstract class ListWidget<T, L, W> : Widget
        where W : Widget
        where L : INotifyList<T>, new()
    {
        static Image upImage = null;
        static Image downImage = null;
        static Image transparentImage = null;

        internal L List;

        PushButton scrollUpButton = null;
        PushButton scrollDownButton = null;

        internal protected ScrollableList<W> Content;

        /// <summary>
        /// Listan alkiot.
        /// </summary>
        public L Items
        {
            get { return List; }
        }

        public ListWidget()
            : this( new L() )
        {
        }

        public ListWidget( L list )
            : base( new HorizontalLayout() )
        {
            Add( Content = new ScrollableList<W> { Color = Color.Transparent } );
            Add( CreateVerticalScrollPanel() );

            Bind( list );

            AddedToGame += AddListeners;
        }

        /// <summary>
        /// Luo annettua alkiota vastaavan listan rivin.
        /// </summary>
        internal protected abstract W CreateWidget( T item );

        private void AddListeners()
        {
            Game.Instance.Keyboard.Listen( Key.Up, ButtonState.Pressed, scrollUp, null ).InContext( this );
            Game.Instance.Keyboard.Listen( Key.Down, ButtonState.Pressed, scrollDown, null ).InContext( this );
        }

        private Widget CreateVerticalScrollPanel()
        {
            Widget scrollPanel = new Widget( new VerticalLayout() ) { Color = Color.Transparent, HorizontalSizing = Sizing.FixedSize };

            if ( upImage == null )
            {
                upImage = Game.LoadImageFromResources( "UpArrow" );
                downImage = Game.LoadImageFromResources( "DownArrow" );
                transparentImage = Image.FromColor( upImage.Width, upImage.Height, Color.Transparent );
            }

            scrollUpButton = new PushButton( transparentImage );
            scrollUpButton.Clicked += scrollUp;
            scrollPanel.Add( scrollUpButton );

            scrollPanel.Add( new VerticalSpacer() );

            scrollDownButton = new PushButton( transparentImage );
            scrollDownButton.Clicked += scrollDown;
            scrollPanel.Add( scrollDownButton );

            return scrollPanel;
        }

        private void Reset()
        {
            Content.Clear();

            foreach ( var item in List )
            {
                Content.Add( CreateWidget( item ) );
            }

            // This is a bit tricky case. Because the clear call above doesn't take
            // effect immediately, calling UpdateLayout() uses old objects when updating
            // layout. Thus, let's just wait for at least one update to occur.
            Timer.SingleShot( 0.1, delegate() { if ( !Content.IsAtBottom ) ShowDownButton(); } );
        }

        private void ShowUpButton()
        {
            scrollUpButton.Image = upImage;
        }

        private void ShowDownButton()
        {
            scrollDownButton.Image = downImage;
        }

        private void Hide(PushButton button)
        {
            button.Image = transparentImage;
        }

        private void scrollDown()
        {
            Content.ScrollDown();

            if ( !Content.IsAtTop )
            {
                ShowUpButton();
            }

            if ( Content.IsAtBottom )
            {
                Hide( scrollDownButton );
            }
        }

        private void scrollUp()
        {
            Content.ScrollUp();

            if ( Content.IsAtTop )
            {
                Hide( scrollUpButton );
            }
        }

        private void listChanged()
        {
            Reset();
        }

        /// <summary>
        /// Sitoo olemassaolevan listan tähän näyttöön.
        /// Kun listaa muutetaan, näytetyt arvot päivittyvät automaattisesti.
        /// </summary>
        /// <param name="list">Lista.</param>
        public void Bind( L list )
        {
            this.List = list;
            list.Changed += listChanged;
            Reset();
        }

        /// <summary>
        /// Poistaa yhteyden olemassaolevaan listaan ja luo oman (sisäisen) listan.
        /// </summary>
        public void Unbind()
        {
            List.Changed -= listChanged;
            Bind( new L() );
        }
    }
}
