﻿#region MIT License
/*
 * Copyright (c) 2009-2011 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tomi Karppinen, Tero Jäntti
 */

using System;
using System.ComponentModel;
using System.Collections.Generic;

namespace Jypeli.Widgets
{
    /// <summary>
    /// Asettelee widgetit päällekäin, järjestyksessä ylhäältä alas.
    /// Jos widgeteille ei ole tarpeeksi tilaa, vain osa niistä asetellaan
    /// paikalleen. Ensimmäisen näkyvän widgetin indeksi asetetaan <c>Update</c>:ssa
    /// propertyyn <c>StartIndex</c> ja viimeisen jälkeisen indeksi propertyyn
    /// <c>EndIndex</c>.
    /// </summary>
    internal class VerticalScrollLayout : ILayout
    {
        private Sizing _horizontalSizing;
        private Sizing _verticalSizing;
        private Vector _preferredSize;
        private double _heightRequestedByExpandingObjects;
        private double _heightRequestedByFixedSizeObjects;
        private double _spacing = 0;
        private double _topPadding = 0;
        private double _bottomPadding = 0;
        private double _leftPadding = 0;
        private double _rightPadding = 0;


        [EditorBrowsable( EditorBrowsableState.Never )]
        public Widget Parent { get; set; }

        /// <summary>
        /// Olioiden väliin jäävä tyhjä tila.
        /// </summary>
        public double Spacing
        {
            get { return _spacing; }
            set { _spacing = value; NotifyParent(); }
        }

        /// <summary>
        /// Yläreunaan jäävä tyhjä tila.
        /// </summary>
        public double TopPadding
        {
            get { return _topPadding; }
            set { _topPadding = value; NotifyParent(); }
        }

        /// <summary>
        /// Alareunaan jäävä tyhjä tila.
        /// </summary>
        public double BottomPadding
        {
            get { return _bottomPadding; }
            set { _bottomPadding = value; NotifyParent(); }
        }

        /// <summary>
        /// Vasempaan reunaan jäävä tyhjä tila.
        /// </summary>
        public double LeftPadding
        {
            get { return _leftPadding; }
            set { _leftPadding = value; NotifyParent(); }
        }

        /// <summary>
        /// Oikeaan reunaan jäävä tyhjä tila.
        /// </summary>
        public double RightPadding
        {
            get { return _rightPadding; }
            set { _rightPadding = value; NotifyParent(); }
        }

        public int StartIndex { get; set; }

        public int EndIndex { get; set; }

        private void NotifyParent()
        {
            if ( Parent != null )
            {
                Parent.NotifyParentAboutChangedSizingAttributes();
            }
        }

        [EditorBrowsable( EditorBrowsableState.Never )]
        public void UpdateSizeHints( IList<Widget> objects )
        {
            if ( objects.Count == 0 )
                return;

            double maxWidth = 0;
            double heightOfExpandingObjects = 0;
            double heightOfFixedSizeObjects = 0;
            Sizing horizontalSizing = Sizing.FixedSize;
            Sizing verticalSizing = Sizing.FixedSize;

            foreach ( var o in objects )
            {
                if ( o.PreferredSize.X > maxWidth )
                {
                    maxWidth = o.PreferredSize.X;
                }

                if ( o.VerticalSizing != Sizing.FixedSize )
                {
                    verticalSizing = Sizing.Expanding;
                    heightOfExpandingObjects += o.PreferredSize.Y;
                }
                else if ( o.VerticalSizing == Sizing.FixedSize )
                {
                    heightOfFixedSizeObjects += o.PreferredSize.Y;
                }

                if ( o.HorizontalSizing != Sizing.FixedSize )
                {
                    horizontalSizing = Sizing.Expanding;
                }
            }

            double preferredHeight = TopPadding + heightOfExpandingObjects + heightOfFixedSizeObjects + ( ( objects.Count - 1 ) * Spacing ) + BottomPadding;
            double preferredWidth = LeftPadding + maxWidth + RightPadding;

            // Just to make sure that any rounding errors don't cause too little space to be
            // allocated. If the last object would not fit entirely, it is not drawn at all.
            preferredHeight += 0.1;

            _horizontalSizing = horizontalSizing;
            _verticalSizing = verticalSizing;
            _heightRequestedByExpandingObjects = heightOfExpandingObjects;
            _heightRequestedByFixedSizeObjects = heightOfFixedSizeObjects;
            _preferredSize = new Vector( preferredWidth, preferredHeight );
        }

        [EditorBrowsable( EditorBrowsableState.Never )]
        public Sizing HorizontalSizing
        {
            get { return _horizontalSizing; }
        }

        [EditorBrowsable( EditorBrowsableState.Never )]
        public Sizing VerticalSizing
        {
            get { return _verticalSizing; }
        }

        [EditorBrowsable( EditorBrowsableState.Never )]
        public Vector PreferredSize
        {
            get { return _preferredSize; }
        }

        [EditorBrowsable( EditorBrowsableState.Never )]
        public void Update( IList<Widget> objects, Vector maximumSize )
        {
            double contentHeight = maximumSize.Y - ( TopPadding + BottomPadding );
            double contentWidth = maximumSize.X - ( LeftPadding + RightPadding );
            double contentBottomLimit = -maximumSize.Y / 2 + BottomPadding;
            double top = maximumSize.Y / 2 - TopPadding;
            int i;

            for ( i = StartIndex; i < objects.Count; i++ )
            {
                Widget o = objects[i];
                double width = o.PreferredSize.X;
                double height = o.PreferredSize.Y;

                if ( ( top - o.PreferredSize.Y ) >= contentBottomLimit )
                {
                    if ( ( o.PreferredSize.X > contentWidth ) || ( o.HorizontalSizing == Sizing.Expanding ) || ( o.HorizontalSizing == Sizing.AlwaysExpanding ) )
                    {
                        width = contentWidth;
                    }

                    o.Size = new Vector( width, height );
                    o.X = ( -maximumSize.X / 2 ) + ( LeftPadding + contentWidth / 2 );
                    o.Y = top - height / 2;

                    top -= height + Spacing;
                }
                else
                {
                    break;
                }
            }

            EndIndex = i;
        }
    }
}
