﻿using Jypeli.Controls;
#region MIT License
/*
 * Copyright (c) 2009-2011 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tero Jäntti, Tomi Karppinen, Janne Nikkanen.
 */

using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;

#if !XBOX && !WINDOWS_PHONE
using System.Runtime.Serialization.Formatters.Binary;
#endif

using Microsoft.Xna.Framework;
using Jypeli.Controls;

namespace Jypeli.Widgets
{
    /// <summary>
    /// Käyttöliittymän komponentti. Samankaltainen kuin <c>GameObject</c>,
    /// mutta toimii vain ruutukoordinaateissa.
    /// </summary>
#if !XBOX && !WINDOWS_PHONE
    [Serializable]
#endif
    public class Widget : ControlContexted, Destroyable, IDisposable
    {
        #region Fields

#if !XBOX && !WINDOWS_PHONE
        [NonSerialized]
#endif
        private Widget parent;

        private Game game = null;
        private Shape shape;
        private Vector size;
        private ILayout layout = null;
        internal List<Widget> childObjects = null;
        private Queue<Widget> childObjectsToRemove = null;
        private Sizing _horizontalSizing = Sizing.FixedSize;
        private Sizing _verticalSizing = Sizing.FixedSize;
        private Vector _preferredSize = new Vector(50, 50);
        private bool _sizeByLayout = true;
        private bool _layoutNeedsToBeUpdated = false;
        private ListenContext _context = new ListenContext();
        internal List<Listener> associatedListeners = new List<Listener>();
        private Color _color = new Color( 250, 250, 250, 240 );

        #endregion

        /// <summary>
        /// Olio, jonka lapsiolio tämä olio on. Jos <c>null</c>, olio ei
        /// ole minkään olion lapsiolio.
        /// </summary>
        public Widget Parent
        {
            get { return parent; }
            private set
            {
                parent = value;
                if ( value != null )
                {
                    _context.dynamicParent = true;
                    _context.parentObject = parent;
                    _context.Active = true;

                    // Parent takes care of updating from now on.
                    _layoutNeedsToBeUpdated = false;
                }
                else
                {
                    _context.dynamicParent = false;
                    _context.parentContext = new ListenContext();
                    _layoutNeedsToBeUpdated = true;
                }
            }
        }

        public ListenContext ControlContext { get { return _context; } }

        /// <summary>
        /// Koon asettaminen vaakasuunnassa, kun olio on
        /// asettelijan sisällä.
        /// </summary>
        public virtual Sizing HorizontalSizing
        {
            get
            {
                if ( SizingByLayout && (_horizontalSizing != Sizing.AlwaysExpanding) && (Layout != null) )
                    return Layout.HorizontalSizing;
                return _horizontalSizing;
            }
            set
            {
                _horizontalSizing = value;
                NotifyParentAboutChangedSizingAttributes();
            }
        }

        /// <summary>
        /// Koon asettaminen pystysuunnassa, kun olio on
        /// asettelijan sisällä.
        /// </summary>
        public virtual Sizing VerticalSizing
        {
            get
            {
                if ( SizingByLayout && ( _verticalSizing != Sizing.AlwaysExpanding ) && (Layout != null) )
                    return Layout.VerticalSizing;
                return _verticalSizing;
            }
            set
            {
                _verticalSizing = value;
                NotifyParentAboutChangedSizingAttributes();
            }
        }

        /// <summary>
        /// Koko, jota oliolla tulisi olla asettelijan sisällä. Todellinen koko voi olla
        /// pienempi, jos tilaa ei ole tarpeeksi.
        /// </summary>
        public virtual Vector PreferredSize
        {
            get
            {
                if ( Layout != null )
                    return Layout.PreferredSize;
                return _preferredSize;
            }
            set
            {
                _preferredSize = value;
                NotifyParentAboutChangedSizingAttributes();
            }
        }

        public bool SizingByLayout
        {
            get { return _sizeByLayout; }
            set { _sizeByLayout = value; }
        }

        /// <summary>
        /// Should be called whenever properties that might affect layouts
        /// are changed.
        /// </summary>
        internal protected void NotifyParentAboutChangedSizingAttributes()
        {
            if (Parent == null)
            {
                _layoutNeedsToBeUpdated = true;
            }
            else
            {
                Parent.NotifyParentAboutChangedSizingAttributes();
            }
        }

        /// <summary>
        /// Asettelija lapsiolioille. Asettaa lapsiolioiden koon sekä paikan.
        /// </summary>
        public ILayout Layout
        {
            get { return layout; }
            set
            {
                if (layout != null)
                {
                    ILayout old = layout;
                    layout = null;
                    old.Parent = null;
                }

                InitializeObjectLists();
                layout = value;
                layout.Parent = this;
                NotifyParentAboutChangedSizingAttributes();
            }
        }

        /// <summary>
        /// Vapaasti asetettava muuttuja.
        /// </summary>
        public object Tag { get; set; }

        /// <summary>
        /// Olion luomisaika.
        /// </summary>
        public TimeSpan CreationTime { get; private set; }

        /// <summary>
        /// Olion elinaika. Lasketaan siitä lähtien, kun olio luodaan.
        /// </summary>
        public TimeSpan Lifetime
        {
            get { return Game.Time.SinceStartOfGame - CreationTime; }
        }

        /// <summary>
        /// Olion suurin mahdollinen elinaika.
        /// Kun <c>Lifetime</c> on suurempi kuin tämä, olio kuolee.
        /// </summary>
        public TimeSpan MaximumLifetime { get; set; }

        /// <summary>
        /// Olion koko pelimaailmassa.
        /// Kertoo olion äärirajat, ei muotoa.
        /// </summary>
        public virtual Vector Size
        {
            get { return size; }
            set { size = value; }
        }

        /// <summary>
        /// Olion leveys (X-suunnassa, leveimmässä kohdassa).
        /// </summary>
        public double Width
        {
            get { return Size.X; }
            set { Size = new Vector( value, Size.Y ); }
        }

        /// <summary>
        /// Olion korkeus (Y-suunnassa, korkeimmassa kohdassa).
        /// </summary>
        public double Height
        {
            get { return Size.Y; }
            set { Size = new Vector( Size.X, value ); }
        }

        /// <summary>
        /// Olion paikka. Jos olio on jonkun toisen peliolion lapsi, paikka on suhteessa
        /// tämän vanhempaan (<c>Parent</c>). Muuten paikka on paikka pelimaailmassa.
        /// </summary>
        public virtual Vector Position { get; set; }

        /// <summary>
        /// Olion absoluuttinen paikka pelimaailmassa.
        /// Jos olio ei ole minkään toisen peliolion lapsiolio,
        /// tämä on sama kuin <c>Position</c>.
        /// </summary>
        public Vector AbsolutePosition
        {
            get
            {
                if ( Parent != null )
                    return Parent.AbsolutePosition + this.Position;
                return Position;
            }
        }

        /// <summary>
        /// Olion absoluuttinen kulma pelimaailmassa.
        /// Jos olio ei ole minkään toisen peliolion lapsiolio,
        /// tämä on sama kuin <c>Angle</c>.
        /// </summary>
        public Angle AbsoluteAngle
        {
            get
            {
                if ( Parent != null )
                    return Parent.AbsoluteAngle + this.Angle;
                return Angle;
            }
        }

        /// <summary>
        /// Olion vasemman reunan x-koordinaatti.
        /// </summary>
        public double Left
        {
            get { return Position.X - Size.X / 2; }
            set { Position = new Vector( value + Size.X / 2, Position.Y ); }
        }

        /// <summary>
        /// Olion oikean reunan x-koordinaatti.
        /// </summary>
        public double Right
        {
            get { return Position.X + Size.X / 2; }
            set { Position = new Vector( value - Size.X / 2, Position.Y ); }
        }

        /// <summary>
        /// Olion yläreunan y-koordinaatti.
        /// </summary>
        public double Top
        {
            get { return Position.Y + Size.Y / 2; }
            set { Position = new Vector( Position.X, value - Size.Y / 2 ); }
        }

        /// <summary>
        /// Olion alareunan y-koordinaatti.
        /// </summary>
        public double Bottom
        {
            get { return Position.Y - Size.Y / 2; }
            set { Position = new Vector( Position.X, value + Size.Y / 2 ); }
        }

        /// <summary>
        /// Olion paikan X-koordinaatti.
        /// </summary>
        public double X
        {
            get
            {
                return Position.X;
            }
            set
            {
                Position = new Vector( value, Position.Y );
            }
        }

        /// <summary>
        /// Olion paikan Y-koordinaatti.
        /// </summary>
        public double Y
        {
            get
            {
                return Position.Y;
            }
            set
            {
                Position = new Vector( Position.X, value );
            }
        }

        /// <summary>
        /// Olion kulma.
        /// Nolla = osoittaa oikealle.
        /// </summary>      
        public virtual Angle Angle { get; set; }

        /// <summary>
        /// Animaatio. Voi olla <c>null</c>, jolloin piirretään vain väri.
        /// </summary>
        public Animation Animation { get; set; }

        /// <summary>
        /// Olion kuva. Voi olla <c>null</c>, jolloin piirretään vain väri.
        /// </summary>
        public virtual Image Image
        {
            get
            {
                if ( Animation != null )
                    return Animation.CurrentFrame;
                return null;
            }
            set
            {
                if ( value == null )
                    Animation = null;
                else
                    Animation = new Animation( value );
            }
        }

        /// <summary>
        /// Määrittää kuinka moneen kertaan kuva piirretään. Esimerkiksi (3.0, 2.0) piirtää
        /// kuvan 3 kertaa vaakasuunnassa ja 2 kertaa pystysuunnassa.
        /// </summary>
        public Vector TextureWrapSize = new Vector( 1.0, 1.0 );

        /// <summary>
        /// Väri, jonka värisenä olio piirretään, jos tekstuuria ei ole määritelty.
        /// </summary>
        public virtual Color Color
        {
            get { return _color; }
            set { _color = value; }
        }

        /// <summary>
        /// Reunojen väri.
        /// </summary>
        public Color BorderColor { get; set; }

        /// <summary>
        /// Määrittää, täyttääkö tekstuuri olion muodon jos olio on suurempi kuin sille
        /// määrätty tekstuuri.
        /// </summary>
        public bool TextureFillsShape { get; set; }

        /// <summary>
        /// Olion muoto.
        /// </summary>
        public virtual Shape Shape
        {
            get { return shape; }
            set { shape = value; }
        }

        /// <summary>
        /// Olion muoto merkkijonona (kenttäeditorin käyttöön)
        /// </summary>
        internal string ShapeString
        {
            get { return Shape.GetType().Name; }
            set { Shape = Shape.FromString( value ); }
        }

        /// <summary>
        /// Peli, johon olio on lisätty. <c>null</c>, jos
        /// oliota ei ole lisätty peliin.
        /// </summary>
        [EditorBrowsable( EditorBrowsableState.Never )]
        public Game Game
        {
            get { return game; }
            internal set
            {
                game = value;
                if (childObjects != null)
                {
                    foreach (var o in childObjects)
                        o.Game = value;
                }
            }
        }

        /// <summary>
        /// Piirretäänkö olio ruudulle.
        /// </summary>
        public bool Visible { get; set; }

        /// <summary>
        /// Jos <c>true</c>, pelin sekä ikkunan alla olevien widgettien
        /// ohjaimet eivät ole käytössä kun ikkuna on näkyvissä.
        /// </summary>
        public bool IsModal { get; set; }

        /// <summary>
        /// Onko olio lisätty peliin.
        /// </summary>
        public bool IsAddedToGame
        {
            get { return Game != null; }
        }


        /// <summary>
        /// Tapahtuu, kun olio lisätään peliin.
        /// </summary>
        public event Action AddedToGame;

        /// <summary>
        /// Tapahtuu, kun olio poistetaan pelistä.
        /// </summary>
        public event Action RemovedFromGame;

        internal void OnAddedToGame()
        {
            if ( AddedToGame != null )
                AddedToGame();

            if (childObjects != null)
            {
                foreach (Widget o in childObjects)
                    o.OnAddedToGame();
            }
        }

        internal void OnRemovedFromGame()
        {
            if ( RemovedFromGame != null )
                RemovedFromGame();

            if ( childObjects != null )
            {
                foreach ( Widget o in childObjects )
                    o.OnRemovedFromGame();
            }
        }

        public Widget( ILayout layout )
            : this( 100, 100 )
        {
            this.Layout = layout;
        }

        public Widget( Animation animation )
            : this( animation.frames[0].Width, animation.frames[0].Height )
        {
            this.Animation = animation;
            this.PreferredSize = new Vector( animation.frames[0].Width, animation.frames[0].Height );
        }

        /// <summary>
        /// Alustaa uuden widgetin.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        public Widget( double width, double height )
            : this( width, height, Shape.Rectangle )
        {
        }

        /// <summary>
        /// Alustaa uuden widgetin.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        /// <param name="shape">Muoto.</param>
        public Widget( double width, double height, Shape shape )
        {
            this.CreationTime = Game.Time.SinceStartOfGame;
            this.MaximumLifetime = TimeSpan.MaxValue;
            this.size = this.PreferredSize = new Vector( width, height );
            this.shape = shape;
            this.BorderColor = Color.Transparent;
            this.Tag = "";
            this.Visible = true;

            if ( ControlContext == null || ControlContext.IsDestroyed )
            {
                _context = new Controls.ListenContext();
            }
        }

#if !XBOX && !WINDOWS_PHONE
        /// <summary>
        /// Tekee oliosta kopion.
        /// </summary>
        /// <returns>
        /// Uusi <see cref="Widget"/>-olio samoilla arvoilla.
        /// </returns>
        internal object Clone()
        {
            MemoryStream ms = new MemoryStream();
            BinaryFormatter bf = new BinaryFormatter();

            bf.Serialize( ms, this );
            ms.Position = 0;

            object obj = bf.Deserialize( ms );
            ms.Close();

            // Copy nonserializable data
            ( (Widget)obj ).Animation = new Animation( this.Animation );

            return obj;
        }
#endif

        /// <summary>
        /// Onko piste <c>p</c> tämän widgetin sisäpuolella.
        /// </summary>
        public bool IsInside(Vector point)
        {
            // TODO: onko lapsiolioiden sisällä?

            Vector p = this.AbsolutePosition;

            return point.X >= (p.X - Width / 2)
                && point.X <= (p.X + Width / 2)
                && point.Y >= (p.Y - Height / 2)
                && point.Y <= (p.Y + Height / 2);
        }

        /// <summary>
        /// Lisää annetun peliolion tämän olion lapseksi. Lapsiolio liikkuu tämän olion mukana,
        /// ja sen paikka ja kulma ilmaistaan suhteessa tähän olioon.
        /// </summary>
        public void Add( Widget childObject )
        {
            InitializeObjectLists();
            childObjects.Add( childObject );
            childObject.Parent = this;
            NotifyParentAboutChangedSizingAttributes();
        }

        /// <summary>
        /// Poistaa lapsiolion, mutta ei tuhoa sitä.
        /// </summary>
        public void Remove( Widget childObject )
        {
            childObjectsToRemove.Enqueue( childObject );
        }

        public virtual void Clear()
        {
            foreach ( var o in childObjects )
                o.Destroy();
        }

        /// <summary>
        /// Antaa widgetin maksimikoon siinä tapauksessa, että kokoa ei
        /// ole annettu rakentajassa (tai tarkemmin sanoen muuttujan <c>SizingByLayout</c>
        /// arvo on <c>true</c>). Olio ei siis automaattisesti kasva tätä isommaksi.
        /// </summary>
        /// <returns></returns>
        protected virtual Vector GetMaximumSize()
        {
            return new Vector( double.PositiveInfinity, double.PositiveInfinity );
        }

        /// <summary>
        /// Päivittää lapsiolioiden paikat ja koot, jos widgetille on asetettu asettelija.
        /// Tätä metodia EI yleensä tarvitse kutsua itse, sillä asettelija
        /// päivitetään automaattisesti jokaisella framella. Tämä metodi on
        /// tarpeellinen esim. silloin, kun widgetille on lisätty lapsiolioita
        /// (tai muutettu niiden ominaisuuksia) ja niiden paikat tarvitsee päivittää
        /// välittömästi lisäyksen jälkeen. Silloinkin tätä tarvitsee kutsua vain kerran,
        /// viimeisimmän muutoksen jälkeen.
        /// </summary>
        public void RefreshLayout()
        {
            if (Layout != null)
            {
                // First, lets ask how big the child objects need to be.
                UpdateSizeHints();

                // Then, lets set the size accordingly, if we are allowed to do so.
                if ( SizingByLayout )
                {
                    Vector newSize = Layout.PreferredSize;
                    Vector maxSize = this.GetMaximumSize();

                    if ( newSize.X > maxSize.X )
                        newSize.X = maxSize.X;
                    if ( newSize.Y > maxSize.Y )
                        newSize.Y = maxSize.Y;

                    Size = newSize;
                }

                // Finally, lets position the child objects into the space we have available
                // for them.
                UpdateLayout(Size);
            }
        }

        /// <summary>
        /// Recursively updates the preferred sizes (and other parameters that
        /// affect the layout) of the object and its child objects.
        /// </summary>
        private void UpdateSizeHints()
        {
            if ( childObjects != null )
            {
                foreach ( var child in childObjects )
                {
                    child.UpdateSizeHints();
                }
            }

            if ( Layout != null )
                Layout.UpdateSizeHints( childObjects );
        }

        /// <summary>
        /// Recursively updates the layouts of the object and
        /// its child objects. <c>UpdateSizeHints()</c> must have
        /// been called before this, because the layout needs to
        /// know how big the objects need to be.
        /// </summary>
        /// <param name="maximumSize">The actual size that is allocated for the layout.</param>
        private void UpdateLayout( Vector maximumSize )
        {
            if ( Layout != null )
                Layout.Update( childObjects, maximumSize );

            if ( childObjects != null )
            {
                foreach ( var child in childObjects )
                {
                    child.UpdateLayout( child.Size );
                }
            }
        }

        private void InitializeObjectLists()
        {
            if (childObjects == null)
            {
                childObjects = new List<Widget>();
                childObjectsToRemove = new Queue<Widget>();
            }
        }

        #region Destroyable

        /// <summary>
        /// Onko olio tuhottu.
        /// </summary>
        /// <returns></returns>
        public bool IsDestroyed
        {
            get { return Lifetime > MaximumLifetime; }
        }

        /// <summary>
        /// Tuhoaa olion. Tuhottu olio poistuu pelistä.
        /// </summary>
        public virtual void Destroy()
        {
            this.MaximumLifetime = new TimeSpan( 0 );
            OnDestroyed();

            ControlContext.Destroy();

            if ( childObjects != null )
            {
                foreach ( Widget child in childObjects )
                    child.Destroy();
            }

            if ( associatedListeners != null )
            {
                foreach ( Listener listener in associatedListeners )
                {
                    listener.Destroy();
                }
            }
        }

        /// <summary> 
        /// Tapahtuu, kun olio tuhotaan. 
        /// </summary> 
        public event Action Destroyed;

        private void OnDestroyed()
        {
            if ( Destroyed != null )
                Destroyed();
        }

        #endregion

        public void Dispose()
        {
            // For garbage collection
            Destroy();
        }

        /// <summary>
        /// Widgetin päivitys.
        /// </summary>
        /// <param name="time">Peliaika.</param>
        [EditorBrowsable( EditorBrowsableState.Never )]
        public virtual void Update( Time time )
        {
            if (childObjects != null)
            {
                foreach (Widget child in childObjects)
                {
                    child.Update(time);
                    if (child.IsDestroyed)
                    {
                        childObjectsToRemove.Enqueue(child);
                    }
                }

                while (childObjectsToRemove.Count > 0)
                {
                    Widget o = childObjectsToRemove.Dequeue();
                    childObjects.Remove(o);
                    o.Game = null;
                    o.Parent = null;
                    NotifyParentAboutChangedSizingAttributes();
                }

                if (_layoutNeedsToBeUpdated)
                {
                    RefreshLayout();
                    _layoutNeedsToBeUpdated = false;
                }
            }
        }

        /// <summary>
        /// Kutsutaan, kun olio piirretään.
        /// </summary>
        internal protected virtual void Draw( Matrix parentTransformation, Matrix transformation )
        {
        }

        internal void Draw( Matrix parentTransformation )
        {
            if ( !Visible ) return;

            Renderer renderer = Game.Renderer;

            Matrix transformation =
                Matrix.CreateScale( (float)Size.X, (float)Size.Y, 1f )
                * Matrix.CreateRotationZ( (float)Angle.Radians )
                * Matrix.CreateTranslation( (float)Position.X, (float)Position.Y, 0f )
                * parentTransformation;

            if ( Image != null && ( !TextureFillsShape ) )
            {
                renderer.DrawImage( Image, ref transformation, TextureWrapSize );
            }
            else if ( Image != null )
            {
                renderer.DrawShape( Shape, ref transformation, ref transformation, Image, TextureWrapSize, Color );
            }
            else
            {
                renderer.DrawShape( Shape, ref transformation, Color );
            }

            Graphics.LineBatch.Begin( ref transformation );
            {
                Vector[] vertices = Shape.Cache.OutlineVertices;
                for ( int i = 0; i < vertices.Length - 1; i++ )
                {
                    Graphics.LineBatch.Draw( vertices[i], vertices[i + 1], BorderColor );
                }
                Graphics.LineBatch.Draw( vertices[vertices.Length - 1], vertices[0], BorderColor );
            }
            Graphics.LineBatch.End();

            Draw( parentTransformation, transformation );

            if ( childObjects != null )
            {
                Matrix childTransformation =
                    Matrix.CreateRotationZ( (float)Angle.Radians )
                    * Matrix.CreateTranslation( (float)Position.X, (float)Position.Y, 0f )
                    * parentTransformation;

                DrawChildObjects( ref parentTransformation, ref childTransformation );
            }
        }

        internal protected virtual void DrawChildObjects( ref Matrix parentTransformation, ref Matrix childTransformation )
        {
            if ( !Visible ) return;

            foreach ( var child in childObjects )
            {
                child.Draw( childTransformation );
            }
        }
    }
}
