﻿#region MIT License
/*
 * Copyright (c) 2009-2011 University of Jyväskylä, Department of Mathematical
 * Information Technology.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#endregion

/*
 * Authors: Tomi Karppinen, Tero Jäntti
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Jypeli.Widgets
{
    /// <summary>
    /// Ikkuna.
    /// </summary>
    public class Window : Widget
    {
        Color _actColor = new Color( 255, 255, 255, 200 );
        Color _inactColor = new Color( 50, 50, 50, 50 );
        Color _actTitle = new Color( 255, 255, 255, 100 );
        Color _inactTitle = new Color( 128, 128, 128, 50 );

        bool moving = false;
        Vector movementCenter = Vector.Zero;

        Widget _clientArea = null;

        /// <summary>
        /// Ikkunan otsikko. Älä käytä.
        /// </summary>
        [Obsolete("Title bar is not used anymore")]
        public Label Title { get; private set; }

        /// <summary>
        /// Ikkunan otsikkopalkin väri, kun ikkuna on aktiivinen. Älä käytä.
        /// </summary>
        [Obsolete( "Title bar is not used anymore" )]
        public Color TitleBarActiveColor
        {
            get { return _actTitle; }
            set { _actTitle = value; }
        }

        /// <summary>
        /// Ikkunan otsikkopalkin väri, kun ikkuna ei ole aktiivinen. Älä käytä.
        /// </summary>
        [Obsolete( "Title bar is not used anymore" )]
        public Color TitleBarInactColor
        {
            get { return _inactTitle; }
            set { _inactTitle = value; }
        }

        public override Color Color
        {
            get
            {
                return base.Color;
            }
            set
            {
                ActiveColor = value;
                InactiveColor = Color.Darker( value, 75 );
                base.Color = value;
            }
        }

        /// <summary>
        /// Ikkunan väri, kun ikkuna on aktiivinen.
        /// </summary>
        public Color ActiveColor
        {
            get { return _actColor; }
            set { _actColor = value; }
        }

        /// <summary>
        /// Ikkunan väri, kun ikkuna ei ole aktiivinen.
        /// </summary>
        public Color InactiveColor
        {
            get { return _inactColor; }
            set { _inactColor = value; }
        }

        /// <summary>
        /// Sulkupainike. Älä käytä.
        /// </summary>
        [Obsolete( "Title bar is not used anymore" )]
        public PushButton CloseButton { get; private set; }

        /// <summary>
        /// Ikkunan piirtoalue.
        /// </summary>
        [Obsolete( "Add childwidgets directly to the window" )]
        public Widget ClientArea 
        { 
            get
            {
                if ( _clientArea == null )
                {
                    _clientArea = new Widget( 0, 0 ) { SizingByLayout = true, Color = Color.Transparent, HorizontalSizing = Sizing.Expanding, VerticalSizing = Sizing.Expanding };
                    _clientArea.Layout = new VerticalLayout() { Spacing = 20, LeftPadding = 20, RightPadding = 20, TopPadding = 20, BottomPadding = 20 };
                    Add( _clientArea );
                }
                return _clientArea;
            }
        }

        /// <summary>
        /// Ikkunatapahtumien käsittelijä.
        /// </summary>
        public delegate void WindowHandler( Window sender );

        /// <summary>
        /// Tapahtuu kun ikkuna suljetaan.
        /// </summary>
        public event WindowHandler Closed;

        private void OnClosed()
        {
            if ( Closed != null )
                Closed( this );
        }

        /// <summary>
        /// Alustaa uuden ikkunan.
        /// </summary>
        public Window()
            : base( new VerticalLayout() )
        {
            initialize();
        }

        /// <summary>
        /// Alustaa uuden ikkunan.
        /// </summary>
        /// <param name="width">Leveys.</param>
        /// <param name="height">Korkeus.</param>
        public Window( double width, double height )
            : base( width, height, Shape.Rectangle )
        {
            SizingByLayout = false;
            Layout = new VerticalLayout();
            initialize();
        }

        /// <summary>
        /// Ikkunalla maksimikoko on siten, että se mahtuu näytölle.
        /// </summary>
        protected override Vector GetMaximumSize()
        {
            var screen = Game.Screen;
            return new Vector( screen.Width * ( 7.0 / 8.0 ), screen.Height * ( 7.0 / 8.0 ) );
        }

        private void initialize()
        {
            CreateTitleBar();

            ControlContext.Activated += Window_Activated;
            ControlContext.Deactivated += Window_Deactivated;

#if WINDOWS
            AddedToGame += AddControls;
#endif

            this.IsModal = true;

            RefreshLayout();
        }

        void AddControls()
        {
            Game.Mouse.ListenOn( this, MouseButton.Left, ButtonState.Pressed, StartMoveWindow, null ).InContext( this );
            Game.Mouse.Listen( MouseButton.Left, ButtonState.Down, MoveWindow, null ).InContext( this );
            Game.Mouse.ListenOn( this, MouseButton.Left, ButtonState.Released, EndMoveWindow, null ).InContext( this );
        }

        void StartMoveWindow()
        {
            if ( Game == null ) return;
            movementCenter = this.Position - Game.Mouse.PositionOnScreen;
            moving = true;
        }

        void MoveWindow()
        {
            if ( !moving ) return;
            this.Position = movementCenter + Game.Mouse.PositionOnScreen;
        }

        void EndMoveWindow()
        {
            moving = false;
        }

        [Obsolete( "Just creates what is necessary to avoid breaking things. Remove in the next release." )]
        private void CreateTitleBar()
        {
            //titleBar = new Widget(new HorizontalLayout()) { SizingByLayout = false, Color = TitleBarActiveColor, HorizontalSizing = Sizing.Expanding, VerticalSizing = Sizing.FixedSize }; 
            //titleBar.Height = 25;
            //Add(titleBar);

            Title = new Label();
            //Title.HorizontalSizing = Sizing.Expanding;
            //Title.VerticalSizing = Sizing.FixedSize;
            //Title.TextColor = Color.White;
            //Title.VerticalAlignment = VerticalAlignment.Center;
            //Title.HorizontalAlignment = HorizontalAlignment.Left;
            //Title.XMargin = 10;
            //titleBar.Add(Title);

            CloseButton = new PushButton( 25, 25, "x" ) { SizingByLayout = false, HorizontalSizing = Sizing.FixedSize, VerticalSizing = Sizing.FixedSize };
            //CloseButton.Clicked += new Action(Close);
            //titleBar.Add(CloseButton);
        }

        void Window_Activated()
        {
            base.Color = ActiveColor;
            //titleBar.Color = TitleBarActiveColor;
        }

        void Window_Deactivated()
        {
            base.Color = InactiveColor;
            //titleBar.Color = TitleBarInactColor;
        }

        /// <summary>
        /// Sulkee ikkunan.
        /// </summary>
        public void Close()
        {
            ControlContext.Destroy();
            OnClosed();
            //Destroy();
            if ( Parent == null )
            {
                if ( Game != null ) Game.Remove( this );
            }
            else
                Parent.Remove( this );
        }
    }
}
