﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using System.IO;
using Jypeli;
using Jypeli.Widgets;

namespace LevelEditor
{
    public class FileWindow : Window
    {
        private InputBox inputBox;
        private Label errorLabel = new Label() { TextColor = Color.Red };
        private Widget completionList = null;
        private bool fileMustExist;

        public event Action<string> PathEntered;

        public FileWindow( string title, string directory, bool mustExist )
        {
            Layout = new VerticalLayout()
            {
                Spacing = 15,
                LeftPadding = 25,
                RightPadding = 25,
                TopPadding = 25,
                BottomPadding = 25,
            };

            Add( new Label( title ) );

            inputBox = new InputBox( 50 );
            inputBox.Text = directory;
            if ( directory[directory.Length - 1] != Path.DirectorySeparatorChar )
                inputBox.Text += Path.DirectorySeparatorChar;
            inputBox.TextChanged += UpdateHints;
            Add( inputBox );

            fileMustExist = mustExist;
            AddedToGame += AddControls;
        }

        void AddControls()
        {
            Game.Keyboard.Listen( Key.Tab, ButtonState.Pressed, CompletePath, null ).InContext( this );
            Game.Keyboard.Listen( Key.Enter, ButtonState.Pressed, TextEntered, null ).InContext( this );
            Game.Keyboard.Listen( Key.Escape, ButtonState.Pressed, Close, null ).InContext( this );
        }

        void UpdateHints( string text )
        {
            if ( errorLabel.Parent == this )
            {
                Remove( errorLabel );
            }
        }

        void CompletePath()
        {
            int i = inputBox.Text.LastIndexOf( Path.DirectorySeparatorChar );
            if ( i < 0 ) return;

            // Path.GetDirectoryName() does the same thing, but fails if the path is invalid.
            string leadingDirectory = inputBox.Text.Substring( 0, i );

            if ( !Directory.Exists( leadingDirectory ) ) return;

            string fileName = Path.GetFileName( inputBox.Text );
            string searchPattern = fileName + '*';
            string longestCommonPrefix = null;
            IEnumerable<string> wordsWithLongestCommonPrefix;

            var entries = from f in Directory.GetFileSystemEntries( leadingDirectory, searchPattern )
                          where !IsHiddenFile( Path.GetFileName( f ) )
                          select Path.GetFileName( f );

            if ( entries.Count() == 1 )
            {
                string completedPath = Path.Combine( leadingDirectory, entries.First() );
                inputBox.Text = completedPath;
                if ( Directory.Exists( completedPath ) )
                {
                    inputBox.Text += Path.DirectorySeparatorChar;
                }
                
                if ( completionList != null )
                    completionList.Clear();
            }
            else if ( HaveLongerCommonPrefix( entries, fileName, out wordsWithLongestCommonPrefix, out longestCommonPrefix ) )
            {
                inputBox.Text = Path.Combine( leadingDirectory, longestCommonPrefix );
                ShowCompletions( wordsWithLongestCommonPrefix );
            }
            else
            {
                ShowCompletions( entries );
            }
        }

        private void ShowCompletions( IEnumerable<string> entries )
        {
            if ( completionList == null )
            {
                completionList = new Widget( new VerticalLayout() ) { Color = Color.Transparent };
                Add( completionList );
            }

            completionList.Clear();
            foreach ( var f in entries )
            {
                completionList.Add( new Label( f ) );
            }
        }

        private static bool IsHiddenFile( string filename )
        {
            return filename.StartsWith( "." );
        }

        private static bool HaveLongerCommonPrefix( IEnumerable<string> words, string initialPrefix, out IEnumerable<string> wordsWithLongestCommonPrefix, out string fullCommonPrefix )
        {
            if ( words.Count() == 0 )
            {
                fullCommonPrefix = initialPrefix;
                wordsWithLongestCommonPrefix = null;
                return false;
            }

            StringBuilder prefix = new StringBuilder( initialPrefix );
            var wordsWithInitialPrefix = from w in words where w.StartsWith( initialPrefix ) select w;
            int i = initialPrefix.Length;

            for ( ; ; )
            {
                string first = words.First();
                if ( first.Length <= i )
                    break;

                char c = first[i];

                foreach ( var word in wordsWithInitialPrefix )
                {
                    if ( ( word.Length <= i ) || ( char.ToUpper( word[i] ) != char.ToUpper( c ) ) )
                    {
                        goto Finish;
                    }
                }

                prefix.Append( c );
                i++;
            }

        Finish:
            fullCommonPrefix = prefix.ToString();
            wordsWithLongestCommonPrefix = wordsWithInitialPrefix;
            return ( fullCommonPrefix.Length > initialPrefix.Length );
        }

        void TextEntered()
        {
            if ( !fileMustExist || File.Exists( inputBox.Text ) )
            {
                if ( PathEntered != null )
                {
                    PathEntered( inputBox.Text );
                }
                Close();
            }
            else
            {
                errorLabel.Text = "File does not exist!";
                if ( errorLabel.Parent == null )
                    Add( errorLabel );
            }
        }
    }
}
