﻿using System;
using System.IO;
using System.Xml;
using Jypeli;
using Jypeli.Controls;
using Jypeli.Widgets;
using Jypeli.LevelEditor;
using LevelEditor.Menus;
using LevelEditor.Tools;
using System.Collections.Generic;

namespace LevelEditor
{
    public class LevelEditor : Game
    {
        public static LevelData LevelData { get; private set; }
        public static TemplateMenu TemplateMenu { get; private set; }
        public static ToolMenu ToolMenu { get; private set; }
        public static Grid Grid { get; private set; }

        static double MoveMargin = 30;
        static double CameraSpeed = 10;

        Label coordLabel = new Label();
        Label zoomLabel = new Label();
        Label gridLabel = new Label();

        static List<ObjectPair> Objects = new List<ObjectPair>();

        string currentLevelFile = null;

        public LevelEditor( string levelFile )
        {
            this.currentLevelFile = levelFile;
        }

        public override void Begin()
        {
            InitializeLevelToDefaults();
            CreateUI();
            SetMouseControls();
            SetKeyboardControls();
            InitLoadLevel();
            ShowControlHelp();

#if DEBUG
            InsertObject( new Vector( 10, 20 ), new Template() );
#endif
        }

        private void CreateUI()
        {
            CreateGrid();
            CreateMenus();
            CreateLabels();           
            
            MessageDisplay.Left = ToolMenu.Right + 60;
        }

        private void CreateGrid()
        {
            Grid = new Grid { Color = Color.DarkMagenta, CellSize = new Vector( 10, 10 ) };
        }

        private void CreateMenus()
        {
            ToolMenu = new ToolMenu();
            ToolMenu.X = Screen.Left + ToolMenu.Width;
            ToolMenu.Y = Screen.Top - ToolMenu.Height;
            Add( ToolMenu );

            TemplateMenu = new TemplateMenu( 100, Screen.Height / 2, LevelData );
            TemplateMenu.X = Screen.Left + TemplateMenu.Width / 2;
            TemplateMenu.Y = Screen.Center.Y;
            TemplateMenu.TemplateAdded += TemplateAdded;
            TemplateMenu.TemplateRemoving += ConfirmRemoveTemplate;
            TemplateMenu.TemplateSelected += StartToInsertObjectsOfSelectedTemplate;
            Add( TemplateMenu );
        }

        void TemplateAdded( Template template )
        {
            LevelData.Templates.Add( template );
        }

        void ConfirmRemoveTemplate( Template template )
        {
            List<ObjectPair> templateObjs = Objects.FindAll( pair => pair.LevelObject.Template == template );

            Action useDefault = delegate
            {
                for ( int i = 0; i < templateObjs.Count; i++ )
                {
                    var gameObj = templateObjs[i].GameObject;
                    var levelObj = templateObjs[i].LevelObject;
                    levelObj.ResetTemplate();
                    levelObj.Template.Properties.Apply( ref gameObj );
                    levelObj.OverridingProperties.Apply( ref gameObj );
                }

                RemoveTemplate( template );
            };

            if ( templateObjs.Count > 0 )
            {
                MultiSelectWindow whatToDo = new MultiSelectWindow(
                    string.Format( "{0} object(s) still use the template \"{1}\"", templateObjs.Count, template.Name ),
                    "Copy properties", "Use default template", "Delete objects", "Cancel" );
                whatToDo.SetButtonColor( Color.DarkRed );
                whatToDo.Buttons[3].Color = Color.DarkGreen;
                whatToDo.AddItemHandler( 0, delegate { CopyTemplateData( templateObjs ); useDefault(); RemoveTemplate( template ); } );
                whatToDo.AddItemHandler( 1, useDefault );
                whatToDo.AddItemHandler( 2, delegate { RemoveObjects( templateObjs ); RemoveTemplate( template ); } );
                Add( whatToDo );
            }
            else
            {
                MultiSelectWindow confirmRemove = new MultiSelectWindow(
                    string.Format( "Remove unused template {0}?", template.Name ),
                    "Remove", "Cancel" );
                confirmRemove.Buttons[0].Color = Color.DarkRed;
                confirmRemove.Buttons[1].Color = Color.DarkGreen;
                confirmRemove.AddItemHandler( 0, delegate { RemoveTemplate( template ); } );
                Add( confirmRemove );
            }
        }

        private void CopyTemplateData( List<ObjectPair> objs )
        {
            foreach ( var pair in objs )
            {
                pair.LevelObject.OverridingProperties.Add( pair.LevelObject.Template.Properties );
            }
        }

        private void RemoveTemplate( Template template )
        {
            TemplateMenu.RemoveTemplate( template );
            LevelData.Templates.Remove( template );
        }

        private void CreateLabels()
        {
            coordLabel.Left = Screen.Right - 250;
            coordLabel.Top = Screen.Top - 50;
            coordLabel.HorizontalAlignment = HorizontalAlignment.Left;
            coordLabel.Width = 250;
            Add( coordLabel );

            zoomLabel.Left = Screen.Right - 250;
            zoomLabel.Top = Screen.Top - 70;
            zoomLabel.HorizontalAlignment = HorizontalAlignment.Left;
            zoomLabel.Width = 250;
            Add( zoomLabel );

            gridLabel.Left = Screen.Right - 250;
            gridLabel.Top = Screen.Top - 90;
            gridLabel.HorizontalAlignment = HorizontalAlignment.Left;
            gridLabel.Width = 250;
            Add( gridLabel );

            UpdateLabels();
        }

        private void UpdateLabels()
        {
            coordLabel.Text = String.Format( "Pos: ({0:0.00}, {1:0.00} )", Camera.Position.X, Camera.Position.Y );
            zoomLabel.Text = String.Format( "Zoom: {0:0.0}", Camera.ZoomFactor );
            gridLabel.Text = String.Format( "Grid size: {0:0.}x{1:0.}", Grid.CellSize.X, Grid.CellSize.Y );
        }

        public static ObjectPair GetObject( GameObject gameObj )
        {
            return Objects.FindLast( pair => pair.GameObject == gameObj );
        }

        public static ObjectPair GetObject( LevelObject levelObj )
        {
            return Objects.FindLast( pair => pair.LevelObject == levelObj );
        }

        public static ObjectPair FindObjectAt( Vector position )
        {
            return Objects.FindLast( pair => pair.GameObject.IsInside( position ) );

            /*foreach ( var pair in Objects )
            {
                var o = pair.GameObject;

                if ( ( o.Left < position.X ) && ( position.X < o.Right ) && ( o.Bottom < position.Y ) && ( position.Y < o.Top ) )
                {
                    return pair;
                }
            }

            return null;*/
        }

        public static GameObject InsertObject( Vector position, Template template )
        {
            LevelObject levelObj = LevelEditor.LevelData.CreateObject( template );
            return InsertObject( position, levelObj );
        }

        private static GameObject InsertObject( Vector position, LevelObject levelObj )
        {
            levelObj.OverridingProperties.SetValue( "Position", position );

            GameObject gameObj = levelObj.ConstructObject();
            gameObj.Position = position;
            Game.Instance.Add( gameObj );

            Objects.Add( new ObjectPair( gameObj, levelObj ) );
            return gameObj;
        }

        public static void RemoveObject( ObjectPair pair )
        {
            Game.Instance.Remove( pair.GameObject );
            LevelData.RemoveObject( pair.LevelObject );
            Objects.Remove( pair );
        }

        public static void RemoveObjects( List<ObjectPair> objs )
        {
            foreach ( var obj in objs )
                RemoveObject( obj );
        }

        public static void RefreshObjectType( ObjectPair obj )
        {
            GameObject newGameObject = obj.LevelObject.ConstructObject();
            RemoveObject( obj );
            obj.GameObject = InsertObject( obj.GameObject.Position, obj.LevelObject );
        }

        public static void UpdateObjects( Template template, PropertySet changedProperties )
        {
            bool typeChanged = changedProperties.Contains( "Type" );

            for ( int i = 0; i < Objects.Count; i++ )
            {
                LevelObject levelObj = Objects[i].LevelObject;

                if ( levelObj.Template != template )
                    continue;

                if ( typeChanged )
                    RefreshObjectType( Objects[i] );

                GameObject gameObj = Objects[i].GameObject;
                changedProperties.Apply( ref gameObj );
                levelObj.OverridingProperties.Apply( ref gameObj );
            }
        }

        private void StartToInsertObjectsOfSelectedTemplate()
        {
            ToolMenu.Select( Tool.Insert );
        }

        private void InitLoadLevel()
        {
            bool levelLoaded = false;

            if (currentLevelFile != null)
            {
                DataStorage.CurrentDirectory = Directory.GetCurrentDirectory();

                if (File.Exists(currentLevelFile))
                {
                    Load(currentLevelFile);
                    levelLoaded = true;
                }
            }

            if ( !levelLoaded )
            {
                TemplateMenu.AddTemplate( new Template() );
            }
        }

        private void InitializeLevelToDefaults()
        {
            LevelData = new LevelData();
            Level.Width = 100;
            Level.Height = 100;
            Camera.ZoomFactor = 10;
        }

        private void SetMouseControls()
        {
            Mouse.IsCursorVisible = true;
           
            Mouse.ListenMovement( -1, AnalogMoveCamera, "Move around the level" );
            Mouse.ListenWheel( delegate() { ZoomCamera( Mouse.WheelChange ); }, "Zoom level" );
        }

        private void SetKeyboardControls()
        {
            Keyboard.Listen( Key.F1, ButtonState.Pressed, ShowControlHelp, "Control help" );
            Keyboard.Listen( Key.S, ButtonState.Pressed, Save, "Save level file / Shift: Save As" );
            Keyboard.Listen( Key.O, ButtonState.Pressed, Open, "Open level file" );
            Keyboard.Listen( Key.G, ButtonState.Pressed, ToggleGrid, "Toggle Grid" );

            Keyboard.Listen( Key.Add, ButtonState.Pressed, AddGridSize, "Add Grid size by one", 1.0 );
            Keyboard.Listen( Key.Subtract, ButtonState.Pressed, AddGridSize, "Reduce Grid size by one", -1.0 );
            Keyboard.Listen( Key.Multiply, ButtonState.Pressed, MultiplyGridSize, "Double Grid size", 2.0 );
            Keyboard.Listen( Key.Divide, ButtonState.Pressed, MultiplyGridSize, "Halve Grid size", 0.5 );

            Keyboard.Listen( Key.Left, ButtonState.Down, MoveCamera, "Move camera left", -1.0, 0.0 );
            Keyboard.Listen( Key.Right, ButtonState.Down, MoveCamera, "Move camera right", 1.0, 0.0 );
            Keyboard.Listen( Key.Up, ButtonState.Down, MoveCamera, "Move camera up", 0.0, 1.0 );
            Keyboard.Listen( Key.Down, ButtonState.Down, MoveCamera, "Move camera down", 0.0, -1.0 );
            Keyboard.Listen( Key.Home, ButtonState.Pressed, ResetCamera, "Reset camera position" );
            Keyboard.Listen( Key.PageUp, ButtonState.Pressed, ZoomCamera, "Zoom out", -1.0 );
            Keyboard.Listen( Key.PageDown, ButtonState.Pressed, ZoomCamera, "Zoom in", 1.0 );
        }

        void MoveCamera( double x, double y )
        {
            Camera.Position.X += x * CameraSpeed / Camera.ZoomFactor;
            Camera.Position.Y += y * CameraSpeed / Camera.ZoomFactor;
            ToolMenu.SelectedTool.UpdatePositions();
            UpdateLabels();
        }

        void AnalogMoveCamera( AnalogState state )
        {
            bool scrollH = Mouse.PositionOnScreen.X < Screen.Left + MoveMargin || Mouse.PositionOnScreen.X > Screen.Right - MoveMargin;
            bool scrollV = Mouse.PositionOnScreen.Y < Screen.Bottom + MoveMargin || Mouse.PositionOnScreen.Y > Screen.Top - MoveMargin;

            if ( !scrollH && !scrollV )
                return;
            if ( scrollH )
                Camera.Position.X += Math.Sign( Mouse.PositionOnScreen.X ) * CameraSpeed / Camera.ZoomFactor;
            if ( scrollV )
                Camera.Position.Y += Math.Sign( Mouse.PositionOnScreen.Y ) * CameraSpeed / Camera.ZoomFactor;

            ToolMenu.SelectedTool.UpdatePositions();
            UpdateLabels();
        }

        void ResetCamera()
        {
            Camera.Position = Vector.Zero;
            ToolMenu.SelectedTool.UpdatePositions();
            UpdateLabels();
        }

        void ZoomCamera( double zoomChange )
        {
            double newZoomFactor = Camera.ZoomFactor + zoomChange;
            if ( newZoomFactor <= 0 || newZoomFactor >= 100 )
                return;

            Camera.ZoomFactor = newZoomFactor;
            ToolMenu.SelectedTool.UpdatePositions();
            UpdateLabels();
        }

        void ToggleGrid()
        {
            var layer = GetLayer( 0 );
            layer.Grid = ( layer.Grid == null ) ? Grid : null;
        }

        void AddGridSize( double delta )
        {
            Grid.CellSize += new Vector( delta, delta );
            UpdateLabels();
        }

        void MultiplyGridSize( double factor )
        {
            Grid.CellSize *= factor;
            UpdateLabels();
        }

        void Save()
        {
            if ( currentLevelFile == null || Keyboard.IsShiftDown() )
            {
                var fileWindow = new FileWindow( "Save level", Directory.GetCurrentDirectory(), false );
                //fileWindow
                Add( fileWindow );
                fileWindow.PathEntered += delegate( string path ) { currentLevelFile = path; DoSave(); };
            }
            else
                DoSave();
        }

        void DoSave()
        {
            DataStorage.Save<LevelData>( LevelData, currentLevelFile );
            ShowMessage( "Saved " + currentLevelFile );
        }

        void Open()
        {
            var fileWindow = new FileWindow( "Open level", Directory.GetCurrentDirectory(), true );
            Add( fileWindow );
            fileWindow.PathEntered += Load;
        }

        void Load( string path )
        {
            LevelData.Clear();
            ClearGameObjects();
            TemplateMenu.Clear();

            //try
            {
                LevelData = new LevelData();
                Factory.FactoryMethod createObject = delegate { return new LevelObject( LevelData, 0 ); };
                Game.Instance.AddFactory<LevelObject>( null, createObject );
                LevelData = DataStorage.Load<LevelData>( LevelData, path );
                Game.Instance.RemoveFactory<LevelObject>( null, createObject );

                foreach ( Template template in LevelData.Templates )
                {
                    TemplateMenu.AddTemplate( template );
                }

                foreach ( LevelObject levelObj in LevelData.Objects )
                {
                    GameObject gameObj = levelObj.ConstructObject();
                    Game.Instance.Add( gameObj );
                    Objects.Add( new ObjectPair( gameObj, levelObj ) );
                }
            }
            /*catch ( Exception e )
            {
                ShowError( "Error while reading file: " + e.Message );
            }*/

            ShowMessage( "Loaded " + path );
        }

        void ShowError( string text )
        {
            MessageWindow w = new MessageWindow( text ) { Color = Color.DarkRed };
            w.OKButton.Color = Color.DarkViolet;
            Add( w );
        }

        public static void ShowMessage( string message )
        {
            Instance.MessageDisplay.Add( message );
        }
    }
}
