﻿using System;
using Jypeli;
using Jypeli.LevelEditor;
using Jypeli.Widgets;

namespace LevelEditor.Menus
{
    public class PropertyMenu : Window
    {
        static readonly Type[] ObjectTypes =
        {
            typeof(GameObject), typeof(PhysicsObject),
            typeof(PlatformCharacter), typeof(PlatformCharacter2)
        };

        PushButton typeButton;
        PushButton colorButton;
        PushButton shapeButton;
        PushButton tagButton;
        PushButton deleteButton;
        Widget shapeModel;
        PropertySet ChangedProperties = new PropertySet();

        public event Action<PropertySet> Finished;
        public event Action Deleted;

        public PropertyMenu( PropertySet properties, bool showDelete )
        {
            Layout = new VerticalLayout();

            typeButton = CreatePropertyRow( "Type", ShowTypeMenu );
            if ( properties.Contains( "Type" ) )
                typeButton.Text = ( (Type)properties.GetValue( "Type" ) ).Name;

            colorButton = CreatePropertyRow( "Color", ShowColorMenu );
            if ( properties.Contains( "Color" ) )
                colorButton.Color = (Color)properties.GetValue( "Color" );

            shapeButton = CreatePropertyRow( "Shape", ShowShapeMenu );
            shapeModel = new Widget( 40, 40 ) { Color = Color.DarkBlue };
            shapeButton.Add( shapeModel );
            if ( properties.Contains( "ShapeString" ) )
            {
                string shapeStr = (string)properties.GetValue( "ShapeString" );
                shapeModel.Shape = Shape.FromString( shapeStr );
            }

            tagButton = CreatePropertyRow( "Tag", ShowTagMenu );
            if ( properties.Contains( "Tag" ) )
                tagButton.Text = (string)properties.GetValue( "Tag" );

            if ( showDelete )
            {
                deleteButton = new PushButton( 200, 50, "Delete" );
                deleteButton.PreferredSize = new Vector( 200, 50 );
                deleteButton.Color = Color.DarkRed;
                deleteButton.TextColor = Color.White;
                deleteButton.Clicked += DeleteObject;
                Add( deleteButton );
            }

            AddedToGame += AddControls;
        }

        PushButton CreatePropertyRow( string propName, Action action )
        {
            PushButton button = new PushButton( 100, 50 );
            button.Clicked += action;
            CreateRow( propName, button );
            return button;
        }

        void AddControls()
        {
            Game.Keyboard.Listen( Key.Escape, ButtonState.Pressed, Close, null ).InContext( this );
            Game.Mouse.Listen( MouseButton.Left, ButtonState.Pressed, Close, null ).InContext( this );
        }

        void Finish()
        {
            if ( Finished != null )
                Finished( ChangedProperties );
            Close();
        }

        Widget CreateRow( string text, Widget w )
        {
            var row = new Widget( new HorizontalLayout() );
            row.Add( new Label( text ) { Color = Color.Transparent, HorizontalAlignment = HorizontalAlignment.Right, XMargin = 15, YMargin = 5 } );
            row.Add( w );
            Add( row );
            return row;
        }

        void ChangeColor( Color color )
        {
            colorButton.Color = color;
            ChangedProperties.SetValue( "Color", color );
            Finish();
        }

        void ChangeShape( string shapeName )
        {
            shapeModel.Shape = Shape.FromString( shapeName );
            ChangedProperties.SetValue( "ShapeString", shapeName );
            Finish();
        }

        void ChangeTag( InputWindow tagWindow )
        {
            string tag = tagWindow.InputBox.Text;
            tagButton.Text = tag;
            ChangedProperties.SetValue( "Tag", tag );
            Finish();
        }

        void ChangeType( int typeIndex )
        {
            Type type = ObjectTypes[typeIndex];
            typeButton.Text = type.Name;
            ChangedProperties.SetValue( "Type", type );
            Finish();
        }

        void DeleteObject()
        {
            if ( Deleted != null )
                Deleted();
            Close();
        }

        void ShowShapeMenu()
        {
            var shapeMenu = new ShapeMenu();
            shapeMenu.ShapeSelected += ChangeShape;
            Game.Add( shapeMenu );
        }

        void ShowColorMenu()
        {
            var colorMenu = new ColorMenu();
            colorMenu.ColorSelected += ChangeColor;
            Game.Add( colorMenu );
        }

        void ShowTagMenu()
        {
            var tagInput = new InputWindow( "Enter a tag" );
            tagInput.TextEntered += ChangeTag;
            Game.Add( tagInput );
        }

        void ShowTypeMenu()
        {
            string[] typeNames = Array.ConvertAll<Type, string>( ObjectTypes, delegate( Type type ) { return type.Name; } );
            var typeMenu = new MultiSelectWindow( "Select type for the object", typeNames );
            typeMenu.ItemSelected += ChangeType;
            Game.Add( typeMenu );
        }
    }
}
