﻿using System;
using System.Collections.Generic;
using Jypeli;
using Jypeli.LevelEditor;
using Jypeli.Widgets;

namespace LevelEditor.Menus
{
    public class TemplateMenu : Widget
    {
        class MenuLabel : Label
        {
            public static readonly Color SelectedColor = Color.Green;
            public static readonly Color DefaultColor = Color.Gray;

            public MenuLabel( string text )
                : base( text )
            {
                Color = MenuLabel.DefaultColor;
                BorderColor = Color.Black;
            }

            public void SetSelected()
            {
                Color = MenuLabel.SelectedColor;
            }

            public void Unselect()
            {
                Color = MenuLabel.DefaultColor;
            }
        }

        class TemplateLabel : MenuLabel
        {
            public Template Template;

            public TemplateLabel( Template t )
                : base( t.Name )
            {
                this.Template = t;
            }
        }


        private List<MenuLabel> widgets = new List<MenuLabel>();
        private LevelData data;

        public Template SelectedTemplate = null;

        public event Action<Template> TemplateAdded;
        public event Action<Template> TemplateRemoving;
        public event Action<Template> TemplateRemoved;
        public event Action TemplateSelected;

        private void OnTemplateAdded( Template template )
        {
            if ( TemplateAdded != null )
                TemplateAdded( template );
        }

        private void OnTemplateRemoving( Template template )
        {
            if ( TemplateRemoving != null )
                TemplateRemoving( template );
        }

        private void OnTemplateRemoved( Template template )
        {
            if ( TemplateRemoved != null )
                TemplateRemoved( template );
        }

        private void OnTemplateSelected()
        {
            if ( TemplateSelected != null )
                TemplateSelected();
        }

        public TemplateMenu( double width, double height, LevelData data )
            : base( width, height )
        {
            this.data = data;
            IsModal = false;

            Layout = new VerticalLayout
            {
                LeftPadding = 25,
                RightPadding = 25,
                TopPadding = 25,
                BottomPadding = 25,
                Spacing = 15,
            };

            DoAdd( new MenuLabel( "<add>" ) );
            AddedToGame += AddControls;
        }

        public void AddTemplate( Template template )
        {
            TemplateLabel newLabel = new TemplateLabel( template );
            MenuLabel addLabel = DoRemoveLast();
            DoAdd(newLabel);

            // TODO: Widget: better synchronicity for adding and removing widgets
            Timer.SingleShot(0, delegate { DoAdd(addLabel); });
        }

        private void RemoveTemplate( TemplateLabel label )
        {
            if ( SelectedTemplate == label.Template )
                SelectedTemplate = null;

            Remove( label );
            widgets.Remove( label );
            OnTemplateRemoved( label.Template );
        }

        public void RemoveTemplate( Template template )
        {
            for ( int i = 0; i < widgets.Count; i++ )
            {
                TemplateLabel label = widgets[i] as TemplateLabel;
                if ( label != null && label.Template == template )
                {
                    RemoveTemplate( label );
                    return;
                }
            }
        }

        void DoAdd( MenuLabel w )
        {
            Add( w );
            widgets.Add( w );
        }

        MenuLabel DoRemoveLast()
        {
            if ( widgets.Count == 0 )
                return null;

            int lastIndex = widgets.Count - 1;
            MenuLabel lastTemplate = widgets[lastIndex];
            Remove( lastTemplate );
            widgets.RemoveAt( lastIndex );
            return lastTemplate;
        }

        void AddControls()
        {
            Game.Mouse.ListenOn( this, MouseButton.Left, ButtonState.Pressed, SelectTemplateThatWasClicked, null ).InContext( this );
            Game.Mouse.ListenOn( this, MouseButton.Right, ButtonState.Pressed, OpenPropertyMenu, null ).InContext( this );
        }

        void OpenPropertyMenu()
        {
            TemplateLabel w = FindWidgetAt( Game.Mouse.PositionOnScreen ) as TemplateLabel;

            if ( w != null )
            {
                bool canDelete = ( widgets.IndexOf( w ) > 0 );
                PropertyMenu menu = new PropertyMenu( w.Template.Properties, canDelete );
                menu.Finished += delegate( PropertySet p ) { SetProperties( w.Template, p ); };
                //menu.Deleted += delegate { RemoveTemplate( w.Template ); };
                menu.Deleted += delegate { OnTemplateRemoving( w.Template ); };
                Game.Add( menu );

            }
        }

        void SetProperties( Template template, PropertySet changedProperties )
        {
            template.Properties.Add( changedProperties );
            LevelEditor.UpdateObjects( template, changedProperties );
        }

        void SelectTemplateThatWasClicked()
        {
            MenuLabel w = FindWidgetAt( Game.Mouse.PositionOnScreen );

            if ( w != null )
            {
                if ( w is TemplateLabel )
                    ChangeSelection( (TemplateLabel)w );
                else
                {
                    // Add a new template
                    InputWindow nameWindow = new InputWindow( "Enter a name for new template" );
                    nameWindow.TextEntered += new InputWindow.InputWindowHandler( WindowAddTemplate );
                    Game.Instance.Add( nameWindow );
                }
            }
        }

        void WindowAddTemplate( InputWindow sender )
        {
            if ( sender.InputBox.Text.Length == 0 )
                return;

            Template newTemplate = new Template( sender.InputBox.Text );
            AddTemplate( newTemplate );
        }

        MenuLabel FindWidgetAt( Vector position )
        {
            return widgets.Find( w => w.IsInside( position ) );
        }

        void ChangeSelection( TemplateLabel selected )
        {
            foreach ( var w in widgets )
                w.Unselect();
            selected.SetSelected();

            SelectedTemplate = selected.Template;
            OnTemplateSelected();
        }

        public override void Clear()
        {
            List<MenuLabel> remove = new List<MenuLabel>();

            foreach ( MenuLabel label in widgets )
            {
                if ( label is TemplateLabel )
                    remove.Add(label);
            }

            foreach ( MenuLabel label in remove )
            {
                Remove( label );
                widgets.Remove( label );
            }
        }
    }
}
