﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Jypeli;
using Jypeli.LevelEditor;
using Jypeli.Widgets;
using LevelEditor.Menus;

namespace LevelEditor.Tools
{
    public sealed class Select : Tool
    {
        public ObjectPair SelectedObject { get; private set; }

        List<Handle> activeHandles = new List<Handle>();
        Handle draggedHandle = null;

        Vector mousecursorDragOffset = Vector.Zero;
        bool draggingGameObject = false;

        public Select()
            : base()
        {
            Deactivated += RemoveHandles;
        }

        protected override void AddControls()
        {
            Mouse.Listen( MouseButton.Left, ButtonState.Pressed, MousePressed, "Move object" ).InContext( this );
            Mouse.Listen( MouseButton.Left, ButtonState.Down, Drag, null ).InContext( this );
            Mouse.Listen( MouseButton.Left, ButtonState.Released, EndDrag, null ).InContext( this );
            Mouse.Listen( MouseButton.Right, ButtonState.Pressed, ShowProperties, "Show object properties" ).InContext( this );
            Keyboard.Listen( Key.Delete, ButtonState.Pressed, DeleteSelected, "Delete selected object" ).InContext( this );
        }

        public override void UpdatePositions()
        {
            UpdateHandles();
            base.UpdatePositions();
        }

        private void MousePressed()
        {
            ObjectPair o = LevelEditor.FindObjectAt( Mouse.PositionOnWorld );
            
            if ( o != null )
            {
                SelectObject( o );
                StartDrag();
            }
            else
                Deselect();
        }

        private void ShowProperties()
        {
            ObjectPair pair = LevelEditor.FindObjectAt( Mouse.PositionOnWorld );
            if ( pair == null ) return;

            GameObject gameObj = pair.GameObject;
            LevelObject levelObj = pair.LevelObject;

            SelectObject( pair );

            var actualProperties = PropertySet.Merge( levelObj.Template.Properties, levelObj.OverridingProperties );
            var propertyMenu = new PropertyMenu( actualProperties, false );
            propertyMenu.Finished += delegate( PropertySet changedProperties )
            {
                levelObj.OverridingProperties.Add( changedProperties );

                if ( changedProperties.Contains( "Type" ) )
                    LevelEditor.RefreshObjectType( pair );
                else
                    levelObj.OverridingProperties.Apply( ref gameObj );
            };
            Game.Instance.Add( propertyMenu );
            MoveNearObject( propertyMenu, gameObj );
        }

        private void StartDraggingHandle( Handle handle )
        {
            draggedHandle = handle;
            mousecursorDragOffset = handle.Position - Mouse.PositionOnScreen;
        }

        private void StartDrag()
        {
            draggingGameObject = true;
            mousecursorDragOffset = SelectedObject.GameObject.Position - Mouse.PositionOnWorld;
        }

        private void Drag()
        {
            if ( SelectedObject == null )
                return;

            if ( draggedHandle != null )
                draggedHandle.OnDrag();

            else if ( draggingGameObject )
                DragObject();
        }

        void DragObject()
        {
            Vector p = mousecursorDragOffset + Mouse.PositionOnWorld;

            if ( Keyboard.IsCtrlDown() )
            {
                Grid grid = LevelEditor.Grid;
                p = Coordinates.FromVector( p, grid ).ToVector( grid );
            }

            SelectedObject.GameObject.Position = p;
        }

        private void EndDrag()
        {
            if ( draggedHandle != null )
                draggedHandle.OnEndDrag();

            if ( draggingGameObject )
                SelectedObject.LevelObject.OverridingProperties.SetValue( "Position", SelectedObject.GameObject.Position );

            draggingGameObject = false;
            draggedHandle = null;

            UpdateHandles();
        }

        private void UpdateHandles()
        {
            if ( SelectedObject == null )
                return;

            foreach ( Handle handle in activeHandles )
            {
                handle.Update();
            }
        }

        private void SelectObject( ObjectPair o )
        {
            if ( SelectedObject == o )
                return;

            SelectedObject = o;
            AddHandles();
            UpdateHandles();
        }

        private void Deselect()
        {
            SelectedObject = null;
            draggedHandle = null;
            draggingGameObject = false;
            RemoveHandles();
        }

        private void DeleteSelected()
        {
            if ( SelectedObject == null )
                return;

            LevelEditor.RemoveObject( SelectedObject );
            Deselect();
        }

        private void AddHandles()
        {
            if ( activeHandles.Count > 0 || SelectedObject == null )
                return;

            activeHandles.Add( new HorizontalSizeHandle( SelectedObject ) );
            activeHandles.Add( new VerticalSizeHandle( SelectedObject ) );
            activeHandles.Add( new AngleHandle( SelectedObject ) );

            foreach ( Handle handle in activeHandles )
            {
                Game.Instance.Add( handle );
                Mouse.ListenOn( handle, MouseButton.Left, ButtonState.Pressed, StartDraggingHandle, null, handle ).InContext( this );
            }
        }

        private void RemoveHandles()
        {
            foreach ( Handle handle in activeHandles )
            {
                handle.Destroy();
            }

            activeHandles.Clear();
        }

        private void MoveNearObject( Widget w, GameObject o )
        {
            Vector p = Camera.WorldToScreen( o.Position );
            if ( ( p.X - w.Width / 2 ) < Screen.Left )
            {
                p.X = Screen.Left + w.Width / 2;
            }
            else if ( Screen.Right < ( p.X + w.Width / 2 ) )
            {
                p.X = Screen.Right - w.Width / 2;
            }
            if ( ( p.Y - w.Height / 2 ) < Screen.Bottom )
            {
                p.Y = Screen.Bottom + w.Height / 2;
            }
            else if ( Screen.Top < p.Y + w.Height / 2 )
            {
                p.Y = Screen.Top - w.Height / 2;
            }
            w.Position = p;
        }
    }
}
