using System;
using System.Collections.Generic;
using Jypeli.Controls;

#if WINDOWS_PHONE
using System.Device.Location;
#endif

namespace Jypeli.WP7
{
    /// <summary>
    /// GPS-luokka.
    /// Authors: Jaakko Kosonen, Tomi Karppinen
    /// </summary>
    public class GPS : IDisposable
    {
#if WINDOWS_PHONE
        GeoCoordinateWatcher sensor;
#endif
        private bool sensorStarted;
        private List<GPSData> GPSDataCollection;

        /// <summary>
        /// Asettaa metrimrn jonka knnykn pit liikkua ennenkuin liikkumisesta tulee tapahtuma.
        /// </summary>
        public float MovementThreshold
        {
            set
            {
#if WINDOWS_PHONE
                if (sensor != null)
                {
                    sensor.MovementThreshold = value;
                }
#endif
            }
        }

        /// <summary>
        /// Kuvaus sensorin tilasta.
        /// </summary>
        public string StatusText { get; private set; }

        /// <summary>
        /// Leveyspiiri.
        /// </summary>
        public GPSData Latitude { get; private set; }

        /// <summary>
        /// Pituuspiiri.
        /// </summary>
        public GPSData Longitude { get; private set; }

        /// <summary>
        /// Nopeus m/s.
        /// </summary>
        public GPSData Speed { get; private set; }

        /// <summary>
        /// Nopeus km/h.
        /// </summary>
        public GPSData SpeedKMH { get; private set; }

        /// <summary>
        /// Nopeus maili/h.
        /// </summary>
        public GPSData SpeedMPH { get; private set; }

        /// <summary>
        /// Suuntakulma, 0 = pohjoinen.
        /// </summary>
        public GPSData Course { get; private set; }

        /// <summary>
        /// Korkeus merenpinnasta metrein.
        /// </summary>
        public GPSData Altitude { get; private set; }

        /// <summary>
        /// Tapahtuma kun GPS-sensorilla on uutta dataa.
        /// </summary>
        public event Handler GpsDataAvailable;

        /// <summary>
        /// Tapahtuma kun GPS-sensorin tila muuttuu.
        /// </summary>
        public event Handler<string> GpsStatusChanged;

        /// <summary>
        /// Konstruktori GPS:lle.
        /// </summary>
        public GPS()
        {
#if WINDOWS_PHONE
            sensor = new GeoCoordinateWatcher(GeoPositionAccuracy.High);
            MovementThreshold = 1f;
            sensor.StatusChanged += new EventHandler<GeoPositionStatusChangedEventArgs>(statusChanged);
            sensor.PositionChanged += new EventHandler<GeoPositionChangedEventArgs<GeoCoordinate>>(positionChanged);
#endif

            GPSDataCollection = new List<GPSData>(5);
            Latitude = new GPSData();
            GPSDataCollection.Add(Latitude);
            Longitude = new GPSData();
            GPSDataCollection.Add(Longitude);
            Speed = new GPSData();
            GPSDataCollection.Add(Speed);
            SpeedKMH = new GPSData();
            GPSDataCollection.Add(SpeedKMH);
            SpeedMPH = new GPSData();
            GPSDataCollection.Add(SpeedMPH);
            Course = new GPSData();
            GPSDataCollection.Add(Course);
            Altitude = new GPSData();
            GPSDataCollection.Add(Altitude);

            StatusText = "GPS not started.";         
        }

        public void Dispose()
        {
            Stop();
        }

        /// <summary>
        /// Kynnist GPS-sensorin.
        /// </summary>
        public void Start()
        {
#if WINDOWS_PHONE
            if (!sensorStarted)
            {
                sensor.Start();
                sensorStarted = true;
                ChangeStatusText("GPS started.");
            }
#endif
        }

        /// <summary>
        /// Pysytt GPS-sensorin.
        /// </summary>
        public void Stop()
        {
#if WINDOWS_PHONE
            if (sensorStarted)
            {
                sensor.Stop();
                sensorStarted = false;
                ChangeStatusText("GPS stopped.");
            }
#endif
        }

        /// <summary>
        /// Nollaa kaikki saatavilla olevat GPS-tiedot.
        /// </summary>
        /// <param name="resetLatestValues">Jos tosi, mys viimeisimmt arvot (Value) nollataan.</param>
        public void ResetAll(bool resetLatestValues)
        {
#if WINDOWS_PHONE
            foreach (GPSData d in GPSDataCollection)
            {
                d.Reset(resetLatestValues);
            }
            if (GpsDataAvailable != null)
            {
                GpsDataAvailable();
            }
#endif
        }

        private void ChangeStatusText(string status)
        {
            StatusText = status;
            if (GpsStatusChanged != null)
            {
                GpsStatusChanged(StatusText);
            }
        }

#if WINDOWS_PHONE
        void statusChanged(object sender, GeoPositionStatusChangedEventArgs e)
        {
            StatusText = e.Status.ToString();
            if (GpsStatusChanged != null)
            {
                GpsStatusChanged(StatusText);
            }
        }

        void positionChanged(object sender, GeoPositionChangedEventArgs<GeoCoordinate> e)
        {                        
            Latitude.Update(e.Position.Location.Latitude);
            
            Longitude.Update(e.Position.Location.Longitude);
            
            double speedMS = e.Position.Location.Speed;
            Speed.Update(speedMS);
            SpeedKMH.Update(speedMS * 3.6);
            SpeedMPH.Update(speedMS * 2.236936);
            
            Course.Update(e.Position.Location.Course);
            
            Altitude.Update(e.Position.Location.Altitude);
                                   
            if (GpsDataAvailable != null)
            {
                GpsDataAvailable();
            }
        }
#endif        
    }
}
