using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Jypeli.Controls;

namespace Jypeli
{
    /// <summary>
    /// Luokka GPS-datalle.
    /// Authors: Jaakko Kosonen
    /// </summary>
    public class GPSData
    {
        private List<Handler<GPSData>> handlers;
        private double value;
        private double sampleCount;
        private double sum;
        private double max;
        private double min;

        /// <summary>
        /// Viimeisin arvo, tai Double.NaN jos GPS-sensorilta ei ole saatu arvoa kertaakaan.
        /// </summary>
        public double Value { get { return value; } }

        /// <summary>
        /// Keskiarvo, tai Double.NaN jos GPS-sensorilta ei ole saatu arvoa kertaakaan.
        /// </summary>
        public double Average { get { return sampleCount == 0 ? Value : sum / sampleCount; } }

        /// <summary>
        /// Maksimiarvo, tai Double.NaN jos ei ole saatu arvoa kertaakaan.
        /// </summary>
        public double Max { get { return max; } }

        /// <summary>
        /// Minimiarvo, tai Double.NaN jos ei ole saatu arvoa kertaakaan.
        /// </summary>
        public double Min { get { return min; } }

        /// <summary>
        /// Konstruktori.
        /// </summary>
        public GPSData()
        {
            handlers = new List<Handler<GPSData>>();
            Reset( true );
        }

        /// <summary>
        /// Pivitt datan.
        /// </summary>
        /// <param name="newValue"></param>
        public void Update( double newValue )
        {
            if ( !Double.IsNaN( newValue ) )
            {
                value = newValue;
                sum += value;
                min = Double.IsNaN( min ) ? newValue : Math.Min( min, newValue );
                max = Double.IsNaN( max ) ? newValue : Math.Max( max, newValue );
                sampleCount++;
                foreach ( Handler<GPSData> h in handlers )
                {
                    h.Invoke( this );
                }
            }
        }

        /// <summary>
        /// Lis tapahtumanksittelijn jota kutsutaan kun GPS-tieto pivitetn.
        /// </summary>
        /// <param name="handler">Tapahtumanksittelij</param>
        public void Listen( Handler<GPSData> handler )
        {
            handlers.Add( handler );
        }

        /// <summary>
        /// Nollaa keski-, minimi- ja maksimiarvot.
        /// </summary>
        /// <param name="resetAll">Jos tosi, mys viimeisin arvo (Value) asetetaan Double.NaN:ksi</param>
        public void Reset( bool resetAll )
        {
            if ( resetAll )
            {
                value = Double.NaN;
            }
            sum = 0;
            sampleCount = 0;
            min = Double.NaN;
            max = Double.NaN;
        }

        /// <summary>
        /// Kertoo onko arvo (Value) validi, ts. onko GPS-sensorlita saatu kunnollista dataa.
        /// </summary>
        /// <returns>Totta jos arvo on validi</returns>
        public bool IsDataAvailable()
        {
            return !Double.IsNaN( value );
        }

        /// <summary>
        /// Merkkijono, jossa lueteltuna kaikki arvot.
        /// </summary>
        /// <returns>Arvot merkkijonossa</returns>
        public override string ToString()
        {
            return "Value: " + String.Format( "{0:0.00}", Value ) + "; " +
                   "Avg: " + String.Format( "{0:0.00}", Average ) + "; " +
                   "Min: " + String.Format( "{0:0.00}", Min ) + "; " +
                   "Max: " + String.Format( "{0:0.00}", Max );
        }
    }
}
