using Microsoft.Xna.Framework;

/*
 * Authors: Jaakko Kosonen 
 */

namespace ParticleComponent
{
    /// <summary>
    /// A single particle for a particle effect.
    /// </summary>
    class Particle
    {
        private float rotationSpeed;
        private Vector2 acceleration;
        private Vector2 velocity;
        
        private Vector2 coordinates;
        /// <summary>
        /// X-coordinate of particle.
        /// </summary>
        public float X
        {
            get { return coordinates.X; }
        }
        /// <summary>
        /// Y-coordinate of particle.
        /// </summary>
        public float Y
        {
            get { return coordinates.Y; }
        }

        private float rotation;
        /// <summary>
        /// Rotation of particle.
        /// </summary>
        public float Rotation
        {
            get { return rotation; }
        }

        private float scale;
        /// <summary>
        /// Scale of particle.
        /// </summary>
        public float Scale
        {
            get { return scale; }
        }

        private float timeSinceActivated;
        /// <summary>
        /// Time since particle was initialized.
        /// </summary>
        public float TimeSinceActivated
        {
            get { return timeSinceActivated; }
        }

        private float lifeTime;
        /// <summary>
        /// Time to live of particle.
        /// </summary>
        public float LifeTime
        {
            get { return lifeTime; }
        }

        private Color color;
        /// <summary>
        /// Colorization of particle.
        /// </summary>
        public Color Color
        {
            get { return color; }
        }

        /// <summary>
        /// True if particle has time to live left.
        /// </summary>
        public bool Active
        {
            get { return timeSinceActivated < lifeTime; }
        }

        /// <summary>
        /// Initializes the particle.
        /// </summary>
        /// <param name="position">Initial position</param>
        /// <param name="velocity">Initial velocity</param>
        /// <param name="acceleration">Acceleration</param>
        /// <param name="timeToLive">Time to live</param>
        /// <param name="rotation">Initial rotation</param>
        /// <param name="rotationSpeed">Rotation speed</param>
        /// <param name="scale">Scale</param>
        /// <param name="color">Colorization</param>
        public void Initialize(Vector2 position, Vector2 velocity, Vector2 acceleration, float timeToLive, float rotation, float rotationSpeed, float scale, Color color)
        {
            coordinates = position;
            timeSinceActivated = 0.0f;
            lifeTime = timeToLive;
            this.rotation = rotation;
            this.rotationSpeed = rotationSpeed;
            this.scale = scale;
            this.acceleration = acceleration;
            this.velocity = velocity;
            this.color = color;
        }

        /// <summary>
        /// Updates:
        /// - velocity
        /// - position
        /// - rotation
        /// </summary>
        /// <param name="diff">Elapsed gametime</param>
        public void Update(float diff)
        {
            velocity += acceleration * diff;
            coordinates += velocity * diff;
            timeSinceActivated += diff;
            rotation += diff * rotationSpeed;
        }
    }
}
