using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Input.Touch;
using Microsoft.Xna.Framework.Media;

/*
 * Authors: Jaakko Kosonen 
 */

namespace ParticleComponent
{
    /// <summary>
    /// Class for a particle effect.
    /// </summary>
    public class ParticleEffect : DrawableGameComponent
    {        
        private SpriteBatch spriteBatch;
        private Texture2D texture;
        private BlendState blendState;
        private Vector2 origin;        
        private Queue<Particle> inActiveParticles;
        private List<Particle> particles;
        private Random random;
        private EffectSettings settings;
        private string settingsFileName;
        private string textureFileName;

        /// <summary>
        /// Total amount of particles used by effect.
        /// </summary>
        public int TotalParticleCount
        {
            get { return particles.Count; }
        }

        /// <summary>
        /// Amount of inactive particles.
        /// </summary>
        public int FreeParticleCount
        {
            get { return inActiveParticles.Count; }
        }

        /// <summary>
        /// Effect settings.
        /// </summary>
        public EffectSettings Settings
        {
            get { return settings; }
        }

        /// <summary>
        /// Constructs a new effect and adds it to the Game.Component of given game.
        /// </summary>
        /// <param name="game">Game that owns the effect</param>
        /// <param name="textureFile">Texture for effect</param>
        /// <param name="settingsFile">Settings for effect</param>
        public ParticleEffect(Game game, string textureFile, string settingsFile) : base(game)
        {
            DrawOrder = 200;            
            textureFileName = textureFile;
            settingsFileName = settingsFile;
            int initialParticles = 10;
            random = new Random();
            inActiveParticles = new Queue<Particle>(initialParticles);
            particles = new List<Particle>(initialParticles);
            AddParticles(initialParticles);
            random = new Random();
            
            blendState = new BlendState
            {
                AlphaSourceBlend = Blend.SourceAlpha,
                ColorSourceBlend = Blend.SourceAlpha,
                AlphaDestinationBlend = Blend.One,
                ColorDestinationBlend = Blend.One
            };

            game.Components.Add(this);
        }        
                      
        /// <summary>
        /// Loads
        /// - texture
        /// - settings
        /// </summary>
        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
            texture = Game.Content.Load<Texture2D>(textureFileName);
            settings = Game.Content.Load<EffectSettings>(settingsFileName);
            
            origin.X = texture.Width / 2;
            origin.Y = texture.Height / 2;
                        
            base.LoadContent();
        }

        /// <summary>
        /// Updates all active particles.
        /// </summary>
        /// <param name="gameTime">Gametime</param>
        public override void Update(GameTime gameTime)
        {
            float diff = (float)gameTime.ElapsedGameTime.TotalSeconds;
            foreach (Particle p in particles)
            {
                if (p.Active)
                {
                    p.Update(diff);
                    if (!p.Active)
                    {
                        inActiveParticles.Enqueue(p);
                    }
                }
            }            
            base.Update(gameTime);
        }

        /// <summary>
        /// Draws all active particles.
        /// </summary>
        /// <param name="gameTime">Gametime</param>
        public override void Draw(GameTime gameTime)
        {
            spriteBatch.Begin(SpriteSortMode.Deferred, blendState);
            foreach (Particle p in particles)
            {
                if (p.Active)
                {
                    float opacity = 4.0f / 100.0f * settings.MaxOpacity;
                    float lifeLived = p.TimeSinceActivated / p.LifeTime;
                    float alpha = opacity * lifeLived * (1 - lifeLived);
                    float scale = p.Scale * (0.75f + 0.25f * lifeLived);
                    Color c = p.Color * alpha;                    
                    spriteBatch.Draw(texture, new Vector2(p.X, p.Y), null, c, p.Rotation, origin, scale, SpriteEffects.None, 0.0f);
                }
            }
            spriteBatch.End();
            base.Draw(gameTime);            
        }

        /// <summary>
        /// Initializes particles for a new effect.
        /// </summary>
        /// <param name="position">Position of new effect</param>
        public void CreateNewEffect(Vector2 position)
        {
            int numberOfParticlesInEffect = random.Next(settings.ParticlesPerEffect.MinIntValue, settings.ParticlesPerEffect.MaxIntValue);
            for (int i = 0; i < numberOfParticlesInEffect; i++)
            {
                if (inActiveParticles.Count == 0)
                {
                    AddParticles(10);
                }
                Particle p = inActiveParticles.Dequeue();
                InitializeParticle(p, position);
            }
        } 

        private void InitializeParticle(Particle p, Vector2 position)
        {
            float rotation = GetRandomFloat(0, 360);
            float rotationSpeed = settings.ParticleRotationSpeed.GetRandomValue(random);
            float scale = settings.ParticleSize.GetRandomValue(random);
            float speed = settings.InitialParticleSpeed.GetRandomValue(random);
            float accelerationScale = settings.ParticleAccelerationScale.GetRandomValue(random);
            float timeToLive = settings.ParticleLifeTime.GetRandomValue(random);

            Vector2 direction = GetRandomDirection();            
            Vector2 velocity = direction * speed;            
            Vector2 acceleration = direction * accelerationScale;

            p.Initialize(position, velocity, acceleration, timeToLive, rotation, rotationSpeed, scale, settings.Colorization);
        }
        private Vector2 GetRandomDirection()
        {
            float angle = settings.ParticleDirection.GetRandomValue(random);
            angle = MathHelper.Clamp(angle, 0, 360);
            angle = MathHelper.ToRadians(angle);
            return new Vector2((float)Math.Cos(angle), (float)Math.Sin(angle));
        }
        private void AddParticles(int count)
        {
            for (int i = 0; i < count; i++)
            {
                Particle p = new Particle();
                particles.Add(p);
                inActiveParticles.Enqueue(p);
            }
        }
        private float GetRandomFloat(float min, float max)
        {
            return min + (float)random.NextDouble() * (max - min);
        }
    }      
}
