using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Input.Touch;
using Microsoft.Xna.Framework.Media;
using ParticleComponent;

/*
 * Authors: Jaakko Kosonen 
 */

namespace WindowsPhoneGame
{    
    /// <summary>
    /// Pong game
    /// </summary>
    public class PongGame : Microsoft.Xna.Framework.Game
    {
        private enum GameState
        {
            WaitingForStart,
            Playing,
            Ended
        }

        #region fields

        private GameState state;
        private int lastStateChange;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;        

        Texture2D backGround;
        SpriteFont font;

        PongLevel level;
        PongBall ball;
        PongPaddle player2Paddle;
        PongPaddle player1Paddle;

        PongPlayer player1;
        PongPlayer player2;

        PongAI ai;
        PongPaddle aiPaddle;
        bool aiEnabled;

        #endregion

        /// <summary>
        /// Constructor.
        /// </summary>
        public PongGame()
        {            
            Content.RootDirectory = "Content";            
            TargetElapsedTime = TimeSpan.FromTicks(333333);
         
            graphics = new GraphicsDeviceManager(this);
            graphics.IsFullScreen = true;
            graphics.SupportedOrientations = DisplayOrientation.LandscapeLeft;
            graphics.ApplyChanges();

            CreateLevel();
            CreateBall();
            CreatePaddles();
            CreatePlayers();

            ai = new PongAI(9);
            aiEnabled = true;

            state = GameState.WaitingForStart;      
        }     

        /// <summary>
        /// Initialization.
        /// </summary>
        protected override void Initialize()
        {                       
            base.Initialize();
        }

        /// <summary>
        /// Loads content.
        /// </summary>
        protected override void LoadContent()
        {            
            spriteBatch = new SpriteBatch(GraphicsDevice);          

            backGround = Content.Load<Texture2D>("img/background");
            font = Content.Load<SpriteFont>("MediumFontBold");
            ball.Image = Content.Load<Texture2D>("img/ball");
            player2Paddle.Image = Content.Load< Texture2D >("img/paddleRed");
            player1Paddle.Image = Content.Load<Texture2D>("img/paddleBlue");
            SoundEffect collision = Content.Load<SoundEffect>("sound/collision");
            level.PaddleCollisionSound = collision;
            level.TopAndBottomLevelBorderSound = collision;
            level.RightAndLeftLevelBorderSound = Content.Load<SoundEffect>("sound/explosion");                        
        }

        /// <summary>
        /// Unload content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Updates pong-objects.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
            {
                this.Exit();
                Microsoft.Devices.VibrateController.Default.Stop();
            }
                                               
            HandleInput();            
            ball.Update();
            player1Paddle.Update();
            player2Paddle.Update();
            UpdateAI();
            player1.Update();
            player2.Update();            
            level.Update();
            UpdateTrailEffect();
            CheckIfGameEnded();            

            base.Update(gameTime);
        }       

        /// <summary>
        /// Draws the game.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.Black);            
            spriteBatch.Begin();

            spriteBatch.Draw(backGround, new Vector2(0, 0), Color.White);
            ball.DrawObject(spriteBatch);
            player2Paddle.DrawObject(spriteBatch);
            player1Paddle.DrawObject(spriteBatch);

            switch (state)
            {
                case GameState.WaitingForStart:
                    spriteBatch.DrawString(font, "Tap screen to start", new Vector2(180, 40), Color.Black, 0, Vector2.Zero, 2, SpriteEffects.None, 0);
                    spriteBatch.DrawString(font, "Tap screen to start", new Vector2(183, 43), Color.Yellow, 0, Vector2.Zero, 2, SpriteEffects.None, 0);
                    break;
                case GameState.Playing:
                    Vector2 offset = new Vector2(3, 3);
                    spriteBatch.DrawString(font, "" + player1.Health, level.Player1HealthPosition, Color.Black, 0, Vector2.Zero, 1.6f, SpriteEffects.None, 0);
                    spriteBatch.DrawString(font, "" + player1.Health, level.Player1HealthPosition + offset, Color.RoyalBlue, 0, Vector2.Zero, 1.6f, SpriteEffects.None, 0);
                    spriteBatch.DrawString(font, "" + player2.Health, level.Player2HealthPosition, Color.Black, 0, Vector2.Zero, 1.6f, SpriteEffects.None, 0);
                    spriteBatch.DrawString(font, "" + player2.Health, level.Player2HealthPosition + offset, Color.OrangeRed, 0, Vector2.Zero, 1.6f, SpriteEffects.None, 0);
                    break;
                case GameState.Ended:
                    bool player1Won = player2.State == PongPlayerState.Defeated;
                    string winner = player1Won ? "Blue wins!" : "Red wins!";
                    Color c = player1Won ? Color.RoyalBlue : Color.OrangeRed;
                    spriteBatch.DrawString(font, winner, level.ScreenCenter - new Vector2(120, 20), Color.Black, 0, Vector2.Zero, 2, SpriteEffects.None, 0);
                    spriteBatch.DrawString(font, winner, level.ScreenCenter - new Vector2(117, 17), c, 0, Vector2.Zero, 2, SpriteEffects.None, 0);
                    break;
                default:
                    break;
            }

            spriteBatch.End();
            base.Draw(gameTime);
        }

        private void CreatePlayers()
        {
            player1 = new PongPlayer();
            level.LeftBoundaryCollisioned += player1.HandleCollision;
            player2 = new PongPlayer();
            level.RightBoundaryCollisioned += player2.HandleCollision;
        }

        private void CreatePaddles()
        {
            Vector2 limits = new Vector2(level.TopLeftY + 3, level.ScreenHeight - 3);
            player2Paddle = new PongPaddle(level.ScreenWidth - 105, level.TopLeftY + 200, 25, 80, 20, limits);
            player1Paddle = new PongPaddle(level.TopLeftX + 80, level.TopLeftY + 200, 25, 80, 20, limits);
            level.AddCollidableObject(player2Paddle);
            level.AddCollidableObject(player1Paddle);
        }

        private void CreateBall()
        {
            ball = new PongBall(level.ScreenCenter, 40, 40, 15);
            ball.TrailEffect = new ParticleEffect(this, "img/pongSmoke", "settings/BallTrailEffectSettings");
            level.Ball = ball;
        }

        private void CreateLevel()
        {
            level = new PongLevel(
                GraphicsDevice.Viewport.X,
                GraphicsDevice.Viewport.Y,
                GraphicsDevice.Viewport.Width,
                GraphicsDevice.Viewport.Height
                );
            ParticleEffect sparks = new ParticleEffect(this, "img/spark", "settings/SparkEffectSettings");
            ParticleEffect explosion = new ParticleEffect(this, "img/explosion", "settings/ExplosionEffectSettings");
            ParticleEffect explosionSmoke = new ParticleEffect(this, "img/explosionSmoke", "settings/ExplosionSmokeEffectSettings");
            level.RightAndLeftLevelBorderCollisionEffects.AddEffect(explosion);
            level.RightAndLeftLevelBorderCollisionEffects.AddEffect(explosionSmoke);
            level.TopAndBottomLevelBorderCollisionEffects.AddEffect(sparks);
            level.PaddleCollisionEffects.AddEffect(sparks);
            level.BallCollision += level.HandleCollision;
        }

        private void CheckIfGameEnded()
        {
            if (player1.State == PongPlayerState.Defeated || player2.State == PongPlayerState.Defeated)
            {
                if (state != GameState.Ended)
                {
                    ball.Direction = Vector2.Zero;
                    ChangeState(GameState.Ended, 10);
                }
            }
        }

        private void UpdateTrailEffect()
        {
            if (ball.TrailEffect == null) return;
            int diff = player1.Health - player2.Health;
            if (diff == 0)
            {
                ball.TrailEffect.Settings.Colorization = Color.Yellow;
            }
            else if (diff < 0)
            {
                ball.TrailEffect.Settings.Colorization = Color.Red;
            }
            else if (diff > 0)
            {                
                ball.TrailEffect.Settings.Colorization = Color.Blue;
            }            
        }

        private void UpdateAI()
        {
            if (aiEnabled && state == GameState.Playing)
            {
                ai.Observe(ball);
                ai.Act(aiPaddle);
            }
        }

        private void HandleInput()
        {
            TouchCollection touchCollection = TouchPanel.GetState();
            foreach (TouchLocation tl in touchCollection)
            {
                if ((tl.State == TouchLocationState.Pressed) || (tl.State == TouchLocationState.Moved))
                {
                    switch (state)
                    {
                        case GameState.WaitingForStart:
                            ChangeState(GameState.Playing, 1000);
                            if (state == GameState.Playing)
                            {
                                StartRound(tl);
                            }
                            break;
                        case GameState.Playing:
                            UpdatePaddleLocations(tl);
                            break;
                        case GameState.Ended:
                            ChangeState(GameState.WaitingForStart, 2000);
                            if (state == GameState.WaitingForStart)
                            {
                                InitializeNewRound();
                            }
                            break;
                        default:
                            break;
                    }                    
                }
            }                                    
        }

        private void InitializeNewRound()
        {                        
            player1.Heal(100);
            player2.Heal(100);
            UpdateTrailEffect();
            ball.Direction = Vector2.Zero;
            ball.Position = level.ScreenCenter;
            player1Paddle.Position = new Vector2(level.TopLeftX + 80, level.TopLeftY + 200);
            player2Paddle.Position = new Vector2(level.ScreenWidth - 105, level.TopLeftY + 200);
            player1Paddle.YcoordinateGoal = player1Paddle.Position.Y;
            player2Paddle.YcoordinateGoal = player2Paddle.Position.Y;
            aiEnabled = true;
        }

        private void StartRound(TouchLocation tl)
        {
            if (tl.Position.X < level.ScreenWidth / 2 && tl.Position.Y < level.ScreenHeight / 2)
            {
                ball.Direction = new Vector2(-1.1f, -1);
                aiPaddle = player2Paddle;
            }
            else if (tl.Position.X < level.ScreenWidth / 2 && tl.Position.Y > level.ScreenHeight / 2)
            {
                ball.Direction = new Vector2(-1.1f, 1);
                aiPaddle = player2Paddle;
            }
            else if (tl.Position.X > level.ScreenWidth / 2 && tl.Position.Y < level.ScreenHeight / 2)
            {
                ball.Direction = new Vector2(1.1f, -1);
                aiPaddle = player1Paddle;
            }
            else if (tl.Position.X > level.ScreenWidth / 2 && tl.Position.Y > level.ScreenHeight / 2)
            {
                ball.Direction = new Vector2(1.1f, 1);
                aiPaddle = player1Paddle;
            }            
        }

        private void UpdatePaddleLocations(TouchLocation tl)
        {
            if (aiEnabled)
            {
                Point p = new Point((Int32)tl.Position.X, (Int32)tl.Position.Y);
                aiEnabled = !aiPaddle.Rectanlge.Contains(p);
            }
            if (tl.Position.X < level.ScreenWidth / 2)
            {
                player1Paddle.YcoordinateGoal = tl.Position.Y - player2Paddle.Height / 2;
            }
            else if (tl.Position.X > level.ScreenWidth / 2)
            {
                player2Paddle.YcoordinateGoal = tl.Position.Y - player2Paddle.Height / 2;
            }            
        }

        private void ChangeState(GameState s, int minTimeSinceLastChange)
        {
            int ticksNow = System.Environment.TickCount;
            if (ticksNow - lastStateChange > minTimeSinceLastChange)
            {
                state = s;
                lastStateChange = ticksNow;
            }
        }
    }
}
