using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using ParticleComponent;
using Microsoft.Xna.Framework.Audio;

/*
 * Authors: Jaakko Kosonen 
 */

namespace WindowsPhoneGame
{
    class PongLevel : CollisionEventHandler
    {
        private Vector2 player1HealthPosition;
        private Vector2 player2HealthPosition;
        private List<PongObject> collidableObjects;
        private PongObject lastCollisionObject;

        /// <summary>
        /// Occurs when the ball collides.
        /// </summary>
        public event CollisionEvent BallCollision;
        /// <summary>
        /// Occurs when the ball hits left border.
        /// </summary>
        public event CollisionEvent LeftBoundaryCollisioned;
        /// <summary>
        /// Occurs when ball hits right border.
        /// </summary>
        public event CollisionEvent RightBoundaryCollisioned;

        /// <summary>
        /// Particle effect for ball-hits-paddle.
        /// </summary>
        public ParticleEffectCollection PaddleCollisionEffects { get; set;  }
        /// <summary>
        /// Particle effect for ball-hits-top-or-bottom-border.
        /// </summary>
        public ParticleEffectCollection TopAndBottomLevelBorderCollisionEffects { get; set;  }
        /// <summary>
        /// Particle effect for ball-hits-left-or-right-border.
        /// </summary>
        public ParticleEffectCollection RightAndLeftLevelBorderCollisionEffects { get; set; }

        /// <summary>
        /// Sound effect for ball-hits-paddle.
        /// </summary>
        public SoundEffect PaddleCollisionSound { get; set; }
        /// <summary>
        /// Sound effect for ball-hits-top-or-bottom-border.
        /// </summary>
        public SoundEffect TopAndBottomLevelBorderSound { get; set; }
        /// <summary>
        /// Sound effect for ball-hits-left-or-right-border.
        /// </summary>
        public SoundEffect RightAndLeftLevelBorderSound { get; set; }        
        
        /// <summary>
        /// Pong ball.
        /// </summary>
        public PongBall Ball { get; set; }        

        /// <summary>
        /// Position for player1 health.
        /// </summary>
        public Vector2 Player1HealthPosition
        {
            get
            {
                return player1HealthPosition;
            }
        }

        /// <summary>
        /// Position for player2 health.
        /// </summary>
        public Vector2 Player2HealthPosition
        {
            get
            {
                return player2HealthPosition;
            }
        }

        public int TopLeftX { get; set; }
        public int TopLeftY { get; set; }
        public int ScreenWidth { get; set; }
        public int ScreenHeight { get; set; }
        
        /// <summary>
        /// Center of screen.
        /// </summary>
        public Vector2 ScreenCenter
        {
            get { return new Vector2(ScreenWidth / 2, ScreenHeight / 2); }
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="topLeftX">Origo-X</param>
        /// <param name="topLeftY">Origo-Y</param>
        /// <param name="screenWidth">Width</param>
        /// <param name="screenHeight">Height</param>
        public PongLevel(int topLeftX, int topLeftY, int screenWidth, int screenHeight)
        {
            TopLeftX = topLeftX;
            TopLeftY = topLeftY;
            ScreenWidth = screenWidth;
            ScreenHeight = screenHeight;
            player1HealthPosition = new Vector2(ScreenWidth / 2 - 15, 15);
            player2HealthPosition = new Vector2(screenWidth / 2 - 15, screenHeight - 50);
            collidableObjects = new List<PongObject>();
            CreateBorders();
            lastCollisionObject = new PongObject(0, 0, 0, 0);
            PaddleCollisionEffects = new ParticleEffectCollection();
            TopAndBottomLevelBorderCollisionEffects = new ParticleEffectCollection();
            RightAndLeftLevelBorderCollisionEffects = new ParticleEffectCollection();
        }

        /// <summary>
        /// Adds an object that should collide.
        /// </summary>
        /// <param name="o">Object that collides</param>
        public void AddCollidableObject(PongObject o)
        {
            collidableObjects.Add(o);
        }

        /// <summary>
        /// Checks for collisions.
        /// </summary>
        public void Update()
        {
            if (Ball == null) return;
            foreach (PongObject o in collidableObjects)
            {
                if (Ball.CheckForCollision(o.Rectanlge) && !o.Equals(lastCollisionObject))
                {                    
                    Ball.Hit(o.Force);
                    lastCollisionObject = o;
                    if (BallCollision != null)
                    {
                        BallCollision(this, new CollisionEventArgs(Ball.Position, o.CollisionType));
                    }
                    break;
                }
            }
        }

        /// <summary>
        /// Handles a collision.
        /// </summary>
        /// <param name="sender">Sender</param>
        /// <param name="e">Collision event arguments</param>
        public override void HandleCollision(object sender, CollisionEventArgs e)
        {
            switch (e.Type)
            {
                case CollisionType.BallAndPaddle:
                    HandlePaddleCollision(e.Position);
                    break;
                case CollisionType.BallAndTopOrBottomLevelBorder:
                    HandleTopOrBottomBorderCollision(e.Position);
                    break;
                case CollisionType.BallAndLeftLevelBorder:
                    HanldeLeftBorderCollision(e.Position);
                    break;
                case CollisionType.BallAndRightLevelBorder:
                    HandleRightBorderCollision(e.Position);
                    break;
                default:
                    break;
            }
            Ball.RotationDirection *= -1;
        }

        private void CreateBorders()
        {
            PongObject topBorder = new PongObject(TopLeftX, TopLeftY, ScreenWidth, 5);
            topBorder.Force = new Vector2(1, -1);
            topBorder.CollisionType = CollisionType.BallAndTopOrBottomLevelBorder;

            PongObject bottomBorder = new PongObject(TopLeftX, ScreenHeight - 5, ScreenWidth, 5);
            bottomBorder.Force = new Vector2(1, -1);
            bottomBorder.CollisionType = CollisionType.BallAndTopOrBottomLevelBorder;

            PongObject leftBorder = new PongObject(TopLeftX, 5, 5, ScreenHeight - 10);
            leftBorder.Force = new Vector2(-1, 1);
            leftBorder.CollisionType = CollisionType.BallAndLeftLevelBorder;

            PongObject rightBorder = new PongObject(ScreenWidth - 5, 5, 5, ScreenHeight - 10);
            rightBorder.Force = new Vector2(-1, 1);
            rightBorder.CollisionType = CollisionType.BallAndRightLevelBorder;

            AddCollidableObject(topBorder);
            AddCollidableObject(bottomBorder);
            AddCollidableObject(leftBorder);
            AddCollidableObject(rightBorder);
        }
        private void HandlePaddleCollision(Vector2 position)
        {
            PaddleCollisionEffects.CreateNewEffect(position + Ball.Origin);
            PaddleCollisionSound.Play();
        }
        private void HandleTopOrBottomBorderCollision(Vector2 position)
        {
            TopAndBottomLevelBorderCollisionEffects.CreateNewEffect(position + Ball.Origin);
            TopAndBottomLevelBorderSound.Play();
        }
        private void HanldeLeftBorderCollision(Vector2 position)
        {
            DoScoringBorderEffects(position + Ball.Origin);
            if (LeftBoundaryCollisioned != null)
            {
                LeftBoundaryCollisioned(this, new CollisionEventArgs(position, CollisionType.BallAndLeftLevelBorder));
            }
        }
        private void HandleRightBorderCollision(Vector2 position)
        {
            DoScoringBorderEffects(position + Ball.Origin);
            if (RightBoundaryCollisioned != null)
            {
                RightBoundaryCollisioned(this, new CollisionEventArgs(position, CollisionType.BallAndRightLevelBorder));
            }
        }
        private void DoScoringBorderEffects(Vector2 position)
        {
            RightAndLeftLevelBorderCollisionEffects.CreateNewEffect(position + Ball.Origin);
            RightAndLeftLevelBorderSound.Play();
            Microsoft.Devices.VibrateController.Default.Start(TimeSpan.FromMilliseconds(500));            
        }
    }
}
