﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Jypeli;

namespace Jypeli.SimplePhysics
{
    public class SimplePhysicsGame : Game
    {
        public static readonly double dt = 1.0 / 1000;

        List<SimplePhysicsObject> physObjects = new List<SimplePhysicsObject>();
        ImpulseCollection globalImpulses = new ImpulseCollection();

        public SimplePhysicsGame()
            : base()
        {
            IsFixedTimeStep = true;
            TargetElapsedTime = TimeSpan.FromSeconds( SimplePhysicsGame.dt );
        }

        public override void Add( GameObject o, int layer )
        {
            if ( o is SimplePhysicsObject )
            {
                physObjects.Add( (SimplePhysicsObject)o );
            }

            base.Add( o, layer );
        }

        protected override void Update( Time time )
        {
            Vector globalForce = globalImpulses.GetNextForce();

            for ( int i = 0; i < physObjects.Count; i++ )
            {
                Vector forces = globalForce + physObjects[i].Impulses.GetNextForce();

                physObjects[i].Acceleration = forces * physObjects[i].MassInv;
                physObjects[i].Velocity += physObjects[i].Acceleration * dt;
                physObjects[i].Position += physObjects[i].Velocity * dt;
            }

            SolveCollisions();

            base.Update( time );
        }

        private void SolveCollisions()
        {
            for ( int i = 0; i < physObjects.Count; i++ )
            {
                bool iStatic = double.IsInfinity( physObjects[i].Mass );

                for ( int j = 0; j < i; j++ )
                {
                    bool jStatic = double.IsInfinity( physObjects[j].Mass );

                    if ( iStatic && jStatic )
                        continue;

                    if ( BoundingRectangle.Intersects( physObjects[i].BoundingRectangle, physObjects[j].BoundingRectangle ) )
                    {
                        // Collision (perfectly elastic)
                        BoundingRectangle intsect = BoundingRectangle.GetIntersection( physObjects[i].BoundingRectangle, physObjects[j].BoundingRectangle );

                        if ( iStatic || jStatic )
                        {
                            int index = iStatic ? j : i;
                            Direction id = BoundingRectangle.GetIntersectionDirection( physObjects[index].BoundingRectangle, intsect );

                            if ( id == Direction.Left || id == Direction.Right )
                                physObjects[index].Velocity = new Vector( physObjects[index].Velocity.X, -physObjects[index].Velocity.Y );
                            else
                                physObjects[index].Velocity = new Vector( -physObjects[index].Velocity.X, physObjects[index].Velocity.Y );
                        }
                        else
                        {
                            Vector p_i = physObjects[i].Mass * physObjects[i].Velocity;
                            Vector p_j = physObjects[j].Mass * physObjects[j].Velocity;
                            Vector p_sum = p_i + p_j;
                            Vector p_diff = p_j - p_i;

                            Vector v_j = ( p_sum - p_diff ) / 2;
                            Vector v_i = p_sum - v_j;

                            physObjects[i].Velocity = v_i;
                            physObjects[j].Velocity = v_j;
                        }
                    }
                }
            }
        }

        //private Bounding

        public void ApplyImpulse( Impulse impulse )
        {
            globalImpulses.Add( impulse );
        }
    }
}
